/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PAL.h"
#include "basic_types.h"
#include "SSPCommon.h"
#include "PluginLib.h"
#include "PALString.h"

#include "mod_ssl.h"

#ifdef __cplusplus 
extern "C" {
#endif

// Constants
static const char MODULENAME[] = "ssp";
static const char DATAKEYINIT[] = "SSPInitialization";

// Structures
typedef struct
{
    // Configuration type variables (read-only).
    ConfigPluginGlobals GlobalConfig;
    const char *        szWorkDir;
    int                 iSharedMemSize;
    int                 iPrivateMemSize;
}
SServerConfig;

// Imported function from mod_ssl.h.
static APR_OPTIONAL_FN_TYPE(ssl_is_https) *pfnIsHttps = NULL;

// Prototypes
static void *InitializeSeverConfig(apr_pool_t *pPool, server_rec *pServerRec);
static void BingFunctionRegistration(apr_pool_t *pPool);
static int PostConfigHandler(apr_pool_t *pPoolConf, apr_pool_t *pPoolLog, apr_pool_t *pPoolTemp, server_rec *pServerRec);
static void ChildInitHandler(apr_pool_t *pPoolChild, server_rec *pServerRec);
static int ContentFilterHandler(ap_filter_t *pfilter, apr_bucket_brigade *pBrigade);
static apr_status_t Cleanup(void *pContext);
static apr_status_t CleanupChild(void *pContext);
static const char *SetCmdStr(cmd_parms *pCmd, void *pStruct, const char *pArg);
static const char *SetCmdPath(cmd_parms *pCmd, void *pStruct, const char *pArg);
static const char *SetCmdInt(cmd_parms *pCmd, void *pStruct, const char *pArg);

static const command_rec ConfigurationDirectives[] =
{
    AP_INIT_TAKE1("SSPWorkDir",
                  (cmd_func) SetCmdPath,
                  (void *) APR_OFFSETOF(SServerConfig, szWorkDir),
                  OR_ALL,
                  "Plugin working directory"),
    AP_INIT_TAKE1("SSPSharedMemSize",
                  (cmd_func) SetCmdInt,
                  (void *) APR_OFFSETOF(SServerConfig, iSharedMemSize),
                  OR_ALL,
                  "Shared memory size"),
    AP_INIT_TAKE1("SSPPrivateMemSize",
                  (cmd_func) SetCmdInt,
                  (void *) APR_OFFSETOF(SServerConfig, iPrivateMemSize),
                  OR_ALL,
                  "Private memory size"),
    { NULL }
};


// The entry point for the module.
module AP_MODULE_DECLARE_DATA ssp_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    InitializeSeverConfig,
    NULL,
    ConfigurationDirectives,
    BingFunctionRegistration
};


static void *InitializeSeverConfig(apr_pool_t *pPool, server_rec *pServerRec)
{
    // Note that Apache performs two-phase initialization. This function is called twice
    // but it should be ok since the pools are cleaned up in between.
    SServerConfig *pServerConfig = (SServerConfig *) apr_pcalloc(pPool, sizeof(SServerConfig));
    if (pServerConfig == NULL)
    {
        ap_log_error(APLOG_MARK, APLOG_CRIT, APR_EGENERAL, pServerRec, "Unable to create server config");
        return NULL;
    }

    pServerConfig->GlobalConfig.SetDefault();
    return pServerConfig;
}

// Function registration.
static void BingFunctionRegistration(apr_pool_t *pPool)
{
    // Register the handler to be called the parent process (post configuration).
    ap_hook_post_config(
        PostConfigHandler,
        NULL,
        NULL,
        APR_HOOK_MIDDLE);

    // Register the handler to be called in each child process.
    ap_hook_child_init(
        ChildInitHandler,
        NULL,
        NULL,
        APR_HOOK_MIDDLE);

    // Register the output filter handler.
    ap_register_output_filter(
        MODULENAME,
        ContentFilterHandler,
        NULL,
        AP_FTYPE_CONTENT_SET) ;
}

static int PostConfigHandler(apr_pool_t *pPoolConf, apr_pool_t *pPoolLog, apr_pool_t *pPoolTemp, server_rec *pServerRec)
{
    // Initialize all the inter-process variables/shared memory.
    apr_status_t status;
    void *pUserData = NULL;
    SServerConfig *pServerConfig;

    // Note that Apache performs two-phase initialization. This function is called twice
    // so we will only initialize the second time. On restart, it is called only once,
    // but enter the 2nd pass since the user data is set from the first time it started.
    status = apr_pool_userdata_get(&pUserData, DATAKEYINIT, pServerRec->process->pool);
    if (status != APR_SUCCESS)
    {
        ap_log_error(APLOG_MARK, APLOG_CRIT, status, pServerRec, "Failed to get user data %s", DATAKEYINIT);
        return status;
    }
    if (pUserData == NULL)
    {
        // 1st pass.
        // Set's the value at DATAKEYINIT so that the next pass will skip this block.
        apr_pool_userdata_set(
            (void *) 1,
            DATAKEYINIT,
            apr_pool_cleanup_null,
            pServerRec->process->pool);

        return OK;
    }

    // 2nd pass
    pServerConfig = (SServerConfig *) ap_get_module_config(pServerRec->module_config, &ssp_module);
    if (pServerConfig == NULL)
    {
        ap_log_error(APLOG_MARK, APLOG_CRIT, APR_EGENERAL, pServerRec, "Empty server config");
        return DECLINED;
    }

    try
    {        
        if (pServerConfig->szWorkDir != NULL)
        {
            TCHAR wszPath[MAX_PATH];
            if (PALString::Convert(wszPath, ARRAYSIZE(wszPath), pServerConfig->szWorkDir, strlen(pServerConfig->szWorkDir)) == 0)
            {
                if (!SetDataFolder(wszPath))
                {
                    ap_log_error(APLOG_MARK, APLOG_WARNING, 0, pServerRec, "Unable to set working directory");
                }
            }
        }
        else
        {
            ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, pServerRec, "Using default working directory");
        }

        if (pServerConfig->iSharedMemSize > 0)
        {
            pServerConfig->GlobalConfig.cbSharedAllocatorSize = pServerConfig->iSharedMemSize;
        }

        if (pServerConfig->iPrivateMemSize > 0)
        {
            pServerConfig->GlobalConfig.cbPrivateAllocatorSize = pServerConfig->iPrivateMemSize;
        }
        
        if (InitPluginGlobals(pServerConfig->GlobalConfig, pPoolConf) == FALSE)
        {
            ap_log_error(APLOG_MARK, APLOG_CRIT, status, pServerRec, "Unable to initialize plugin globals");
            return DECLINED;
        }
    }
    catch (int e)
    {
        ap_log_error(APLOG_MARK, APLOG_CRIT, e, pServerRec, "PostConfigHandler: Exception thrown");
        return DECLINED;
    }

    // Note that we do not call apr_pool_cleanup_register() because subpools
    // are destroyed before this call which causes some APR objectes
    // to be destroyed (ie, apr_thread_pool_t) prematurely.
    apr_pool_pre_cleanup_register(
        pPoolConf,
        pPoolConf,
        Cleanup);

    pfnIsHttps = APR_RETRIEVE_OPTIONAL_FN(ssl_is_https);

    return OK;
}

static void ChildInitHandler(apr_pool_t *pPoolChild, server_rec *pServerRec)
{
    // Called at every forked child process.
    SServerConfig *pServerConfig = (SServerConfig *) ap_get_module_config(pServerRec->module_config, &ssp_module);

    if (pServerConfig == NULL)
    {
        ap_log_error(APLOG_MARK, APLOG_CRIT, APR_EGENERAL, pServerRec, "Empty server config");
        return;
    }

    try
    {
        if (InitPluginGlobalsChild(pPoolChild) == FALSE)
        {
            ap_log_error(APLOG_MARK, APLOG_CRIT, APR_EGENERAL, pServerRec, "Unable to initialize plugin globals in child process");
            return;
        }
    }
    catch (int e)
    {
        ap_log_error(APLOG_MARK, APLOG_CRIT, e, pServerRec, "ChildInitHandler: Exception thrown");
        return;
    }

    // Note that we do not call apr_pool_cleanup_register() because subpools
    // are destroyed before this call which causes some APR objectes
    // to be destroyed (ie, apr_thread_pool_t) prematurely.
    apr_pool_pre_cleanup_register(
        pPoolChild,
        pPoolChild,
        CleanupChild);
}

static int ContentFilterHandler(ap_filter_t *pFilter, apr_bucket_brigade *pBrigade)
{
    WSFilter *pWSFilter = NULL;

    // If connection is HTTPS, then skip the processing.
    if (pfnIsHttps != NULL && pfnIsHttps(pFilter->r->connection))
    {
        goto Done;
    }

    try
    {
        if (pFilter->ctx == NULL)
        {
            pFilter->ctx = (WSFilter *) new WSFilter();
            if (pFilter->ctx == NULL)
            {
                ap_log_perror(APLOG_MARK, APLOG_CRIT, 0, pFilter->r->pool, "ContentFilterHandler: Unable to allocate WSFilter");
            }
        }

        pWSFilter = (WSFilter *)pFilter->ctx;
        pWSFilter->SetFilter(pFilter, pBrigade);
        ProcessResponse(pWSFilter);
    }
    catch (int e)
    {
        ap_log_perror(APLOG_MARK, APLOG_CRIT, e, pFilter->r->pool, "ContentFilterHandler: Exception int thrown");
    }
    catch (...)
    {
        ap_log_perror(APLOG_MARK, APLOG_CRIT, 0, pFilter->r->pool, "ContentFilterHandler: Exception thrown");
    }

Done:
    // If the response is complete, clear the context, since ProcessResponse() does not
    // clear the completed response.
    if (pWSFilter != NULL)
    {
        for (apr_bucket *pBucket = APR_BRIGADE_FIRST(pBrigade);
            pBucket != APR_BRIGADE_SENTINEL(pBrigade);
            pBucket = APR_BUCKET_NEXT(pBucket))
        {
            if (APR_BUCKET_IS_EOS(pBucket))
            {
                delete pWSFilter;
                pFilter->ctx = NULL;
                break;
            }
        }
    }

    return ap_pass_brigade(pFilter->next, pBrigade);
}

static apr_status_t Cleanup(void *pContext)
{
    apr_pool_t *pPool = (apr_pool_t *) pContext;

    try
    {
        FreePluginGlobals(pPool);
    }
    catch (int e)
    {
        ap_log_perror(APLOG_MARK, APLOG_CRIT, e, pPool, "Cleanup: Exception thrown");
        return e;
    }

    return APR_SUCCESS;
}

static apr_status_t CleanupChild(void *pContext)
{
    apr_pool_t *pPool = (apr_pool_t *) pContext;

    try
    {
        FreePluginGlobalsChild(pPool);
    }
    catch (int e)
    {
        ap_log_perror(APLOG_MARK, APLOG_CRIT, e, pPool, "Cleanup Child: Exception thrown");
        return e;
    }

    return APR_SUCCESS;
}

static const char *SetCmdStr(cmd_parms *pCmd, void *pStruct, const char *pArg)
{
    int iOffset = (int) ((long) pCmd->info);
    SServerConfig *pServerConfig = (SServerConfig *) ap_get_module_config(pCmd->server->module_config, &ssp_module);
 
    // It's ok to set NULL. We will validate later.
    *(const char **)(((char *) pServerConfig) + iOffset) = pArg;

    return NULL;
}

static const char *SetCmdPath(cmd_parms *pCmd, void *pStruct, const char *pArg)
{
    int iOffset = (int) ((long) pCmd->info);
    SServerConfig *pServerConfig = (SServerConfig *) ap_get_module_config(pCmd->server->module_config, &ssp_module);
    const char *szPath = ap_server_root_relative(pCmd->pool, pArg);

    // It's ok to set NULL. We will validate later.
    *(const char **)(((char *) pServerConfig) + iOffset) = szPath;

    return NULL;
}

static const char *SetCmdInt(cmd_parms *pCmd, void *pStruct, const char *pArg)
{
    int iOffset = (int) ((long) pCmd->info);
    SServerConfig *pServerConfig = (SServerConfig *) ap_get_module_config(pCmd->server->module_config, &ssp_module);

    if (pArg != NULL)
    {
        int iVal;
        iVal = atoi(pArg);
        *(int *)(((char *) pServerConfig) + iOffset) = iVal;
    }

    return NULL;
}

#ifdef __cplusplus 
} // extern "C"
#endif
