/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PAL.h"
#include "PALString.h"
#include "PALFile.h"
#include "SSPCommon.h"
#include "bloomfilter.h"
#include "DeltaMerge.h"

void PrintHelp()
{
    printf("sspservice [directory] [publishing mode] [help ssp mode]\n");
    printf("  [directory] Location of the sitemap generated data.\n");
    printf("  [publishing mode] The publishing modes: 0-preview, 1-sitemap & robots.txt (default).\n");
    printf("  [help ssp mode] The help ssp modes: 0-no, 1-yes (default).\n");
    printf("Or for cross-server merge:\n");
    printf("  sspservice %s <args>\n", MERGE_OPTION);
    printf("  (use \"sspservice %s\" for help on the %s option)\n", MERGE_OPTION, MERGE_OPTION);
}


#if APR_HAVE_WINDOWS_H
int _cdecl _tmain(int argc, TCHAR* argv[])
#else
int main(int argc, char *argv[])
#endif
{
    int iRetValue = 0;
    PALFind findFile;
    TCHAR wszFolder[MAX_PATH];
    BYTE pbBloom[2000]; // probability of collision 1/200k for 1000 entries
    BloomFilter bf;
    // Default mode is 2.
    int iSiteMapPublishMode = 2;
    int iSiteMapHelpSSPMode = 1;

    if ((argc > 1) && (0 == PALString::CompareInsensitive( argv[1], TEXT(MERGE_OPTION))))
    {
        return CommandLineMerge(argc - 2, argv + 2);
    }

    if (!bf.Init(pbBloom, sizeof(pbBloom)))
    { // not that this can actually fail
        iRetValue = -1;
        goto Done;
    }

    if (!PAL_Initialize(TRUE))
    {
        iRetValue = -1;
        goto Done;
    }
    
    if (argc == 4)
    {
        // Get the data directory.
#if APR_HAVE_WINDOWS_H
        if (PALString::Copy(wszFolder, ARRAYSIZE(wszFolder), argv[1]) != 0)
        {
            iRetValue = -1;
            goto Done;
        }
#else
        if (PALString::Convert(wszFolder, ARRAYSIZE(wszFolder), argv[1], strlen(argv[1])) != 0)
        {
            iRetValue = -1;
            goto Done;
        }
#endif

        // Get the publishing mode.
#if APR_HAVE_WINDOWS_H
        int iMode = _tstoi(argv[2]);
        int iHelpSSP = _tstoi(argv[3]);
#else
        int iMode = atoi(argv[2]);
        int iHelpSSP = atoi(argv[3]);
#endif

        iSiteMapPublishMode = iMode;
        iSiteMapHelpSSPMode = iHelpSSP;
    }
    else
    {
        PrintHelp();
        iRetValue = -1;
        goto Done;
    }

    if (!SetSiteMapPublishMode((DWORD) iSiteMapPublishMode) || !SetSiteMapHelpSSPMode((DWORD) iSiteMapHelpSSPMode))
    {
        iRetValue = -1;
        goto Done;
    }

    if (findFile.Initialize(wszFolder, TEXT("site_*")) == FALSE)
    {
        iRetValue = -1;
        goto Done;
    }

    if (findFile.Find() != PALFind::Success)
    {
        iRetValue = -1;
        goto Done;
    }

    do
    {
        if (findFile.GetAttrib() == PALFind::Directory)
        {
            const TCHAR* tszSubfolder = findFile.GetName();

            // We find folders with "_<digits>" suffix, but DoOneSite() wants the portion without that suffix
            Size_t cchPrefix = FolderPrefixSize(tszSubfolder);
            if (!cchPrefix)
            {
                continue;
            }

            // DoOneSite() only needs a given prefix once - it iterates over different "_<digits>" suffixes internally
            if (bf.Contains(sizeof(TCHAR) * cchPrefix, tszSubfolder))
            {
                continue;
            }
            bf.Add(sizeof(TCHAR) * cchPrefix, tszSubfolder);

            TCHAR wszSubfolder[MAX_PATH];
            if (SNTPRINTF_S(wszSubfolder, ARRAYSIZE(wszSubfolder), _TRUNCATE,
                STRFMT_TLS STRFMT_TDIRSLASH STRFMT_NTLS,
                wszFolder, (int) cchPrefix, tszSubfolder) > 0)
            {
                DoOneSite(wszSubfolder, NULL, NULL, NULL);
            }
        }
    }
    while (findFile.Find() == PALFind::Success);

Done:
    findFile.Destroy();

    PAL_Terminate();
    
    if (iRetValue != 0)
    {
        printf("Merge processing error.\n");
    }

    return iRetValue;
}
