/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "SSPCommon.h"
#include "SSPConfig.h"
#include "DeltaMerge.h"
#include "PALString.h"
#include "SimpleHashTable.h"
#include "SSPHash.h"
#include "Counters.h"
#include "PALFile.h"

#include "iiscnfg.h" // MD_ & IIS_MD_ #defines
#include <initguid.h>
#include "iadmw.h" // MSAdminBase
#include <ahadmin.h> // IIS7 config


static FILE* g_fpGlobalLog = NULL;
static DWORD g_fLogLevel = 0;

#define SERVICE_LOG(...) { \
  struct tm tms; \
  PluginGetLocalTime(&tms); \
  fprintf(g_fpGlobalLog, "%04u/%02u/%02u %02u:%02u:%02u ", 1900 + tms.tm_year, 1 + tms.tm_mon, tms.tm_mday, tms.tm_hour, tms.tm_min, tms.tm_sec); \
  fprintf(g_fpGlobalLog, __VA_ARGS__); \
}

static TCHAR* MyStrDup(const TCHAR* sz, size_t cch)
{
    size_t cb = (cch + 1) * sizeof(TCHAR);
    TCHAR* szRet = (TCHAR*)malloc(cb);  if (!szRet) return NULL;
    memcpy_s(szRet, cb, sz, cb);
    szRet[cch] = '\0';
    return szRet;
}

static TCHAR* MyStrDup(const TCHAR* sz)
{
    return MyStrDup(sz, _tcslen(sz));
}

static unsigned IISVersion()
{
    unsigned uRet = 0;
    DWORD dwType;
    HKEY hKey = NULL;
    DWORD dwVersion, cbBuf = sizeof(dwVersion);

    if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("SYSTEM\\CurrentControlSet\\services\\W3SVC\\Parameters"), 0, KEY_QUERY_VALUE, &hKey)) goto Exit;
    if (ERROR_SUCCESS != RegQueryValueEx(hKey, TEXT("MajorVersion"), NULL, &dwType, (BYTE*)&dwVersion, &cbBuf)) goto Exit;
    uRet = dwVersion;
Exit:
    if (hKey) RegCloseKey(hKey);
    return uRet;
}

typedef struct
{
    TCHAR* wszSitemapFolder; // table key - where the sitemap goes
    TCHAR* wszRobotsFolder; // host root known to IIS (where robots.txt lives) - may be aliased to wszSitemapFolder
    TCHAR* wszDataFolder; // location of delta and snapshot files
    char* szSubPath; // path of SitemapFolder relative to RobotsFolder.  I.e., wszSitemapFolder == wszRobotsFolder + szSubPath
    UInt64 cUrls;
} SiteEntry;

typedef struct
{
    TCHAR* wszFolderPrefix;
    UInt64 cUrls;
} PreliminarySiteEntry;

// Table indexed by SitemapFolder, serves two purposes (both implemented in AddSite() ):
// (1) avoid using WWWRoot not in IIS config (security consideration)
// (2) if multiple data folders point to the same SitemapFolder, process only the one with the most cUrls +
// The table can have a mix of 2 types of entries - those from EnumerateIIS*Sites() and those from AddSubSite()
class SiteTable
{
    SimpleHashTable<SiteEntry> m_ht;
    SimpleHashTable<PreliminarySiteEntry> m_htPreliminary;
    
public:
    ~SiteTable() { Free(); }
    BOOL Initialize();
    BOOL AddWWWRoot(const TCHAR* wszWWWRoot);
    SiteEntry* FindSitemapPath(TCHAR* wszSitemapPath, BOOL fOkToAdd);
    void AddSite(const TCHAR* wszFolderPrefix, UInt64 cUrls);
    void AddDataFolder(const TCHAR* wszFolder);
    SiteEntry* AddSubSite(TCHAR* wszPath, Size_t cchRobotsPath, Size_t cchSitemapPath);
    void Process(PALEvent *pStopEvent);
    void Finalize();
    void Free();
};

BOOL SiteTable::Initialize()
{
    return m_ht.Initialize(10000) && m_htPreliminary.Initialize(10000);
}

static BOOL FreeSiteEntry(void* pvContext, SiteEntry* p)
{
    UNREFERENCED_PARAMETER(pvContext);
    if (p->wszRobotsFolder)
    {
        if (p->wszRobotsFolder != p->wszSitemapFolder)
        {
            free(p->wszRobotsFolder);
        }
        p->wszRobotsFolder = NULL;
    }
    if (p->wszSitemapFolder)
    {
        free(p->wszSitemapFolder);
        p->wszSitemapFolder = NULL;
    }
    if (p->wszDataFolder)
    {
        free(p->wszDataFolder);
        p->wszDataFolder = NULL;
    }
    if (p->szSubPath)
    {
        free(p->szSubPath);
        p->szSubPath = NULL;
    }
    return TRUE;
}

static BOOL FreePreliminarySiteEntry(void* pvContext, PreliminarySiteEntry* p)
{
    UNREFERENCED_PARAMETER(pvContext);
    if (p->wszFolderPrefix)
    {
        free(p->wszFolderPrefix);
        p->wszFolderPrefix = NULL;
    }
    return TRUE;
}

void SiteTable::Free()
{
    m_ht.Enumerate(NULL, FreeSiteEntry);
    m_htPreliminary.Enumerate(NULL, FreePreliminarySiteEntry);
}

static UInt64 FolderStringHash(const TCHAR* wszWWWRoot, size_t cch)
{
    TCHAR temp[MAX_PATH];
    if (cch >= ARRAYSIZE(temp)) return 0;
    memcpy_s(temp, sizeof(temp), wszWWWRoot, (cch + 1) * sizeof(TCHAR));
    _tcslwr_s(temp, ARRAYSIZE(temp));
    return SSPHash(temp, cch * sizeof(TCHAR));
}

static UInt64 FolderStringHash(const TCHAR* wszWWWRoot)
{
    return FolderStringHash(wszWWWRoot, _tcslen(wszWWWRoot));
}

BOOL SiteTable::AddWWWRoot(const WCHAR* wszWWWRoot)
{
    UInt64 hash = FolderStringHash(wszWWWRoot);  if (!hash) return FALSE;
    WCHAR* wsz = MyStrDup(wszWWWRoot);
    SiteEntry e;
    memset(&e, 0, sizeof(e));
    e.wszSitemapFolder = wsz;
    e.wszRobotsFolder = wsz; // updated in AddSubSite()
    BOOL fRet = m_ht.Add(hash, e);
    if (!fRet)
    {
        free(wsz);
    }
    return fRet;
}

static void MakeSlashesForward(__inout_ecount(cch) char* s, size_t cch)
{
    for (size_t i = 0; i < cch; i++)
    {
        if ('\\' == s[i])
        {
            s[i] = '/';
        }
    }
}

SiteEntry* SiteTable::AddSubSite(TCHAR* wszPath, size_t cchPath, size_t cchRobotsPath)
{
    char* szSubPath = NULL;
    TCHAR* wszRobotsFolder = NULL;
    size_t cb;
    SiteEntry* p = NULL;

    cb = cchPath - cchRobotsPath + 1;
    szSubPath = (char*)malloc(cb);
    if (!szSubPath)
    {
        goto Exit;
    }
    if (0 != PALString::Convert(szSubPath, cb, wszPath + cchRobotsPath, cchPath - cchRobotsPath))
    {
        goto Exit;
    }
    MakeSlashesForward(szSubPath, cb - 1); // used in robots.txt path to sitemap

    cb = (cchRobotsPath + 1) * sizeof(TCHAR);
    wszRobotsFolder = (TCHAR*)malloc(cb);
    if (!wszRobotsFolder)
    {
        goto Exit;
    }
    memcpy(wszRobotsFolder, wszPath, cb);
    wszRobotsFolder[cchRobotsPath] = '\0';

    if (!AddWWWRoot(wszPath))
    {
        goto Exit;
    }
    p = m_ht.Get(FolderStringHash(wszPath));

Exit:
    if (p)
    {
        p->wszRobotsFolder = wszRobotsFolder;
        p->szSubPath = szSubPath;
    }
    else
    {
        if (szSubPath) free(szSubPath);
        if (wszRobotsFolder) free(wszRobotsFolder);
    }
    SERVICE_LOG("%s %s as subfolder of " STRFMT_TS "\n", p ? "added" : "failed to add", szSubPath, wszRobotsFolder);
    return p;
}

SiteEntry* SiteTable::FindSitemapPath(TCHAR* wszSitemapPath, BOOL fOkToAdd)
{
    SiteEntry* p = m_ht.Get(FolderStringHash(wszSitemapPath));
    if (p && 0 == _wcsicmp(wszSitemapPath, p->wszSitemapFolder)) // _wcsicmp guards against hash collision attack
    {
        return p;
    }
    
    if (!fOkToAdd)
    {
        return NULL;
    }
    
    // IIS has cases where a site's APPL_PHYSICAL_PATH is a subfolder of the real host root.
    // GetSitesKnownToIIS() only gives host roots, so to handle sub-folder APPL_PHYSICAL_PATH cases,
    // we look for an entry whose host root is a path prefix of the passed in one.
    // If found, add a new entry for the passed in one, noting the extra suffix (szSubFolder) of the of the passed in one
    // relative to the host root.  APPL_PHYSICAL_PATH is the sitemap location and the host root is the robots location.
    size_t cchSitemapPath = _tcslen(wszSitemapPath);
    size_t cch = cchSitemapPath;
    while (cch)
    {   // remove last subdirectory and try again
        while (cch)
        {
            cch--;
            if ('\\' == wszSitemapPath[cch] || '/' == wszSitemapPath[cch])
            {
                break;
            }
        }
        if (!cch)
        {
            break;
        }

        SiteEntry* p = m_ht.Get(FolderStringHash(wszSitemapPath, cch));
        // Check for match against wszRobotsFolder rather than wszSitemapFolder because,
        // in case we have multiple levels of nested sites (e.g., "a", "a\b", "a\b\c"),
        // when looking for the parent of "a\b\c", we need to find "a" - using "a\b" as
        // the parent of "a\b\c" would give an incorrect robots.txt location "a\b"
        // (robots.txt lives in the host's root - i.e., "a").
        if (p && 0 == _tcsncicmp(wszSitemapPath, p->wszRobotsFolder, cch))
        { // found a prefix of wszSitemapPath but not wszSitemapPath itself
            return AddSubSite(wszSitemapPath, cchSitemapPath, cch);
        }
    }
    return NULL;
}

static BOOL ReadConfig(const TCHAR* wszDataFolderPrefix, __out_ecount(cchWWWRoot) TCHAR* wszWWWRoot, size_t cchWWWRoot)
{
    SSPConfig config;
    TCHAR wszConfigPath[MAX_PATH], wszDataFolder[MAX_PATH];
    if (!GetDataFolderWithIndex(wszDataFolderPrefix, 0, wszDataFolder, ARRAYSIZE(wszDataFolder)))
    {
        return FALSE;
    }
    if (!GetConfigFile(wszDataFolder, wszConfigPath, ARRAYSIZE(wszConfigPath)))
    {
        return FALSE;
    }
    if (!ReadSSPConfig(wszConfigPath, &config))
    {
        return FALSE;
    }
    size_t cch = _tcslen(config.wszwwwroot);
    if (cch >= cchWWWRoot)
    {
        return FALSE;
    }
    memcpy_s(wszWWWRoot, cchWWWRoot * sizeof(TCHAR), config.wszwwwroot, (cch + 1) * sizeof(TCHAR));
    return TRUE;
}

static UInt64 ReadUrlCount(const TCHAR* wszDataFolder)
{
    UInt64 uRet = 0;

    TCHAR wszCountersFile[MAX_PATH];
    Counters counters;
    if (GetAllTimeCountersFile(wszDataFolder, wszCountersFile, ARRAYSIZE(wszCountersFile)) && ReadCounters(wszCountersFile, &counters))
    {
        uRet += counters.cBeginResponse.m_val;
    }

    ResetCounters(&counters);
    if (GetPluginCountersFile(wszDataFolder, wszCountersFile, ARRAYSIZE(wszCountersFile)) && ReadCounters(wszCountersFile, &counters))
    {
        uRet += counters.cBeginResponse.m_val;
    }

    return uRet;
}

static void LogDedupe(UInt64 cUrlsBad, UInt64 cUrlsGood, const TCHAR* wszFolderBad, const TCHAR* wszFolderGood, const TCHAR* wszWWWRoot)
{
    if (g_fLogLevel)
    {
        SERVICE_LOG( STRFMT_TS " (%I64u) skipped in favor of " STRFMT_TS " (%I64u) - both point to " STRFMT_TS "\n", 
                wszFolderBad, cUrlsBad, wszFolderGood, cUrlsGood, wszWWWRoot);
    }
}

// wszFolder is really a prefix (without the plugin index)
void SiteTable::AddSite(const TCHAR* wszFolder, UInt64 cUrls)
{
    TCHAR wszWWWRoot[MAX_PATH];
    if (!ReadConfig(wszFolder, wszWWWRoot, ARRAYSIZE(wszWWWRoot)))
    {
        SERVICE_LOG( STRFMT_TS ": cannot read wwwroot from config.ini\n", wszFolder);
        return;
    }

    SiteEntry* p = FindSitemapPath(wszWWWRoot, TRUE);
    if (!p)
    {
        if (g_fLogLevel)
        {
            SERVICE_LOG( STRFMT_TS " points to wwwroot " STRFMT_TS " not known to IIS\n", wszFolder, wszWWWRoot);
        }
        return;
    }

    if (p->wszDataFolder)
    {
        if (p->cUrls < cUrls)
        { // new one is better - kick out existing one
            LogDedupe(p->cUrls, cUrls, p->wszDataFolder, wszFolder, wszWWWRoot);
            delete[] p->wszDataFolder;
        }
        else
        { // existing one is better - skip the new one
            LogDedupe(cUrls, p->cUrls, wszFolder, p->wszDataFolder, wszWWWRoot);
            return;
        }
    }

    p->wszDataFolder = MyStrDup(wszFolder);
    p->cUrls = cUrls;
}

// This method is fed folders of the form "<prefix>_<pluginindex>", its job is to
// track unique values of <prefix> and accumualte the URL count per <prefix>
void SiteTable::AddDataFolder(const TCHAR* wszFolder)
{
    size_t cchPrefix = FolderPrefixSize(wszFolder);
    if (!cchPrefix)
    {
        SERVICE_LOG( STRFMT_TS " does not end with '_<integer>' pattern - skipped\n", wszFolder);
        return;
    }

    UInt64 cUrls = ReadUrlCount(wszFolder);

    UInt64 hash = FolderStringHash(wszFolder, cchPrefix);
    PreliminarySiteEntry* p = m_htPreliminary.Get(hash);
    if (p)
    {
        if (_tcsncicmp(p->wszFolderPrefix, wszFolder, cchPrefix))
        {
            SERVICE_LOG( STRFMT_TS " skipped due to hash collision with " STRFMT_TS "\n", wszFolder, p->wszFolderPrefix);
            return;
        }
        p->cUrls += cUrls;
    }
    else
    {
        PreliminarySiteEntry e;
        e.wszFolderPrefix = MyStrDup(wszFolder, cchPrefix);
        e.cUrls = cUrls;
        if (!m_htPreliminary.Add(hash, e))
        {
            SERVICE_LOG( STRFMT_TS " skipped - hash table full\n", wszFolder);
        }
    }
}

static BOOL AddSite(void* pvContext, PreliminarySiteEntry* p)
{
    ((SiteTable*)(pvContext))->AddSite(p->wszFolderPrefix, p->cUrls);
    return TRUE; // continue trying to add other entries even if this one failed
}

void SiteTable::Finalize()
{
    m_htPreliminary.Enumerate(this, ::AddSite);
}

typedef struct
{
    SiteTable* pSites;
    PALEvent *pStopEvent;
} ProcessContext;

// Usually returns TRUE so that the hashtable enumeration continnues (so other, innocent sites can be processed)
static BOOL ProcessSite(void* pvContext, SiteEntry* p)
{
    ProcessContext* pCtx = (ProcessContext*)pvContext;
    if (pCtx->pStopEvent && (PALEvent::Signaled == pCtx->pStopEvent->Wait(0)))
    {
        SERVICE_LOG("service is stopping\n");
        return FALSE; // terminate hashtable enumeration
    }

    if (!p->wszDataFolder) // normal - table contains all sites from IIS config, most might not have our plugin enabled
    {
        return TRUE;
    }

    TCHAR wszWWWRoot[MAX_PATH];
    if (!ReadConfig(p->wszDataFolder, wszWWWRoot, ARRAYSIZE(wszWWWRoot))) // should normally be caught during AddSite()
    {
        SERVICE_LOG( STRFMT_TS ": cannot read wwwroot from config.ini, even though we read it ok earlier\n", p->wszDataFolder);
        return TRUE;
    }

    // make sure wszWWWRoot still points to the entry it pointed to during AddSite()
    if (pCtx->pSites->FindSitemapPath(wszWWWRoot, FALSE) != p)
    {
        SERVICE_LOG( STRFMT_TS " wwwroot changed after it was validated\n", p->wszDataFolder);
        return TRUE;
    }

    SERVICE_LOG( STRFMT_TS " processing to " STRFMT_TS " (" STRFMT_TS ")\n", p->wszDataFolder, p->wszSitemapFolder, p->wszRobotsFolder);
    DoOneSite(p->wszDataFolder, p->wszRobotsFolder, p->szSubPath, pCtx->pStopEvent);
    return TRUE;
}

void SiteTable::Process(PALEvent *pStopEvent)
{
    ProcessContext ctx = { this, pStopEvent };
    m_ht.Enumerate(&ctx, ProcessSite);
}

#define CHECK_HR_EXIT(a) { hr = a; if (FAILED(hr)) goto Exit; }
#define CHECK_HR_LOOP(x) { hr = x; if (FAILED(hr)) goto Loop; }
#define RELEASE(p) if (p) p->Release()

static BOOL EnumerateIIS6Sites(SiteTable* pSites)
{
    BOOL fRet = FALSE, fCoInitialized = FALSE;
    IMSAdminBase* pIMeta = NULL;
    METADATA_HANDLE hKey = NULL;
    HRESULT hr;

    CHECK_HR_EXIT(CoInitialize(NULL));  fCoInitialized = TRUE;
    CHECK_HR_EXIT(CoCreateInstance(CLSID_MSAdminBase, NULL, CLSCTX_ALL, IID_IMSAdminBase, (void **) &pIMeta));
    CHECK_HR_EXIT(pIMeta->OpenKey(METADATA_MASTER_ROOT_HANDLE, L"/LM", METADATA_PERMISSION_READ, 20, &hKey));
    
    for (DWORD iKey = 0; ; iKey++)
    {
        WCHAR SubKeyName[METADATA_MAX_NAME_LEN], SubKeyPath[METADATA_MAX_NAME_LEN];
        const WCHAR* path = L"/W3SVC";
        hr = pIMeta->EnumKeys(hKey, path, SubKeyName, iKey);
        if ((HRESULT_FROM_WIN32(ERROR_NO_MORE_ITEMS) == hr) || (HRESULT_FROM_WIN32(ERROR_PATH_NOT_FOUND) == hr) || (iKey > 100000))
        {
            break;
        }
        if (FAILED(hr))
        {
            continue;
        }
        if (0 >= swprintf_s(SubKeyPath, ARRAYSIZE(SubKeyPath), L"%s/%s/root", path, SubKeyName))
        {
            continue;
        }

        for (DWORD iData = 0; ;iData++)
        {
            METADATA_RECORD r;
            memset(&r, 0, sizeof(r));
            unsigned char temp[1024];
            r.pbMDData = temp;
            r.dwMDDataLen = sizeof(temp);
            DWORD dw;
            hr = pIMeta->EnumData(hKey, SubKeyPath, &r, iData, &dw);
            if ((HRESULT_FROM_WIN32(ERROR_NO_MORE_ITEMS) == hr) || (HRESULT_FROM_WIN32(ERROR_PATH_NOT_FOUND) == hr) || (iData > 1000))
            {
                break;
            }
            if (FAILED(hr))
            {
                continue;
            }
    
            if (MD_VR_PATH == r.dwMDIdentifier && STRING_METADATA == r.dwMDUserType)
            { // BUGBUG: is r.pbMDData guaranteed NULL-terminated ?
                pSites->AddWWWRoot((WCHAR*)r.pbMDData); // BUGBUG: ignore failures
            }
        }
    }

    fRet = TRUE;
Exit:
    //release the hKey
    if (hKey) pIMeta->CloseKey(hKey);
    if (pIMeta) pIMeta->Release();
    if (fCoInitialized) CoUninitialize();

    return fRet;
}

static BOOL EnumerateIIS7Sites(SiteTable* pSites)
{
    HRESULT hr;
    BOOL fRet = FALSE;
    BOOL fCoInitialized = FALSE;
	IAppHostAdminManager *      pAdminManager    = NULL;
    IAppHostElement *           pAdmin           = NULL;
    IAppHostElementCollection * pAdminCollection = NULL;
    BSTR bstrConfigPath       = NULL;
    BSTR bstrSitesSectionName = NULL;

    bstrConfigPath       = SysAllocString( L"MACHINE/WEBROOT/APPHOST" );       if (!bstrConfigPath) goto Exit;
    bstrSitesSectionName = SysAllocString( L"system.applicationHost/sites" );  if (!bstrSitesSectionName) goto Exit;

    CHECK_HR_EXIT(CoInitialize(NULL));  fCoInitialized = TRUE;
    CHECK_HR_EXIT(CoCreateInstance( CLSID_AppHostAdminManager, NULL, CLSCTX_INPROC_SERVER, IID_IAppHostAdminManager, (VOID **)&pAdminManager ));
    CHECK_HR_EXIT(pAdminManager->GetAdminSection( bstrSitesSectionName, bstrConfigPath, &pAdmin));
    CHECK_HR_EXIT(pAdmin->get_Collection(&pAdminCollection));
    DWORD     dwCount   = 0;
    CHECK_HR_EXIT(pAdminCollection->get_Count(&dwCount));
    
    fRet = TRUE;
    for(DWORD index = 0; index < dwCount; index ++)
    {
        IAppHostElement *           siteElement           = NULL;
        IAppHostProperty *          siteName              = NULL;
        IAppHostProperty *          siteId                = NULL;
        IAppHostElementCollection * pSiteCollection       = NULL;
        IAppHostElement *           rootApplication       = NULL;
        IAppHostElementCollection * vdirCollection        = NULL;
        IAppHostElement *           rootVdir              = NULL;
        IAppHostProperty *          sitePath              = NULL;
        VARIANT varPath;  VariantInit(&varPath);

        VARIANT varElementIndex; VariantInit(&varElementIndex); varElementIndex.vt = VT_I4; varElementIndex.lVal = index;
        CHECK_HR_LOOP(pAdminCollection->get_Item(varElementIndex, &siteElement));
        CHECK_HR_LOOP(siteElement->get_Collection(&pSiteCollection));
        VARIANT  var0; VariantInit(&var0);  var0.vt = VT_I2; var0.lVal = 0;
        CHECK_HR_LOOP(pSiteCollection->get_Item(var0, &rootApplication));
        CHECK_HR_LOOP(rootApplication->get_Collection(&vdirCollection));
        CHECK_HR_LOOP(vdirCollection->get_Item(var0, &rootVdir));
        CHECK_HR_LOOP(rootVdir->GetPropertyByName(L"physicalPath", &sitePath));
        CHECK_HR_LOOP(sitePath->get_Value(&varPath)); 
        
        pSites->AddWWWRoot(varPath.bstrVal); // BUGBUG: ignore failures

    Loop:
        VariantClear(&varPath);
        RELEASE(siteElement);
        RELEASE(siteName);
        RELEASE(siteId);
        RELEASE(pSiteCollection);
        RELEASE(rootApplication);
        RELEASE(vdirCollection);
        RELEASE(rootVdir);
        RELEASE(sitePath);
    }

Exit:
    RELEASE(pAdminCollection);
    RELEASE(pAdmin);
    RELEASE(pAdminManager);
     
    if (bstrConfigPath) SysFreeString(bstrConfigPath);
    if (bstrSitesSectionName) SysFreeString(bstrSitesSectionName);
 
    if (fCoInitialized) CoUninitialize();

    return fRet;
}

static BOOL GetSitesKnownToIIS(SiteTable* pSites)
{
    unsigned ver = IISVersion();
    if (ver >= 7)
    {
        return EnumerateIIS7Sites(pSites);
    }
    else if (ver == 6)
    {
        return EnumerateIIS6Sites(pSites);
    }
    else
    {
        SERVICE_LOG("unsupported IIS version %u\n", ver);
        return FALSE;
    }
}

static void EnumerateAndAddFolders(const TCHAR* wszPrefix, SiteTable* pSites)
{
    for (unsigned u = 0; ;u++)
    {
        TCHAR wszFolder[MAX_PATH];
        if (!GetDataFolderWithIndex(wszPrefix, u, wszFolder, ARRAYSIZE(wszFolder)))
        {
            return;
        }
        if (!PALDirectory::PathExists(wszFolder))
        {
            break; // pluginlib.cpp doesn't create N+1 unless N already exists
        }

        pSites->AddDataFolder(wszFolder);
    }
}

void DoWork(PALEvent *pStopEvent)
{
    HANDLE hFind = NULL;
    HKEY hKey = NULL;
    SiteTable sites;
    if (!sites.Initialize())
    {
        return;
    }

    TCHAR wszFolder[MAX_PATH];
    if (!GetDataFolder(wszFolder, ARRAYSIZE(wszFolder)))
    {
        return;
    }

    DWORD dw = 0;
    if (PluginRegQueryDWORD(NULL, TEXT("ServiceLogLevel"), &dw))
    {
        g_fLogLevel = dw;
    }

    TCHAR wszLog[MAX_PATH];
    if (0 >= swprintf_s(wszLog, ARRAYSIZE(wszLog), TEXT("%s\\ServiceLog.txt"), wszFolder))
    {
        return;
    }

    if (!OpenMergeServiceLog(wszLog, &g_fpGlobalLog))
    {
        return;
    }

    if (!GetSitesKnownToIIS(&sites))
    {
        SERVICE_LOG("cannot enumerate IIS sites\n");
        return;
    }

    if (ERROR_SUCCESS == RegOpenKeyEx(SSP_REGISTRY_HIVE, SSP_REGISTRY_PATH, 0, KEY_ENUMERATE_SUB_KEYS, &hKey))
    {
        for (DWORD dwIndex = 0; ;dwIndex++)
        {
            TCHAR wszName[MAX_PATH];
            LONG lRet = RegEnumKey(hKey, dwIndex, wszName, ARRAYSIZE(wszName));
            if (ERROR_MORE_DATA == lRet) // key name too long
            {
                continue;
            }
            if (ERROR_SUCCESS != lRet)
            {
                break;
            }

            TCHAR wszFolder[MAX_PATH];
            if (!PluginRegQueryValue(wszName, SSP_REGISTRY_FOLDERPATH, wszFolder, ARRAYSIZE(wszFolder)))
            {
                continue;
            }

            EnumerateAndAddFolders(wszFolder, &sites);
        }

        RegCloseKey(hKey);  hKey = NULL;
    }
    else
    {
        SERVICE_LOG("no keys under " STRFMT_TS "\n", SSP_REGISTRY_PATH);
    }

    
    TCHAR wszSearch[MAX_PATH];
    if (0 >= swprintf_s(wszSearch, ARRAYSIZE(wszSearch), TEXT("%s\\site_*"), wszFolder))
    {
        return;
    }

    WIN32_FIND_DATA finddata;
    hFind = FindFirstFile(wszSearch, &finddata);
    if (hFind == INVALID_HANDLE_VALUE)
    {
        SERVICE_LOG("no files matching " STRFMT_TS "\n", wszSearch);
        goto Exit;
    }
    do
    {
        if (!(FILE_ATTRIBUTE_DIRECTORY & finddata.dwFileAttributes))
        {
            continue;
        }

        TCHAR wszSubfolder[MAX_PATH];
        if (0 >= swprintf_s(wszSubfolder, ARRAYSIZE(wszSubfolder), TEXT("%s\\%s"), wszFolder, finddata.cFileName))
        {
            continue;
        }

        sites.AddDataFolder(wszSubfolder);
    } while (FindNextFile(hFind, &finddata));

    sites.Finalize();
    sites.Process(pStopEvent);

Exit:
    if (hFind != INVALID_HANDLE_VALUE) FindClose(hFind);
    if (hKey) RegCloseKey(hKey);
    CloseMergeServiceLog(g_fpGlobalLog);
}

