/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PALMemory.h"
#include "PALString.h"
#include "shared.h"
#include "diskio.h"
#include "CompressingWriter.h"

/*************************************************************************
**  Method: CompressingFileWriter::CompressingFileWriter
**
**  Description: Constructs a CompressingFileWriter object
**
**************************************************************************/
void CompressingFileWriter::Construct()
{
    m_fCompression = false;
    m_compressedBuffer = NULL;
}

CompressingFileWriter::CompressingFileWriter()
{
    m_pOwnMemory = NULL;
    Construct();
}

CompressingFileWriter::~CompressingFileWriter()
{
    DoDispose();
    if (m_pOwnMemory) free(m_pOwnMemory); // don't do this in DoDispose() because m_pOwnMemory might be uninitialized
}

void CompressingFileWriter::ResetCompression()
{
    if (m_fInitialized && m_fCompression)
    {
        deflateReset(&m_strmCompress);
        ResetCompressedInput();
        ResetCompressedOutput();
    }
}

void CompressingFileWriter::ResetCompressedOutput()
{
    m_strmCompress.next_out = m_compressedBuffer;
    m_strmCompress.avail_out = c_compressedBufferSize;
}

void CompressingFileWriter::ResetCompressedInput()
{
    m_strmCompress.next_in = (Bytef *)NULL;
    m_strmCompress.avail_in = (UInt32)0;
}

DWORD32 CompressingFileWriter::Close( void )
{
    DWORD32 dwRet = NO_ERROR;
    if ( m_fInitialized && m_fOpened )
    {
        if (m_fCompression)
        {
            m_strmCompress.avail_in = 0;

            UInt32 len;
            int err;
            bool done = false;
            for ( ; ; ) 
            {
                len = c_compressedBufferSize - m_strmCompress.avail_out;

                if (len != 0) 
                {
                    DWORD32 ec = m_seqwrite.Write(m_compressedBuffer, len);
                    if (ec != NO_ERROR)
                    {
                        Log(LogID_Common, LogLevel_Error, "CompressingFileWriter::~CompressingFileWriter",
                            "Write failed with ec = %u",
                            ec);
                        dwRet = ec;
                        break;
                    }
                    
                    ResetCompressedOutput();
                }
                if (done) 
                {
                    break;
                }
                err = deflate(&m_strmCompress, Z_FINISH);

                /* Ignore the second of two consecutive flushes: */
                if (len == 0 && err == Z_BUF_ERROR) 
                {
                    err = Z_OK;
                }

                /* deflate has finished flushing only when it hasn't used up
                * all the available space in the output buffer: 
                */
                done = (m_strmCompress.avail_out != 0 || err == Z_STREAM_END);
            
                if (err != Z_OK && err != Z_STREAM_END) 
                {
                    break;
                }
            }


        }
        dwRet = m_seqwrite.Close();
        m_fOpened = false;
    }
    return dwRet;
}

DWORD32 CompressingFileWriter::DoDispose()
{
    DWORD32 dwRet = NO_ERROR;

    if (m_fInitialized)
    {
        if (m_fOpened)
        {
            dwRet = Close(); // flush
        }

        if (m_fCompression)
        {
            int err = deflateEnd(&m_strmCompress); // no-op - we're using arena memory and all it does is free some blocks
            if (err != Z_OK)
            {
                Log(LogID_Common, LogLevel_Error, "CompressingFileWriter::~CompressingFileWriter",
                    "deflateEnd failed with err = %d",
                    err);
            }
            
            // m_compressedBuffer it's arena-allocated from parent object, so no need to free
            m_compressedBuffer = NULL;
        }

        m_seqwrite.DoDispose();

        m_fInitialized = false;
    }
    return dwRet;
}

/*************************************************************************
**  Method: CompressingFileWriter::InitializeCompression
**
**  Description: Initializes the compression subobjects.  Allocate memory, 
**      and initialize the z_stream
**
**  Parameters:
**      compressionLevel        desired compression level 
**                              default is Z_DEFAULT_COMPRESSION which is -1 but maps to level 6
**                              level 0 is no compression
**                              level 1 is the maximum speed option for compression
**                              level 9 provides the best compression ratio
**
**  Returns:
**      NO_ERROR=0      
**      ERROR_OUTOFMEMORY 
**      ERROR_ENCRYPTION_FAILED
**
**  Notes:
**
**************************************************************************/

static int DeflateInitWrap(z_stream* pstream, BOOL fGZip)
{
    return deflateInit2(pstream, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15 + (fGZip ? 16 : 0), 8, Z_DEFAULT_STRATEGY);
}

DWORD32
CompressingFileWriter::InitializeCompression(IAllocator* pAlloc, BOOL fGZip)
{
    m_fCompression = true;

    // setup buffer for compressed data
    // (reuse of this instance is not supported)
    LogAssert(m_compressedBuffer == NULL);
    m_compressedBuffer = (UInt8 *)pAlloc->Allocate(c_compressedBufferSize);
    if (!m_compressedBuffer)
    {
        return ERROR_OUTOFMEMORY;
    }

    // initialize compression object
    memset(&m_strmCompress, 0, sizeof(m_strmCompress));
    m_strmCompress.zalloc = zlib_alloc;
    m_strmCompress.zfree = zlib_free;
    m_strmCompress.opaque = pAlloc;

    ResetCompressedInput();
    ResetCompressedOutput();

    if (Z_OK != DeflateInitWrap(&m_strmCompress, fGZip))
    {
        return ERROR_ENCRYPTION_FAILED;
    }
    
    m_strmCompress.opaque = NULL; // zlib doesn't appear to allocate memory after initialization
    return NO_ERROR;
}

/*************************************************************************
**  Method: CompressingFileWriter::DoInit
**
**  Description: Initializes the object.  Allocate memory, open file and 
**      prepare for writing.
**
**  Parameters:
**      szFile          name of file to open
**      cbWrite         size of each write buffer
**      maxWrites       max number of writes
**
**  Returns:
**      NO_ERROR=0      
**      ERROR_INVAID_PARAMTER 
**      ERROR_OUTOFMEMORY
**
**  Notes:
**      File is always created and will overwrite existing files.
**
**************************************************************************/
DWORD32 
CompressingFileWriter::DoInit(
                 BOOL fCompression,
                 BOOL fGZip,
                 IAllocator* pAlloc,
                 int         cbWrite,
                 int         maxWrites)
{
    Construct();
    
    DWORD32 ec = NO_ERROR;
    ArenaAllocator a;
    if (!pAlloc)
    {
        size_t cb = MemoryNeeded(fCompression) + c_compressedBufferSize;
        void* pMem = malloc(cb);
        if (!pMem)
        {
            ec = ERROR_OUTOFMEMORY;
            goto Exit;
        }
        m_pOwnMemory = pMem;
        a.Initialize(pMem, cb);
        pAlloc = &a;

        // Don't pass the arena allocator because it will allocate
        // additional memory and then will not be enough for zlib.
        ec = m_seqwrite.DoInit(NULL, cbWrite, maxWrites);
    }
    else
    {
        ec = m_seqwrite.DoInit(pAlloc, cbWrite, maxWrites);
    }

    if ( ec != NO_ERROR )
    {
        goto Exit;
    }

    if (fCompression)
    {
        ec = InitializeCompression(pAlloc, fGZip);
        if ( ec != NO_ERROR )
        {
            goto Exit;
        }
    }

    m_fOpened = false;
    m_fInitialized = true;
Exit:
    return( ec );
}

DWORD32
CompressingFileWriter::Open(const TCHAR* wszFile)
{
    Close();
    ResetCompression();

    DWORD32 ec = m_seqwrite.Open(wszFile);
    if (NO_ERROR != ec)
    {
        return ec;
    }
    m_fOpened = true;
    return NO_ERROR;
}


typedef struct 
{
    void* pb;
    size_t cbSize;
    size_t cbUsed;
} DummyAllocator;

static void *dummy_alloc(void *opaque, UINT items, UINT size)
{
    DummyAllocator* p = (DummyAllocator*)opaque;
    size_t cb = items * size;
    if (p->cbUsed + cb > p->cbSize)
    {
        return NULL;
    }
    void* pvRet = (char*)(p->pb) + p->cbUsed;
    p->cbUsed += cb;
    return pvRet;
}

size_t CompressingFileWriter::DeflateMemorySize()
{
    size_t cbRet = 0;

    DummyAllocator a;
    a.cbSize = 1048576;
    a.cbUsed = 0;
    a.pb = malloc(a.cbSize);
    if (!a.pb)
    {
        goto Exit;
    }

    z_stream stream;
    memset(&stream, 0, sizeof(stream));
    stream.zalloc = dummy_alloc;
    stream.opaque = &a;

    if (Z_OK != DeflateInitWrap(&stream, FALSE)) // fGZip doesn't seem to make a difference for memory size
    {
        goto Exit;
    }
    // assuming there is no point calling deflateEnd()

    cbRet = a.cbUsed;
Exit:
    if (a.pb)
    {
        free(a.pb);
    }
    return cbRet;
}

size_t CompressingFileWriter::MemoryNeeded(BOOL fCompression)
{
    if (fCompression)
    {
        return DeflateMemorySize();
    }
    else
    {
        return 1; // Windows uses VirtualAlloc, but TODO for Apache implement buffering for the non-compressed case
    }
}

DWORD32 CompressingFileWriter::InitAndOpen(const TCHAR* wszFile, BOOL fCompression, BOOL fGZip)
{
    DWORD32 dw;
    dw = DoInit(fCompression, fGZip);
    if (NO_ERROR != dw)
    {
        return dw;
    }

    return Open(wszFile);
}

CompressingFileWriter* CompressingFileWriter::Create(const TCHAR* wszFile, BOOL fCompression, BOOL fGZip)
{
    CompressingFileWriter *pWriter = new CompressingFileWriter;
    if (!pWriter)
    {
        return NULL;
    }

    if (NO_ERROR != pWriter->InitAndOpen(wszFile, fCompression, fGZip))
    {
        delete pWriter;
        return NULL;
    }

    return pWriter;
}

/*************************************************************************
**  Method: CompressingFileWriter::Write
**
**  Description: Calls WriteInternal for the right size buffers.
**
**  Parameters:
**      pbWrite         data to write
**      cbWrite         number of bytes of data to write
**
**  Returns:
**      NO_ERROR=0      
**      ERROR_ENCRYPTION_FAILED
**      GetLastError()  if read fails
**
**  Notes:
**
**************************************************************************/
DWORD32 
CompressingFileWriter::Write(const void* pbWrite, DWORD cbWrite)
{
    DWORD ec = NO_ERROR;
    if (!m_fCompression)
    {
        ec = m_seqwrite.Write(pbWrite, cbWrite);
    }
    else
    {
        // compress and then pass to m_seqwrite
        int zlibError = Z_OK;
        m_strmCompress.next_in = (Bytef *)pbWrite;
        m_strmCompress.avail_in = cbWrite;

        while (m_strmCompress.avail_in != 0) 
        {
            if (m_strmCompress.avail_out == 0) 
            {
                m_strmCompress.next_out = m_compressedBuffer;
                ec = m_seqwrite.Write(m_compressedBuffer, c_compressedBufferSize);
                if (ec != NO_ERROR)
                {
                    return ec;
                }
                m_strmCompress.avail_out = c_compressedBufferSize;
            }
            zlibError = deflate(&m_strmCompress, Z_NO_FLUSH);
            if (zlibError != Z_OK) 
            {
                break;
            }
        }

        if (zlibError != Z_OK)
        {
            return ERROR_ENCRYPTION_FAILED;
        }
    }
    return ec;
}


/*************************************************************************
**  Method: CompressingFileWriter::Print
**
**  Description: Print formatted data into an internal buffer and then
**      write the data.
**
**  Parameters:
**      szFormat        format string
**      ...             params for print statement
**
**  Returns:
**      NO_ERROR=0 
**      ERROR_ENCRYPTION_FAILED
**      GetLastError()  if read fails
**
**  Notes:
**
**************************************************************************/
DWORD32
CompressingFileWriter::Print(const char* szFormat, ...)
{
    char    rgch[SEQWRITE::c_cbWriteMax + 1];
    int cbWrite;
    va_list args;
    va_start(args, szFormat);

    cbWrite = VSNPRINTF_S( rgch, _countof(rgch), _TRUNCATE, szFormat, args);

    if (cbWrite < 0)
    {
        return ERROR_INVALID_PARAMETER;
    }

    rgch[cbWrite] = '\0';

    return Write(rgch, cbWrite);
}

void *
CompressingFileWriter::zlib_alloc(void *opaque, UINT size, UINT count)
{
    if (!opaque) return NULL;
    //printf("zlib_alloc(%u)\n", size * count);
    return ((ArenaAllocator*)opaque)->Allocate(size * count);
}

void
CompressingFileWriter::zlib_free(void *opaque, void *data)
{ // arena allocator - nothing to do
    UNREFERENCED_PARAMETER(opaque);
    UNREFERENCED_PARAMETER(data);
}

DWORD32
CompressingFileWriter::CompressFile(const TCHAR *pszUncompressed, const TCHAR *pszCompressed)
{
    const char *pszLogTitle = "CompressingFileWriter::CompressFile";

    SEQREAD reader;
    CompressingFileWriter writer;

    DWORD32 dwRet = reader.DoInit(pszUncompressed);
    if (dwRet != NO_ERROR)
    {
        Log(LogID_Common, LogLevel_Error, pszLogTitle,
            "Failed to open uncompressed file %s with error %u", pszUncompressed, dwRet);
        return dwRet;
    }

    UInt64 inputFileRemaining = reader.FileSize();

    dwRet = writer.InitAndOpen(pszCompressed);
    if (dwRet != NO_ERROR)
    {
        Log(LogID_Common, LogLevel_Error, pszLogTitle,
            "Failed to open output file %s with error %u", pszCompressed, dwRet);
        return dwRet;
    }

    UInt8 *pBuffer;
    while (inputFileRemaining > 0)
    {
        UInt32 cbRequestSize = (UInt32)inputFileRemaining;
        if (cbRequestSize > SEQREAD::c_readBufSize)
            cbRequestSize = SEQREAD::c_readBufSize;
        dwRet = reader.GetDataPointer((void **)&pBuffer, cbRequestSize);
        if (dwRet != NO_ERROR)
        {
            Log(LogID_Common, LogLevel_Error, pszLogTitle,
                    "GetDataPointer failed for %s with error %u", pszUncompressed, dwRet);
            return dwRet;
        }

        inputFileRemaining -= cbRequestSize;
        dwRet = writer.Write(pBuffer, cbRequestSize);
        if (dwRet != NO_ERROR)
        {
            Log(LogID_Common, LogLevel_Error, pszLogTitle,
                    "Write to %s failed with error %u", pszCompressed, dwRet);
            return dwRet;
        }
    }

    return NO_ERROR;
}
