/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "CritSec.h"
#include "PALString.h"

#ifdef __TARGET_IIS__
CRITSEC::CRITSEC() : m_hMutex(NULL), m_fCritSecInitialized(FALSE)
{
}

CRITSEC::~CRITSEC()
{
}

void CRITSEC::DoInit(const TCHAR *tszName, void *pPlatformMemPool, BOOL fWorkAcrossProcesses)
{
    UNREFERENCED_PARAMETER(pPlatformMemPool);

    m_hMutex = NULL;
    if (fWorkAcrossProcesses)
    {   // "CreateMutex opens a mutex if it exists and creates it if it does not" (http://msdn.microsoft.com/en-us/library/windows/desktop/ms684315(v=vs.85).aspx)
        m_hMutex = CreateMutex(NULL, FALSE, tszName);
    }

    if (!m_hMutex)
    { // can't get cross-process synchronization - fall back on in-process
        InitializeCriticalSection(&m_critsec);
        m_fCritSecInitialized = TRUE;
    }
}

void CRITSEC::DoInitChild(const TCHAR * tszName, void *pPlatformMemPool)
{
    UNREFERENCED_PARAMETER(tszName);
    UNREFERENCED_PARAMETER(pPlatformMemPool);
    // Empty functions to mimic Apache equivalent.
}

void CRITSEC::DoDispose()
{
    // http://msdn.microsoft.com/en-us/library/windows/desktop/ms682411(v=vs.85).aspx says
    // "The system closes the handle automatically when the process terminates"
    // So I suppose calling CloseHandle(m_hMutex) here would be small risk with 0 benefit
    if (m_fCritSecInitialized)
    {
        DeleteCriticalSection(&m_critsec);
        m_fCritSecInitialized = FALSE;
    }
}

void CRITSEC::Enter()
{
    if (m_hMutex)
    {
        WaitForSingleObject(m_hMutex, INFINITE);
    }
    else
    {
        EnterCriticalSection(&m_critsec);
    }
}

BOOL CRITSEC::TryEnter()
{
    if (m_hMutex)
    {
        return WAIT_OBJECT_0 == WaitForSingleObject(m_hMutex, 0);
    }
    else
    {
        return TryEnterCriticalSection(&m_critsec);
    }

}

void CRITSEC::Leave()
{
    if (m_hMutex)
    {
        ReleaseMutex(m_hMutex);
    }
    else
    {
        LeaveCriticalSection(&m_critsec);
    }
}

BOOL CRITSEC::WorkAcrossProcesses()
{
    return NULL != m_hMutex;
}

#else

CRITSEC::CRITSEC() :
    m_pCritsec(NULL),
    m_pPool(NULL)
{
}

CRITSEC::~CRITSEC()
{
}

const char* CRITSEC::MultibyteName(const TCHAR* tszName, char* szTemp, Size_t cchTemp)
{
#ifdef UNICODE
    if (PALString::Convert(szTemp, cchTemp, tszName, TSTRLEN(tszName)) != 0)
    {
        throw APR_EGENERAL;
    }
    return szTemp;
#else
    return tszName;
#endif
}

void CRITSEC::DoInit(const TCHAR *tszName, void *pPlatformMemPool, BOOL fWorkAcrossProcesses)
{
    if (m_pCritsec == NULL)
    {
        m_pPool = (apr_pool_t *) pPlatformMemPool;

        char szTemp[MAX_PATH];
        apr_status_t status = apr_proc_mutex_create(
            &m_pCritsec,
            MultibyteName(tszName, szTemp, ARRAYSIZE(szTemp)),
            APR_LOCK_DEFAULT,
            m_pPool);

        if (status != APR_SUCCESS)
        {
            DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to create process mutex");
            throw status;
        }

#ifdef AP_NEED_SET_MUTEX_PERMS
        // Must set permissions for certain systems.
        status = unixd_set_proc_mutex_perms(m_pCritsec);
        if (status != APR_SUCCESS)
        {
            DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to set process mutex permission");
            throw status;
        }
#endif
    }
    else
    {
        throw APR_EGENERAL;
    }
}

void CRITSEC::DoInitChild(const TCHAR * tszName, void *pPlatformMemPool)
{
    m_pPool = (apr_pool_t *) pPlatformMemPool;
    
    char szTemp[MAX_PATH];
    apr_status_t status = apr_proc_mutex_child_init(
        &m_pCritsec,
        MultibyteName(tszName, szTemp, ARRAYSIZE(szTemp)),
        m_pPool);

    if (status != APR_SUCCESS)
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to initialize process mutex in child");
        throw status;
    }
}

void CRITSEC::DoDispose()
{
    if (m_pCritsec != NULL)
    {
        apr_status_t status = apr_proc_mutex_destroy(m_pCritsec);

        if (status != APR_SUCCESS)
        {
            DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to destroy process mutex");
            throw status;
        }
        
        m_pCritsec = NULL;
        m_pPool = NULL;
    }
}

void CRITSEC::Enter()
{
    apr_status_t status = apr_proc_mutex_lock(m_pCritsec);

    if (status != APR_SUCCESS)
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Process mutex lock failed");
        throw status;
    }
}

BOOL CRITSEC::TryEnter()
{
    apr_status_t status = apr_proc_mutex_trylock(m_pCritsec);

    if (status != APR_SUCCESS && !APR_STATUS_IS_EBUSY(status))
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Process mutex trylock failed");
        throw status;
    }

    return !APR_STATUS_IS_EBUSY(status);
}

void CRITSEC::Leave()
{
    apr_status_t status = apr_proc_mutex_unlock(m_pCritsec);

    if (status != APR_SUCCESS)
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Process mutex unlock failed");
        throw status;
    }
}

BOOL CRITSEC::WorkAcrossProcesses()
{
    return TRUE;
}



CRITSECTHREAD::CRITSECTHREAD()
    : m_pCritsec(NULL), m_pPool(NULL)
{
}

CRITSECTHREAD::~CRITSECTHREAD()
{
}

void CRITSECTHREAD::DoInit(const char *szName, void *pPlatformMemPool)
{
    UNREFERENCED_PARAMETER(szName);

    if (m_pCritsec == NULL)
    {
        m_pPool = (apr_pool_t *) pPlatformMemPool;

        apr_status_t status = apr_thread_mutex_create(
            &m_pCritsec,
            APR_THREAD_MUTEX_DEFAULT,
            m_pPool);

        if (status != APR_SUCCESS)
        {
            throw status;
        }
    }
}

void CRITSECTHREAD::DoDispose()
{
    if (m_pCritsec != NULL)
    {
        apr_status_t status = apr_thread_mutex_destroy(m_pCritsec);

        if (status != APR_SUCCESS)
        {
            throw status;
        }

        m_pCritsec = NULL;
        m_pPool = NULL;
    }
}

void CRITSECTHREAD::Enter()
{
    apr_status_t status = apr_thread_mutex_lock(m_pCritsec);

    if (status != APR_SUCCESS)
    {
        throw status;
    }
}

BOOL CRITSECTHREAD::TryEnter()
{
    apr_status_t status = apr_thread_mutex_trylock(m_pCritsec);

    if (status != APR_SUCCESS && !APR_STATUS_IS_EBUSY(status))
    {
        throw status;
    }

    return !APR_STATUS_IS_EBUSY(status);
}

void CRITSECTHREAD::Leave()
{
    apr_status_t status = apr_thread_mutex_unlock(m_pCritsec);

    if (status != APR_SUCCESS)
    {
        throw status;
    }
}
#endif
