/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#pragma once
#include "PAL.h"
#include "basic_types.h"

class ICriticalSection
{
public:
    virtual void Enter() = 0;
    virtual BOOL TryEnter() = 0;
    virtual void Leave() = 0;
};

class ICritSecDispenser
{
public:
    virtual int GetCS() = 0;
};

//This is an interface to disallow heap construction.
class INoHeapInstance
{
private:
    void* operator new(Size_t);
};

//Smart wrapper around CRITSEC so that we dont need to call enter/leave
class AutoCriticalSection : public INoHeapInstance
{
private:
    ICriticalSection *m_pCritSec;
public:
    AutoCriticalSection(ICriticalSection *pCritSec) : m_pCritSec(pCritSec)
    {
        this->m_pCritSec->Enter();
    }
    ~AutoCriticalSection()
    {
        this->m_pCritSec->Leave();
    }
};


class CRITSEC : public ICriticalSection
{
public:
    CRITSEC();

    virtual ~CRITSEC();

    void DoInit(const TCHAR *tszName, void *pPlatformMemPool, BOOL fWorkAcrossProcesses);

    void DoInitChild(const TCHAR * tszName, void *pPlatformMemPool);

    void DoDispose();

    void Enter();

    BOOL TryEnter();

    void Leave();

    BOOL WorkAcrossProcesses();

private:
#if defined ( DISABLE_COPY_OF_CRIT_SECTION ) 

    // if you see "cannot access private member declared in class 'CRITSEC'" error this mean that compiler is trying to generate copy constructor. 
    // Please, check the code which is trying to copy CRITSEC (or copy some class which have CRITSEC member) and fix it.

    // We don't need a copy constructor - you cannot copy CRITICAL_SECTION struct
    CRITSEC(const CRITSEC&);
    // We don't need an assignment operator - you cannot copy CRITICAL_SECTION struct
    CRITSEC& operator=(const CRITSEC&);
#endif

#ifdef __TARGET_IIS__
    HANDLE m_hMutex;
    CRITICAL_SECTION m_critsec; // used only if m_hMutex is NULL
    BOOL m_fCritSecInitialized;
#else
    apr_proc_mutex_t   *m_pCritsec;
    apr_pool_t         *m_pPool;

    const char* MultibyteName(const TCHAR* tszName, char* szTemp, Size_t cchTemp);
#endif
};


#ifdef __TARGET_APACHE__
class CRITSECTHREAD : public ICriticalSection
{
public:
    CRITSECTHREAD();

    virtual ~CRITSECTHREAD();

    void DoInit(const char *szName, void *pPlatformMemPool);

    void DoInitChild(const char * szName, void *pPlatformMemPool);

    void DoDispose();

    void Enter();

    BOOL TryEnter();

    void Leave();

private:
#if defined ( DISABLE_COPY_OF_CRIT_SECTION ) 

    // if you see "cannot access private member declared in class 'CRITSEC'" error this mean that compiler is trying to generate copy constructor. 
    // Please, check the code which is trying to copy CRITSEC (or copy some class which have CRITSEC member) and fix it.

    // We don't need a copy constructor - you cannot copy CRITICAL_SECTION struct
    CRITSEC(const CRITSEC&);
    // We don't need an assignment operator - you cannot copy CRITICAL_SECTION struct
    CRITSEC& operator=(const CRITSEC&);
#endif

    apr_thread_mutex_t *m_pCritsec;
    apr_pool_t         *m_pPool;
};
#endif
