/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "DiskSpaceTracker.h"
#include "SSPCommon.h"
#include "PALString.h"
#include "PALFile.h"

DiskSpaceTracker::DiskSpaceTracker()
{
    Construct();
}

DiskSpaceTracker::~DiskSpaceTracker()
{
    DoDispose();
}

void DiskSpaceTracker::Construct()
{
    m_uCheckPeriodSec = 0;
    m_uMinMB = 0;
    m_uMaxMB = 0;
    m_tLastCheck = 0;
    m_fHappy = 0;
    m_ulClusterSize = 1;
    m_wszFolder[0] = '\0';
#if !APR_HAVE_WINDOWS_H
    m_szFolder[0] = '\0';
#endif
}

BOOL DiskSpaceTracker::DoInit(const TCHAR* wszFolder, unsigned uCheckPeriodSec, unsigned uMinMB, unsigned uMaxMB)
{
    Construct();

    Size_t cchFolder = TSTRLEN(wszFolder);
    if (cchFolder >= ARRAYSIZE(m_wszFolder))
    {
        return FALSE;
    }
    memcpy(m_wszFolder, wszFolder, (cchFolder + 1) * sizeof(TCHAR));

    m_uCheckPeriodSec = uCheckPeriodSec;
    m_uMinMB = uMinMB;
    m_uMaxMB = uMaxMB;

#if APR_HAVE_WINDOWS_H
    {
        unsigned long BytesPerSector, SectorsPerCluster;
        if (GetDiskFreeSpace(wszFolder, &SectorsPerCluster, &BytesPerSector, NULL, NULL))
        {
            m_ulClusterSize = SectorsPerCluster * BytesPerSector;
        }
    }
#else
    {
        struct statfs64 filestat;

        PALString::Convert(m_szFolder, ARRAYSIZE(m_szFolder), wszFolder, TSTRLEN(wszFolder));

        if (statfs64(m_szFolder, &filestat) != 0)
        {
            return 0;
        }

        m_ulClusterSize = filestat.f_bsize;
    }
#endif

    return TRUE;
}

BOOL DiskSpaceTracker::DoDispose()
{
    return TRUE;
}

BOOL DiskSpaceTracker::CheckSpace()
{
    time_t tNow = PluginTime();

    if (tNow > m_tLastCheck + (time_t)m_uCheckPeriodSec)
    { // probably some race condition possible
        m_tLastCheck = tNow;
        m_fHappy = IsHappy();
    }
    return m_fHappy;
}

ULONGLONG DiskSpaceTracker::BytesAllocated(ULONGLONG used) const
{
  return ((used + m_ulClusterSize - 1) / m_ulClusterSize) * m_ulClusterSize;
}

BOOL DiskSpaceTracker::IsHappy() const
{
    UInt64 u64FreeSpace = 0;

#if APR_HAVE_WINDOWS_H
    {
        ULARGE_INTEGER uli;
        if (!GetDiskFreeSpaceEx(m_wszFolder, &uli, NULL, NULL))
        {
            return FALSE;
        }

        u64FreeSpace = uli.QuadPart;
    }
#else
    {
        struct statfs64 filestat;

        if (statfs64(m_szFolder, &filestat) != 0)
        {
            return FALSE;
        }

        u64FreeSpace = filestat.f_bavail * filestat.f_bsize;
    }
#endif

    if (u64FreeSpace > 1048576 * (ULONGLONG)m_uMaxMB)
    {
        return TRUE;
    }
    if (u64FreeSpace < 1048576 * (ULONGLONG)m_uMinMB)
    {
        return FALSE;
    }

    ULONGLONG cbUsed = 0;
    if (!CalcSpaceUsed(m_wszFolder, &cbUsed))
    {
        return FALSE;
    }
    return u64FreeSpace > cbUsed; // somewhat arbitrary - allow ourselves to use up to 1/2 of the space that would be available if we weren't using it
}

BOOL DiskSpaceTracker::RecurseSpaceUsed(__inout_ecount(cchPathBuf) TCHAR *rwchPath, Size_t cchCurPath, const Size_t cchPathBuf, unsigned depth, ULONGLONG* pcchTotal) const
{
    BOOL fOK = TRUE;
    PALFind findFile;
    PALFind::EStatus findStatus;

    // rwchPath is modified inside the function but restored before it returns
    if (depth > 10)
    {
        return FALSE;
    }

    *pcchTotal += m_ulClusterSize; // assume directory uses 1 cluster

    if (cchCurPath + 1 > cchPathBuf)
    {
        return FALSE;
    }
    rwchPath[cchCurPath] = '\0';
    
    if (findFile.Initialize(rwchPath, TEXT("*")) == FALSE)
    {
        return FALSE;
    }

    findStatus = findFile.Find();

    if (findStatus == PALFind::Error_FileNotFound || findStatus == PALFind::Error_DirectoryNotFound)
    {
        return TRUE;
    }

    if (findStatus != PALFind::Success)
    {
        return FALSE;
    }

    do
    {
        if (!PALString::Compare(findFile.GetName(), TEXT(".")) ||
            !PALString::Compare(findFile.GetName(), TEXT("..")))
        {
            continue;
        }

        if (findFile.GetAttrib() == PALFind::Directory)
        {
            Size_t cchFile = TSTRLEN(findFile.GetName());
            if (cchCurPath + cchFile + 1 >= cchPathBuf)
            {
                fOK = FALSE;
                break;
            }
            memcpy(rwchPath + cchCurPath, findFile.GetName(), cchFile * sizeof(TCHAR));
            rwchPath[cchCurPath + cchFile] = CHAR_DIRSLASH;
            fOK = RecurseSpaceUsed(rwchPath, cchCurPath + cchFile + 1, cchPathBuf, depth + 1, pcchTotal);
        }
        else
        {
            *pcchTotal += BytesAllocated(findFile.GetSize());
        }
    }
    while (fOK && findFile.Find() == PALFind::Success);

    return fOK;
}

BOOL DiskSpaceTracker::CalcSpaceUsed(const TCHAR* wszFolder, ULONGLONG* pcch) const
{
    TCHAR rwchPath[MAX_PATH];

    Size_t cchFolder = TSTRLEN(wszFolder);
    if (cchFolder + 2 > ARRAYSIZE(rwchPath))
    {
        return FALSE;
    }
    memcpy(rwchPath, wszFolder, cchFolder * sizeof(TCHAR));
    if (rwchPath[cchFolder - 1] != CHAR_DIRSLASH)
    {
        rwchPath[cchFolder] = CHAR_DIRSLASH;
        cchFolder++;
    }

    *pcch = 0;
    return RecurseSpaceUsed(rwchPath, cchFolder, ARRAYSIZE(rwchPath), 0, pcch);
}
