/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "DocSignature.h"
#include "DocSignatureUtil.h"
#include "SSPHash.h"
#include "DocSignatureHash.h"

Signature ComputeSignature(const char* pbDoc, Size_t cbDoc)
{
    const char* pbEnd           = pbDoc + cbDoc;
    const char* token           = pbDoc;

    bool        fValidDoc       = false;
    Size_t      bodyLocation    = 0;
    Size_t      position        = 0;
    Size_t      charScanned     = 0;
    Size_t      charIgnored     = 0;
    Size_t      tokenSize       = 0;


    // add token count and anchor count in signature to avoid the case
    // tokens/anchors are added/removed but signatures are accidentally same
    Size_t      tokenCount      = 0;
    Size_t      anchorCount     = 0;

    DocSignatureHash docHash;

    // First, find where body starts
    // <body> is required tag, if body doesn't exist, return 0 as Signature
    for (; bodyLocation < cbDoc; bodyLocation++)
    {
        if (IsMatch(pbDoc+bodyLocation, pbEnd, TAG_BODY_OPEN))
        {
            fValidDoc = true;
            break;
        }
    }

    if (!fValidDoc)
    {
        return 0;
    }

    // 1. handle title tag, it will be in <head> part
    // NOTE: if there are comments in title tag, we will just take it. This case unlikely happens.
    // TODO: Handle the comments within title strings in V2 release

    for (position = 0; position < bodyLocation; position++)
    {
        if(IsMatch(pbDoc + position, pbDoc + bodyLocation, TAG_TITLE_OPEN))
        {
            token = ExtractContent(pbDoc + position, pbEnd, &tokenSize, &charScanned, &charIgnored);
            if(token != NULL)
            {            
                if(tokenSize > charIgnored)
                {
                    docHash.AddToken(tokenSize-charIgnored);
                    tokenCount++;
                }
                position += charScanned;
                break;
            }
        }
    }

    // set the doc start from <body>
    pbDoc += bodyLocation;

    if (pbDoc >= pbEnd)
    {
        return 0;
    }
    cbDoc = pbEnd - pbDoc;

    for (position = 0; position < cbDoc; )
    {
        charIgnored = SkipCommentOrScript(pbDoc + position, pbEnd);
      
        if(charIgnored)
        {
            position += charIgnored;
            continue;
        }

        // check if it is anchor
        if(IsMatch(pbDoc + position, pbEnd, TAG_ANCHOR_OPEN))
        {
            token = ExtractAnchor(pbDoc + position, pbEnd, &tokenSize);
            if(token != NULL)
            {
                // add the length of anchor to hash structure
                docHash.AddToken(tokenSize);
                anchorCount++;
            }
        }

        token = ExtractContent(pbDoc + position, pbEnd, &tokenSize, &charScanned, &charIgnored);
        position += charScanned;

        if((token != NULL) && (tokenSize > charIgnored))
        {
            // add the length of text to hash structure
            docHash.AddToken(tokenSize-charIgnored);
            tokenCount++;            
        }

        // this is unlikely to happen, but just make sure we keep moving within loop
        if(charScanned == 0)
        {
            position++;
        }
    }

    docHash.AddToken(tokenCount);
    docHash.AddToken(anchorCount);

    docHash.CalculateHash();
    Signature ret = docHash.GetHashValue();
    return ret;
}

const char* ExtractCanonicalLink(const char* pbDoc, Size_t cbDoc, Size_t* cbCanonicalLink)
{
    const char* pbEnd           = pbDoc + cbDoc;
    const char* token           = pbDoc;

    bool        fHead           = false;

    Size_t      charScanned     = 0;
    Size_t      position        = 0;
    Size_t      tokenSize       = 0;

    *cbCanonicalLink = 0;

    // First, find where head starts and ends
    // canonical tag has to be in <head> part, if we cannot find it, return NULL

    for (; position < cbDoc; )
    {
        position += SkipCommentOrScript(pbDoc + position, pbEnd);

        if (IsMatch(pbDoc + position, pbEnd, TAG_HEAD_OPEN))
        {
            fHead = true;
            position += GetSizeOfTag(TAG_HEAD_OPEN);
            continue;
        }

        // canonical tag has to be in head part in html
        // if it is <link > tag, then we will try to find canonical tag
        if (fHead && IsMatch(pbDoc + position, pbEnd, TAG_LINK_OPEN))
        {
            token = ExtractCanonicalLink(pbDoc + position, pbEnd, &tokenSize);
            if(token != NULL)
            {   
                // if there are more than 1 canonical links, always return first one
                // assume we just need 1 canonical link
                *cbCanonicalLink = tokenSize;
                return token;
            }
        }

        // if we see </head> before we find canonical tag, we can stop
        if (IsMatch(pbDoc + position, pbEnd, TAG_HEAD_CLOSE))
        {
            break;
        }

        position++;
    }

    return NULL;
}

int DocSignatureCppLineCount()
{
    return __LINE__;
}
