/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "DocSignatureUtil.h"
#include "SSPHash.h"
#include "PALString.h"

bool IsMatch(const char* begin, const char* end, const char* str)
{
    Size_t length = GetSizeOfTag(str);
    if(length == 0)
    {
        return 0;
    }
    if (begin + length > end)
    {
        return false;
    }
    if (!PALString::CompareInsensitive(begin, str, length))
    {
        return true;
    }
    return false;
}

Size_t GetSizeOfTag(const char* str)
{
    if(str == NULL || *str == '\0')
    {
        return 0;
    }

    return strlen(str);
}

const char* ExtractContent(const char* input, const char* end, Size_t* tokenLength, Size_t* charScanned, Size_t* charIgnored)
{
    // every time the extract content will be called when we detect there is tag, ie, start with '<'
    const char* tokenStart = input;
    const char* tokenEnd = input;
    Size_t startPos = 0;
    Size_t endPos = 0;

    *tokenLength = 0;
    *charScanned = 1;
    *charIgnored = 0;

    Size_t ignoredCharsCount = 0;

    // must be something wrong here
    if(!Extract(input, end, &startPos, &endPos, &ignoredCharsCount))
    {
        *charScanned = 1;
        *tokenLength = 0;
        return NULL;
    }

    tokenStart = input + startPos;
    tokenEnd = input + endPos;

    *charScanned = endPos;
    *tokenLength = tokenEnd - tokenStart;
    *charIgnored = ignoredCharsCount;

    if(*tokenLength == 0 || *tokenLength <= ignoredCharsCount)
    {
        *tokenLength = 0;
        return NULL;
    }
    else
    {
        return tokenStart;
    }

}

bool Extract(const char* input, const char* end, Size_t* startPos, Size_t* endPos, Size_t* ignoredCharsCount)
{
    // every time the extract content will be called when we detect there is tag, ie, start with '<'
    *startPos = 0;
    *endPos = 0;
    Size_t position = 1;
    Size_t inputLength = end - input;
    bool isContent = false;
    Size_t ignoredChars = 0;

    // iterate the string, and find content between '>' and '<'
    // start from position = 1, assume position = 0 is always '<' here
    for(; position < inputLength; position++ )
    {
        // Assume there is no way that html can have this format:
        // <tag1 <tag2> attributes>
        // The chance < in quote is small (nested '<'), and we can take it as error for now
        if(IsMatch(input + position, end, TAG_CLOSE))
        {
            *startPos = position + 1;
            isContent = true;
            continue;
        }

        if(IsMatch(input + position, end, TAG_OPEN))
        {
            break;
        }

        if(isContent && (IsSpace(input[position]) || IsPunct(input[position])))
        {
            ignoredChars++;
        }

    }

    // return false if we couldn't find content
    if(position == 0 || !isContent)
    {
        return false;
    }

    *endPos = position;
    *ignoredCharsCount = ignoredChars;
    return true;
}

const char* ExtractAnchor(const char* input, const char* end, Size_t* tokenLength)
{
    const char* tokenStart = input;
    Size_t startPos = 0;
    Size_t endPos = 0;
    Size_t position = 1;
    Size_t inputLength = end - input;

    for(; position < inputLength; position++ )
    { 
        // stop when we see close tag '>'
        if(IsMatch(input + position, end, TAG_CLOSE))
        {
            break;
        }

        if(IsMatch(input + position, end, TAG_HREF) && !IsMatch(input + position, end, " hreflang"))
        {
            // skip rest of attribute name
            position = position + GetSizeOfTag(TAG_HREF);

            if(!ExtractAttributeValue(input + position, end, &startPos, &endPos))
            {
                return NULL;
            }
            
            tokenStart = input + position + startPos;

            // iterate the string
            // process anchor to cut anything after # but not #! or ?
            Size_t i = 0;
            for(; i < endPos - startPos; i++)
            {
                if((IsMatch(tokenStart + i, end, "#") && !IsMatch(tokenStart + i, end, "#!")) || IsMatch(tokenStart + i, end, "?"))
                {
                    break;
                }
            }

            if(i == 0)
            {
                return NULL;
            }

            *tokenLength = i;
            return tokenStart;
        }
    }

    return NULL;
}

const char* ExtractCanonicalLink(const char* input, const char* end, Size_t* tokenLength)
{
    const char* tokenStart = input;
    const char* relValueStart = input;
    Size_t startPos = 0;
    Size_t endPos = 0;
    Size_t position = 1;
    Size_t inputLength = end - input;
    bool fCanonical = false;
    bool fLink = false;

    for(; position < inputLength; position++ )
    {
        // stop when we see close tag '>' or when canonical tag is found
        if(IsMatch(input + position, end, TAG_CLOSE) || (fCanonical && fLink))
        {
            break;
        }

        if(IsMatch(input + position, end, TAG_REL))
        {
            position = position + GetSizeOfTag(TAG_REL);

            if(!ExtractAttributeValue(input + position, end, &startPos, &endPos))
            {
                return NULL;
            }

            relValueStart = input + position + startPos;

            // check if the rel is canonical
            if(!IsMatch(relValueStart, end, "canonical"))
            {
                return NULL;
            }

            fCanonical = true;
            continue;
        }
      
        else if(IsMatch(input + position, end, TAG_HREF) && !IsMatch(input + position, end, " hreflang"))
        {
            // skip rest of attribute name
            position = position + GetSizeOfTag(TAG_HREF);
            if(!ExtractAttributeValue(input + position, end, &startPos, &endPos))
            {
                return NULL;
            }

            tokenStart = input + position + startPos;

            *tokenLength = endPos - startPos;
            fLink = true;
        }
    }

    // validate URL, since it is canonical URL, it has to be start with http: or https:
    if(!IsMatch(tokenStart, end, "http:") && !IsMatch(tokenStart, end, "https:"))
    {
        fLink = false;
    }
        
    if(fCanonical && fLink)
    {
        // We treat Canonical url as an authoritative URL, so no post-processing on canonical url (ie, trim after # and etc)
        // return url extracted as it is
        return tokenStart;
    }
    return NULL;
}

bool ExtractAttributeValue(const char* input, const char* end, Size_t* startPos, Size_t* endPos)
{
    *startPos = 0;
    *endPos = 0;
    Size_t position = 1;
    Size_t inputLength = end - input;

    bool isValue = false;

    for(; position < inputLength; position++)
    {
        // stop when we see close tag '>'
        if(IsMatch(input + position, end, TAG_CLOSE))
        {
            break;
        }

        // if we found href attribute, skip all meaningless chars such as " ", "\"", "=" before the value of attribute
        if(!IsSkipAttributeChar(input[position]))
        {
            if(!isValue)
            {
                *startPos = position;
                isValue = true;
                continue;
            }
        }
        // Attributes value ends at these chars: '' ', '\"'
        if(isValue && IsEndOfAttributeChar(input[position]))
        {
            break;
        }
    }

    *endPos = position;
 
    if( *startPos == 0
        || *endPos == 0
        || (*endPos <= *startPos)        
        )
    {
        return false;
    }
    return isValue;

}

Size_t SkipTo(const char* input, const char* end, const char* skipToTag)
{
    Size_t position = 0;
    Size_t inputLength = end - input;

    for(; position < inputLength; )
    {
        if (IsMatch(input + position, end, skipToTag))
        {
            position += GetSizeOfTag(skipToTag);
            break;
        }
        else
        {
            position++;
        }
    }
    return position;
}

Size_t SkipCommentOrScript(const char* input, const char* end)
{
    Size_t position = 0;
    Size_t inputLength = end - input;


    // skip comments
    if(IsMatch(input + position, end, TAG_COMMENT_OPEN))
    {
        return SkipTo(input + position, end, TAG_COMMENT_CLOSE);
    }

    // skip script
    if(IsMatch(input + position, end, TAG_SCRIPT_OPEN))
    {
        return SkipTo(input + position, end, TAG_SCRIPT_CLOSE);
    }
    

    return 0;

}
bool IsSpace(char c)
{
    if(c == ' ' 
        || c == '\n' 
        || c == '\t'
        || c == '\r'
        || c == '\f'
        || c == '\v')
    {
        return true;
    }
    return false;
}

bool IsPunct(char c)
{
    if((c >= 0x21 && c <= 0x2f) ||
       (c >= 0x3a && c <= 0x40) ||
       (c >= 0x5b && c <= 0x60) ||
       (c >= 0x7b && c <= 0x7e))
    {
        return true;
    }
    return false;
}

bool IsAlNum(char c)
{
    if((c >= 0x30 && c <= 0x39) ||
        (c >= 0x41 && c <= 0x5a) ||
        (c >= 0x61 && c <= 0x7a))
    {
        return true;
    }
    return false;
}

bool IsSkipAttributeChar(char c)
{
    if(c == ' ' 
        || c == '\"' 
        || c == '=')
    {
        return true;
    }
    return false;
}

bool IsEndOfAttributeChar(char c)
{
    if( c == ' ' 
        || c == '\"' )
    {
        return true;
    }
    return false;
}

int DocSignatureUtilCppLineCount()
{
    return __LINE__;
}
