/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "bloomfilter.h"
#include "DoubleBloomFilter.h"
#include "SSPCommon.h"
#include "PALSync.h"

BOOL DoubleBloomFilter::Initialize(
    unsigned uOneSlotDurationSec, // how often we flip between the 2 filters (e.g., 24 hours)
    size_t cbOneSlotBytes, // size of each filter in bytes.  Total allocated memory is ~2x this.
    IAllocator* pAlloc,
    unsigned cSlots
)
{
    if (cSlots > 2) return FALSE;
    m_cSlots = cSlots;
    m_cbSlotSize = cbOneSlotBytes;
    for (unsigned u = 0; u < cSlots; u++)
    {
        void* pMem = pAlloc->Allocate(cbOneSlotBytes);
        if (!pMem)
        {
            return FALSE;
        }
        if (!m_filters[u].Init(pMem, cbOneSlotBytes))
        {
            return FALSE;
        }
    }

    m_uDurationSec = uOneSlotDurationSec;
    m_lCurSlot = 0;
    m_cNewSlotEntries = 0;
    m_lock = 0;
    m_tLastReset = PluginTime();

    return TRUE;
}

BOOL DoubleBloomFilter::Check(
    size_t cbUrl, const char* pbUrl // url - not necessarily NULL-terminated
)
{
    // this isn't 100% thread safe, but it's just a bloomfilter - what's the worst that can happen..
    if (0 == PALInterlocked::CompareSet(&m_lock, 1, 0))
    {
        time_t tNow = PluginTime();
        if ((m_cNewSlotEntries > (LONG64)m_cbSlotSize) || (tNow > m_tLastReset + (time_t)m_uDurationSec))
        {
            if (tNow > m_tLastReset + 2 * (time_t)m_uDurationSec)
            { // reset both slots
                m_filters[m_lCurSlot & (m_cSlots - 1)].Reset();
            }
            m_tLastReset = tNow;
            m_cNewSlotEntries = 0;
            m_lCurSlot ^= 1;
            // reset the slot that used to be the old slot (and has now become the new slot)
            m_filters[m_lCurSlot & (m_cSlots - 1)].Reset();
        }

        m_lock = 0;
    }

    BOOL fRet = FALSE;
    for (unsigned u = 0; u < m_cSlots; u++)
    {
        if (m_filters[u].Contains(cbUrl, pbUrl))
        {
            return TRUE;
        }
    }
    
    m_filters[m_lCurSlot & (m_cSlots - 1)].Add(cbUrl, pbUrl);
    m_cNewSlotEntries++; // not the end of the world if we get a race
    return FALSE;
}

UInt64 DoubleBloomFilter::Count()
{
    return m_cNewSlotEntries;
}

