/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "FileAccess.h"

#if APR_HAVE_WINDOWS_H

BOOL SetFileAccess(const TCHAR* wszFilePath, const char* szGroupName, DWORD dwAccessMask)
{
    // In order to access via HTTP request, the ACL needs to allow the read access 
    // to Everyone.

    // we always assume the result is false.
    BOOL result = FALSE;
    
    // need to free up memory for these at the end
    void* pSid = NULL;
    PACL pNewAcl = NULL;
    PSECURITY_DESCRIPTOR pFileSD = NULL;

    //
    // 1. Look up SID for the group
    //
    if(!GetAccountSID(szGroupName, &pSid))
    {
        goto Exit;
    }

    //
    // 2. Get SD of the file
    //
    SECURITY_INFORMATION secInfo = DACL_SECURITY_INFORMATION;
    if(!GetFileSD(wszFilePath, &pFileSD, secInfo))
    {
        goto Exit;
    }
    
    //
    // 3. Initialize new SD
    //
    SECURITY_DESCRIPTOR  newSD;
    if(!InitializeSecurityDescriptor(&newSD, SECURITY_DESCRIPTOR_REVISION))
    {
        goto Exit;
    }

    //
    // 4. Get DACL from the old SD
    //
    BOOL fDaclPresent;
    BOOL fDaclDefaulted;
    PACL pAcl = NULL;
    if(!GetSecurityDescriptorDacl(pFileSD, &fDaclPresent, &pAcl, &fDaclDefaulted))
    {
        goto Exit;
    }

    //
    // 5. Get size info for DACL
    //
    ACL_SIZE_INFORMATION AclInfo;
    
    AclInfo.AceCount = 0;
    AclInfo.AclBytesFree = 0;
    AclInfo.AclBytesInUse = 0;

    if(pAcl == NULL)
    {
        fDaclPresent = FALSE;
    }
    // if DACL presents, get the size info
    if(fDaclPresent && !GetAclInformation(pAcl, &AclInfo, sizeof(ACL_SIZE_INFORMATION), AclSizeInformation))
    {
        goto Exit;
    }

    //
    // 6. Initialize new ACL
    //

    // calculate the new size
    // remove SidStart as it is part of both ACE and SID
    // http://msdn.microsoft.com/en-us/library/windows/hardware/ff552151(v=vs.85).aspx
    DWORD cbNewACL = AclInfo.AclBytesInUse + sizeof(ACCESS_ALLOWED_ACE) + GetLengthSid(pSid) - sizeof(DWORD);   

    pNewAcl = (PACL)malloc(cbNewACL);
    if(!pNewAcl)
    {
        goto Exit;
    }
    // ACL_REVISION2 is needed for managing/manipulating on-the-wire.
    if(!InitializeAcl(pNewAcl, cbNewACL, ACL_REVISION2))
    {
        goto Exit;
    }

    //
    // 7. Copy ACEs from the current DACL to new DACL object
    //
    void* pTmpAcl = NULL;
    
    // Copy non-inherited ACEs from the old DACL to the new DACL to conform Windows 2000 preferred order
    if(fDaclPresent && (AclInfo.AceCount > 0))
    {
        if(!CopyACEs(pAcl, pNewAcl, pSid, AclInfo.AceCount))
        {
            goto Exit;
        }
    }

    // add the access-allowed ACE to the new DACL, immediately after all existing non-inherited ACEs.
    if(!AddAccessAllowedAce(pNewAcl, ACL_REVISION2, dwAccessMask, pSid))
    {
        goto Exit;
    }

    // Copy inherited ACEs from the old DACL to the new DACL to conform Windows 2000 preferred order
    if(fDaclPresent && (AclInfo.AceCount > 0))
    {
        if(!CopyACEs(pAcl, pNewAcl, pSid, AclInfo.AceCount, TRUE))
        {
            goto Exit;
        }
    }

    // 
    // 8. Set the new DACL to the new SD
    if(!SetSecurityDescriptorDacl(&newSD, TRUE, pNewAcl, FALSE))
    {
        goto Exit;
    }

    //
    // 9. Copy the old security descriptor control flags regarding DACL automatic inheritance
    //
    if(!CopyAutomaticInheritSecurityDescriptorControlFlags(pFileSD, &newSD))
    {
        goto Exit;
    }

    //
    // 10. Set the new SD to the file
    //
    if (!SetFileSecurity(wszFilePath, secInfo, &newSD))
    {
        goto Exit;
    }

    //
    // Done
    //
    result = TRUE;

Exit:
    if(pSid)
    {
        free(pSid);
        pSid = NULL;
    }
    if(pFileSD)
    {
        free(pFileSD);
        pFileSD = NULL;
    }
    if(pNewAcl)
    {
        free(pNewAcl);
        pNewAcl = NULL;
    }
    return result;
}

BOOL GetAccountSID(const char* szGroupName, void** ppSid, DWORD* pcbSid)
{
    BOOL fRet = FALSE;
    char* szFoundDomain = NULL;
    *ppSid = NULL;

    DWORD cbSid = 0, cbFoundDomain = 0;
    SID_NAME_USE eUse;
    const char* szDomain = "";
    if (LookupAccountNameA(szDomain, szGroupName, NULL, &cbSid, NULL, &cbFoundDomain, &eUse))
    {
        goto Exit;
    }
    if (ERROR_INSUFFICIENT_BUFFER != GetLastError())
    {
        goto Exit;
    }
    *ppSid = malloc(cbSid);  
    if (!*ppSid) 
    {
        goto Exit;
    }
    szFoundDomain = (char*)malloc(cbFoundDomain);  
    if (!szFoundDomain) 
    {
        goto Exit;
    }
    if (!LookupAccountNameA(szDomain, szGroupName, *ppSid, &cbSid, szFoundDomain, &cbFoundDomain, &eUse)) 
    { 
        goto Exit;
    }

    if (pcbSid) 
    {
        *pcbSid = cbSid;
    }
    fRet = TRUE;

Exit:
    if (szFoundDomain) free(szFoundDomain);
    if (!fRet && *ppSid)
    {
        free(*ppSid);  *ppSid = NULL;
    }
    return fRet;
}

BOOL GetFileSD(const TCHAR* wszFilePath, PSECURITY_DESCRIPTOR* pFileSD, const SECURITY_INFORMATION secInfo)
{
    BOOL fRet = FALSE;
    DWORD cbFileSD = 0;

    // first to find out the buffer size needed. This is expected to fail with insufficient buffer.
    if(GetFileSecurity(wszFilePath, secInfo, *pFileSD, 0, &cbFileSD))
    {
        goto Exit;
    }
    if(GetLastError() != ERROR_INSUFFICIENT_BUFFER)
    {
        goto Exit;
    }
    
    // allocate the needed buffer once we know the size needed
    *pFileSD = (PSECURITY_DESCRIPTOR)malloc(cbFileSD);
    if(!pFileSD)
    {
        goto Exit;
    }

    // try again with the allocated buffer
    if(!GetFileSecurity(wszFilePath, secInfo, *pFileSD, cbFileSD, &cbFileSD))
    {
        goto Exit;
    }

    fRet = TRUE;
    
Exit:
    if(!fRet && *pFileSD)
    {
        free(*pFileSD);
        *pFileSD = NULL;
    }
    return fRet;
}

BOOL CopyACEs(const PACL pAcl, const PACL pNewAcl, const PSID pSid, DWORD aceCount, BOOL fInheritACEs)
{
    void* pTmpAce = NULL;
    for (DWORD currentAceIndex = 0; currentAceIndex < aceCount; currentAceIndex++)
    {
        if(!GetAce(pAcl, currentAceIndex, &pTmpAce))
        {
            return FALSE;
        }

        
        ACCESS_ALLOWED_ACE* pAllowedAce = (ACCESS_ALLOWED_ACE*)pTmpAce;
        
        // Skip inherited ACE for copying non-inherited ACE
        if(!fInheritACEs && (pAllowedAce->Header.AceFlags & INHERITED_ACE))
        {
            continue;
        }

        // Skip non-inherited ACE for copying inherited ACE
        if(fInheritACEs && !(pAllowedAce->Header.AceFlags & INHERITED_ACE))
        {
            continue;
        }

        // if it is the same SID, skip
        if(EqualSid(pSid, &(pAllowedAce->SidStart)))
        {
            continue;
        }

        // add the ACE to the new ACL
        if(!AddAce(pNewAcl, ACL_REVISION, MAXDWORD, pTmpAce, ((PACE_HEADER) pTmpAce)->AceSize))
        {
            return FALSE;
        }
    }

    return TRUE;
}

BOOL CopyAutomaticInheritSecurityDescriptorControlFlags(const PSECURITY_DESCRIPTOR pFileSD, SECURITY_DESCRIPTOR* newSD)
{
    // assuming the code is running from XP+, 
    
    SECURITY_DESCRIPTOR_CONTROL controlBitsOfInterest = 0;
    SECURITY_DESCRIPTOR_CONTROL controlBitsToSet = 0;
    SECURITY_DESCRIPTOR_CONTROL oldControlBits = 0;
    DWORD dwRevision = 0;

    if (!GetSecurityDescriptorControl(pFileSD, &oldControlBits, &dwRevision)) 
    {
        return FALSE;
    }

    if (oldControlBits & SE_DACL_AUTO_INHERITED) {
        controlBitsOfInterest =
            SE_DACL_AUTO_INHERIT_REQ |
            SE_DACL_AUTO_INHERITED;
        controlBitsToSet = controlBitsOfInterest;
    }
    else if (oldControlBits & SE_DACL_PROTECTED) {
        controlBitsOfInterest = SE_DACL_PROTECTED;
        controlBitsToSet = controlBitsOfInterest;
    }

    if (controlBitsOfInterest) 
    {
        if (!SetSecurityDescriptorControl(newSD, controlBitsOfInterest, controlBitsToSet)) 
        {
            return FALSE;
        }
    }

    return TRUE;
}

#else

BOOL SetFileAccess(const TCHAR* wszFilePath)
{
    // TODO: Research and add linux port
    return TRUE;
}

#endif
