/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "HeapInstance.h"
#include "PALMemory.h"

#ifdef __TARGET_IIS__
HeapInstance::HeapInstance()
{
    m_fInitialized = false;
    Initialize(0, 1024 * 1024, 0);
}

HeapInstance::HeapInstance(DWORD options, Size_t initialSize, Size_t maximumSize)
{
    m_fInitialized = false;
    Initialize(options, initialSize, maximumSize);
}

HeapInstance::~HeapInstance()
{
    if (m_fInitialized)
    {
        Deinitialize();
    }
}

void HeapInstance::Deinitialize(void)
{
    if (!m_fInitialized) return;
    if (m_numCurrentAllocations != 0 || m_sizeHeap != 0)
    {
        Log(LogID_Common, LogLevel_Warning,
            "HeapInstance",
            "HANDLE %p detected leak at time of destruction: %Iu bytes in %u allocations",
            m_hHeap,
            m_sizeHeap,
            m_numCurrentAllocations);
#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
        HeapInstanceAllocationData *pCurrent = m_pallocdataHead;
        while (pCurrent)
        {
            Log(LogID_Common, LogLevel_Warning,
                "HeapInstance",
                "Leaked memory at location %p of size %Iu",
                pCurrent->location,
                pCurrent->cbAlloc);
            pCurrent = pCurrent->pnext;
        }
#endif
    }
    if (!HeapDestroy(m_hHeap))
    {
        Log(LogID_Common, LogLevel_Warning,
            "HeapInstance",
            "HANDLE %p: HeapDestroy failed with GetLastError = %u", 
            m_hHeap,
            GetLastError());
    }

    m_csHeapStats.DoDispose();
    m_fInitialized = false;
}

Size_t HeapInstance::GetSize(void)
{
    Size_t sizeHeap;

    m_csHeapStats.Enter();
    sizeHeap = m_sizeHeap;
    m_csHeapStats.Leave();

    return sizeHeap;
}

LPVOID HeapInstance::Allocate(DWORD dwFlags, Size_t sizeBytes)
{
    bool fLogStats = false;
    Size_t currentSize = 0;

    m_csHeapStats.Enter();
    m_sizeHeap += sizeBytes;
    m_numCurrentAllocations++;
    m_numTotalAllocations++;
    if ((m_numTotalAllocations & 0xFFFF) == 0)
    {
        fLogStats = true;
        currentSize = m_sizeHeap;
    }
    m_csHeapStats.Leave();

    if (fLogStats)
    {
        Log(LogID_Common, LogLevel_Info, "HeapInstance",
            "HANDLE %p: currently allocated = %Iu; current allocatons = %u; total allocations = %u",
            m_hHeap, 
            currentSize,
            m_numCurrentAllocations,
            m_numTotalAllocations);
    }
    LPVOID pvRet = HeapAlloc(m_hHeap, dwFlags | HEAP_GENERATE_EXCEPTIONS, sizeBytes);
#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
    // add the allocation to the head of the list
    if (pvRet)
    {
        HeapInstanceAllocationData *pdata = (HeapInstanceAllocationData *)malloc(sizeof(HeapInstanceAllocationData));
        if (!pdata) return NULL;
        pdata->location = pvRet;
        pdata->cbAlloc = sizeBytes;
        pdata->pnext = m_pallocdataHead;
        pdata->pprev = NULL;
        if (m_pallocdataHead)
        {
            m_pallocdataHead->pprev = pdata;
        }
        m_pallocdataHead = pdata;
    }
#endif
    return pvRet;
}

LPVOID HeapInstance::ReAllocate(DWORD dwFlags, LPVOID lpMem, Size_t sizeBytes)
{
    Size_t oldSize = HeapSize(m_hHeap, 0, lpMem);

    m_csHeapStats.Enter();
    m_sizeHeap = (m_sizeHeap - oldSize) + sizeBytes;
    m_numTotalAllocations++;
    m_csHeapStats.Leave();

    LPVOID pvRet = HeapReAlloc(m_hHeap, dwFlags | HEAP_GENERATE_EXCEPTIONS, lpMem, sizeBytes);
#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
    if (pvRet)
    {
        // locate the previous allocation in the list
        HeapInstanceAllocationData *pCurrent = m_pallocdataHead;
        while (pCurrent)
        {
            if (pCurrent->location == lpMem)
            {
                // update the location and size
                pCurrent->location = pvRet;
                pCurrent->cbAlloc = sizeBytes;
                break;
            }
            pCurrent = pCurrent->pnext;
        }
    }
#endif
    return pvRet;
}

BOOL HeapInstance::Free(DWORD dwFlags, LPVOID lpMem)
{
    // this happens when the heap for news and deletes is itself deleted
    if (!m_fInitialized)
    {
        if (lpMem != this) return FALSE;
        free(lpMem);
        return TRUE;
    }

    Size_t sizeAlloc = HeapSize(m_hHeap, 0, lpMem);

    m_csHeapStats.Enter();
    m_sizeHeap -= sizeAlloc;
    m_numCurrentAllocations--;
    m_csHeapStats.Leave();

    BOOL fRet = HeapFree(m_hHeap, dwFlags, lpMem);
    if (!fRet)
    {
        Log(LogID_Common, LogLevel_Warning,
            "HeapInstance",
            "HANDLE %p: HeapFree failed with GetLastError = %u", 
            m_hHeap,
            GetLastError());
    }
#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
    // remove the element from the list
    HeapInstanceAllocationData *pCurrent = m_pallocdataHead;
    while (pCurrent)
    {
        if (pCurrent->location == lpMem)
        {
            // remove the element from the list
            if (pCurrent->pprev)
            {
                pCurrent->pprev->pnext = pCurrent->pnext;
            }
            if (pCurrent->pnext)
            {
                pCurrent->pnext->pprev = pCurrent->pprev;
            }
            if (pCurrent == m_pallocdataHead)
            {
                m_pallocdataHead = pCurrent->pnext;
                if (m_pallocdataHead)
                {
                    m_pallocdataHead->pprev = NULL;
                }
            }
            free(pCurrent);
            pCurrent = NULL;
            break;
        }
        pCurrent = pCurrent->pnext;
    }
#endif
    return fRet;
}

void HeapInstance::Initialize(DWORD options, Size_t initialSize, Size_t maximumSize)
{
    // will throw an exception on failure
    m_hHeap = HeapCreate(options | HEAP_GENERATE_EXCEPTIONS, initialSize, maximumSize);
    ULONG ulLowFragmentationHeapOption = 2;
    if (!HeapSetInformation (m_hHeap, HeapCompatibilityInformation, 
        &ulLowFragmentationHeapOption, sizeof(ULONG)))
    {
        Log(LogID_Common, LogLevel_Warning,
            "HeapInstance",
            "HANDLE %p: failed to select low fragmentation heap GetLastError = %u", 
            m_hHeap,
            GetLastError());
    }
    m_fInitialized = true;
    m_sizeHeap = 0;
    m_numCurrentAllocations = 0;
    m_numTotalAllocations = 0;

    m_csHeapStats.DoInit(NULL, NULL, FALSE);

#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
    m_pallocdataHead = NULL;
#endif
}

#else
HeapInstance::HeapInstance()
{
    m_fInitialized = false;
    Initialize(0, 1024 * 1024, 0);
}

HeapInstance::HeapInstance(DWORD options, Size_t initialSize, Size_t maximumSize)
{
    m_fInitialized = false;
    Initialize(options, initialSize, maximumSize);
}

HeapInstance::~HeapInstance()
{
    if (m_fInitialized)
    {
        Deinitialize();
    }
}

void HeapInstance::Deinitialize(void)
{
    if (!m_fInitialized) return;
    if (m_numCurrentAllocations != 0 || m_sizeHeap != 0)
    {
        Log(LogID_Common, LogLevel_Warning,
            "HeapInstance",
            "Pool %p detected leak at time of destruction: %Iu bytes in %u allocations",
            (void *) m_pPool,
            m_sizeHeap,
            m_numCurrentAllocations);
#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
        HeapInstanceAllocationData *pCurrent = m_pallocdataHead;
        while (pCurrent)
        {
            Log(LogID_Common, LogLevel_Warning,
                "HeapInstance",
                "Leaked memory at location %p of size %Iu",
                pCurrent->location,
                pCurrent->cbAlloc);
            pCurrent = pCurrent->pnext;
        }
#endif
    }
    
    m_csHeapStats.DoDispose();
    apr_pool_destroy(m_pPool);
    m_pPool = NULL;

    m_fInitialized = false;
}

LPVOID HeapInstance::Allocate(DWORD dwFlags, Size_t sizeBytes)
{
    bool fLogStats = false;
    Size_t currentSize = 0;

    m_csHeapStats.Enter();
    m_sizeHeap += sizeBytes;
    m_numCurrentAllocations++;
    m_numTotalAllocations++;
    if ((m_numTotalAllocations & 0xFFFF) == 0)
    {
        fLogStats = true;
        currentSize = m_sizeHeap;
    }
    m_csHeapStats.Leave();

    if (fLogStats)
    {
        Log(LogID_Common, LogLevel_Info, "HeapInstance",
            "Pool %p: currently allocated = %Iu; current allocatons = %u; total allocations = %u",
            m_pPool, 
            currentSize,
            m_numCurrentAllocations,
            m_numTotalAllocations);
    }

    LPVOID pvRet = PALMemory::Allocate(sizeBytes, m_pPool);
    
#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
    // add the allocation to the head of the list
    if (pvRet)
    {
        HeapInstanceAllocationData *pdata = (HeapInstanceAllocationData *)malloc(sizeof(HeapInstanceAllocationData));
        if (!pdata) return NULL;
        pdata->location = pvRet;
        pdata->cbAlloc = sizeBytes;
        pdata->pnext = m_pallocdataHead;
        pdata->pprev = NULL;
        if (m_pallocdataHead)
        {
            m_pallocdataHead->pprev = pdata;
        }
        m_pallocdataHead = pdata;
    }
#endif
    return pvRet;
}

LPVOID HeapInstance::ReAllocate(DWORD dwFlags, LPVOID lpMem, Size_t sizeBytes)
{
    Size_t oldSize = PALMemory::Size(lpMem);

    m_csHeapStats.Enter();
    m_sizeHeap = (m_sizeHeap - oldSize) + sizeBytes;
    m_numTotalAllocations++;
    m_csHeapStats.Leave();

    LPVOID pvRet = PALMemory::Reallocate(lpMem, sizeBytes);

#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
    if (pvRet)
    {
        // locate the previous allocation in the list
        HeapInstanceAllocationData *pCurrent = m_pallocdataHead;
        while (pCurrent)
        {
            if (pCurrent->location == lpMem)
            {
                // update the location and size
                pCurrent->location = pvRet;
                pCurrent->cbAlloc = sizeBytes;
                break;
            }
            pCurrent = pCurrent->pnext;
        }
    }
#endif
    return pvRet;
}

BOOL HeapInstance::Free(DWORD dwFlags, LPVOID lpMem)
{
    // this happens when the heap for news and deletes is itself deleted
    if (!m_fInitialized)
    {
        // BUGBUG: Not sure if this should be calling free(). This will depend on
        // the memory allocation model of the rest of the application.
        if (lpMem != this) return FALSE;
        free(lpMem);
        return TRUE;
    }

    Size_t sizeAlloc = PALMemory::Size(lpMem);

    m_csHeapStats.Enter();
    m_sizeHeap -= sizeAlloc;
    m_numCurrentAllocations--;
    m_csHeapStats.Leave();

    PALMemory::Free(lpMem);

#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
    // remove the element from the list
    HeapInstanceAllocationData *pCurrent = m_pallocdataHead;
    while (pCurrent)
    {
        if (pCurrent->location == lpMem)
        {
            // remove the element from the list
            if (pCurrent->pprev)
            {
                pCurrent->pprev->pnext = pCurrent->pnext;
            }
            if (pCurrent->pnext)
            {
                pCurrent->pnext->pprev = pCurrent->pprev;
            }
            if (pCurrent == m_pallocdataHead)
            {
                m_pallocdataHead = pCurrent->pnext;
                if (m_pallocdataHead)
                {
                    m_pallocdataHead->pprev = NULL;
                }
            }
            free(pCurrent);
            pCurrent = NULL;
            break;
        }
        pCurrent = pCurrent->pnext;
    }
#endif
    return TRUE;
}

Size_t HeapInstance::GetSize(void)
{
    Size_t sizeHeap;

    m_csHeapStats.Enter();
    sizeHeap = m_sizeHeap;
    m_csHeapStats.Leave();

    return sizeHeap;
}

void HeapInstance::Initialize(DWORD options, Size_t initialSize, Size_t maximumSize)
{
    apr_status_t status = apr_pool_create(&m_pPool, NULL);
    if (status != APR_SUCCESS)
    {
        throw status;
    }

    m_fInitialized = true;
    m_sizeHeap = 0;
    m_numCurrentAllocations = 0;
    m_numTotalAllocations = 0;
#if defined HEAPINSTANCE_EXTENSIVE_TRACKING
    m_pallocdataHead = NULL;
#endif

    m_csHeapStats.DoInit("SSPHeapInstanceLock", (void *) m_pPool);
}
#endif
