/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#pragma once
// HeapInstance provides a layer in between callers and the heap
// in order to provide for central tracking of the total allocated size
// and to provide for the possibility of identifying leaks

// This class will be used by the crawler's download queue
#include "basic_types.h"
#include "CritSec.h"
#include "shared.h"
#include "logging.h"

#if defined(HEAPINSTANCE_EXTENSIVE_TRACKING)
struct HeapInstanceAllocationData
{
    void *location;
    Size_t cbAlloc;
    HeapInstanceAllocationData *pnext;
    HeapInstanceAllocationData *pprev;
};
#endif


class HeapInstance
{
public:
    // both constructors will throw an exception if HeapCreate fails
    // default constructor sets no options, an initial size of 1MB and no max size
    HeapInstance();

    // constructor with parameters
    HeapInstance(DWORD options, Size_t initialSize, Size_t maximumSize);

    // destructor destroys the heap
    // the heap is created in the constructor
    // which throws an exception on failure
    ~HeapInstance();

    void Deinitialize(void);

    // Returns the total size of all allocations from the heap
    // Note that it is not aware of the level of fragmentation
    Size_t GetSize(void);

    // Allocates a block of memory from the heap
    // Will throw an exception on allocation failure
    LPVOID Allocate(DWORD dwFlags, Size_t sizeBytes);

    // Reallocates a block of memory from the heap
    // Will throw an exception on failure
    LPVOID ReAllocate(DWORD dwFlags, LPVOID lpMem, Size_t sizeBytes);

    // Frees memory from the heap
    BOOL Free(DWORD dwFlags, LPVOID lpMem);

protected:
    // creates the heap (exception on failure) and sets low fragmentation option
    void Initialize(DWORD options, Size_t initialSize, Size_t maximumSize);

    bool m_fInitialized;

#ifdef __TARGET_IIS__
    // critical section used to control access to heap statistics
    // held for very short periods of time
    CRITSEC m_csHeapStats;

    // the heap handle
    HANDLE m_hHeap;
#else
    CRITSECTHREAD  m_csHeapStats;
    apr_pool_t    *m_pPool;
#endif

    // the total size of all blocks allocated from the heap
    Size_t m_sizeHeap;

    // the number of currently outstanding allocations
    UInt32 m_numCurrentAllocations;

    // the number of allocations ever performed by this heap
    UInt32 m_numTotalAllocations;

#if defined(HEAPINSTANCE_EXTENSIVE_TRACKING)
    HeapInstanceAllocationData *m_pallocdataHead;
#endif
};

