/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "SSPCommon.h"
#include "HostMapping.h"
#include "PALFile.h"
#include "PALString.h"

// represents a length of either key or value of host mapping
// 2 * MAX_PATH for hostname and wwwroot path
// PORT_STR_SIZE for port number
// +2 for delimiters ':' and ','
#define MAPPING_ELEMENT_LENGTH 2 * MAX_PATH + PORT_STR_SIZE + 2

class HostMappingFileReader
{
public:
	HostMappingFileReader(){HostMappingFileReader::m_fp = NULL;};
	~HostMappingFileReader(){HostMappingFileReader::Close();};

	BOOL Initialize(const TCHAR* wszMappingFile);
	void Close();

	BOOL GetMapping(__out_ecount(cchKey) TCHAR* wszKey, Size_t cchKey, __out_ecount(cchValue) TCHAR* wszValue, Size_t cchValue);

private:
	FILE* m_fp;

};

BOOL HostMappingFileReader::Initialize(const TCHAR* wszMappingFile)
{
	BOOL fRet = FALSE;

	if(m_fp)
	{
		goto Exit;
	}

	// assuming the configuration file is ANSI encoded
	if(PALFile::Open(&m_fp, wszMappingFile, TEXT("rt, ccs=UNICODE")) != 0)
	{
		goto Exit;
	}

	fRet = TRUE;
Exit:
	return fRet;
}

void HostMappingFileReader::Close()
{
	if(!m_fp)
	{
		return;
	}

	fclose(m_fp);
	m_fp = NULL;
}

BOOL HostMappingFileReader::GetMapping(__out_ecount(cchKey) TCHAR* wszKey, Size_t cchKey, __out_ecount(cchValue) TCHAR* wszValue, Size_t cchValue)
{
	BOOL fRet = FALSE;

    // +2 at the end for '=' and '\0'
    TCHAR wszLine[2 * (MAPPING_ELEMENT_LENGTH) + 2] = {0};
    const TCHAR* pcEq = NULL;
    int iRet = 0;
    const TCHAR* wszTmpValue;

	// caller owns the buffer
	if(!wszKey || !wszValue)
	{
		goto Exit;
	}

	if(!m_fp)
	{
		goto Exit;
	}

	if(feof(m_fp))
	{
		goto Exit;
	}

	if(!FGETWS(wszLine, ARRAYSIZE(wszLine), m_fp))
	{
		goto Exit;
	}

	for (Size_t i = 0; i < ARRAYSIZE(wszLine); i++)
	{
        if (wszLine[i] == '\0')
        {
            break;
        }
		if (wszLine[i] == '=')
		{
			pcEq = wszLine + i;
		}
		if (wszLine[i] == '\r' || wszLine[i] == '\n')
		{
			wszLine[i] = '\0';
            break;
		}
	}

	if(!pcEq)
	{
		goto Exit;
	}


	iRet = PALString::Copy(wszKey, cchKey, wszLine, pcEq - wszLine);
	if(iRet != 0)
	{
		goto Exit;
	}
	wszTmpValue = pcEq + 1;
	iRet = PALString::Copy(wszValue, cchValue, wszTmpValue, TSTRLEN(wszTmpValue));
	if(iRet != 0)
	{
		goto Exit;
	}

	fRet = TRUE;
Exit:
	return fRet;
}


BOOL HostMapping::Entry::Match(const TCHAR* wszHost, const TCHAR* wszPort, const TCHAR* wszApplPath)
{
    if (!wszHost || !wszPort || !wszApplPath)
    {
        return FALSE;
    }

    if (m_wszHost && PALString::CompareInsensitive(m_wszHost, wszHost))
    {
        return FALSE;
    }
    if (m_wszPort && PALString::CompareInsensitive(m_wszPort, wszPort))
    {
        return FALSE;
    }

    if (m_wszApplPath && PathStringCompare(m_wszApplPath, wszApplPath))
    {
        return FALSE;
    }

    return TRUE;
}

HostMapping::Entry* HostMapping::Entry::NextEntry()
{
	return m_pNext;
}

BOOL HostMapping::Entry::CopyToSharedMemory(IAllocator* pAlloc, __out_bcount(cbDest) TCHAR** wszDest, Size_t cbDest, const TCHAR* wszSrc)
{
    TCHAR* wszTmp = (TCHAR*)pAlloc->Allocate(cbDest);
	if(!wszTmp)
	{
		return FALSE;
	}
	if(PALString::Copy(wszTmp, cbDest, wszSrc) != 0)
	{
		return FALSE;
	}
    *wszDest = wszTmp;
    return TRUE;
}

BOOL HostMapping::Entry::SetKey(IAllocator* pAlloc, const TCHAR* wszHost, const TCHAR* wszPort, const TCHAR* wszApplPath)
{
	BOOL fRet = FALSE;
    Size_t cbHost = 0;
    Size_t cbPort = 0;
    Size_t cbApplPath = 0;

    if(!wszHost || !wszPort || !wszApplPath)
    {
        goto Exit;
    }

	cbHost = (TSTRLEN(wszHost) + 1) * sizeof(*wszHost);
	cbPort = (TSTRLEN(wszPort) + 1 ) * sizeof(*wszPort);
	cbApplPath = (TSTRLEN(wszApplPath) + 1) * sizeof(*wszApplPath);

    if(cbHost == 0 || cbApplPath == 0)
    {
        goto Exit;
    }

    if(!CopyToSharedMemory(pAlloc, const_cast<TCHAR**>(&m_wszHost), cbHost, wszHost))
    {
        goto Exit;
    }
	
	if(wszPort[0] == '\0')
	{
		m_wszPort = NULL;
	}
	else
	{
        if(!CopyToSharedMemory(pAlloc, const_cast<TCHAR**>(&m_wszPort), cbPort, wszPort))
        {
            goto Exit;
        }
	}

    if(!CopyToSharedMemory(pAlloc, const_cast<TCHAR**>(&m_wszApplPath), cbApplPath, wszApplPath))
    {
        goto Exit;
    }


	fRet = TRUE;
Exit:
	return fRet;
}

BOOL HostMapping::Entry::SetValue(IAllocator* pAlloc, const TCHAR* wszOutHostAndPort, const TCHAR* wszOutApplPath)
{
	BOOL fRet = FALSE;
    Size_t cbOutHostAndPort = 0;
    Size_t cbOutApplPath = 0;

    if(!wszOutHostAndPort || !wszOutApplPath)
    {
        goto Exit;
    }

	cbOutHostAndPort = (TSTRLEN(wszOutHostAndPort) + 1 ) * sizeof(*wszOutHostAndPort);
	cbOutApplPath = (TSTRLEN(wszOutApplPath) + 1) * sizeof(*wszOutApplPath);

    if(cbOutHostAndPort == 0 || cbOutApplPath == 0)
    {
        goto Exit;
    }

    if(!CopyToSharedMemory(pAlloc, const_cast<TCHAR**>(&m_wszOutHostAndPort), cbOutHostAndPort, wszOutHostAndPort))
    {
        goto Exit;
    }

    if(!CopyToSharedMemory(pAlloc, const_cast<TCHAR**>(&m_wszOutApplPath), cbOutApplPath, wszOutApplPath))
    {
        goto Exit;
    }

	fRet = TRUE;
Exit:
	return fRet;
}

BOOL HostMapping::Entry::GetHostAndPort(__out_ecount(cchOutHostAndPort) TCHAR* wszOutHostAndPort, Size_t cchOutHostAndPort)
{
    if (PALString::Copy(wszOutHostAndPort, cchOutHostAndPort, m_wszOutHostAndPort) != 0)
	{
		return FALSE;
	}

	return TRUE;
}

BOOL HostMapping::Entry::GetApplPath(__out_ecount(cchOutApplPath) TCHAR* wszOutApplPath, Size_t cchOutApplPath)
{
    Size_t cch = TSTRLEN(m_wszOutApplPath);
    if (cch >= cchOutApplPath)
    {
        return FALSE;
    }
    
    if(PALString::Copy(wszOutApplPath, cchOutApplPath, m_wszOutApplPath) != 0)
    {
		return FALSE;
	}

    return TRUE;
}

const TCHAR* HostMapping::Extract(__out_ecount(cchExtracted) TCHAR* wszExtracted, Size_t cchExtracted, const TCHAR* wszCombined, const TCHAR* delim)
{    

    const TCHAR* pwszNextElement = TCSSTR(wszCombined, delim);
    if(pwszNextElement == NULL)
    {
        goto Exit;
    }
 
    if(!wszExtracted || cchExtracted == 0)
    {
        goto Exit;
    }


    if(PALString::Copy(wszExtracted, cchExtracted, wszCombined, pwszNextElement - wszCombined) != 0)
    {
        goto Exit;
    }
    
    pwszNextElement += TSTRLEN(delim);

Exit:
    return pwszNextElement;
}

BOOL HostMapping::ExtractElements(__out_ecount(cchHost) TCHAR* wszHost, Size_t cchHost, 
                                         __out_ecount(cchPort) TCHAR* wszPort, Size_t cchPort,
                                         TCHAR** wszApplPath,
                                         const TCHAR* wszCombined)
{
    BOOL fRet = FALSE;
    Size_t cchHostOrPort = cchHost + cchPort + 1;
    const TCHAR* wszApplPathPos;
    const TCHAR* wszPortPos;
	TCHAR* wszHostOrPort = new TCHAR[cchHostOrPort];
	memset(wszHostOrPort, 0, cchHostOrPort);

    // start from getting wwwroot since wwwroot contains ':' in Windows
	wszApplPathPos = Extract(wszHostOrPort, cchHostOrPort, wszCombined, TEXT(","));
    if(!wszApplPathPos)
    {
        goto Exit;
    }
    *wszApplPath = const_cast<TCHAR*>(wszApplPathPos);

    wszPortPos = Extract(wszHost, cchHost, wszHostOrPort, TEXT(":"));
    if(wszPortPos)
    {
		if(PALString::Copy(wszPort, cchPort, wszPortPos) != 0)
        {
            goto Exit;
        }
    }
    else
    {
        if(PALString::Copy(wszHost, cchHost, wszHostOrPort) != 0)
        {
            goto Exit;
        }
    }

    fRet = TRUE;
Exit:
    delete[] wszHostOrPort;
    return fRet;
}

BOOL HostMapping::AddEntry(IAllocator* pAlloc, const TCHAR* wszKey, const TCHAR* wszValue)
{
	BOOL fRet = FALSE;

    TCHAR wszInHost[MAX_PATH] = {0};
    TCHAR wszInPort[PORT_STR_SIZE + 1] = {0};
    TCHAR* wszInApplPath = NULL;
    TCHAR wszOutHost[MAX_PATH] = {0};
    TCHAR wszOutPort[PORT_STR_SIZE + 1] = {0};
    TCHAR* wszOutApplPath = NULL;
    TCHAR wszOutHostAndPort[MAX_PATH + PORT_STR_SIZE + 2] = {0};

    Entry* newEntry = NULL;

	// Key
    if(!ExtractElements(wszInHost, ARRAYSIZE(wszInHost), wszInPort, ARRAYSIZE(wszInPort), &wszInApplPath, wszKey))
    {
        goto Exit;
    }

    // Value
    if(!ExtractElements(wszOutHost, ARRAYSIZE(wszOutHost), wszOutPort, ARRAYSIZE(wszOutPort), &wszOutApplPath, wszValue))
    {
        goto Exit;
    }
    
	if((TSTRLEN(wszOutPort) == 0) || (wszOutPort && (PALString::Compare(wszOutPort, TEXT("80")) == 0)))
    {
        if(SNTPRINTF_S(wszOutHostAndPort, ARRAYSIZE(wszOutHostAndPort), _TRUNCATE, STRFMT_TLS, wszOutHost) == -1)
        {
            goto Exit;
        }
    }
    else
    {
        if(SNTPRINTF_S(wszOutHostAndPort, ARRAYSIZE(wszOutHostAndPort), _TRUNCATE, STRFMT_TLS TEXT(":") STRFMT_TLS, wszOutHost, wszOutPort) == -1)
        {
            goto Exit;
        }
    }

	newEntry = (Entry*)pAlloc->Allocate(sizeof(Entry));
    if(!newEntry)
    {
        goto Exit;
    }

	if(!newEntry->SetKey(pAlloc, wszInHost, wszInPort, wszInApplPath))
	{
		goto Exit;
	}
	if(!newEntry->SetValue(pAlloc, wszOutHostAndPort, wszOutApplPath))
	{
		goto Exit;
	}

    newEntry->m_pNext = m_pEntries;
	m_pEntries = newEntry;

	fRet = TRUE;

Exit:
	return fRet;
}

BOOL HostMapping::Load(IAllocator* pAlloc, const TCHAR* wszMappingFile)
{
	BOOL fRet = FALSE;

	HostMappingFileReader hostMappingFileReader;

    TCHAR wszKey[MAPPING_ELEMENT_LENGTH] = {0};
    TCHAR wszValue[MAPPING_ELEMENT_LENGTH] = {0};
	BOOL fContinue = FALSE;


	if(!hostMappingFileReader.Initialize(wszMappingFile))
	{
		goto Exit;
	}

	while(true)
    {
		fContinue = hostMappingFileReader.GetMapping(wszKey, ARRAYSIZE(wszKey), wszValue, ARRAYSIZE(wszValue));
		if(!fContinue)
		{
			break;
		}
		AddEntry(pAlloc, wszKey, wszValue);
    }

	hostMappingFileReader.Close();

	fRet = TRUE;

Exit:
	return fRet;
    
}

HostMapping::Entry* HostMapping::Map(const TCHAR* wszHost, const TCHAR* wszPort, const TCHAR* wszApplPath)
{
    Entry *p = m_pEntries;
	
    while (p)
    { // not expecting a large number of mapping entries perf-wise
        if (p->Match(wszHost, wszPort, wszApplPath))
        {
            break;
        }
        p = p->NextEntry();
    }
    return p;
}

BOOL HostMapping::Enumerate(void* pvContext, BOOL (*pfnCallback)(void* pvContext, void* pUser))
{
    Entry *p = m_pEntries;
    while (p)
    {
        if (!pfnCallback(pvContext, p->m_pUser))
        {
            return FALSE;
        }
		p = p->NextEntry();
    }
    
    return TRUE;
}

