/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
// Platform Abstraction Layer
//
// This is the main header to target different platforms: IIS or Apache
// The default target is IIS. A third define is used specifically to
// target the merge service used by Apache in Linux for Win32.
//
// Compile with the following defintion:
// IIS: TARGET_IIS
// Apache: TARGET_APACHE
// Apache merge service for Win32: TARGET_APACHE_SERVICE_WIN32
#pragma once

#if (defined(TARGET_IIS) && defined(TARGET_APACHE)) || \
    (defined(TARGET_IIS) && defined(TARGET_APACHE_SERVICE_WIN32)) || \
    (defined(TARGET_APACHE) && defined(TARGET_APACHE_SERVICE_WIN32))
#error Can not target multiple platforms.
#endif


// Use __TARGET_IIS__ or __TARGET_APACHE__ or __PUBLISH_MODE_MANUAL__ to help with 
// application development of different platforms.
#if defined(TARGET_APACHE)
#define __TARGET_APACHE__
#undef __TARGET_IIS__
#undef __PUBLISH_MODE_MANUAL__
#elif defined(TARGET_APACHE_SERVICE_WIN32)
#define __TARGET_IIS__
#define __PUBLISH_MODE_MANUAL__
#undef __TARGET_APACHE__
#else
#define __TARGET_IIS__
#undef __TARGET_APACHE__
#undef __PUBLISH_MODE_MANUAL__
#endif


#ifdef __TARGET_IIS__
// IIS Definitions
#include <Windows.h>
#include "PALSAL.h"
#include <limits.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <ctype.h>
#include <time.h>
#include <Shlobj.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utime.h>

#ifndef Z_PREFIX
#define Z_PREFIX
#endif

//#define _SINGLE_PROC_MODE

// Define APR_HAVE_WINDOWS_H as it may be
// used to determine a WIN32 machine.
#ifndef APR_HAVE_WINDOWS_H
#ifdef WIN32
#define APR_HAVE_WINDOWS_H 1
#else
#define APR_HAVE_WINDOWS_H 0
#endif
#endif

struct PAL_Parameters
{
    int Version;
};

#else


// apr.h only includes the Windows XP subset of <windows.h> definitions.
// We will include <Windows.h> first so that all the latest functions can be called.
#ifdef WIN32
#include <Windows.h>
#include <Shlobj.h>
#endif

// Apache Definitions
#include <apr.h>

// Older versions of APR do not have APR_HAVE_WINDOWS_H definition
#ifndef APR_HAVE_WINDOWS_H
#ifdef WIN32
#define APR_HAVE_WINDOWS_H 1
#else
#define APR_HAVE_WINDOWS_H 0
#endif
#endif

#if APR_HAVE_WINDOWS_H
#define _SINGLE_PROC_MODE
#endif

#if !APR_HAVE_STDIO_H
#error Platform does not have <stdio.h>
#endif

#if !APR_HAVE_STDLIB_H
#error Platform does not have <stdlib.h>
#endif

#if !APR_HAVE_STRING_H
#error Platform does not have <string.h>
#endif

#if !APR_HAVE_CTYPE_H
#error Platform does not have <ctype.h>
#endif

#if !APR_HAS_THREADS
#error Platform is not multi-threaded
#endif

#include "PALSAL.h"
#include <apr_atomic.h>
#include <apr_file_io.h>
#include <apr_fnmatch.h>
#include <apr_pools.h>
#include <apr_strings.h>
#include <apr_thread_mutex.h>
#include <apr_thread_cond.h>
#include <apr_thread_pool.h>
#include <apr_proc_mutex.h>
#include <apr_portable.h>
#include <apr_rmm.h>
#include <apr_shm.h>
#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>

#ifdef AP_NEED_SET_MUTEX_PERMS
#ifndef _IGNORE_MUTEX_PERM
extern "C" apr_status_t unixd_set_proc_mutex_perms(apr_proc_mutex_t *);
#else
#define unixd_set_proc_mutex_perms(x) APR_SUCCESS
#endif
#endif

#if APR_HAVE_TIME_H
#include <time.h>
#if !APR_HAVE_WINDOWS_H
#include <utime.h>
#else
#include <sys/utime.h>
#endif
#else
#if APR_HAVE_SYS_TIME_H
#include <sys/time.h>
#if !APR_HAVE_WINDOWS_H
#include <utime.h>
#else
#include <sys/utime.h>
#endif
#else
#error Platform does not have <time.h> or <sys/time.h>
#endif
#endif

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#else
#if !APR_HAVE_WINDOWS_H
#error Platform does not have <unistd.h>
#endif
#endif

#if APR_HAVE_WINDOWS_H
#include <tchar.h>
#else
// Linux specific includes.
#include <wchar.h>
#include <sys/vfs.h>
#endif

#include "PALDefs.h"

#ifndef _X86_
#if !(defined(__LP64__) || defined(_LP64))
#define _X86_
#endif
#endif

#if !APR_HAVE_WINDOWS_H
extern "C++"
template <typename T, size_t N>
char (*RtlpNumberOf( T (&)[N]))[N];

#define ARRAYSIZE(A)    (sizeof(*RtlpNumberOf(A)))

#define __cdecl
#define __stdcall
#endif


struct PAL_Parameters
{
    int         Version;
    BOOL        bPreInit;
    apr_pool_t *pMemPool;
};


#endif


// Global variable that is used internally by the application.
extern PAL_Parameters g_PALParams;

// This initializes g_PALParams and should be called once prior to
// any of the common functions. If bPreIntialize is true, then
// APR initialize functions will also be called. In APR modules, this
// should be false since Apache already calls the initialize functions.
BOOL PAL_Initialize(BOOL bPreIntialize = FALSE);

// This performs cleanup of g_PALParams and should be called
// before termination.
void PAL_Terminate(void);
