/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PALFile.h"
#include "PALString.h"
#include "SSPCommon.h"

BOOL PALFile::s_fFakeGeneratedFileTimes = FALSE;

BOOL PALFile::Move(const TCHAR *ptszSrc, const TCHAR *ptszDest)
{
#if APR_HAVE_WINDOWS_H
    return MoveFileEx(ptszSrc, ptszDest, MOVEFILE_REPLACE_EXISTING | MOVEFILE_COPY_ALLOWED );

#else
    char szFrom[MAX_PATH];
    char szTo[MAX_PATH];

    if (ptszSrc == NULL || ptszDest == NULL)
    {
        return FALSE;
    }

    PALString::Convert(szFrom, ARRAYSIZE(szFrom), ptszSrc, TSTRLEN(ptszSrc));
    PALString::Convert(szTo, ARRAYSIZE(szTo), ptszDest, TSTRLEN(ptszDest));

     if (rename(szFrom, szTo) == 0)
     {
         return TRUE;
     }
     else
     {
         return FALSE;
     }
#endif
}

BOOL PALFile::Delete(const TCHAR *ptszFile)
{
#if APR_HAVE_WINDOWS_H
    return DeleteFile(ptszFile);
#else

    char szFile[MAX_PATH];

    if (ptszFile == NULL)
    {
        return FALSE;
    }

    PALString::Convert(szFile, ARRAYSIZE(szFile), ptszFile, TSTRLEN(ptszFile));

    if (remove(szFile) == 0)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
#endif
}

int PALFile::Open(FILE **ppFile, const TCHAR *ptszFilename, const TCHAR *ptszMode)
{
#if APR_HAVE_WINDOWS_H
    return _tfopen_s(ppFile, ptszFilename, ptszMode);
#else
    char szFile[MAX_PATH];
    char szMode[MAX_PATH];

    if (ppFile == NULL || ptszFilename == NULL || ptszMode == NULL)
    {
        return EINVAL;
    }

    PALString::Convert(szFile, ARRAYSIZE(szFile), ptszFilename, TSTRLEN(ptszFilename));
    PALString::Convert(szMode, ARRAYSIZE(szMode), ptszMode, TSTRLEN(ptszMode));

    *ppFile = fopen(szFile, szMode);

    if (*ppFile == NULL)
    {
        return EINVAL;
    }
    
    return 0;

#endif
}

void PALFile::SetModifiedTime(const TCHAR* wszFile)
{
    if (s_fFakeGeneratedFileTimes)
    {
#if APR_HAVE_WINDOWS_H
        struct _utimbuf buf;
        buf.modtime = PluginTime();
        buf.actime = buf.modtime;
        _tutime(wszFile, &buf);
#else
        char szFile[MAX_PATH];
        struct utimbuf buf;
        buf.modtime = PluginTime();
        buf.actime = buf.modtime;
    
        PALString::Convert(szFile, ARRAYSIZE(szFile), wszFile, TSTRLEN(wszFile));
        utime(szFile, &buf);
#endif
    }
}


BOOL PALDirectory::CreateDirectory(const TCHAR *ptszFolder)
{
#if APR_HAVE_WINDOWS_H
    return ::CreateDirectory(ptszFolder, NULL);
#else
    BOOL bRetVal = FALSE;
    apr_pool_t *pPool = NULL;
    char szPath[MAX_PATH];
    apr_status_t status;

    if (PALString::Convert(szPath, ARRAYSIZE(szPath), ptszFolder, TSTRLEN(ptszFolder)) != 0)
    {
        goto Done;
    }

    status = apr_pool_create(&pPool, NULL);
    if (status != APR_SUCCESS)
    {
        goto Done;
    }

    status = apr_dir_make(szPath, APR_FPROT_OS_DEFAULT, pPool);
    if (status != APR_SUCCESS)
    {
        goto Done;
    }

    bRetVal = TRUE;    
Done:
    if (pPool != NULL)
    {
        apr_pool_destroy(pPool);
    }

    return bRetVal;
#endif
}

BOOL PALDirectory::DeleteDirectory(const TCHAR *ptszFolder)
{
#if APR_HAVE_WINDOWS_H
    return RemoveDirectory(ptszFolder);
#else
    char szFolder[MAX_PATH];

    if (ptszFolder == NULL)
    {
        return FALSE;
    }

    if (PALString::Convert(szFolder, ARRAYSIZE(szFolder), ptszFolder, TSTRLEN(ptszFolder)) != 0)
    {
        return FALSE;
    }

    if (rmdir(szFolder) == 0)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
#endif
}


BOOL PALDirectory::EnsureDirectoryExists(const TCHAR* startdirname)
{
#ifdef __TARGET_IIS__
    DWORD attr = GetFileAttributes(startdirname);
    //
    // If the directory does not exist, we will try to create it.
    //
    if (attr == INVALID_FILE_ATTRIBUTES)
    {
        TCHAR dirname[_MAX_PATH];
        _tcscpy_s(dirname,ARRAYSIZE(dirname),startdirname);
        //
        // Remove the trailing path separator from the directory.
        //
        Size_t dirLength = TSTRLEN(dirname);
        while (dirLength && (dirname[dirLength - 1] == '\\' || dirname[dirLength - 1] == '/'))
        {
            dirname[--dirLength] = '\0';
        }
        //
        // Treat root directory as always there.
        //
        if (!dirLength)
        {
            return TRUE;
        }
        //
        // Try to make the parent directory since this directory is not there.
        //
        TCHAR drive[_MAX_DRIVE];
        TCHAR dir[_MAX_DIR];
        TCHAR fname[_MAX_FNAME];
        TCHAR ext[_MAX_EXT];
        _tsplitpath_s(dirname,drive,_MAX_DRIVE,dir, _MAX_DIR,fname,_MAX_FNAME,ext,_MAX_EXT);
        //
        // Treat root directory as always there.
        //
        if (!dir[0] || (!dir[1] && (dirname[dirLength - 1] == '\\' || dirname[dirLength - 1] == '/')))
        {
            return TRUE;
        }
        //
        // Compose the parent directory.
        //
        TCHAR parentdir[_MAX_PATH];
        _tmakepath_s(parentdir, _MAX_PATH, drive, dir, 0, 0);
        if (!parentdir[0])
        {
            return TRUE;
        }
        if (EnsureDirectoryExists(parentdir))
        {
            //
            // Since the parent directory already exists or was created,
            // we can attempt to create the target directory.
            //

            if (::CreateDirectory(dirname,0))
            {
                return TRUE;
            }
            // Comments: Change by Zhaoyi & Dafan, to change the logic of create directory to return true when folder already existed
            DWORD ec = GetLastError();
            // some other thread/process created the directory
            // between the time we checked for this directory
            // and created the parentdirectory
            if (ec == ERROR_ALREADY_EXISTS)
            {
                return TRUE;
            }
        }
        return (false);
    }
    
    // The directory already exists.
    if(attr & FILE_ATTRIBUTE_DIRECTORY)
    {
        return TRUE;
    }
    
    // This is a regular file.
    return FALSE;

#else
    char szPath[MAX_PATH];
    apr_pool_t *pPool = NULL;
    BOOL bRetVal = FALSE;

    // This is may be an expensive process, we should find another alternative.
    if (apr_pool_create(&pPool, NULL) != APR_SUCCESS)
    {
        goto Exit;
    }

    if (PALString::Convert(szPath, ARRAYSIZE(szPath), startdirname, TSTRLEN(startdirname)) != 0)
    {
        goto Exit;
    }

    if (apr_dir_make_recursive(szPath, APR_FPROT_OS_DEFAULT, pPool) != APR_SUCCESS)
    {
        goto Exit;
    }

    bRetVal = TRUE;

Exit:
    if (pPool != NULL)
    {
        apr_pool_destroy(pPool);
    }

    return bRetVal;
#endif
}

BOOL PALDirectory::PathExists(const TCHAR *ptszPath)
{
#if APR_HAVE_WINDOWS_H != 0
    DWORD attr = GetFileAttributes(ptszPath);
    if (attr == INVALID_FILE_ATTRIBUTES)
    {
        DWORD dwError=GetLastError();
        if (dwError == ERROR_PATH_NOT_FOUND || dwError == ERROR_FILE_NOT_FOUND)
        {
            return FALSE;
        }
    }
    return TRUE;
#else
    apr_pool_t *pPool = NULL;
    BOOL bRetValue = FALSE;
    char szPath[MAX_PATH];

    if (PALString::Convert(szPath, ARRAYSIZE(szPath), ptszPath, TSTRLEN(ptszPath)) != 0)
    {
        goto Done;
    }
    
    if (apr_pool_create(&pPool, NULL) != APR_SUCCESS)
    {
        goto Done;
    }

    apr_finfo_t finfo;
    if (apr_stat(&finfo, szPath, APR_FINFO_TYPE, pPool) != APR_SUCCESS)
    {
        goto Done;
    }

    if (finfo.filetype == APR_DIR || finfo.filetype == APR_REG)
    {
       bRetValue = TRUE;
       goto Done;
    }
    
Done:
    if (pPool != NULL)
    {
        apr_pool_destroy(pPool);
    }

    return bRetValue;
#endif
}

BOOL PALDirectory::GetWorkingDirectory(__out_ecount(len) TCHAR *ptszPath, Size_t len)
{
#if APR_HAVE_WINDOWS_H
    if (len > ((Size_t) ((DWORD) -1)))
    {
        return FALSE;
    }

    DWORD val = GetCurrentDirectory((DWORD) len, ptszPath);

    if (val == 0 || val > len)
    {
        return FALSE;
    }

    return TRUE;
#else
    char szPath[MAX_PATH];
    
    if (getcwd(szPath, ARRAYSIZE(szPath)) == NULL)
    {
        return FALSE;
    }

    Size_t lenPath = strlen(szPath);
    if (lenPath + 1 > len)
    {
        return FALSE;
    }
    
    if (PALString::Convert(ptszPath, len, szPath, lenPath) != 0)
    {
        return FALSE;
    }

    return TRUE;
#endif
}


#if APR_HAVE_WINDOWS_H
PALFind::PALFind() :
    m_bInit(false),
    m_hFind(INVALID_HANDLE_VALUE)
{
    memset(&m_FindFileData, 0, sizeof(m_FindFileData));
}

PALFind::~PALFind()
{
    Destroy();
}

void PALFind::Destroy()
{
    if (m_hFind != INVALID_HANDLE_VALUE)
    {
        FindClose(m_hFind);
        m_hFind = INVALID_HANDLE_VALUE;
    }

    memset(&m_FindFileData, 0, sizeof(m_FindFileData));
    
    m_bInit = false;
}

BOOL PALFind::Initialize(const TCHAR *ptszPath, const TCHAR *ptszSearchPattern)
{
    BOOL bRetVal = TRUE;

    Destroy();

    if (ptszPath == NULL || ptszSearchPattern == NULL)
    {
        bRetVal = FALSE;
        goto Done;
    }

    if (SNTPRINTF_S(
        m_ptszSearchPath,
        ARRAYSIZE(m_ptszSearchPath),
        ARRAYSIZE(m_ptszSearchPath) - 1,
        STRFMT_TLS STRFMT_TDIRSLASH STRFMT_TLS,
        ptszPath, 
        ptszSearchPattern) <= 0)
    {
        bRetVal = FALSE;
        goto Done;
    }

    m_bInit = true;

Done:
    return bRetVal;
}

PALFind::EStatus PALFind::LastError()
{
    switch (GetLastError())
    {
    case ERROR_FILE_NOT_FOUND:
        return Error_FileNotFound;
    case ERROR_PATH_NOT_FOUND:
        return Error_DirectoryNotFound;
    case ERROR_NO_MORE_FILES:
        return NoMoreResults;
    }

    return Error_Other;
}

PALFind::EStatus PALFind::Find()
{
    EStatus status = Success;

    if (!m_bInit)
    {
        status = Error_NotInitialized;
        goto Done;
    }

    if (m_hFind == INVALID_HANDLE_VALUE)
    {
        m_hFind = FindFirstFile(m_ptszSearchPath, &m_FindFileData);

        if (m_hFind == INVALID_HANDLE_VALUE)
        {
            status = LastError();
            goto Done;
        }
    }
    else
    {
        if (FindNextFile(m_hFind, &m_FindFileData) == FALSE)
        {
            status = LastError();
            goto Done;
        }
    }
    
Done:
    return status;
}

PALFind::EAttrib PALFind::GetAttrib()
{
    if (m_FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
        return Directory;
    }
    else if (m_FindFileData.dwFileAttributes & FILE_ATTRIBUTE_VIRTUAL ||
        m_FindFileData.dwFileAttributes & FILE_ATTRIBUTE_OFFLINE ||
        m_FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DEVICE)
    {
        return Unknown;
    }
    else
    {
        return File;
    }

}

const TCHAR *PALFind::GetName()
{
    return m_FindFileData.cFileName;
}

UInt64 PALFind::GetSize()
{
    return ((UInt64) m_FindFileData.nFileSizeHigh * ((UInt64) MAXDWORD + 1)) + (UInt64) m_FindFileData.nFileSizeLow;
}
#else

PALFind::PALFind() :
    m_bInit(false),
    m_pPool(NULL),
    m_pFolder(NULL)
{
}

PALFind::~PALFind()
{
    Destroy();
}

void PALFind::Destroy()
{
    if (m_pFolder != NULL)
    {
        apr_dir_close(m_pFolder);
        m_pFolder = NULL;
    }

    if (m_pPool != NULL)
    {
        apr_pool_destroy(m_pPool);
        m_pPool = NULL;
    }

    memset(&m_finfo, 0, sizeof(m_finfo));

    m_szPath[0] = '\0';
    m_szPattern[0] = '\0';
    m_tszName[0] = '\0';

    m_bInit = false;
}

BOOL PALFind::Initialize(const TCHAR *ptszPath, const TCHAR *ptszSearchPattern)
{
    BOOL bRetVal = TRUE;
    Size_t len;

    Destroy();

    if (ptszPath == NULL || ptszSearchPattern == NULL)
    {
        bRetVal = FALSE;
        goto Done;
    }

    len = TSTRLEN(ptszPath);
    if (ARRAYSIZE(m_szPath) < len + 1)
    {
        bRetVal = FALSE;
        goto Done;
    }

    PALString::Convert(m_szPath, ARRAYSIZE(m_szPath), ptszPath, len);

    len = TSTRLEN(ptszSearchPattern);
    if (ARRAYSIZE(m_szPattern) < len + 1)
    {
        bRetVal = FALSE;
        goto Done;
    }
    PALString::Convert(m_szPattern, ARRAYSIZE(m_szPattern), ptszSearchPattern, len);

    if (apr_pool_create(&m_pPool, NULL) != APR_SUCCESS)
    {
        bRetVal = FALSE;
        goto Done;
    }

    m_bInit = true;

Done:
    return bRetVal;
}

PALFind::EStatus PALFind::Find()
{
    bool bFirstFind = false;
    bool bMatchFound = false;
    EStatus status = Success;

    m_tszName[0] = '\0';

    if (!m_bInit)
    {
        status = Error_NotInitialized;
        goto Done;
    }

    if (m_pFolder == NULL)
    {
        if (apr_dir_open(&m_pFolder, m_szPath, m_pPool) != APR_SUCCESS)
        {
            status = Error_DirectoryNotFound;
            goto Done;
        }

        bFirstFind = true;
    }

    
    for (;;)
    {
        if (apr_dir_read(&m_finfo, APR_FINFO_MIN | APR_FINFO_NAME, m_pFolder) != APR_SUCCESS)
        {
            status = NoMoreResults;
            break;
        }

        if (apr_fnmatch(m_szPattern, m_finfo.name , APR_FNM_CASE_BLIND) == APR_SUCCESS)
        {
            // Match found.
            PALString::Convert(m_tszName, ARRAYSIZE(m_tszName), m_finfo.name, strlen(m_finfo.name));
            bMatchFound = true;
            break;
        }
    }

    if (bFirstFind && !bMatchFound)
    {
        status = Error_FileNotFound;
    }

Done:
    return status;
}

PALFind::EAttrib PALFind::GetAttrib()
{
    switch (m_finfo.filetype)
    {
    case APR_DIR:
        return Directory;
    case APR_REG:
        return File;
    }

    return Unknown;
}

const TCHAR *PALFind::GetName()
{
    return m_tszName;
}

UInt64 PALFind::GetSize()
{
    return m_finfo.size;
}
#endif
