/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PALMemory.h"

#define MEM_ALIGN 16

#if defined(__TARGET_IIS__)
void *PALMemory::Allocate(Size_t cbSrcLen)
{
    return Allocate(cbSrcLen, NULL);
}

void *PALMemory::Allocate(Size_t cbSrcLen, void *pPool)
{
    UNREFERENCED_PARAMETER(pPool);

    void *pvRet = _aligned_malloc(cbSrcLen, MEM_ALIGN);
    if (pvRet == NULL)
    {
        throw STATUS_NO_MEMORY;
    }

    return pvRet;
}

void *PALMemory::CAllocate(Size_t cbSrcLen)
{
    return CAllocate(cbSrcLen, NULL);
}

void *PALMemory::CAllocate(Size_t cbSrcLen, void *pPool)
{
    void *pvRet = Allocate(cbSrcLen, pPool);
    if (pvRet == NULL)
    {
        throw ERROR_OUTOFMEMORY;
    }

    memset(pvRet, 0, cbSrcLen);

    return pvRet;
}

void *PALMemory::Reallocate(void *pMem, Size_t cbSrcLen)
{
    void *pvRet = _aligned_realloc(pMem, cbSrcLen, MEM_ALIGN);
    if (pvRet == NULL)
    {
        throw ERROR_OUTOFMEMORY;
    }

    return pvRet;
}

void PALMemory::Free(void *pMem)
{
    _aligned_free(pMem);
}

Size_t PALMemory::Size(void *pMem)
{
    pMem;
    // Not implemented.
    throw ERROR_CALL_NOT_IMPLEMENTED;
}
#else
void *PALMemory::Allocate(Size_t cbSrcLen)
{
    return Allocate(cbSrcLen, NULL);
}

void *PALMemory::Allocate(Size_t cbSrcLen, void *pPool)
{
    // We will create sub-pools so that memory can be re-allocated and de-allocated.
    apr_pool_t *pSubpool;
    apr_status_t status = apr_pool_create(&pSubpool, (apr_pool_t *) pPool);
    if (status != APR_SUCCESS)
    {
        throw status;
    }
    
    // Allocate the desired space in addition to storage
    // for the subpool pointer and size.
    Size_t cbAllocHeaderSize = sizeof(apr_pool_t *) + sizeof(Size_t);

    // Align to MEM_ALIGN (16-byte) boundaries by rounding up. This assumes apr_palloc()
    // returns an already aligned pointer.
    cbAllocHeaderSize = ((cbAllocHeaderSize + MEM_ALIGN - 1) / MEM_ALIGN) * MEM_ALIGN;

    void *pvRet = apr_palloc(pSubpool, cbSrcLen + cbAllocHeaderSize);
    if (pvRet == NULL)
    {
        apr_pool_destroy(pSubpool);
        throw APR_ENOMEM;
    }
       
    memcpy(pvRet, &pSubpool, sizeof(apr_pool_t *));
    memcpy((BYTE *) pvRet + sizeof(apr_pool_t *), &cbSrcLen, sizeof(Size_t));
    pvRet = (BYTE *) pvRet + cbAllocHeaderSize;

    return pvRet;
}

void *PALMemory::CAllocate(Size_t cbSrcLen)
{
    return CAllocate(cbSrcLen, NULL);
}

void *PALMemory::CAllocate(Size_t cbSrcLen, void *pPool)
{
    void *pvRet = Allocate(cbSrcLen, pPool);

    memset(pvRet, 0, cbSrcLen);

    return pvRet;
}

void *PALMemory::Reallocate(void *pMem, Size_t cbSrcLen)
{
    apr_pool_t *pSubpool;
    Size_t cbAllocHeaderSize = sizeof(apr_pool_t *) + sizeof(Size_t);
    cbAllocHeaderSize = ((cbAllocHeaderSize + MEM_ALIGN - 1) / MEM_ALIGN) * MEM_ALIGN;

    memcpy(&pSubpool, (BYTE *) pMem - cbAllocHeaderSize, sizeof(apr_pool_t *));

    apr_pool_clear(pSubpool);

    void *pvRet = apr_palloc(pSubpool, cbSrcLen + cbAllocHeaderSize);
    if (pvRet == NULL)
    {
        apr_pool_destroy(pSubpool);
        throw APR_ENOMEM;
    }
    
    memcpy(pvRet, &pSubpool, sizeof(apr_pool_t *));
    memcpy((BYTE *) pvRet + sizeof(apr_pool_t *), &cbSrcLen, sizeof(Size_t));
    pvRet = (BYTE *) pvRet + cbAllocHeaderSize;

    return pvRet;
}

void PALMemory::Free(void *pMem)
{
    apr_pool_t *pSubpool;
    Size_t cbAllocHeaderSize = sizeof(apr_pool_t *) + sizeof(Size_t);
    cbAllocHeaderSize = ((cbAllocHeaderSize + MEM_ALIGN - 1) / MEM_ALIGN) * MEM_ALIGN;

    memcpy(&pSubpool, (BYTE *) pMem - cbAllocHeaderSize, sizeof(apr_pool_t *));
    
    apr_pool_destroy(pSubpool);
}

Size_t PALMemory::Size(void *pMem)
{
    Size_t oldSize;
    Size_t cbAllocHeaderSize = sizeof(apr_pool_t *) + sizeof(Size_t);
    cbAllocHeaderSize = ((cbAllocHeaderSize + MEM_ALIGN - 1) / MEM_ALIGN) * MEM_ALIGN;

    memcpy(&oldSize, (BYTE *) pMem - cbAllocHeaderSize + sizeof(apr_pool_t *), sizeof(Size_t));

    return oldSize;
}
#endif
