/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PALSync.h"

#ifdef __TARGET_IIS__
UInt32 PALInterlocked::Increment(volatile UInt32 *pMem)
{
    return (UInt32) InterlockedIncrement((LONG *) pMem);
}

UInt32 PALInterlocked::Decrement(volatile UInt32 *pMem)
{
    return (UInt32) InterlockedDecrement((LONG *) pMem);
}

UInt32 PALInterlocked::Add(volatile UInt32 *pMem, UInt32 Value)
{
    return InterlockedExchangeAdd((LONG *) pMem, Value);
}

void PALInterlocked::Set(volatile UInt32 *pMem, UInt32 Value)
{
    InterlockedExchange((LONG *) pMem, Value);
}

UInt32 PALInterlocked::CompareSet(volatile UInt32 *pMem, UInt32 Value,  UInt32 Comparand)
{
    return InterlockedCompareExchange((LONG *) pMem, Value, Comparand);
}

PALEvent::PALEvent() :
    m_hEvent(NULL)
{
}

PALEvent::~PALEvent()
{
    Deinitialize();
}

BOOL PALEvent::Initialize()
{
    if (m_hEvent != NULL)
    {
        return FALSE;
    }

    // bManual reset is set to TRUE  since we are using
    // the example code, but this will behave differently
    // from the Linux implementation. For this purposes
    // of our service app, we shouldn't see any difference.
    m_hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

    return (m_hEvent != NULL) ? TRUE : FALSE;
}

void PALEvent::Deinitialize()
{
    if (m_hEvent != NULL)
    {
        CloseHandle(m_hEvent);
        m_hEvent = NULL;
    }
}

PALEvent::EWaitState PALEvent::Wait(DWORD dwMilliSeconds)
{
    if (m_hEvent == NULL)
    {
        return Failed;
    }

    switch(WaitForSingleObject(m_hEvent, dwMilliSeconds))
    {
    case WAIT_OBJECT_0:
        return Signaled;
        break;
    case WAIT_TIMEOUT:
        return Timeout;
        break;
    default:
        return Failed;
        break;
    }
}

BOOL PALEvent::Signal()
{
    if (m_hEvent == NULL)
    {
        return FALSE;
    }

    return SetEvent(m_hEvent);
}

#else
UInt32 PALInterlocked::Increment(volatile UInt32 *pMem)
{
    // apr_atomic_inc32() returns the old value.
    return apr_atomic_inc32(pMem) + 1;
}

UInt32 PALInterlocked::Decrement(volatile UInt32 *pMem)
{
    // We don't use apr_atomic_dec32() because it returns TRUE/FALSE condition.
    return apr_atomic_add32(pMem, -1) - 1;
}

UInt32 PALInterlocked::Add(volatile UInt32 *pMem, UInt32 Value)
{
    return apr_atomic_add32(pMem, Value);
}

void PALInterlocked::Set(volatile UInt32 *pMem, UInt32 Value)
{
    apr_atomic_set32(pMem, Value);
}

UInt32 PALInterlocked::CompareSet(volatile UInt32 *pMem, UInt32 Value,  UInt32 Comparand)
{
    return apr_atomic_cas32(pMem, Value, Comparand);
}


PALEvent::PALEvent() :
    m_pPool(NULL),
    m_pMutex(NULL),
    m_pCondVar(NULL)
{
}

PALEvent::~PALEvent()
{
    Deinitialize();
}

BOOL PALEvent::Initialize()
{
    BOOL bRetValue = TRUE;

    if (m_pCondVar != NULL)
    {
        BOOL bRetValue = FALSE;
        goto Done;
    }

    if (apr_pool_create(&m_pPool, NULL) != APR_SUCCESS)
    {
        BOOL bRetValue = FALSE;
        goto Done;
    }

    if (apr_thread_mutex_create(&m_pMutex, APR_THREAD_MUTEX_DEFAULT, m_pPool) != APR_SUCCESS)
    {
        BOOL bRetValue = FALSE;
        goto Done;
    }

    if (apr_thread_cond_create(&m_pCondVar, m_pPool) != APR_SUCCESS)
    {
        BOOL bRetValue = FALSE;
        goto Done;
    }

Done:
    if (bRetValue == FALSE)
    {
        Deinitialize();
    }

    return bRetValue;
}

void PALEvent::Deinitialize()
{
    if (m_pCondVar != NULL)
    {
        apr_thread_cond_destroy(m_pCondVar);
        m_pCondVar = NULL;
    }

    if (m_pMutex != NULL)
    {
        apr_thread_mutex_destroy(m_pMutex);
        m_pMutex = NULL;
    }

    if (m_pPool != NULL)
    {
        apr_pool_destroy(m_pPool);
        m_pPool = NULL;
    }
}

PALEvent::EWaitState PALEvent::Wait(DWORD dwMilliSeconds)
{
    apr_interval_time_t tMicroSeconds = 1000 * (apr_interval_time_t) dwMilliSeconds;

    if (m_pCondVar == NULL)
    {
        return Failed;
    }

    if (apr_thread_mutex_lock(m_pMutex) != APR_SUCCESS)
    {
        return Failed;
    }

    switch (apr_thread_cond_timedwait(m_pCondVar, m_pMutex, tMicroSeconds))
    {
    case APR_SUCCESS:
        apr_thread_mutex_unlock(m_pMutex);
        return Signaled;
        break;
    case APR_TIMEUP:
        apr_thread_mutex_unlock(m_pMutex);
        return Timeout;
        break;
    default:
        return Failed;
        break;
    }
}

BOOL PALEvent::Signal()
{
    apr_status_t status;

    if (m_pCondVar == NULL)
    {
        return FALSE;
    }

    if (apr_thread_mutex_lock(m_pMutex) != APR_SUCCESS)
    {
        return FALSE;
    }

    status = apr_thread_cond_signal(m_pCondVar);
    
    apr_thread_mutex_unlock(m_pMutex);

    return (status == APR_SUCCESS) ? TRUE : FALSE;
}
#endif
