/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PALThread.h"

PALThread::PFN PALThread::s_pfn = NULL;

void PALThread::StaticSetCallbackFunction(PFN pfn)
{
    s_pfn = pfn;
}

#ifdef __TARGET_IIS__

typedef PTP_POOL (*PFNCreateThreadpool)(PVOID);
typedef void (*PFNCloseThreadpool)(PTP_POOL);
typedef PTP_WORK (*PFNCreateThreadpoolWork)(PTP_WORK_CALLBACK, PVOID, PTP_CALLBACK_ENVIRON);
typedef void (*PFNSubmitThreadpoolWork)(PTP_WORK);
typedef void (*PFNCloseThreadpoolWork)(PTP_WORK pwk);

// making these static members of PALThread would pollute PALThread.h with platform-specific stuff
class ThreadpoolFunctionPointers
{
public:
    PFNCreateThreadpool pfnCreateThreadpool;
    PFNCloseThreadpool pfnCloseThreadpool;
    PFNCreateThreadpoolWork pfnCreateThreadpoolWork;
    PFNSubmitThreadpoolWork pfnSubmitThreadpoolWork;
    PFNCloseThreadpoolWork pfnCloseThreadpoolWork;

    ThreadpoolFunctionPointers() { Zero(); }
    void Zero() { memset(this, 0, sizeof(*this)); }
    BOOL Valid(); 
} s_pfns;

BOOL ThreadpoolFunctionPointers::Valid()
{
    return pfnCreateThreadpool
        && pfnCloseThreadpool
        && pfnCreateThreadpoolWork
        && pfnSubmitThreadpoolWork
        && pfnCloseThreadpoolWork;
}

PALThread::PALThread()
#ifdef __VISTA_THREADPOOL__
    : m_threadPool(NULL)
#endif
{
}

PALThread::~PALThread()
{
    Deinitialize();
}

BOOL PALThread::Initialize(void *pPlatformMemPool)
{
    UNREFERENCED_PARAMETER(pPlatformMemPool);

#ifdef __VISTA_THREADPOOL__
    HMODULE hKernel32 = LoadLibraryA("kernel32.dll"); // BUGBUG: refcount leak, but it's not like kernel32 could ever hope to get unloaded anyway
    if (!hKernel32)
    { // this would be unexpected on any version of Windows
        return FALSE;
    }

    s_pfns.pfnCreateThreadpool = (PFNCreateThreadpool)GetProcAddress(hKernel32, "CreateThreadpool");
    s_pfns.pfnCloseThreadpool = (PFNCloseThreadpool)GetProcAddress(hKernel32, "CloseThreadpool");
    s_pfns.pfnCreateThreadpoolWork = (PFNCreateThreadpoolWork)GetProcAddress(hKernel32, "CreateThreadpoolWork");
    s_pfns.pfnSubmitThreadpoolWork = (PFNSubmitThreadpoolWork)GetProcAddress(hKernel32, "SubmitThreadpoolWork");
    s_pfns.pfnCloseThreadpoolWork = (PFNCloseThreadpoolWork)GetProcAddress(hKernel32, "CloseThreadpoolWork");

    if (!s_pfns.Valid())
    {
        m_threadPool = NULL;
        return TRUE; // pre-Vista - use old threadpool API
    }

    m_threadPool = s_pfns.pfnCreateThreadpool(NULL);
    if (m_threadPool == NULL)
    {
        return FALSE;
    }

    // TODO: Not enabled at the momemnt. This should be set to optimal values
    // and perhaps be configurable.
    //SetThreadpoolThreadMaximum(m_threadPool, SomeMaxValue);
    //SetThreadpoolThreadMinimum(m_threadPool, SomeMinValue);

    InitializeThreadpoolEnvironment(&m_callbackEnv);
    SetThreadpoolCallbackPool(&m_callbackEnv, m_threadPool);
#endif
    return TRUE;
}

BOOL PALThread::Deinitialize()
{
#ifdef __VISTA_THREADPOOL__
    if (m_threadPool != NULL)
    {
        s_pfns.pfnCloseThreadpool(m_threadPool);
        m_threadPool = NULL;
    }
#endif
    return TRUE;
}

static DWORD WINAPI ProcessWork(_In_ LPVOID lpParameter)
{
    PALThread::s_pfn(lpParameter);
    return TRUE;
}

static VOID CALLBACK ProcessWorkVista(PTP_CALLBACK_INSTANCE Instance, PVOID context, PTP_WORK work)
{
    UNREFERENCED_PARAMETER(Instance);
    if (work)
    {
        s_pfns.pfnCloseThreadpoolWork(work);
    }
    ProcessWork(context);
}

BOOL PALThread::DispatchWork(void *pContext)
{
#ifdef __VISTA_THREADPOOL__
    if (m_threadPool)
    {
        PTP_WORK work = s_pfns.pfnCreateThreadpoolWork(ProcessWorkVista, pContext, &m_callbackEnv);
        if (work == NULL)
        {
            return FALSE;
        }
        s_pfns.pfnSubmitThreadpoolWork(work);
        return TRUE;
    }
    else
#endif
    {
        return ::QueueUserWorkItem(ProcessWork, pContext, WT_EXECUTEDEFAULT);
    }
}

#else

void APR_THREAD_FUNC ProcessWork(apr_thread_t *pThread, void *pContext)
{
    PALThread::s_pfn(pContext);
}

// TODO: To be configurable
#define THREADPOOLMAXTHREAD 32

PALThread::PALThread() :
    m_pThreadPool(NULL),
    m_pPool(NULL)
{

}

PALThread::~PALThread()
{
    Deinitialize();
}

BOOL PALThread::Initialize(void *pPlatformMemPool)
{    
    apr_status_t status;

    if (m_pThreadPool != NULL)
    {
        return FALSE;
    }

    if (pPlatformMemPool == NULL)
    {
        return FALSE;
    }
        
    m_pPool = (apr_pool_t *) pPlatformMemPool;
    
    // TODO: Need to determine thread pool sizes later.
    status = apr_thread_pool_create(&m_pThreadPool, 1, THREADPOOLMAXTHREAD, m_pPool);
    if (status != APR_SUCCESS || m_pThreadPool == NULL)
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to create threadpool");
        return FALSE;
    }

    return TRUE;
}

BOOL PALThread::Deinitialize()
{
    apr_status_t status = APR_SUCCESS;

    if (m_pThreadPool != NULL)
    {
        // NOTE: There is a slight difference in behavior between the two versions.
        // This function will wait a certain amount of time for tasks to complete.
        // Win32 will close asynchronously once all thread completes.
        int iMaxRetries = 0;
        Size_t iTotalCount = apr_thread_pool_busy_count(m_pThreadPool) + apr_thread_pool_tasks_count(m_pThreadPool);
        Size_t iPrevCount = iTotalCount;
        while (iTotalCount > 0)
        {
            // If the count has reduced, then reset the retries.
            if (iTotalCount < iPrevCount)
            {
                iMaxRetries = 0;
            }
            else
            {
                iMaxRetries++;
            }

            if (iMaxRetries > 30)
            {
                // Get out of the loop since it has exceeded the retries when count hasn't 
                // reduced.
                DEBUG_ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, m_pPool, "Max retries reached, will not wait.");
                break;
            }
            DEBUG_ap_log_perror(APLOG_MARK, APLOG_NOTICE, 0, m_pPool, "Waiting for worker to complete. Workers: %d Retry: %d", iTotalCount, iMaxRetries);
            Sleep(100);
            iPrevCount = iTotalCount;
            iTotalCount = apr_thread_pool_busy_count(m_pThreadPool) + apr_thread_pool_tasks_count(m_pThreadPool);
        }

        status = apr_thread_pool_destroy(m_pThreadPool);
        if (status != APR_SUCCESS)
        {
            DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to destroy threadpool");
        }

        m_pThreadPool = NULL;
    }

    return status == APR_SUCCESS ? TRUE : FALSE;
}

BOOL PALThread::DispatchWork(void *pContext)
{
    apr_status_t status;

    if (m_pThreadPool == NULL)
    {
        return FALSE;
    }

    status = apr_thread_pool_push(
        m_pThreadPool,
        (apr_thread_start_t) ProcessWork,
        pContext,
        APR_THREAD_TASK_PRIORITY_NORMAL,
        NULL);
    
    if (status != APR_SUCCESS)
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_ERR, status, m_pPool, "Unable to add worker to threadpool");
        return FALSE;
    }

    return TRUE;
}

#endif
