/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "SSPConfig.h"
#include "SSPCommon.h"
#include "PALFile.h"
#include "PALString.h"

static BOOL StrEq(const char* pcStart, const char* pcEnd, const char* szRef)
{
    Size_t cb = strlen(szRef);
    return (pcStart + cb == pcEnd) && !PALString::CompareInsensitive(pcStart, szRef, cb);
}

static void ReadUnsigned(const char* sz, unsigned* pu)
{
    unsigned u;
    if (1 == SSCANF_S(sz, "%u", &u))
    {
        *pu = u;
    }
}

static void ReadFloat(const char* sz, float* pf)
{
    float f;
    if (1 == SSCANF_S(sz, "%f", &f))
    {
        *pf = f;
    }
}

static const char* c_szDefaultRobotsTxtFileName = "robots.txt";

BOOL g_fGZipSitemapsByDefault = TRUE; // set to FALSE by automated test
BOOL g_fWriteToRobotsByDefault = FALSE;
BOOL g_fPublishSitemapByDefault = TRUE; 
BOOL g_fPingSearchEngines = TRUE;
BOOL g_fHelpSSPByDefault = TRUE; 

BOOL g_fUseRegistryConfig = TRUE;

void InitializeOptions()
{
    if (g_fUseRegistryConfig)
    {
        DWORD dw;
#if defined(__TARGET_APACHE__) || defined(__PUBLISH_MODE_MANUAL__)
        dw = GetSiteMapPublishMode();
#else
        if (PluginRegQueryDWORD(NULL, TEXT("SitemapPublishMode"), &dw))
#endif
        {
            switch (dw)
            {
            case 0: g_fPublishSitemapByDefault = FALSE; g_fWriteToRobotsByDefault = FALSE; break;
            case 1: g_fPublishSitemapByDefault = TRUE;  g_fWriteToRobotsByDefault = FALSE; break;
            case 2: g_fPublishSitemapByDefault = TRUE;  g_fWriteToRobotsByDefault = TRUE;  break;
            // default mode is same as case 1
            default: g_fPublishSitemapByDefault = TRUE;  g_fWriteToRobotsByDefault = FALSE; break; 
            }
        }

#if defined(__TARGET_APACHE__) || defined(__PUBLISH_MODE_MANUAL__)
        dw = GetSiteMapPingBackMode();
#else
        if (PluginRegQueryDWORD(NULL, TEXT("SitemapPingBackMode"), &dw))
#endif
        {
            switch (dw)
            {
            case 0: g_fPingSearchEngines = FALSE; break;
            case 1: g_fPingSearchEngines = TRUE; break;
            // default mode is same as case 1
            default: g_fPingSearchEngines = TRUE;  break; 
            }
        }

#if defined(__TARGET_APACHE__) || defined(__PUBLISH_MODE_MANUAL__)
        dw = GetSiteMapHelpSSPMode();
#else
        if (PluginRegQueryDWORD(NULL, TEXT("SitemapHelpSSPMode"), &dw))
#endif
        {
            switch (dw)
            {
            case 0: g_fHelpSSPByDefault = FALSE; break;
            case 1: g_fHelpSSPByDefault = TRUE; break;
                // default mode is same as case 1
            default: g_fHelpSSPByDefault = TRUE;  break; 
            }
        }
    }
}

// setting default constants
#define TCHAR_PARAM(name,default) // most strings don't have defaults
#define CHAR_PARAM(name,default)  // most strings don't have defaults
#define BOOL_PARAM(name,default)           static const BOOL c_fDefault##name = default;
#define FLOAT_PARAM(name,default)          static const float c_flDefault##name = default;
#define UINT_PARAM(name,default,lmin,lmax) static const unsigned c_uDefault##name = default;
#include "SSPConfigList.h"
#undef TCHAR_PARAM
#undef CHAR_PARAM
#undef BOOL_PARAM
#undef UINT_PARAM
#undef FLOAT_PARAM

void WriteDefaultConfig(FILE* fpOut)
{

    // We write default config and config selected in installer UI to config.ini so it will be easy for user to update
    // if fPublishSitemap==FALSE, fWriteToRobots is implicitly FALSE, we also need to document this in documentation
    InitializeOptions();

    fprintf(fpOut, "%s=%u\n", c_szPublishSitemap, g_fPublishSitemapByDefault);
    fprintf(fpOut, "%s=%u\n", c_szWriteToRobots, g_fWriteToRobotsByDefault);
    fprintf(fpOut, "%s=%u\n", c_szgzip, g_fGZipSitemapsByDefault);
    fprintf(fpOut, "%s=%u\n", c_szDecompress, c_fDefaultDecompress);
    fprintf(fpOut, "%s=%u\n", c_szPingSearchEngines, g_fPingSearchEngines);
    fprintf(fpOut, "%s=%u\n", c_szHelpSSP, g_fHelpSSPByDefault);

    fprintf(fpOut, "%s=%u\n", c_szMaxSnapshotFileMB, c_uDefaultMaxSnapshotFileMB);
    fprintf(fpOut, "%s=%u\n", c_szMaxMemoryCacheMB, c_uDefaultMaxMemoryCacheMB);
    fprintf(fpOut, "%s=%u\n", c_szSitemapGenerationPeriodHours, c_uDefaultSitemapGenerationPeriodHours);
    fprintf(fpOut, "%s=%u\n", c_szDiskMinFreeMB, c_uDefaultDiskMinFreeMB);
    fprintf(fpOut, "%s=%u\n", c_szMergeCacheMemoryMB, c_uDefaultMergeCacheMemoryMB);
    fprintf(fpOut, "%s=%u\n", c_szVisitTimeoutSec, c_uDefaultVisitTimeoutSec);


}

SSPConfig::SSPConfig()
{
    Initialize();
}

void SSPConfig::Initialize()
{
    memset(this, 0, sizeof(*this));
#define TCHAR_PARAM(name,default)
#define CHAR_PARAM(name,default)
#define BOOL_PARAM(name,default) f##name = c_fDefault##name;
#define FLOAT_PARAM(name,default) fl##name = c_flDefault##name;
#define UINT_PARAM(name,default,lmin,lmax) u##name = c_uDefault##name;
#include "SSPConfigList.h"
#undef TCHAR_PARAM
#undef CHAR_PARAM
#undef BOOL_PARAM
#undef UINT_PARAM
#undef FLOAT_PARAM
    fgzip = g_fGZipSitemapsByDefault;
    InitializeOptions();
    fPublishSitemap = g_fPublishSitemapByDefault;
    fWriteToRobots = g_fWriteToRobotsByDefault;
    fHelpSSP = g_fHelpSSPByDefault;
    fPingSearchEngines = g_fPingSearchEngines;

    PALString::Convert(wszRobotsTxtFilename, ARRAYSIZE(wszRobotsTxtFilename), c_szDefaultRobotsTxtFileName, strlen(c_szDefaultRobotsTxtFileName));
}

static void KillCRLF(char* p)
{
    while (*p != '\r' && *p != '\n' && *p != '\0') p++;
    *p = '\0';
}

#ifdef APR_HAVE_WINDOWS_H
#pragma warning(disable:4296) // "expression is always false" - happens in UINT_PARAM when lmin is 0
#endif

BOOL ReadSSPConfig(const TCHAR* wszFile, SSPConfig* pConfig)
{
    FILE* fp;

    if (0 != PALFile::Open(&fp, wszFile, TEXT("rt")))
    {
        return FALSE;
    }

    BOOL fWWWRootFound = FALSE, fWWWHostFound = FALSE;

    while (!feof(fp))
    {
        char s[1024];
        s[0] = '\0';
        if (!fgets(s, sizeof(s), fp))
        {
            break;
        }
        KillCRLF(s);
        const char* pcEq = strchr(s, '=');
        if (!pcEq)
        {
            continue;
        }

        if (StrEq(s, pcEq, c_szwwwroot))
        {
            const char* szRoot = pcEq + 1;

            if (PALString::Convert(pConfig->wszwwwroot, ARRAYSIZE(pConfig->wszwwwroot), szRoot, strlen(szRoot)) != 0)
            {
                return FALSE;
            }

            fWWWRootFound = TRUE;
        }
        if (StrEq(s, pcEq, "wwwhost"))
        {
            const char* szRoot = pcEq + 1;
            Size_t cch = strlen(szRoot);
            if (cch >= ARRAYSIZE(pConfig->szwwwhost))
            {
                return FALSE;
            }
            memcpy(pConfig->szwwwhost, szRoot, cch + 1);
            fWWWHostFound = TRUE;
        }
        else if (StrEq(s, pcEq, c_szRobotsTxtFilename))
        {
            const char* szrobotsTxtFileName = pcEq + 1;

            if (PALString::Convert(pConfig->wszRobotsTxtFilename, ARRAYSIZE(pConfig->wszRobotsTxtFilename), szrobotsTxtFileName, strlen(szrobotsTxtFileName)) != 0)
            {
                return FALSE;
            }
        }

#define TCHAR_PARAM(name,default) // strings handled case-by-case above
#define CHAR_PARAM(name,default) // strings handled case-by-case above
#define BOOL_PARAM(name,default) \
        else if (StrEq(s, pcEq, c_sz##name)) \
        { \
            unsigned u = pConfig->f##name; \
            ReadUnsigned(pcEq + 1, &u); \
            pConfig->f##name = u; \
        }
#define FLOAT_PARAM(name,default) \
        else if (StrEq(s, pcEq, c_sz##name)) \
        { \
            ReadFloat(pcEq + 1, &pConfig->fl##name); \
        }
#define UINT_PARAM(name,default,lmin,lmax) \
        else if (StrEq(s, pcEq, c_sz##name)) \
        { \
            ReadUnsigned(pcEq + 1, &pConfig->u##name); \
            if (pConfig->u##name < lmin) \
                pConfig->u##name = lmin; \
            if (pConfig->u##name > lmax && lmax) \
                pConfig->u##name = lmax; \
        }
#include "SSPConfigList.h"
#undef TCHAR_PARAM
#undef CHAR_PARAM
#undef BOOL_PARAM
#undef UINT_PARAM
#undef FLOAT_PARAM
    }

    fclose(fp);

    return fWWWRootFound && fWWWHostFound;
}

BOOL UpdateSSPConfig(const TCHAR* wszFile, const char* szName, const char* szValue)
{
    BOOL fRet = FALSE;
    FILE *fpIn = NULL, *fpOut = NULL;
    TCHAR wszTmpFile[MAX_PATH];
    if (0 >= STPRINTF_S(wszTmpFile, ARRAYSIZE(wszTmpFile), STRFMT_TLS TEXT(".tmp"), wszFile))
    {
        goto Exit;
    }

    if ((0 != PALFile::Open(&fpIn, wszFile, TEXT("rt"))) || !fpIn)
    {
        goto Exit;
    }
    if ((0 != PALFile::Open(&fpOut, wszTmpFile, TEXT("wt"))) || !fpOut)
    {
        goto Exit;
    }

    while (!feof(fpIn))
    {
        char s[1024];
        s[0] = '\0';
        if (!fgets(s, sizeof(s), fpIn))
        {
            break;
        }
        KillCRLF(s);
        const char* pcEq = strchr(s, '=');
        if (pcEq && StrEq(s, pcEq, szName))
        {
            fprintf(fpOut, "%s=%s\n", szName, szValue);
        }
        else
        {
            fprintf(fpOut, "%s\n", s);
        }
    }
    fclose(fpIn);  fpIn = NULL;
    fclose(fpOut); fpOut = NULL;

    if (!PALFile::Move(wszTmpFile, wszFile))
    {
        goto Exit;
    }

    fRet = TRUE;
Exit:
    if (fpIn) fclose(fpIn); fpIn = NULL;
    if (fpOut) fclose(fpOut); fpOut = NULL;
    return fRet;
}

void PrintSSPConfig(FILE* fp, const SSPConfig* pConfig)
{
    if (!pConfig) return;

#define TCHAR_PARAM(name,default)
#define CHAR_PARAM(name,default)
#define BOOL_PARAM(name,default)  fprintf(fp, "<!-- Setting_%s=%u -->\n", #name, pConfig->f##name);
#define FLOAT_PARAM(name,default) fprintf(fp, "<!-- Setting_%s=%f -->\n", #name, pConfig->fl##name);
#define UINT_PARAM(name,default,lmin,lmax)  fprintf(fp, "<!-- Setting_%s=%u -->\n", #name, pConfig->u##name);
#include "SSPConfigList.h"
#undef TCHAR_PARAM
#undef CHAR_PARAM
#undef BOOL_PARAM
#undef UINT_PARAM
#undef FLOAT_PARAM
}

int SSPConfigCppLineCount()
{
    return __LINE__;
}
