/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "SSPHash.h"

//
// Mix(): cause every bit of a,b,c to affect 32 bits of a,b,c
// both forwards and in reverse.  Same for pairs of bits in a,b,c.
// This can be used along with Final() to hash a fixed number of
// 4-byte integers, for example see the implementation of Guid().
//
inline static void Mix( UInt32& a, UInt32& b, UInt32& c)
{
    a -= c;  a ^= _rotl(c, 4);  c += b;
    b -= a;  b ^= _rotl(a, 6);  a += c;
    c -= b;  c ^= _rotl(b, 8);  b += a;
    a -= c;  a ^= _rotl(c,16);  c += b;
    b -= a;  b ^= _rotl(a,19);  a += c;
    c -= b;  c ^= _rotl(b, 4);  b += a;
}

//
// Final: cause every bit of a,b,c to affect every bit of c, only forward.
// Same for pairs of bits in a,b,c.  It also causes b to be an OK hash.
// This is a good way to hash 1 or 2 or 3 integers:
//   a = k1; b = k2; c = 0;
//   CsHash32::Final(a,b,c);
// Use c (and maybe b) as the hash value
//
inline static void Final( UInt32& a, UInt32& b, UInt32& c)
{
    c ^= b; c -= _rotl(b,14);
    a ^= c; a -= _rotl(c,11);
    b ^= a; b -= _rotl(a,25);
    c ^= b; c -= _rotl(b,16);
    a ^= c; a -= _rotl(c,4);
    b ^= a; b -= _rotl(a,14);
    c ^= b; c -= _rotl(b,24);
}

// Compute2: Compute two hashes of an array of bytes.
// The first hash is slightly better mixed than the second hash.
static void Compute2 (
    const void *pData,  // byte array to hash; may be null if uSize==0
    Size_t      uSize,  // length of pData
    UInt32      uSeed1, // first seed
    UInt32      uSeed2, // second seed
    UInt32     *uHash1, // OUT: first hash (may not be null)
    UInt32     *uHash2) // OUT: second hash (may not be null)
{
    UInt32 a,b,c;
    const UInt32 *k = (const UInt32 *)pData;       // read 32-bit chunks
    const UInt8  *k1;

    // Set up the internal state
    a = b = c = 0xdeadbeef + ((UInt32)uSize) + uSeed1;
    c += uSeed2;

    // all but last block: aligned reads and affect 32 bits of (a,b,c)
    while (uSize > 12)
    {
        a += k[0];
        b += k[1];
        c += k[2];
        Mix(a,b,c);
        uSize -= 12;
        k += 3;
    }
    
    // handle the last (probably partial) block
    k1 = (const UInt8 *)k;
    switch(uSize)
    {
    case 12: c+=k[2]; b+=k[1]; a+=k[0]; break;
    case 11: c+=((UInt32)k1[10])<<16; __fallthrough;
    case 10: c+=((UInt32)k1[9])<<8; __fallthrough;
    case 9 : c+=(UInt32)k1[8]; __fallthrough;
    case 8 : b+=k[1]; a+=k[0]; break;
    case 7 : b+=((UInt32)k1[6])<<16; __fallthrough;
    case 6 : b+=((UInt32)k1[5])<<8; __fallthrough;
    case 5 : b+=((UInt32)k1[4]); __fallthrough;
    case 4 : a+=k[0]; break;
    case 3 : a+=((UInt32)k1[2])<<16; __fallthrough;
    case 2 : a+=((UInt32)k1[1])<<8; __fallthrough;
    case 1 : a+=k1[0]; break;
    case 0 : 
        *uHash1 = c;
        *uHash2 = b;
        return;
    }

    Final(a,b,c);
    *uHash1 = c;
    *uHash2 = b;
    return;
}

UInt64 SSPHash(const void* pData, Size_t uSize, UInt64 uSeed)
{
    UInt32 uHash1, uHash2;
    Compute2( pData, uSize, (UInt32) uSeed, (UInt32) (uSeed >> 32), &uHash1, &uHash2);
    return uHash1 | (((UInt64)uHash2) << 32);
}

UInt32 SSPHash32(const void* pData, Size_t uSize, UInt64 uSeed)
{
    UInt32 uHash1, uHash2;
    Compute2( pData, uSize, (UInt32) uSeed, (UInt32) (uSeed >> 32), &uHash1, &uHash2);
    return uHash1;
}