/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "SSPTypes.h"
#include "PAL.h"

#define myassert(cond) { if (!(cond)) { printf("%u %s", __LINE__, #cond); exit(-1); } }

typedef unsigned Ptr; // 32-bit offset to save memory on 64-bit machine
typedef unsigned SizeT;

template <class T> class LRUCache
{
    typedef struct
    {
        Ptr nextInBucket; // hash table
        Ptr newer; // doubly linked list
        Ptr older; // doubly linked list
        Hash hash;
        T data;
    } Node;

    Ptr m_oldest;
    Ptr m_newest;

    Ptr* m_table;
    SizeT m_cTable;

    Node* m_nodes;
    SizeT m_cNodes;
    SizeT m_cUsed;

public:
    bool Initialize(void* pbMemory, unsigned cbMemory);
    // if fCreate is false, Access returns NULL and do not create a new entry
    T* Access(Hash hash, BOOL fCreate = TRUE);
    SizeT Count() { return m_cUsed - 1; }
    void Enumerate(void* ctx, void (*pfn)(void* ctx, Hash, T*));
    void Verify(bool verbose);
};

template <class T> bool LRUCache<T>::Initialize(void* pbMemory, SizeT cbMemory)
{
    memset(this, 0, sizeof(*this));
    memset(pbMemory, 0, cbMemory);

    const SizeT cAvgNodesPerBucket = 4;
    SizeT NodeToTableMemoryRatio = cAvgNodesPerBucket * sizeof(Node) / sizeof(Ptr);
    SizeT cbTable = cbMemory / (NodeToTableMemoryRatio + 1);
    SizeT cbNodes = cbMemory - cbTable;
    m_cTable = cbTable / sizeof(Ptr);
    m_cNodes = cbNodes / sizeof(Node);
    //printf("%u nodes\n", m_cNodes);
    
    m_table = (Ptr*)pbMemory;
    m_nodes = (Node*)(m_table + m_cTable);
    myassert((char*)(m_nodes + m_cNodes) <= (char*)pbMemory + cbMemory);

    m_cUsed = 1; // first slot wasted to let 0 offset mean "null"
    m_oldest = 0;
    m_newest = 0;
    
    return true;
}

#define NODEPTR(index) (index ? (m_nodes + index) : NULL)

template <class T> T* LRUCache<T>::Access(Hash hash, BOOL fCreate)
{
    // look for existing node
    Ptr node = m_table[hash % m_cTable];
    while (node)
    {
        Node* p = NODEPTR(node);
        if (p->hash == hash)
        {
            break;
        }
        node = p->nextInBucket;
    }

    if (node) // found - remove from current location
    {
        myassert(m_newest);
        myassert(m_oldest);
        Node* p = NODEPTR(node);

        if (p->newer)
        {
            NODEPTR(p->newer)->older = p->older;
        }
        else
        {    
            myassert(m_newest == node);
            m_newest = p->older;
        }

        if (p->older)
        {
            NODEPTR(p->older)->newer = p->newer;
        }
        else
        {
            myassert(m_oldest == node);
            m_oldest = p->newer;
        }
    }
    else // not found - allocate or recycle
    {
        if(!fCreate)
        {
            // shouldn't create a new entry if not found
            return NULL;
        }

        if (m_cUsed < m_cNodes) // allocate new
        {
            node = m_cUsed++;
        }
        else // kick out the oldest
        {
            myassert(m_oldest);
            node = m_oldest;
            Node* p = NODEPTR(node);

            // remove from hash table
            Hash deleteHash = p->hash;
            SizeT deleteSlot = deleteHash % m_cTable;
            Ptr deleteNode = m_table[deleteSlot];
            Ptr prev = (Ptr) NULL;
            while (deleteNode)
            {
                Node* pDelete = NODEPTR(deleteNode);
                if (pDelete->hash == deleteHash)
                {
                    break;
                }
                prev = deleteNode;
                deleteNode = pDelete->nextInBucket;
            }
            myassert(deleteNode);
            Ptr nextInBucket = NODEPTR(deleteNode)->nextInBucket;
            if (prev)
            {
                NODEPTR(prev)->nextInBucket = nextInBucket;
            }
            else
            {
                m_table[deleteSlot] = nextInBucket;
            }


            // remove from doubly linked list
            myassert(!p->older);
            myassert(p->newer);
            Node* pNewer = NODEPTR(p->newer);
            myassert(pNewer->older == m_oldest);
            pNewer->older = (Ptr) NULL;
            m_oldest = p->newer;
        }

        // initialize allocated/recycled node
        Node* p = NODEPTR(node);
        p->hash = hash;
        memset(&(p->data), 0, sizeof(p->data));

        // insert into hashtable
        SizeT slot = hash % m_cTable;
        p->nextInBucket = m_table[slot];
        m_table[slot] = node;
    }

    Node* p = NODEPTR(node);
    myassert(p->hash == hash);

    // insert in front
    p->newer = (Ptr) NULL;
    p->older = m_newest;
    if (m_newest)
    {
        NODEPTR(m_newest)->newer = node;
    }
    m_newest = node;
    if (!m_oldest) m_oldest = node;

    return &p->data;
}

template <class T> void LRUCache<T>::Enumerate(void* ctx, void (*pfn)(void* ctx, Hash, T*))
{
    Ptr node = m_oldest;
    while (node)
    {
        Node* p = NODEPTR(node);
        pfn(ctx, p->hash, &p->data);
        node = p->newer;
    }
}

template <class T> void LRUCache<T>::Verify(bool verbose)
{
    Ptr node = m_oldest;
    SizeT count = 0;
    while (node)
    {
        count++;
        Node* p = NODEPTR(node);
        Ptr newer = p->newer;
        if (!newer)
        {
            myassert(node == m_newest);
            break;
        }
        myassert(NODEPTR(newer)->older == node);
        node = newer;
    }
    myassert(m_cUsed == count + 1);

    SizeT maxChain = 0;
    count = 0;
    for (SizeT u = 0; u < m_cTable; u++)
    {
        SizeT chain = 0;
        Ptr node = m_table[u];
        while (node)
        {
            chain++;
            Node* p = NODEPTR(node);
            myassert(p->hash % m_cTable == u);
            node = p->nextInBucket;
        }
        count += chain;
        if (maxChain < chain) maxChain = chain;
    }
    myassert(m_cUsed == count + 1);
    if (verbose) printf("%u nodes, %u longest chain\n", count, maxChain);
}

#if 0

unsigned Log2(unsigned u)
{
    unsigned ret = 0;
    while (u)
    {
        u >>= 1;
        if (u) ret++;
    }
    return ret;
}

void ValidateSequence(void* ctx, Hash hash, unsigned* pval)
{
    unsigned* pu = (unsigned*)ctx;
    //printf("%u %u\n", *pval, *pu + 1);
    myassert(*pval >= *pu + 1);
    *pu = *pval;
}

int __cdecl main()
{
    LRUCache<unsigned> cache;
    bool fRet = cache.Initialize(65536);  myassert(fRet);
    cache.Verify(true);
    for (unsigned u = 0; u < 1000000000; u++)
    {
        bool verbose = (0 == u % 100000);

        Hash h = rand() /* % (1 + Log2(u) / 2) */;
        if (verbose) printf("insert %u %u\n", u, h);
        unsigned* pu = cache.Access(h);
        *pu = u;
        
        SizeT c = cache.Count();
        unsigned tmp = -1; // last enumerated value
        cache.Enumerate(&tmp, ValidateSequence);
        cache.Verify(verbose);
        if (verbose) printf("%u\n", u);
    }
}

#endif
