/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "SSPRobots.h"
#include "basic_types.h"
#include "PALFile.h"
#include "PALString.h"

static BOOL CheckFileForLines(const TCHAR* wszFile, BOOL* pfMissingEOL, unsigned cLines, const char** rszLines, BOOL* rfFound)
{
    FILE* fp = NULL;

    if (0 != PALFile::Open(&fp, wszFile, TEXT("rt")) || !fp)
    {
        return FALSE;
    }

    memset(rfFound, 0, sizeof(rfFound[0]) * cLines);
    *pfMissingEOL = FALSE;
    while (!feof(fp))
    {
        char s[1024];
        if (!fgets(s, sizeof(s), fp)) continue;
        if (!s[0]) break;
        size_t cch = strlen(s);
        *pfMissingEOL = ('\n' != s[cch-1]);

        for (unsigned u = 0; u < cLines; u++)
        {
            size_t cchLine = strlen(rszLines[u]);
            if (cch >= cchLine && !memcmp(rszLines[u], s, cchLine)) // prefix check - s[] likely contains trailing \r\n
            {
                rfFound[u] = TRUE;
            }
        }
    }
    fclose(fp);
    return TRUE;
}

BOOL UpdateBlacklistFile(const TCHAR* wszFile)
{
    const int cEntries = 2;
    BOOL rfFound[cEntries];
    const char* rszEntries[] = { "Disallow: /*.css", "Disallow: /*.js" };

    BOOL fMissingEOL = FALSE;
    BOOL fRet = FALSE;
    FILE* fp = NULL;

    if (!CheckFileForLines(wszFile, &fMissingEOL, cEntries, rszEntries, rfFound))
    {
        return FALSE;
    }
    
    BOOL fNeedUpdate = FALSE;
    for (unsigned u = 0; u < cEntries; u++)
    {
        if (!rfFound[u])
        {
            fNeedUpdate = TRUE;
            break;
        }
    }
    
    if (!fNeedUpdate)
    {
        return TRUE;
    }

    if ((0 != PALFile::Open(&fp, wszFile, TEXT("at"))) || !fp)
    {
        goto Exit;
    }

    if (fMissingEOL)
    {
        if (0 >= fprintf(fp, "\n"))
        {
            goto Exit;
        }
    }

    for (unsigned u = 0; u < cEntries; u++)
    {
        if (!rfFound[u])
        {
            if (0 >= fprintf(fp, "%s\n", rszEntries[u]))
            {
                goto Exit;
            }
        }
    }
    fRet = TRUE;

Exit:
    if (fp) fclose(fp);
    return fRet;

}

void WriteBlacklistFile(const TCHAR* wszFile)
{
    FILE* fp = NULL;

    if (0 != PALFile::Open(&fp, wszFile, TEXT("wt")) || !fp)
    {
        goto Error;
    }

    fputs("# This file configures the URL filter used by the sitemap generation plugin for sitemap generation purposes.\n", fp);
    fputs("# If the site has robts.txt, the plugin will read it and ignore URLs disallowed by robots.txt for the default (*) user-agent.\n", fp);
    fputs("# If you want to exclude additional URLs from being processed by the plugin without affecting robots.txt, you can put additional exclusion rules in this file.\n", fp);
    fputs("# The format of this file is similar to robots.txt, except there are no user-agent sections - all Disallow: and Allow: lines are loaded.\n", fp);
    fputs("# E.g., a line 'Disallow: /cart' will prevent all URLs starting with the prefix '/cart' from being processed - unless overridden by a more specific rule like 'Allow: /cartography'.\n", fp);
    fputs("# A line starting with a pound sign is treated as a comment.\n", fp);
    fputs("\n", fp);
    fputs("# Changes to this file take effect when the plugin is restarted (e.g. by restarting the parent web server process).\n", fp);

Error:
    if (fp) fclose(fp);
}



#ifndef max
#define max(a,b) (((a)>(b)) ? (a) : (b))
#endif

BOOL Robots::RobotsData::Finalize()
{
    if (!m_allowSpace || !m_disallowSpace)
    {
        return FALSE;
    }
    if (!m_disallowOffset)
    {
        return FALSE;
    }

    m_cbAllow = m_allowOffset;
    m_cbDisallow = m_disallowOffset;

    return TRUE;
}


// Locates the end of a line, whether it ends with \r\n, \n, or \r
// Fills in the address of the end of line character, or the address of the terminating NULL
// and also fills in the address of a comment indicator if one is found, or the address of the
static void DetermineLineExtent(const char **ppchLine, 
                                const char *pchMax, 
                                const char **ppchEndOfRead, 
                                const char **ppchNextLine)
{
    // find the first occurrence of either \r or \n
    // if \r is found and is followed by \n, then return pointing at \n

    // first skip white space, it is acceptance to skip whole lines of white space
    while (*ppchLine < pchMax && isspace((unsigned char)**ppchLine))
    {
        (*ppchLine)++;
    }

    const char *pchTemp = *ppchLine;
    bool fFoundComment = false;
    while (pchTemp < pchMax && *pchTemp != '\r' && *pchTemp != '\n')
    {
        if (!fFoundComment && *pchTemp == '#')
        {
            fFoundComment = true;
            *ppchEndOfRead = pchTemp;
        }
        pchTemp++;
    }

    if (!fFoundComment)
    {
        *ppchEndOfRead = pchTemp;
    }

    // set location at the byte before the comment
    // in preparation for the trailing space check
    // it can move to one byte less than the beginning of the line
    // if the line begins with a comment
    (*ppchEndOfRead)--;

    // seek back over white space from **ppchEndOfRead
    while (*ppchEndOfRead >= *ppchLine && isspace((unsigned char)**ppchEndOfRead))
    {
        (*ppchEndOfRead)--;
    }

    // put the end of line marker at the byte just past the readable space
    (*ppchEndOfRead)++;

    // pass back a pointer to the character on which the next line begins
    while (pchTemp < pchMax && (*pchTemp == '\r' || *pchTemp == '\n'))
    {
        pchTemp++;
    }

    *ppchNextLine = pchTemp;
}


const char * Robots::RobotsData::szDisallow = "disallow";
const char * Robots::RobotsData::szAllow = "allow";
const char * Robots::RobotsData::szUserAgent = "user-agent";
const char * Robots::RobotsData::szCrawlDelay = "crawl-delay";
const char * Robots::RobotsData::szSitemap = "sitemap";
const char Robots::RobotsData::s_rgszUserAgents[Robots::RobotsData::c_maxRobotsUserAgents][256] = { "*" } ;
const size_t Robots::RobotsData::s_rgcbUserAgents[Robots::RobotsData::c_maxRobotsUserAgents] = { 1 };

Robots::RobotsData::RobotsData()
   : m_disallowSpace(NULL), m_allowSpace(NULL), m_fOwnMemory(FALSE)
{
}

Robots::RobotsData::~RobotsData()
{
    if (m_fOwnMemory)
    {
        if (m_allowSpace) free(m_allowSpace);
        if (m_disallowSpace) free(m_disallowSpace);
    }
}

BOOL Robots::RobotsData::Initialize(IAllocator* pAlloc, size_t cbData)
{
    MallocAllocator a;
    memset(this, 0, sizeof(*this)); // don't add virtuals..
    
    m_fOwnMemory = FALSE; // constructor not called in shared memory scenario
    if (!pAlloc)
    {
        pAlloc = &a;
        m_fOwnMemory = TRUE;
    }

    m_cbAllow = m_cbDisallow = cbData;
    m_allowOffset = m_disallowOffset = 0;
    m_allowSpace = (char*)pAlloc->Allocate(m_cbAllow);
    m_disallowSpace = (char*)pAlloc->Allocate(m_cbDisallow);
    return m_allowSpace && m_disallowSpace;
}

// Parses a downloaded robots.txt file to find DISALLOW and ALLOW rules.
// Writes the rules to m_workspace and sets pointers in probotsdata.
// This must be called with a null-terminated string.

// rgszSitemapUrls must point to an array of c_maxSitemapUrls buffers of size INTERNET_MAX_URL_LENGTH + 1
// The number of sitemap URLs found is returned in *pnumSitemapUrls
void Robots::RobotsData::ParseRobotsRules(const char *pchBuffer, size_t cbData)
{
    const char *pchMax = pchBuffer + cbData;
    //LogAssert(*pchMax == '\0');

    // Read the buffer line by line
    const char *pchLine = pchBuffer;
    const char *pchEndOfRead;
    const char *pchNextLine;

    const UInt32 c_maxBlocks = 10;
    const char *rgpchUserAgentRulesStart[c_maxRobotsUserAgents][c_maxBlocks];
    const char *rgpchUserAgentRulesEnd[c_maxRobotsUserAgents][c_maxBlocks];
    memset(rgpchUserAgentRulesStart, 0, sizeof(rgpchUserAgentRulesStart));
    memset(rgpchUserAgentRulesEnd, 0, sizeof(rgpchUserAgentRulesEnd));

    // Flags for whether the last found block is stored for each agent; it is needed to determine whether we need to terminiate 
    // the last block when a new block is found; it also helps prevent duplicated rules blocks when the same agent shows up 
    // multiple times in the same UserAgent name list block
    bool rgfLastBlockStored[c_maxRobotsUserAgents]; 
    memset(rgfLastBlockStored, 0, sizeof(rgfLastBlockStored));
    
    UInt32 rgNumBlocks[c_maxRobotsUserAgents];
    memset(rgNumBlocks, 0, sizeof(rgNumBlocks));

    // flag for whether pchLine is in a continuous UserAgent name block like "UserAgent: Bot1\r\nUserAgent:Bot2\r\n...". 
    // Will be used to determine whether to terminate the previous rules block
    bool fInUserAgentNamesBlock = false; 
    
    // on the first pass, we will identify sitemap URLs (which can occur anywhere in the file)
    // and the user-agent section relevant to our crawl

    // locate our user-agent, even if surrounded by quotes
    while (pchLine < pchMax)
    {
        DetermineLineExtent(&pchLine, pchMax, &pchEndOfRead, &pchNextLine);
        // skip empty lines
        if (pchLine >= pchEndOfRead)
        {
            pchLine = pchNextLine;
            continue;
        }

        size_t cbLine = pchEndOfRead - pchLine;
        bool fSitemap = false;
        bool fUserAgent = false;
        if (cbLine > strlen(szSitemap)
            && PALString::CompareInsensitive(pchLine, szSitemap, strlen(szSitemap)) == 0)
        {
            pchLine += strlen(szSitemap);
            fSitemap = true;
        }
        else if (cbLine > strlen(szUserAgent)
            && PALString::CompareInsensitive(pchLine, szUserAgent, strlen(szUserAgent)) == 0)
        {
            pchLine += strlen(szUserAgent);
            fUserAgent = true;
        }

        if (!fUserAgent)
        {
            // the previous continuous UserAgent name block ends
            fInUserAgentNamesBlock = false; 
        }

        if (!fSitemap && !fUserAgent)
        {
            pchLine = pchNextLine;
            continue;
        }

        bool fValid = true;
        // if something other than a tab or space is found before 
        // the colon, abort.
        while (pchLine < pchEndOfRead && *pchLine != ':')
        {
            if (!isspace((unsigned char)*pchLine))
            {
                fValid = false;
                break;
            }
            pchLine++;
        }
        if (!fValid)
        {
            pchLine = pchNextLine;
            continue;
        }

        // move past the colon
        pchLine++;
        // skip spaces, tabs, and quotes again
        while (pchLine < pchEndOfRead 
            && (isspace((unsigned char)*pchLine) || *pchLine == '\"' || *pchLine == '\''))
        {
            pchLine++;
        }
        if (fSitemap)
        {
            // we have reached the URL
            // size_t cbUrl = pchEndOfRead - pchLine;
        }
        else if (fUserAgent)
        {
            // the last UserAgent list and its rules block ends when a new "UserAgent:" block is found
            if (!fInUserAgentNamesBlock) 
            {
                for (UInt32 i = 0; i < s_numUserAgents; i++)
                {
                    // terminate the last rules blocks
                    if (rgfLastBlockStored[i])
                    {
                        rgpchUserAgentRulesEnd[i][rgNumBlocks[i] - 1] = pchNextLine;
                    }
                    // reset flags in prep for new block; this also prevents unnecessary terminates which could insert not-applicable rules 
                    // from invalid agents
                    rgfLastBlockStored[i] = false;
                }
            }                  
            fInUserAgentNamesBlock = true;
            // we have reached the user-agent string
            // first, is there a prefix match on one of the listed user agents?
            for (UInt32 i = 0; i < s_numUserAgents; i++)
            {
                // loop over all UserAgent names in each NULL separated list
                const char *pchAgentName = s_rgszUserAgents[i];
                while (pchAgentName < s_rgszUserAgents[i] + s_rgcbUserAgents[i])
                {
                    size_t cbAgentName = strlen(pchAgentName);
                    fValid = false;
                    if ( PALString::CompareInsensitive(pchLine, pchAgentName, cbAgentName) == 0)
                    {
                         // must preserve pchLine for the next iteration through the loop
                        const char *pchTemp = pchLine + cbAgentName;

                        // we have a prefix match
                        // are there trailing characters that invalidate the match?
                        // If we encounter a character other than a single or double quote or '*' prior to hitting a
                        // space, a tab, a slash (either direction), an open parenthesis, or the end of the readable line
                        // the match is invalidated.
                        fValid = true;
                        while (fValid && pchTemp < pchEndOfRead && 
                            !isspace((unsigned char)*pchTemp) && *pchTemp != '(' && *pchTemp != '/' && *pchTemp != '\\')
                        {
                            if (*pchTemp != '\"' && *pchTemp != '\'' && *pchTemp != '*')
                            {
                                fValid = false;
                            }
                            pchTemp++;
                        }
                        if (fValid) 
                        {
                            // all names in the list are equivalent, we only need to find one name to match this UserAgent
                            break; 
                        }
                    }
                    // move on to the next name in the list
                    pchAgentName += cbAgentName + 1; 
                }
                if (fValid) 
                {
                    // found a valid UserAgent name we are interested in
                    m_fAgentFound[i] = true;
                    // store the rules block for this agent if it is not already stored (i.e., this agent has no duplicated entry in earlier 
                    // part of the same agent name lists) and there is space for it
                    if (!rgfLastBlockStored[i] && rgNumBlocks[i] < c_maxBlocks) 
                    {
                        rgpchUserAgentRulesStart[i][rgNumBlocks[i]] = pchNextLine;
                        rgpchUserAgentRulesEnd[i][rgNumBlocks[i]] = NULL;
                        rgNumBlocks[i]++;
                        rgfLastBlockStored[i] = true;
                    }
                    
                    // no need to look at other UserAgent for this line; break from the for loop
                    break; 
                }
            }
        }

        pchLine = pchNextLine;
    }

    // terminate the very last blocks
    for (UInt32 i = 0; i < s_numUserAgents; i++)
    {
        if (rgfLastBlockStored[i] && rgpchUserAgentRulesEnd[i][rgNumBlocks[i] - 1] == NULL)
        {
            rgpchUserAgentRulesEnd[i][rgNumBlocks[i] - 1] = pchMax;
        }
    }

    for (UInt32 i = 0; i < s_numUserAgents; i++)
    {
        for (UInt32 j = 0; j < rgNumBlocks[i]; j++)
        {
            AddRules(i, rgpchUserAgentRulesStart[i][j], rgpchUserAgentRulesEnd[i][j]);
        }
    }
}

// Given a block of rules, add them to our internal structures
// It is guaranteed that the string beginning at pchBuffer will be null-terminated
// either at or before pchMax.
// The rules for MsnBot must be added before *, so the CrawlDelay for MsnBot will be stored if available
void Robots::RobotsData::AddRules(const UInt32 idUserAgent, const char *pchBuffer, const char *pchMax)
{
    if (pchBuffer >= pchMax)
    {
        return;
    }

    m_fAgentFound[idUserAgent] = true;

    // pchBuffer points to the beginning of the range to search
    // The range is null-terminated, possibly at pchMax
    const char *pchLine = pchBuffer;
    const char *pchEndOfRead;
    const char *pchNextLine;
    //bool fFoundCrawlDelay = false;

    while (pchLine < pchMax)
    {
        DetermineLineExtent(&pchLine, pchMax, &pchEndOfRead, &pchNextLine);
        // skip empty lines
        if (pchLine >= pchEndOfRead)
        {
            pchLine = pchNextLine;
            continue;
        }

        size_t cbLine = pchEndOfRead - pchLine;
        if (cbLine > strlen(szAllow)
            && PALString::CompareInsensitive(pchLine, szAllow, strlen(szAllow)) == 0)
        {
            pchLine += strlen(szAllow);
            AddOneRule(idUserAgent, pchLine, pchEndOfRead, ALLOW);
        }
        else if (cbLine > strlen(szDisallow)
            && PALString::CompareInsensitive(pchLine, szDisallow, strlen(szDisallow)) == 0)
        {
            pchLine += strlen(szDisallow);
            AddOneRule(idUserAgent, pchLine, pchEndOfRead, DISALLOW);
        }
#if 0
        // there should be only one of these in each block
        // if there are more than one we'll just use the first one in the last block
        // The function call for * should be after MsnBot so its CrawlDelay will be used only if MsnBot does not have one
        else if ((idUserAgent == s_MsnBotIndex || idUserAgent == s_StarBotIndex) 
            && m_crawlDelay == c_invalidCrawlDelay
            && !fFoundCrawlDelay && cbLine > strlen(szCrawlDelay)
            && PALString::CompareInsensitive(pchLine, szCrawlDelay, strlen(szCrawlDelay)) == 0)
        {
            pchLine += strlen(szCrawlDelay);
            // skip white space and the colon
            while (pchLine < pchEndOfRead && (*pchLine == ':' || isspace((unsigned char)*pchLine)))
            {
                pchLine++;
            }
            // absorb the decimal number that follows
            m_crawlDelay = 0;
            while (pchLine < pchEndOfRead && isdigit((unsigned char)*pchLine))
            {
                m_crawlDelay *= 10;
                m_crawlDelay += *pchLine - '0';
                pchLine++;
            }
            fFoundCrawlDelay = true;
            
        }
#endif
        pchLine = pchNextLine;
    }
}

// Locates the extent of the rule string and adds it to the data set
// This function assumes it is called from within the critical section
// This function must be called with all rules for same UserAgent grouped together
void Robots::RobotsData::AddOneRule(const UInt32 idUserAgent, const char *pchBuffer, const char *pchMax, RuleType type)
{
    if (pchBuffer >= pchMax)
    {
        return;
    }

    // now find the beginning of the rule
    // pchBuffer points at the char immediately after the end of allow
    // or disallow, so skip over : and whitespace
    const char *pchRule = pchBuffer;
    while (pchRule < pchMax && (*pchRule == ':' || isspace((unsigned char)*pchRule)))
    {
        pchRule++;
    }

    // it is possible for there to be no rule here
    if (pchRule >= pchMax)
    {
        return;
    }

    // pchRule now points at the first byte of the rule
    // note that this function is only called from AddRules and we
    // are therefore still within the critical section
    // copy the rule into the workspace and update probotsdata
    size_t cbRule = pchMax - pchRule;

    // limit the maximum length of a rule
    if (cbRule > c_maxRuleLength)
    {
        cbRule = c_maxRuleLength;
    }

    if (type == ALLOW)
    {
        // +1 for the possibility of adding a / to the front
        // and another +1 for the null terminator
        if (m_allowOffset + cbRule + 1 + 1 < m_cbAllow 
            && m_countAllow[idUserAgent] < (ULONG) c_maxRulesPerHost)
        {
            if (m_countAllow[idUserAgent] == 0)
            {
                m_pchAllow[idUserAgent] = m_allowSpace + m_allowOffset;
            }

            // append / to the front if it is not already there
            if (*pchRule != '/')
            {
                m_allowSpace[m_allowOffset] = '/';
                m_allowOffset++;
            }

            memcpy(m_allowSpace + m_allowOffset, pchRule, cbRule);
            *(m_allowSpace + m_allowOffset + cbRule) = '\0';

            m_countAllow[idUserAgent]++;

            m_allowOffset += cbRule + 1;
        }
    }
    else if (type == DISALLOW)
    {
        // +1 for the possibility of adding a / to the front
        // and another +1 for the null terminator
        if (m_disallowOffset + cbRule + 1 + 1 < m_cbDisallow
            && m_countDisallow[idUserAgent] < (ULONG) c_maxRulesPerHost)
        {
            if (m_countDisallow[idUserAgent] == 0)
            {
                m_pchDisallow[idUserAgent] = m_disallowSpace + m_disallowOffset;
            }

            // append / to the front if it is not already there
            if (*pchRule != '/')
            {
                m_disallowSpace[m_disallowOffset] = '/';
                m_disallowOffset++;
            }

            memcpy(m_disallowSpace + m_disallowOffset, pchRule, cbRule);
            *(m_disallowSpace + m_disallowOffset + cbRule) = '\0';

            m_countDisallow[idUserAgent]++;

            m_disallowOffset += cbRule + 1;
        }
    }
}


// This works only for ASCII characters.
// Trivial implementaion of tolower
static inline int MyTolower(int ch)
{
    if (ch >= 'A' && ch <= 'Z')
    {
        ch += 'a' - 'A';
    }
    return(ch);
}

// Case-insensitive strstr with a buffer size for the string to be searched
static char * FindStringIN(const char * str1, size_t cchBuffer, const char * str2)
{
    const char *cp = str1;
    const char *s1;
    const char *s2;

    const char *s1Max = cp + cchBuffer;

    if ( !*str2 )
    {
        return((char *)str1);
    }

    while (cp < s1Max && *cp)
    {
        s1 = cp;
        s2 = str2;

        while (s1 < s1Max && *s1 && *s2)
        {
            // Convert characters to lowercase before comparing them
            int chBigString = MyTolower(*s1);
            int chLittleString = MyTolower(*s2);
            
            if (chBigString == chLittleString)
            {
                s1++;
                s2++;
            }
            else
            {
                break;    // strings don't match
            }        
        }

        if (!*s2)
        {
            return((char *)cp);
        }

        cp++;
    }

    return(NULL);
}

// Case-insensitive strstr
static char * FindStringI(const char * str1, const char * str2)
{
    return(FindStringIN(str1, strlen(str1), str2));
}



// Checks the url path against all robots rules to determine if
// the url is allowed for MsnBot and also set *pRobotsAllowed
// if the corresponding agent is allowed.
// This must be called from within the critical section if probotsdata
// points to memory within the hash table.
bool Robots::RobotsData::UrlAllowed(const char * const szRelativeUrl)
{
    // return value to be passed to *pRobotsAllowed if it is not NULL
    UInt32 retRobotsAllowed = 0; 

    UInt32 numAgentsJudged = 0;
    for (UInt32 iUserAgent = 0; iUserAgent < s_numUserAgents; iUserAgent++)
    {
        // no need to look at rules for UserAgent==* if all other bots have been judged
        //if (iUserAgent == s_StarBotIndex && numAgentsJudged == s_numUserAgents - 1) 
        //{
        //    continue; 
        //}
        
        // test if there is any rules block found for this Agent
        if (m_fAgentFound[iUserAgent]) 
        {
            numAgentsJudged++;
        }
        else
        {
            // set * to allow if no rules block is found for it; note this behavior does not apply to other bots: if no rules block is found for 
            // other bots, they should use the result for * instead of set to allowed 
            if (iUserAgent == s_StarBotIndex)
            {
                retRobotsAllowed |= 0x1 << iUserAgent; 
            }
            continue;
        }
        
        bool fAllowed = false;
        bool fDisallowed = false;
        // Need to store the longest applicable allow and disallow rule lengths to determine which one is more specific in case of conflict
        // Disallow rules with wildcard will counted as 0 length, i.e., least specific
        size_t cbMaxAllowedRuleLength = 0; 
        size_t cbMaxDisallowedRuleLength = 0;
        
        ULONG i;
        char *pchRule = m_pchAllow[iUserAgent];
        for (i = 0; i < m_countAllow[iUserAgent]; i++)
        {
            size_t cbRule = strlen(pchRule);
            // no need to match with this Allow rule if it cannot make Allow more specific
            if (fAllowed && cbMaxAllowedRuleLength >= cbRule) 
            {
                pchRule += cbRule + 1;
                continue;
            }
            // be as respectful as we can -- do a case sensitive compare on these
            // and a case-insensitive compare on disallow
            if (strncmp(szRelativeUrl, pchRule, cbRule) == 0)
            {
                // we have a match, and the url is allowed
                fAllowed = true;
                cbMaxAllowedRuleLength = max(cbMaxAllowedRuleLength, cbRule);
            }
            pchRule += cbRule + 1;
        }

        pchRule = m_pchDisallow[iUserAgent];
        for (i = 0; i < m_countDisallow[iUserAgent]; i++)
        {
            size_t cbRule = strlen(pchRule);
            const char *pchStar = strchr(pchRule, '*');

            // no need to match with this Disallow rule if it cannot make Disallow more specific
            if (fDisallowed && cbMaxDisallowedRuleLength >= (pchStar ? 0 : cbRule)) 
            {
                pchRule += cbRule + 1;
                continue;
            }

            
            if (NULL == pchStar) 
            {
                // the easy case: no '*' in the rule
                // relative URL just need to start with the rule string to match this rule
                if (PALString::CompareInsensitive(pchRule, szRelativeUrl, cbRule) == 0)  
                {
                    fDisallowed = true;
                    cbMaxDisallowedRuleLength = max(cbMaxDisallowedRuleLength, cbRule);
                }
            }
            else 
            {
                // the hard case: there is '*' wildcard in the rule
                bool fMatch = true;
                // first match the beginning of the rule to the beginning of the URL with sized comparison 
                size_t cbPattern = pchStar - pchRule;
                if (cbPattern > 0 && PALString::CompareInsensitive(pchRule, szRelativeUrl, cbPattern) != 0)
                {
                    fMatch = false;
                }

                // next match the patterns separated by '*'s with substring search (except for the last pattern if the rule ends with '$')
                // rule pointer for next pattern
                const char *pchPatternStart = pchRule + cbPattern; 
                // URL pointer for next match
                const char *pchUrlStart = szRelativeUrl + cbPattern; 
                while (fMatch && pchStar < pchRule + cbRule)
                {
                    // find next '*'
                    while (*pchPatternStart == '*')
                    {               
                        pchPatternStart++;
                    }
                    pchStar = strchr(pchPatternStart, '*');
                    if (pchStar == NULL) 
                    {
                        // now we are looking at the last part of the rule
                        if (pchRule[cbRule - 1] == '$') 
                        {
                            // the last part of the rule must be exact match if it ends with '$'
                            cbPattern = pchRule + cbRule - pchPatternStart - 1;
                            size_t cbRemainingUrl = strlen(pchUrlStart);
                            if (cbPattern > 0 &&
                                (cbRemainingUrl < cbPattern || PALString::CompareInsensitive(pchPatternStart, pchUrlStart + cbRemainingUrl - cbPattern, cbPattern) != 0))
                            {
                                fMatch = false; 
                            }
                            // no more pattern match needed
                            break; 
                        }
                        else 
                        {
                            // rule not ended with '$'; continue pattern match as if there is an implicit '*' at the end of the rule
                            pchStar = pchRule + cbRule;  
                        }
                    }
                    // match the pattern separated by '*' with substring search
                    cbPattern = pchStar - pchPatternStart;
                    if (cbPattern > 0)
                    {
                        // make a NULL terminated copy of the pattern as required by FindStringI
                        char szPattern[c_maxRuleLength + 1];
                        if (cbPattern >= sizeof(szPattern)) // assert
                        {
                            return false;
                        }
                        memcpy(szPattern, pchPatternStart, cbPattern);
                        szPattern[cbPattern] = '\0';
                        pchUrlStart = FindStringI(pchUrlStart, szPattern);

                        if (NULL == pchUrlStart)
                        {
                            fMatch = false;             
                        }
                        else
                        {
                            // move rule and URL pointers to prepare for next substring match
                            pchPatternStart = pchStar;
                            pchUrlStart += cbPattern; 
                        }
                    }
                }
                if (fMatch) 
                {
                    // matched this disallow rule containing '*' wildcard; note we don't modify cbMaxDisallowedRuleLength since we consider rule 
                    // with * wildcard 0 length becase they are the least specific
                    fDisallowed = true;
                }
            }            
            pchRule += cbRule + 1;
        }
        // now determine whether Allow or Disallow wins for this bot
        // we require explicit Disallow and it must be more specific than all applicable Allow rules to disallow this URL
        if (!fDisallowed || (fAllowed && cbMaxAllowedRuleLength >= cbMaxDisallowedRuleLength)) 
        {
            retRobotsAllowed |= 0x1 << iUserAgent;
        }
    }
#if 0
    // if ther is no rules block for a certain Agent, use the value from UserAgent==*
    for (UInt32 iUserAgent = 0; iUserAgent < s_numUserAgents - 1; iUserAgent++)
    {
        if (!m_fAgentFound[iUserAgent])
        {
             if * is allowed; set bots with missing rules blocks to allow as well
            if (retRobotsAllowed & (0x1 << s_StarBotIndex)) 
            {
                retRobotsAllowed |= 0x1 << iUserAgent;
            }
        }
    }

    if (pRobotsAllowed)
    {
        // clean the * bit before we pass back retRobotsAllowed
       retRobotsAllowed -= retRobotsAllowed & (0x1 << s_StarBotIndex);
        *pRobotsAllowed = retRobotsAllowed;
    }
#endif
    return 0 != (retRobotsAllowed /* & (0x1 << s_MsnBotIndex) */);
}


BOOL Robots::RobotsData::Load(IAllocator* pAlloc, const TCHAR* wszFile, BOOL fRobots)
{
    FILE* fp = NULL;
    RobotsData* p = NULL;
    char* buf = NULL;
    BOOL fRet = FALSE;
    size_t cb = 0;
    
    if (0 != PALFile::Open(&fp, wszFile, TEXT("rb")))
    {
        goto Exit;
    }

    if (fseek (fp, 0, SEEK_END) != 0)
    {
        goto Exit;
    }

    cb = ftell (fp);
    fseek (fp, 0, SEEK_SET);

    if (cb > 100 * 1048576)
    {
        goto Exit;
    }
    
    buf = (char*)malloc(cb); // don't need IAllocator because this is temporary - for the duration of loading only
    if (!buf)
    {
        goto Exit;
    }
    if (cb != fread(buf, 1, cb, fp))
    {
        goto Exit;
    }

    fclose(fp);  fp = NULL;
    
    if (!Initialize(pAlloc, cb))
    {
        goto Exit;
    }
    if (fRobots)
    {
        ParseRobotsRules(buf, cb);
    }
    else
    {
        AddRules(s_StarBotIndex, buf, buf + cb);
    }
    if (!Finalize())
    {
        goto Exit;
    }

    fRet = TRUE;
Exit:
    if (fp) fclose(fp);
    if (buf) free(buf);
    return fRet;
}


BOOL Robots::Load(IAllocator* pAlloc, const TCHAR* wszRobotsFile, const TCHAR* wszNonRobotsFile)
{
    m_fRobots = m_robots.Load(pAlloc, wszRobotsFile, TRUE);
    m_fNonRobots = m_nonRobots.Load(pAlloc, wszNonRobotsFile, FALSE);
    return TRUE; // BUGBUG: not distinguishing non-existence (normal) from load failure (concerning)
}

BOOL Robots::Allowed(const char* szUrl)
{
    bool fRobotsAllowed = true, fNonRobotsAllowed = true;

    if (m_fNonRobots)
    {
        fNonRobotsAllowed = m_nonRobots.UrlAllowed(szUrl);
    }
    if (m_fRobots)
    {
        fRobotsAllowed = m_robots.UrlAllowed(szUrl);
    }
    return fNonRobotsAllowed && fRobotsAllowed;
}

BOOL Robots::Allowed(const char* pcUrl, size_t cbUrl)
{
    char temp[1024];
    if (cbUrl >= ARRAYSIZE(temp))
    {
        return FALSE; // crazy long URL not likely to be useful
    }
    memcpy(temp, pcUrl, cbUrl);
    temp[cbUrl] = '\0';
    return Allowed(temp);
}

