/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#pragma once

template <class T> class SimpleHashTable
{
    class Node
    {
    public:
        T val;
        ULONGLONG hash;
        Node* pNext;
    };
    Node* m_rNodes;
    size_t m_cNodesUsed;
    size_t m_cMaxNodes;

    Node** m_rTable;
    unsigned m_cSpineBits;
    unsigned m_uMask;

    BOOL m_fExternalAllocator;

    unsigned Index(ULONGLONG hash);

public:
    SimpleHashTable();
    ~SimpleHashTable();
    
    static unsigned Log2Ceil(UInt64 u);
    static unsigned SpineBits(size_t cMaxElements);
    static size_t BytesPerEntry();
    static size_t MemoryNeeded(size_t cMaxElements);
    static size_t MemoryNeeded(size_t cMaxElements, unsigned cSpineBits);

    void Reset();
    BOOL Initialize(size_t cMaxElements, IAllocator* pAllocator = NULL);
    BOOL Initialize(size_t cMaxElements, unsigned cSpineBits, IAllocator* pAllocator = NULL);
    BOOL Add(ULONGLONG hash, T val);
    T* Get(ULONGLONG hash);
    BOOL Enumerate(void* pvContext, BOOL (*pfnCallback)(void* pvContext, T* pVal));
    size_t Entries() { return m_cNodesUsed; }
    size_t SpaceLeft() { return m_cMaxNodes - m_cNodesUsed; }
    void Free();
};

template <class T>
SimpleHashTable<T>::SimpleHashTable()
{
    m_rNodes = NULL;
    m_rTable = NULL;
    m_fExternalAllocator = FALSE;
}
template <class T>
SimpleHashTable<T>::~SimpleHashTable()
{
    Free();
}
template <class T>
void SimpleHashTable<T>::Free()
{
    if (m_rNodes) { if (!m_fExternalAllocator) delete[] m_rNodes; m_rNodes = NULL; }
    if (m_rTable) { if (!m_fExternalAllocator) delete[] m_rTable; m_rTable = NULL; }
}

template <class T>
unsigned SimpleHashTable<T>::Log2Ceil(UInt64 u)
{
    unsigned uRet = 0;
    do
    {
        u >>= 1;
        uRet++;
    } while (u);
    return uRet;
}

template <class T>
unsigned SimpleHashTable<T>::SpineBits(size_t cMaxElements)
{
    return Log2Ceil(cMaxElements >> 2);
}

template <class T>
size_t SimpleHashTable<T>::BytesPerEntry()
{
    return sizeof(Node) + (sizeof(Node*) >> 2);
}

template <class T>
size_t SimpleHashTable<T>::MemoryNeeded(size_t cMaxElements)
{
    return MemoryNeeded(cMaxElements, SpineBits(cMaxElements));
}

template <class T>
size_t SimpleHashTable<T>::MemoryNeeded(size_t cMaxElements, unsigned cSpineBits)
{
    return cMaxElements * sizeof(Node) + (1ULL << cSpineBits) * sizeof(Node*);
}

template <class T>
BOOL SimpleHashTable<T>::Initialize(size_t cMaxElements, IAllocator* pAllocator)
{
    return Initialize(cMaxElements, SpineBits(cMaxElements), pAllocator);
}

template <class T>
void SimpleHashTable<T>::Reset()
{
    m_cNodesUsed = 0;
    memset(m_rTable, 0, (1ULL << m_cSpineBits) * sizeof(Node*));
}

template <class T>
BOOL SimpleHashTable<T>::Initialize(size_t cMaxElements, unsigned cSpineBits, IAllocator* pAllocator)
{
    m_cSpineBits = cSpineBits;
    m_uMask = (1 << cSpineBits) - 1;
    m_cMaxNodes = cMaxElements;
    size_t cTable = 1ULL << m_cSpineBits;
    if (pAllocator)
    {
        m_fExternalAllocator = TRUE;
        m_rNodes = (Node*)pAllocator->Allocate(sizeof(Node) * m_cMaxNodes);
        m_rTable = (Node**)pAllocator->Allocate(sizeof(Node*) * cTable);
    }
    else
    {
        m_rNodes = new Node[m_cMaxNodes];
        m_rTable = new Node*[cTable];
    }
    if (!m_rNodes || !m_rTable) return FALSE;
    Reset();
    return TRUE;
}

template <class T>
unsigned SimpleHashTable<T>::Index(ULONGLONG hash)
{
    return hash & m_uMask;
}

template <class T>
BOOL SimpleHashTable<T>::Add(ULONGLONG hash, T val)
{
    if (m_cNodesUsed >= m_cMaxNodes)
    {
        return FALSE;
    }
    
    Node** ppNode = &m_rTable[Index(hash)];
    while (*ppNode)
    {
        ppNode = &((*ppNode)->pNext);
    }
    *ppNode = &m_rNodes[m_cNodesUsed++];
    (*ppNode)->val = val;
    (*ppNode)->hash = hash;
    (*ppNode)->pNext = NULL;

    return TRUE;
}

template <class T>
T* SimpleHashTable<T>::Get(ULONGLONG hash)
{
    Node* pNode = m_rTable[Index(hash)];
    while (pNode)
    {
        if (pNode->hash == hash)
        {
            return &(pNode->val);
        }
        pNode = pNode->pNext;
    }
    return NULL;
}

template <class T>
BOOL SimpleHashTable<T>::Enumerate(void* pvContext, BOOL (*pfnCallback)(void* pvContext, T* pVal))
{
    if (m_rTable == NULL)
    {
        return FALSE;
    }

    for (size_t u = 0; u < (1ULL << m_cSpineBits); u++)
    {
        Node* p = m_rTable[u];
        while (p)
        {
            if (!pfnCallback(pvContext, &(p->val)))
            {
                return FALSE;
            }
            p = p->pNext;
        }
    }
    return TRUE;
}

