/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PALMemory.h"
#include "shared.h"
#include "diskio.h"
#include "UncompressingReader.h"


/*************************************************************************
**  Method: UncompressingFileReader::UncompressingFileReader
**
**  Description: Constructs a UncompressingFileReader object
**
**************************************************************************/
UncompressingFileReader::UncompressingFileReader()
{
    m_pAllocator = new HeapInstance();
    m_fDeleteAllocator = true;

    m_fCompression = false;
    m_compressedBuffer = NULL;
    m_uncompressedBuffer = NULL;
}

/*************************************************************************
**  Method: UncompressingFileReader::UncompressingFileReader
**
**  Description: Constructs a UncompressingFileReader object
**  using an existing HeapInstance  
**
**************************************************************************/
UncompressingFileReader::UncompressingFileReader( HeapInstance* pAllocator )
{
    m_pAllocator = pAllocator;
    m_fDeleteAllocator = false;

    m_fCompression = false;
    m_compressedBuffer = NULL;
    m_uncompressedBuffer = NULL;
}

/*************************************************************************
**  Method: UncompressingFileReader::~UncompressingFileReader
**
**  Description: Destructs and frees memory for a UncompressingFileReader object
**
**************************************************************************/
UncompressingFileReader::~UncompressingFileReader() 
{
    ResetUncompression();
    if( m_fDeleteAllocator )
    {
        delete m_pAllocator;
        m_pAllocator = NULL;
    }
}

/*************************************************************************
**  Method: CompressingFileReader::InitializeUncompression
**
**  Description: Initializes the compression subobjects.  Allocate memory, 
**      and initialize the z_stream
**
**  Parameters: none
**
**  Returns:
**      NO_ERROR=0      
**      ERROR_OUTOFMEMORY 
**      ERROR_DECRYPTION_FAILED
**
**  Notes:
**
**************************************************************************/
DWORD32
UncompressingFileReader::InitializeUncompression(void)
{
    m_fCompression = true;

    // setup buffer for compressed data
    // (reuse of this instance is not supported)
    LogAssert(m_compressedBuffer == NULL);

#ifdef __TARGET_IIS__
    m_compressedBuffer = (UInt8 *)VirtualAlloc(NULL, c_compressedBufferSize, MEM_COMMIT, PAGE_READWRITE);
    if (!m_compressedBuffer)
    {
        return ERROR_OUTOFMEMORY;
    }
#else
    try
    {
        m_compressedBuffer = (UInt8 *)PALMemory::Allocate(c_compressedBufferSize);
    }
    catch(...)
    {
        return ERROR_OUTOFMEMORY;
    }
#endif

    LogAssert(m_uncompressedBuffer == NULL);

#ifdef __TARGET_IIS__
    m_uncompressedBuffer = (UInt8 *)VirtualAlloc(NULL, c_uncompressedBufferSize, MEM_COMMIT, PAGE_READWRITE);
    if (!m_uncompressedBuffer)
    {
        return ERROR_OUTOFMEMORY;
    }
#else
    try
    {
        m_uncompressedBuffer = (UInt8 *)PALMemory::Allocate(c_uncompressedBufferSize);
    }
    catch(...)
    {
        return ERROR_OUTOFMEMORY;
    }
#endif
    
    // initialize compression object
    memset(&m_strmUncompress, 0, sizeof(m_strmUncompress));
    m_strmUncompress.zalloc = zlib_alloc;
    m_strmUncompress.zfree = zlib_free;
    m_strmUncompress.opaque = m_pAllocator;

    m_strmUncompress.next_in = m_compressedBuffer;
    m_strmUncompress.avail_in = 0;
    m_strmUncompress.next_out = m_uncompressedBuffer;
    m_strmUncompress.avail_out = 0;

    int nRet = inflateInit(&m_strmUncompress);
    if (nRet != Z_OK)
    {
        return ERROR_DECRYPTION_FAILED;
    }
    return NO_ERROR;
}
/*************************************************************************
**  Method: CompressingFileReader::ResetUncompression
**
**  Description: this function may be called at the reset time to undo what 
**      InitializeUncompression does ( so that InitializeUncompression can be again called at the appropriate time)
**
**  Parameters: none
**
**  Returns:
**      NO_ERROR=0      
**      ERROR_DECRYPTION_FAILED 
**
**  Notes:
**
**************************************************************************/
DWORD32 
UncompressingFileReader::ResetUncompression(void)
{
    if (m_fCompression)
    {
        int err = inflateEnd(&m_strmUncompress);
        if (m_compressedBuffer)
        {
#ifdef __TARGET_IIS__
            VirtualFree(m_compressedBuffer, 0, MEM_RELEASE);
#else
            PALMemory::Free(m_compressedBuffer);
#endif
            m_compressedBuffer = NULL;
        }
        if (m_uncompressedBuffer)
        {
#ifdef __TARGET_IIS__
            VirtualFree(m_uncompressedBuffer, 0, MEM_RELEASE);
#else
            PALMemory::Free(m_uncompressedBuffer);
#endif
            m_uncompressedBuffer = NULL;
        }
        m_fCompression = false;
        if (err != Z_OK)
        {
            Log(LogID_Common, LogLevel_Error, "UncompressingFileReader::ResetUncompression",
                "inflateEnd failed with err = %d",
                err);
            return ERROR_DECRYPTION_FAILED;
        }
    }
    return NO_ERROR;
}
/*************************************************************************
**  Method: UncompressingFileReader::DoInit
**
**  Description: Initializes the object.  Allocate memory, open file and 
**      issue initial reads.
**
**  Parameters:
**      szFile          name of file to open
**      maxReads        max number of outstanding reads
**
**  Returns:
**      NO_ERROR=0      
**      ERROR_INVAID_PARAMTER 
**      ERROR_OUTOFMEMORY
**
**  Notes:
**
**************************************************************************/
DWORD32 
UncompressingFileReader::DoInit(
                const TCHAR*       wszFile,
                int         maxReads,
                DWORD32     readBufSize,
                BOOL        readwrite)
{
    DWORD ec = m_seqread.DoInit(wszFile, maxReads, readBufSize, readwrite);

    if (ec == NO_ERROR)
    {
        m_fileSize = m_seqread.FileSize();
        m_bytesRemaining = m_fileSize;
    }

    return ec;
}

/*************************************************************************
**  Method: UncompressingFileReader::Reset
**
**  Description: Resets the sequential read to begining of file and 
**      reissues the initial reads.
**
**  Parameters:
**
**  Returns:
**      NO_ERROR=0      
**      GetLastError()  if read fails
**
**  Notes:
**
**************************************************************************/
DWORD32 
UncompressingFileReader::Reset(void)
{
    m_bytesRemaining = m_fileSize;
    DWORD32 ec = ResetUncompression();
    if ( ec != NO_ERROR )
    {
        return ec;
    }
    return m_seqread.Reset(0);
}

DWORD32
UncompressingFileReader::DoDispose()
{
    m_seqread.DoDispose();
    return ResetUncompression();
}

/*************************************************************************
**  Method: UncompressingFileReader::GetDataPointer
**
**  Description: Gets a pointer to bytes read.  When crossing read buffers
**      this pointer goes to a temp buffer.  Data is only copied when a 
**      read crosses the read buffers.
**
**  Parameters:
**      ppv     returns pointer to buffer with bytes of data.
**      cb      number of bytes requested.
**
**  Returns:
**      NO_ERROR=0
**      ERROR_INVALID_PARAMETER
**      ERROR_HANDLE_EOF
**      GetLastError()  if read fails
**
**  Notes:
**
**************************************************************************/
DWORD32 
UncompressingFileReader::GetDataPointer( void ** ppv, DWORD cb, DWORD* pcbRead )
{
    DWORD ec = NO_ERROR;
    if (!m_fCompression)
    {
        ec = m_seqread.GetDataPointer(ppv, cb, pcbRead);
        DWORD cbRead;
        if (pcbRead)
        {
            cbRead = *pcbRead;
        }
        else if (ec == NO_ERROR)
        {
            cbRead = cb;
        }
        else
        {
            cbRead = 0;
        }
        if (cbRead > m_bytesRemaining)
        {
            Log(LogID_Common, LogLevel_Error, "UncompressingFileReader::GetDataPointer",
                "File larger than expected:  cbRead (%u) > m_bytesRemaining (%u)",
                cbRead, m_bytesRemaining);
            m_bytesRemaining = 0;
        }
        else
        {
            m_bytesRemaining -= cbRead;
        }
    }
    else
    {
        
        m_strmUncompress.next_out = (Bytef*)m_uncompressedBuffer;
        m_strmUncompress.avail_out = cb;

        int zlibError = Z_OK;
        while (m_strmUncompress.avail_out != 0) 
        {
            if (m_strmUncompress.avail_in == 0) 
            {
                UInt32 bytesToRead = c_compressedBufferSize;
                if (m_bytesRemaining < c_compressedBufferSize)
                {
                    bytesToRead = (UInt32) m_bytesRemaining;
                }

                ec = m_seqread.GetData(m_compressedBuffer, bytesToRead);
                if (ec != NO_ERROR)
                {
                    return ec;
                }
                m_bytesRemaining -= bytesToRead;
                m_strmUncompress.avail_in = bytesToRead;
                if (m_strmUncompress.avail_in == 0) 
                {
                    return ERROR_HANDLE_EOF;
                }
                m_strmUncompress.next_in = (Bytef *)m_compressedBuffer;
            }
            zlibError = inflate(&m_strmUncompress, Z_NO_FLUSH);

            if (zlibError != Z_OK) 
            {
                if (zlibError != Z_STREAM_END)
                {
                   ec = ERROR_INVALID_DATA;
                }
                break;
            }
        }

        if (pcbRead)
        {
            *pcbRead = cb - m_strmUncompress.avail_out;
        }
        *ppv = m_uncompressedBuffer;
    }
    return ec;
}
/*************************************************************************
**  Method: UncompressingFileReader::Skip
**
**  Description: skips next dwNumBytes bytes.
**
**  Parameters:
**      dwNumBytes      number of bytes to be skipped.
**
**  Returns:
**      NO_ERROR=0
**      errors from GetData
**
**  Notes:
**
**************************************************************************/

DWORD32 UncompressingFileReader::Skip(DWORD dwNumBytes)
{
    return( GetData( NULL, dwNumBytes ));
}

/*************************************************************************
**  Method: UncompressingFileReader::GetData
**
**  Description: Copies read bytes into passed buffer.
**
**  Parameters:
**      pv      buffer to copy data into.
**      cb      number of bytes requested.
**
**  Returns:
**      NO_ERROR=0
**      errors from GetDataPointer
**
**  Notes:
**
**************************************************************************/
DWORD32 
UncompressingFileReader::GetData( __out_bcount_opt(cb) void * pv, DWORD cb, DWORD * pcbRead )
{
    DWORD32 ec = NO_ERROR;
    BYTE*   pb;
    DWORD   cbRead = 0;
    while ( cb > 0 )
    {
        DWORD readBytes = cb;
        if (cb > SEQREAD::c_readBufSize)
            cb = SEQREAD::c_readBufSize;
        DWORD cbReadGoal = readBytes;
        ec = GetDataPointer((void**)&pb, readBytes, &readBytes);
        if (ec)
        {
            return ec;
        }
        if ( pv != NULL )
        {
            memcpy(pv, pb, readBytes);
            pv = (BYTE *) pv + readBytes;        
        }
        cb -= readBytes;
        cbRead += readBytes;
        
        if (readBytes < cbReadGoal) // happens at the end of inflated streams sometimes
        {
            break;
        }
    }
    if ( pcbRead != NULL )
    {
        *pcbRead = cbRead;
    }
    return ec;
}

void *
UncompressingFileReader::zlib_alloc(void *opaque, UINT size, UINT count)
{
    HeapInstance *pAlloc = (HeapInstance *)opaque;
    return pAlloc->Allocate(0, size * count);
}

void
UncompressingFileReader::zlib_free(void * opaque, void *data)
{
    HeapInstance *pAlloc = (HeapInstance *)opaque;
    pAlloc->Free(0, data);
}

DWORD32
UncompressingFileReader::UncompressFile(const TCHAR *pszCompressed, const TCHAR *pszUncompressed)
{
    const char *pszLogTitle = "UncompressingFileReader::UncompressFile";

    UncompressingFileReader reader;
    SEQWRITE writer;

    DWORD32 dwRet = reader.DoInit(pszCompressed);
    if (dwRet == NO_ERROR)
    {
        dwRet = reader.InitializeUncompression();
    }
    if (dwRet != NO_ERROR)
    {
        Log(LogID_Common, LogLevel_Error, pszLogTitle,
            "Failed to open compressed file %s with error %u", pszCompressed, dwRet);
        return dwRet;
    }

    dwRet = writer.DoInit();
    if (NO_ERROR == dwRet)
    {
        dwRet = writer.Open(pszUncompressed);
    }
    if (dwRet != NO_ERROR)
    {
        Log(LogID_Common, LogLevel_Error, pszLogTitle,
            "Failed to open uncompressed output file %s with error %u", pszUncompressed, dwRet);
        return dwRet;
    }

    UInt32 cbRequest = UncompressingFileReader::c_uncompressedBufferSize;
    DWORD cbRead;
    UInt8 *pBuffer;
    bool fEOF = false;
    while (!fEOF)
    {
        cbRead = 0;
        dwRet = reader.GetDataPointer((void **)&pBuffer, cbRequest, &cbRead);
        if (dwRet == ERROR_HANDLE_EOF)
        {
            fEOF = true;
        }                
        else if (dwRet != NO_ERROR)
        {
            Log(LogID_Common, LogLevel_Error, pszLogTitle, "Read from file %s failed with error %u", pszCompressed, dwRet);
            return dwRet;
        }

        dwRet = writer.Write(pBuffer, cbRead);
        if (dwRet != NO_ERROR)
        {
            Log(LogID_Common, LogLevel_Error, pszLogTitle, "Write to uncompressed file %s failed with error %u", pszUncompressed, dwRet);
            return dwRet;
        }

    }
    return NO_ERROR;
}
