/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "UrlNormalizer.h"
#include "bloomfilter.h"
#include "SSPCommon.h"
#include "PALFile.h"

void WriteNormalizerFile(const TCHAR* wszFile)
{
    FILE* fp = NULL;

    if (0 != PALFile::Open(&fp, wszFile, TEXT("wt")) || !fp)
    {
        goto Error;
    }

    fputs("# This file configures URL normalization for the sitemap generation plugin.\n", fp);
    fputs("# By default all URL parameters are removed before a URL is processed for sitemap generation purposes.\n", fp);
    fputs("# Parameters listed in this file are kept.\n", fp);
    fputs("# The format is simply one parameter name per line, no whitespace or tabs or anything.\n", fp);
    fputs("# E.g., to keep parameters of the form '&ProductID=12345', add the string 'ProductID' as a line.\n", fp);
    fputs("# Parameter names are case-sensitive.\n", fp);
    fputs("# Number of parameter names listed in this file should not exceed about 10000.\n", fp);
    fputs("# A line starting with a pound sign is treated as a comment.\n", fp);
    fputs("\n", fp);
    fputs("# Changes to this file take effect when the plugin is restarted (e.g. by restarting the parent web server process).\n", fp);

Error:
    if (fp) fclose(fp);
}

BOOL UrlNormalizer::Load(IAllocator* pAlloc, const TCHAR* wszFile, unsigned uMaxEntries)
{
    FILE* fp = NULL;
    const size_t cb = uMaxEntries * 4; // bytes per entry => false positive rate: 1 bpe => 1/150, 2 bpe => 1/150,000 , 4 bpe => forget about it
    void* pMem = pAlloc->Allocate(cb);
    if (!pMem)
    {
        goto Error;
    }
    if (!m_bloomfilter.Init(pMem, cb))
    {
        goto Error; 
    }
    m_bloomfilter.Add(0, NULL); // 0-length name parameter

    if (wszFile && PALDirectory::PathExists(wszFile))
    {
        if (0 != PALFile::Open(&fp, wszFile, TEXT("rt")) || !fp)
        {
            goto Error;
        }
        for (unsigned cEntries = 0; cEntries < uMaxEntries; cEntries++)
        while (!feof(fp))
        {
            char s[1024];
            if (!fgets(s, sizeof(s), fp)) break;
            if (!s[0]) break;
            if ('#' == s[0]) continue;
            unsigned u;
            for (u = 0; (u < sizeof(s) - 1) && ('\r' != s[u]) && ('\n' != s[u]) && ('\t' != s[u]); u++);
            s[u] = '\0';
            m_bloomfilter.Add(u, s);
        }
        fclose(fp);
    }
    return TRUE;
Error:
    if (fp) fclose(fp);
    return FALSE;
}

BOOL UrlNormalizer::NormalizeUrl(const char* szUrl, __out_bcount_part(*pcbNormalized,*pcbNormalized) char* pcNormalized, size_t *pcbNormalized)
{
    const char* pLastSeparator = NULL, *pNameEnd = NULL;
    size_t cbOut = 0;
    BOOL fAteQuestionMark = FALSE;
    for (const char* p = szUrl; ; p++)
    {
        BOOL isSeparator = ('?' == *p) || ('&' == *p) || (',' == *p) || (';' == *p) || ('\0' == *p) || ('#' == *p);
        if (pLastSeparator)
        {
            if ('=' == *p && !pNameEnd)
            {
                pNameEnd = p;
            }
            else if (isSeparator)
            {
                if (!pNameEnd)
                {
                    pNameEnd = p;
                }
                const char *pNameStart = pLastSeparator + 1;
                if (m_bloomfilter.Contains( pNameEnd - pNameStart, pNameStart))
                { // keep this parameter
                    size_t cb = p - pLastSeparator;
                    if (cbOut + cb > *pcbNormalized)
                    {
                        return FALSE;
                    }
                    memcpy(pcNormalized + cbOut, pLastSeparator, cb);
                    if (fAteQuestionMark)
                    { // we dropped the first parameter ?name=value, so for the first &name=value we keep replace & with ?
                        fAteQuestionMark = FALSE;
                        pcNormalized[cbOut] = '?';
                    }
                    cbOut += cb;
                }
                else
                { // drop this parameter
                    if ('?' == *pLastSeparator)
                    {
                        fAteQuestionMark = TRUE;
                    }
                }

                pLastSeparator = p;
                pNameEnd = NULL;
            }
        }
        else if (isSeparator)
        {
            pLastSeparator = p;
        }
        else
        {
            if (cbOut >= *pcbNormalized)
            {
                return FALSE;
            }
            pcNormalized[cbOut++] = *p;
        }

        if (!*p || ('#' == *p && '!' != *p)) break;
    }

    *pcbNormalized = cbOut;
    return TRUE;
}

int UrlNormalizerCppLineCount()
{
    return __LINE__;
}

#if 0

// test hook
static void AddWhitelistParameter(void* pNormalizer, size_t cb, char* pc)
{
    ((BloomFilter*)pNormalizer)->Add(cb, pc);
}

static void Verify(void* pNormalizer, const char* szIn, const char* szOut)
{
    char pcOut[1024];
    size_t cbOut = sizeof(pcOut);
    if (NormalizeUrl(pNormalizer, szIn, pcOut, &cbOut))
    {
        size_t cbTarget = strlen(szOut);
        if (cbTarget == cbOut && !memcmp(szOut, pcOut, cbTarget))
        {
            printf("OK '%s' => '%.*s'\n", szIn, cbOut, pcOut);
        }
        else
        {
            printf("FAIL '%s' => '%.*s' (expected '%s')\n", szIn, cbOut, pcOut, szOut);
        }
    }
    else
    {
        printf("failed to normalize '%s'\n", szIn);
    }
}

int __cdecl main()
{
    void* p = CreateNormalizer(NULL, 1024);
    AddWhitelistParameter(p, 4, "keep");
    Verify(p, "url", "url");
    Verify(p, "url?drop", "url");
    Verify(p, "url?drop=", "url");
    Verify(p, "url?drop=foo", "url");
    Verify(p, "url?keep", "url?keep");
    Verify(p, "url?keep=", "url?keep=");
    Verify(p, "url?keep=foo", "url?keep=foo");
    Verify(p, "url?drop=foo&keep=bar", "url?keep=bar");
    Verify(p, "url?drop=&keep=", "url?keep=");
    Verify(p, "url?drop&keep", "url?keep");
    Verify(p, "url?keep=foo&drop=bar", "url?keep=foo");
    Verify(p, "url?keep=&drop=", "url?keep=");
    Verify(p, "url?keep&drop", "url?keep");
    Verify(p, "url&keep&drop", "url&keep");
    Verify(p, "url&&keep&&drop", "url&&keep&");
    Verify(p, "url&&drop&&keep", "url&&&keep");
}

#endif


