/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "bloomfilter.h"
#include "SSPHash.h"

BloomFilter::BloomFilter()
{
    this->bit = NULL;
}

BloomFilter::~BloomFilter()
{
}


void BloomFilter::Reset()
{
    ::memset(this->bit, 0, this->cUInt64*sizeof(*this->bit));   
}

bool BloomFilter::Init(void* p, size_t cb, unsigned cHashIn)
{
    cBit = 8 * cb;
    cHash = cHashIn;

    this->cUInt64 = (this->cBit/(8*sizeof(UInt64)));
    this->bit = (UInt64*)p;
    if (!this->bit)
    {
        return false;
    }
    Reset();
    return true;
}

bool BloomFilter::Contains(size_t length, const void* buffer) const
{
    UInt64 currHash = 0;
    for (UInt8 i = 0; i < this->cHash; ++i)
    {
        currHash = this->CalculateHash(currHash, length, buffer);
        
        if (!this->GetBit(currHash % this->cBit))
        {
            return false;
        }        
    }
    return true;
}

void BloomFilter::Add(size_t length, const void* buffer)
{
    UInt64 currHash = 0;
    for (UInt8 i = 0; i < this->cHash; ++i)
    {
        currHash = this->CalculateHash(currHash, length, buffer);
        this->SetBit(currHash % this->cBit);
    }
}
    
const static UInt8 LOG_BITS_PER_UINT64 = 6;
const static UInt64 ALL_ONE_UINT64 = ((UInt64)1 << LOG_BITS_PER_UINT64) - 1;

bool BloomFilter::GetBit(const UInt64 bitIndex) const
{
    return (0 != (this->bit[bitIndex >> LOG_BITS_PER_UINT64] & 
        ((UInt64)1 << (bitIndex & ALL_ONE_UINT64))));
}

void BloomFilter::SetBit(const UInt64 bitIndex)
{
    this->bit[bitIndex >> LOG_BITS_PER_UINT64] |= 
        ((UInt64)1 << (bitIndex & ALL_ONE_UINT64));
}

UInt64 BloomFilter::CalculateHash(const UInt64 base, const size_t length, 
    const void* buffer) const
{
    return SSPHash(buffer, length, base);
}

