/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "counters.h"
#include "SSPCommon.h"
#include "PALString.h"
#include "PALFile.h"

#define COUNTER(name) Counter g_counter##name;
#include "counterlist.h"
#undef COUNTER

typedef void (*PFNReceiveCounterValue)(void* pvContext, const char* szName, unsigned uVal);

BOOL ReadCountersFile(const TCHAR* wszFile, void* pvContext, PFNReceiveCounterValue pfnReceiveCounter)
{
    FILE* fp = NULL;

    if (0 != PALFile::Open(&fp, wszFile, TEXT("rt")) || !fp)
    {
        return FALSE;
    }

    while (!feof(fp))
    {
        char s[100];
        s[0] = '\0';
        if (!fgets(s, sizeof(s), fp)) break;
        if (!s[0]) break;
        char* pchTab = strchr(s, '\t');
        if (!pchTab) continue;
        UInt32 uVal;
        if (1 != SSCANF_S(pchTab + 1, STRFMT_I32 "u", &uVal)) continue;
        *pchTab = '\0';
        pfnReceiveCounter(pvContext, s, uVal);
    }

    fclose(fp);
    return TRUE;
}

static void ReceivePluginCounter(void* pvContext, const char* s, unsigned uVal)
{
    Counters* pCounters = (Counters*)pvContext;
    if ('\0' == s[0])
    { // never happens - this is just to start the if-else-if ladder
    }
#define COUNTER(name) \
    else if (!strcmp(#name, s)) \
    { \
       pCounters->c##name.m_val = uVal; \
    }
#include "counterlist.h"
#undef COUNTER
}

BOOL ReadCounters(const TCHAR* wszFile, Counters* pCounters)
{
    return ReadCountersFile(wszFile, pCounters, ReceivePluginCounter);
}


static BOOL WriteCounters(const TCHAR* wszFile, const Counters* pCounters)
{
    FILE* fp = NULL;

    if (0 != PALFile::Open(&fp, wszFile, TEXT("wb")) || !fp)
    { // sharing violation possible
        return FALSE;
    }

#define COUNTER(name) if (fprintf(fp, "%s\t" STRFMT_I32 "u\r\n", #name, pCounters->c##name.m_val) < 0) goto Error;
#include "counterlist.h"
#undef COUNTER

    fclose(fp);
    return TRUE;

Error:
    fclose(fp);
    return FALSE;
}

void AddCounters(Counters* pCounters, const Counters *pAdd)
{
#define COUNTER(name) pCounters->c##name.m_val += pAdd->c##name.m_val;
#include "counterlist.h"
#undef COUNTER
}

void ResetCounters(Counters* pCounters)
{
#define COUNTER(name) pCounters->c##name.Reset();
#include "counterlist.h"
#undef COUNTER
}

BOOL AddCountersToFile(const TCHAR* wszFile, const Counters* pCounters)
{
    // BUGBUG: race with the process reading the file, could be avoided by opening the file "r+b" and keeping it open until we save new values
    Counters counters;
    if (PALDirectory::PathExists(wszFile))
    {
        if (!ReadCounters(wszFile, &counters))
        {
            return FALSE;
        }
    }
    AddCounters(&counters, pCounters);
    if (!WriteCounters(wszFile, &counters))
    {
        return FALSE;
    }

    return TRUE;
}

NotificationCounters::NotificationCounters()
{
    memset(this, 0, sizeof(*this));
}

void NotificationCounters::Inc(unsigned u)
{
    if (u < ARRAYSIZE(m_c))
    {
        m_c[u]++; // race - don't really care
    }
}


void NotificationCounters::Tick()
{
    time_t t = PluginTime();
    if (!m_t)
    { // Only try this once - don't keep calling Initialize() if it Initialize() fails
        Initialize();
    }
    else if (t - m_t > 60)
    {
        Save();
    }
    m_t = t;
}

static void ReceiveNotificationCounter(void* pvContext, const char* s, unsigned uVal)
{
    NotificationCounters* pnc = (NotificationCounters*)pvContext;
    unsigned u;
    if (1 != SSCANF_S(s, "%u", &u))
    {
        return;
    }
    if (u >= ARRAYSIZE(pnc->m_c))
    {
        return;
    }
    pnc->m_c[u] = uVal;
}

void NotificationCounters::Initialize()
{
    TCHAR wszFolder[MAX_PATH];
    if (!GetDataFolder(wszFolder, ARRAYSIZE(wszFolder)))
    {
        return;
    }
    if (0 >= STPRINTF_S(m_wszFile, ARRAYSIZE(m_wszFile), STRFMT_TLS STRFMT_TDIRSLASH TEXT("counters%u.tsv"), wszFolder, GetCurrentProcessId()))
    {
        return;
    }

    if (PALDirectory::PathExists(m_wszFile))
    {
        if (!ReadCountersFile(m_wszFile, this, ReceiveNotificationCounter))
        {
            return;
        }
    }

    m_fInitialized = TRUE;
}

void NotificationCounters::Save()
{
#ifdef __TARGET_IIS__
    if (!m_fInitialized)
    {
        return;
    }

    DWORD dw = 0;
    if (!PluginRegQueryDWORD(NULL, TEXT("log"), &dw) || !dw)
    {
        return;
    }

    FILE* fp = NULL;
    if (0 != PALFile::Open(&fp, m_wszFile, TEXT("wt")) || !fp)
    {
        return;
    }
    for (unsigned u = 0; u < ARRAYSIZE(m_c); u++)
    {
        if (m_c[u])
        {
            fprintf(fp, STRFMT_I32 "u\t" STRFMT_I32 "u\n", u, m_c[u]);
        }
    }
    fclose(fp);
#else
    // would need something instead of registry on Linux..
    // For now, P3 item as this is used solely for debugging.
#endif
}


