/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "shared.h"
#include "diskio.h"
#include "PALFile.h"
#include "PALString.h"

#ifndef APR_SUCCESS
#define APR_SUCCESS NO_ERROR
#define APR_ENOMEM ERROR_OUTOFMEMORY
#define APR_BADARG ERROR_INVALID_PARAMETER
#endif

#if APR_HAVE_WINDOWS_H

/*************************************************************************
**  Method: SEQREAD::SEQREAD
**
**  Description: Constructs a SEQREAD object
**
**************************************************************************/
SEQREAD::SEQREAD()
{
    memset(m_rgReads, 0, sizeof(m_rgReads));
    m_hfile = INVALID_HANDLE_VALUE;
    m_cbLeft = 0;
    m_pbRead = NULL;
    m_cbSkip = 0;
    m_rgb = NULL;
    m_numReads = 0;
    m_offsetNext.ddw = 0;
    m_size.ddw = 0;
    m_readBufSize = 0;
    m_iRead = -1;
}

/*************************************************************************
**  Method: SEQREAD::~SEQREAD
**
**  Description: Destructs and frees memory for a SEQREAD object
**
**************************************************************************/
SEQREAD::~SEQREAD() 
{
    DoDispose();
}

/*************************************************************************
**  Method: SEQREAD::DoDispose
**
**  Description: Cleans up a SEQREAD object.
**
**************************************************************************/

void SEQREAD::DoDispose() 
{
    if (m_hfile != INVALID_HANDLE_VALUE)
    {
        // Wait all asyncRead call to complete
        WaitOutstandingAsyncRead();
   
        CloseHandle(m_hfile);
        m_hfile = INVALID_HANDLE_VALUE;
    }

    for (int iRead=0; iRead< c_maxReads; iRead++)
    {
        if (m_rgReads[iRead].m_overlap.hEvent)
            CloseHandle(m_rgReads[iRead].m_overlap.hEvent);

        if (m_rgReads[iRead].m_pbRead)
            VirtualFree(m_rgReads[iRead].m_pbRead, 0, MEM_RELEASE);
    }
    memset(m_rgReads, 0, sizeof(m_rgReads));

    if (m_rgb)
    {
        VirtualFree(m_rgb, 0, MEM_RELEASE);
        m_rgb = NULL;
    }

    m_cbLeft = 0;
    m_pbRead = NULL;
    m_cbSkip = 0;
    m_numReads = 0;
    m_offsetNext.ddw = 0;
    m_size.ddw = 0;
    m_readBufSize = 0;
    m_iRead = -1;
}



/*************************************************************************
**  Method: SEQREAD::DoInit
**
**  Description: Initializes the object.  Allocate memory, open file and 
**      issue initial reads.
**
**  Paramaters:
**      szFile          name of file to open
**      maxReads        max number of outstanding reads
**
**  Returns:
**      NO_ERROR=0      
**      ERROR_INVAID_PARAMTER 
**      ERROR_OUTOFMEMORY
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::DoInit(
                const TCHAR*       wszFile,
                int         maxReads,
                DWORD32     readBufSize,
                BOOL        readwrite)
{
    DWORD32     ec = NO_ERROR;
    int         iRead;

    this->m_readBufSize = readBufSize;
    // initialize params based on passed values
    //
    if ((maxReads <= SEQREAD::c_maxReads) && (maxReads > 1))
        m_numReads = maxReads;
    else
    {
        ec = ERROR_INVALID_PARAMETER;
        goto Done;
    }

    // allocate memory for the temp buffer

    m_rgb = (BYTE *)VirtualAlloc(NULL, c_cbReadCopy, MEM_COMMIT, PAGE_READWRITE);
    if (m_rgb == NULL)
    {
        ec = ERROR_OUTOFMEMORY;
        goto Done;
    }

    // allocate events and memory for each io buffer
    //
    for (iRead=0; iRead< m_numReads; iRead++)
    {
        if ((m_rgReads[iRead].m_overlap.hEvent = 
            CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL)
        {
            ec = ERROR_OUTOFMEMORY;
            goto Done;
        }

        if ((m_rgReads[iRead].m_pbRead = 
            VirtualAlloc(NULL, m_readBufSize, MEM_COMMIT, PAGE_READWRITE)) == NULL)
        {
            ec = ERROR_OUTOFMEMORY;
            goto Done;
        }
        m_rgReads[iRead].m_fWait=FALSE;
    }

    // open file
    //
    DWORD dwRetries = 0;
    do
    {
        m_hfile = CreateFile(wszFile, FILE_READ_DATA,
                              readwrite ? FILE_SHARE_READ | FILE_SHARE_DELETE | FILE_SHARE_WRITE
                              : FILE_SHARE_READ | FILE_SHARE_DELETE,
                              NULL, 
                              OPEN_EXISTING, 
                              FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, NULL);
        
        if (m_hfile != INVALID_HANDLE_VALUE)
        {
            break;
        }
        DWORD dwErr = GetLastError();
        m_hfile = NULL;
        
        if (dwErr == ERROR_SHARING_VIOLATION && 
            (dwRetries < c_dwMaxSharingViolationRetries))
        {
            dwRetries++;
            Sleep(c_dwSharingViolationDelay);
        }
        else
        {
            ec = GetLastError();
            Log(LogID_Common, LogLevel_Error, "DISKIO",
                "error code - %u", ec);
            goto Done;          
        }
    } while (1);

    // get size of file
    //
    m_size.dw.low = GetFileSize(m_hfile, &m_size.dw.high);
    if (m_size.dw.low == INVALID_FILE_SIZE)
    {
        ec = GetLastError();
        Log(LogID_Common, LogLevel_Error, "DISKIO",
            "error code - %u", ec);
        if (ec)
            goto Done;
    }

    ec = Reset();
Done:
    return ec;
}

/*************************************************************************
**  Method: SEQREAD::IssueRead
**
**  Description: Issues a read for the next position and advance position. 
**
**  Paramaters:
**      pread           read block to use for read
**
**  Returns:
**      NO_ERROR=0      
**      ERROR_HANDLE_EOF 
**      GetLastError()  if read fails
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::IssueRead(READBUFFER *pread)
{
    // check to see if we have passed the end of the file
    //
    pread->m_fNotEof = FALSE;
    pread->m_fWait=FALSE;
    if (m_offsetNext.ddw >= m_size.ddw)
    {
        return ERROR_HANDLE_EOF;
    }

    // setup read postion
    //
    pread->m_overlap.Offset = m_offsetNext.dw.low;
    pread->m_overlap.OffsetHigh = m_offsetNext.dw.high;

    // read data
    //
    if (!ReadFile(m_hfile, pread->m_pbRead, m_readBufSize, 
        &pread->m_cbRead, &pread->m_overlap))
    {
        if (GetLastError() == ERROR_IO_PENDING)
        {
            pread->m_fWait=TRUE;
        }
        else
        {
            DWORD ec = GetLastError(); 
            Log(LogID_Common, LogLevel_Error, "DISKIO",
                "error code - %u", ec);
            return ec;
        }
    }

    // this was not past the end of the file
    //
    pread->m_fNotEof = TRUE;

    // setup for next read
    //
    m_offsetNext.ddw += m_readBufSize;

    return NO_ERROR;
}


/*************************************************************************
**  Method: SEQREAD::ReadNext
**
**  Description: Retrieves the next buffer and the number of bytes read. 
**
**  Paramaters:
**      ppbRead     pointer to return pointer buffer
**      pcbRead     pointer to return count of bytes set to 0 at eof
**
**  Returns:
**      NO_ERROR=0      
**
**  Notes:
**      PcbRead should only be different than the read buf size at the end
**      of the file.
**
**************************************************************************/
DWORD32 
SEQREAD::ReadNext(BYTE** ppbRead, DWORD *pcbRead)
{
    READBUFFER *    pread;

    if (m_iRead >= 0)
    {
        DWORD32     ec = NO_ERROR;

        // issue read for previous block since we are done with it
        //
        ec = IssueRead(m_rgReads+m_iRead);
        if (ec != ERROR_HANDLE_EOF && ec != NO_ERROR)
        {
            return ec;
        }

        // advance the read postion
        m_iRead++;
        if (m_iRead >= m_numReads)
            m_iRead = 0;
    }
    else
    {
        // m_iRead will be < 0 first time through
        // there is no previous read to issue since
        // they were done at init time.
        m_iRead = 0;
    }

    pread = &m_rgReads[m_iRead];

    // see if we are past the end of file
    BYTE* pbReadRet = (BYTE*)pread->m_pbRead;
    if (!pread->m_fNotEof)
    {
        pread->m_cbRead = 0;
        pbReadRet = NULL;
    }
    else if (pread->m_fWait)
    {
        pread->m_fWait = FALSE;
        if (!GetOverlappedResult(m_hfile, &pread->m_overlap, 
            &pread->m_cbRead, TRUE))
        {
            DWORD ec = GetLastError(); 
            Log(LogID_Common, LogLevel_Error, "DISKIO",
                "error code - %u", ec);

            return ec;
        }
    }

    *pcbRead = pread->m_cbRead;
    *ppbRead = pbReadRet;
    return NO_ERROR;
}

/*************************************************************************
**  Method: SEQREAD::WaitOutstandingAsyncRead
**
**  Description: Wait all outstanding asyncRead call to complete
**
**  Paramaters:
**
**  Returns:
**      NO_ERROR=0      
**      GetLastError()  if read fails
**
**  Notes:
**
**************************************************************************/
DWORD32 SEQREAD::WaitOutstandingAsyncRead()
{
    DWORD32     ec = NO_ERROR;
    int         iRead;

     // Make sure that there are no outstanding IOs
     //
     for (iRead=0; iRead< m_numReads; iRead++)
     {
        if (m_rgReads[iRead].m_fWait)
        {
            if (!GetOverlappedResult(m_hfile, &m_rgReads[iRead].m_overlap, 
                &m_rgReads[iRead].m_cbRead, TRUE))
            {
                ec = GetLastError();
                Log(LogID_Common, LogLevel_Error, "DISKIO",
                    "error code - %u", ec);

                goto Done;
            }
        }
    }

  Done:
    return ec;
}

/*************************************************************************
**  Method: SEQREAD::Reset
**
**  Description: Resets the sequential read to begining of file and 
**      reissues the initial reads.
**
**  Paramaters:
**
**  Returns:
**      NO_ERROR=0      
**      GetLastError()  if read fails
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::Reset(DWORD64 pos)
{
    DWORD32     ec = NO_ERROR;
    int         iRead;

    m_cbSkip = (DWORD)(pos%(DWORD64)m_readBufSize);
    m_offsetNext.ddw = pos - (DWORD64)m_cbSkip;

    ec = WaitOutstandingAsyncRead();
    if (ec != NO_ERROR)
    {
        goto Done;
    }

    // do initial reads
    //
    for (iRead=0; iRead< m_numReads; iRead++)
    {
        ec = IssueRead(m_rgReads+iRead);
        if (ec != NO_ERROR)
        {
            if (ec == ERROR_HANDLE_EOF)
            {
                ec = NO_ERROR;
                break;
            }
            else
                goto Done;
        }
    }

    // setup read index
    //
    m_iRead = -1;

    m_cbLeft = 0;
    m_pbRead = NULL;    

Done:
    return ec;
}

/*************************************************************************
**  Method: SEQREAD::GetDataPointer
**
**  Description: Gets a pointer to bytes read.  When crossing read buffers
**      this pointer goes to a temp buffer.  Data is only copied when a 
**      read crosses the read buffers.
**
**  Paramaters:
**      ppv     returns pointer to buffer with bytes of data.
**      cb      number of bytes requested.
**
**  Returns:
**      NO_ERROR=0
**      ERROR_INVALID_PARAMETER
**      ERROR_HANDLE_EOF
**      GetLastError()  if read fails
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::GetDataPointer( void ** ppv, DWORD cb, DWORD* pcbRead )
{
    DWORD32 ec = NO_ERROR;
    DWORD   cbSaved = 0;

    // If we have more left over data than the temp buffer size, we
    // can't use the temp buffer. If the user requested more than
    // left over data, just return the left over data.
    if (m_cbLeft < cb && m_cbLeft > c_cbReadCopy)
    {
        // if caller is not getting the bytes read then fail since
        // size returned will be different
        if (pcbRead == NULL)
        {
            return ERROR_INVALID_PARAMETER;
        }
        cb = m_cbLeft;
    }
    
    if (m_cbLeft < cb)
    {
        // save left over from previous buffer if needed
        //
        cbSaved = m_cbLeft;
        if (cbSaved > 0)
        {
            if (cb > c_cbReadCopy)
            {
                if (pcbRead == NULL)
                {
                    // if caller is not getting the bytes read then fail since
                    // size returned will be different
                    return ERROR_INVALID_PARAMETER;
                }
                
                // if we need to use the temp space then limit read size
                // to size of temp buffer
                cb = c_cbReadCopy;
            }
            
            // copy remaining data to internal buffer
            memcpy(m_rgb, m_pbRead, cbSaved);
        }

        // read next buffer
        //
        ec = ReadNext(&m_pbRead, &m_cbLeft);
        if (ec != NO_ERROR)
        {
            return ec;
        }

        // handle any skip setup duing reset
        //
        if (m_cbSkip)
        {
            m_cbLeft -=m_cbSkip;
            m_pbRead +=m_cbSkip;
            m_cbSkip = 0;
        }

        if ((m_cbLeft+cbSaved) < cb)
        {
            cb = m_cbLeft+cbSaved;
            if (!pcbRead || (cb==0))
                return ERROR_HANDLE_EOF;
        }

        if (pcbRead)
            *pcbRead = cb;

        if (cbSaved > 0)
        {
            // if there were saved bytes then we return a pointer to copy
            //
            m_cbLeft -= (cb-cbSaved);
            memcpy(m_rgb+cbSaved, m_pbRead, cb-cbSaved);
            *ppv = m_rgb;
            m_pbRead+= (cb-cbSaved);
            return NO_ERROR;
        }
    }

    if (pcbRead)
        *pcbRead = cb;
    *ppv = m_pbRead;
    m_pbRead+=cb;
    m_cbLeft -=cb;
    return NO_ERROR;
}


/*************************************************************************
**  Method: SEQREAD::GetData
**
**  Description: Copies read bytes into passed buffer.
**
**  Paramaters:
**      pv      buffer to copy data into.
**      cb      number of bytes requested.
**
**  Returns:
**      NO_ERROR=0
**      errors from GetDataPointer
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::GetData( __out_bcount(dwNumBytes) void * pv,DWORD cb )
{
    DWORD32 ec = NO_ERROR;
    BYTE*   pb;

    ec = GetDataPointer((void**)&pb, cb);
    if (ec)
        return ec;
    memcpy(pv, pb, cb);

    return ec;
}

/*************************************************************************
**  Method: SEQREAD::Skip
**
**  Description: Skips bytes in file.
**
**  Paramaters:
**      dwNumBytes      number of bytes to skip.
**
**  Returns:
**      NO_ERROR=0
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::Skip(DWORD dwNumBytes)
{
    // if skip in current buffer then just advance m_cbLeft & m_pbRead
    //
    if (m_cbLeft > dwNumBytes)
    {
        m_cbLeft -= dwNumBytes;
        m_pbRead += dwNumBytes;
    }
    // if skip in next buffer then just advance the read
    //
    else if (dwNumBytes < m_readBufSize)
    {
        void *  pv;

        dwNumBytes -= m_cbLeft;
        m_cbLeft = 0;

        return GetDataPointer( &pv, dwNumBytes );
    }
    // skip is large and just reset to the new location
    //
    else
    {
        return Reset( m_offsetNext.ddw+m_cbLeft+dwNumBytes );
    }

    return NO_ERROR;
}



#else
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/*************************************************************************
**  Method: SEQREAD::SEQREAD
**
**  Description: Constructs a SEQREAD object
**
**************************************************************************/
SEQREAD::SEQREAD() :
    m_pPool(NULL),
    m_pFile(NULL),
    m_size(0),
    m_rgReadBuf(NULL),
    m_readBufSize(0)
{
}

/*************************************************************************
**  Method: SEQREAD::~SEQREAD
**
**  Description: Destructs and frees memory for a SEQREAD object
**
**************************************************************************/
SEQREAD::~SEQREAD() 
{
    DoDispose();
}

/*************************************************************************
**  Method: SEQREAD::DoDispose
**
**  Description: Cleans up a SEQREAD object.
**
**************************************************************************/

void SEQREAD::DoDispose() 
{
    if (m_pFile != NULL)
    {
        apr_file_close(m_pFile);
        m_pFile = NULL;
    }

    if (m_pPool != NULL)
    {
        apr_pool_destroy(m_pPool);
        m_pPool = NULL;
    }

    m_rgReadBuf = NULL;
    m_readBufSize = 0;
    m_size = 0;
}



/*************************************************************************
**  Method: SEQREAD::DoInit
**
**  Description: Initializes the object.  Allocate memory, open file and 
**      issue initial reads.
**
**  Paramaters:
**      szFile          name of file to open
**      maxReads        max number of outstanding reads
**
**  Returns:
**      
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::DoInit(
                const TCHAR* wszFile,
                int         maxReads,
                DWORD32     readBufSize,
                BOOL        readwrite)
{
    apr_status_t ec = APR_SUCCESS;
    char         szFile[MAX_PATH];

    if (wszFile == NULL)
    {
        ec = APR_BADARG;
        goto Done;
    }

    if (m_pPool != NULL)
    {
        DoDispose();
    }

    ec = apr_pool_create(&m_pPool, NULL);
    if (ec != APR_SUCCESS)
    {
        goto Done;
    }

    // Allocate memory
    m_readBufSize = readBufSize;
    m_rgReadBuf = apr_palloc(m_pPool, m_readBufSize);
 
    if (m_rgReadBuf == NULL)
    {
        ec = APR_ENOSHMAVAIL;
        goto Done;
    }
    
    PALString::Convert(szFile, ARRAYSIZE(szFile), wszFile, TSTRLEN(wszFile));

    // Get the file size:
    apr_finfo_t finfo;
    
    ec = apr_stat(&finfo, szFile, APR_FINFO_SIZE | APR_FINFO_TYPE, m_pPool);
    if (ec != APR_SUCCESS)
    {
        goto Done;
    }

    if (finfo.filetype != APR_REG)
    {
        ec = APR_EGENERAL;
        goto Done;
    }

    m_size = finfo.size;

    // Open file
    ec = apr_file_open(
        &m_pFile,
        szFile,
        readwrite ? APR_READ | APR_WRITE | APR_BUFFERED : APR_READ | APR_BUFFERED,
        APR_OS_DEFAULT,
        m_pPool);
    if (ec != APR_SUCCESS)
    {
        goto Done;
    }

Done:
    return (DWORD32) ec;
}

/*************************************************************************
**  Method: SEQREAD::Reset
**
**  Description: Resets the sequential read to begining of file and 
**      reissues the initial reads.
**
**  Paramaters:
**
**  Returns:
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::Reset(DWORD64 pos)
{
    apr_status_t ec = APR_SUCCESS;
    apr_off_t offset = pos;

    if (m_pFile == NULL)
    {
        ec = APR_EGENERAL;
        goto Done;
    }
    
    ec =  apr_file_seek(m_pFile, APR_SET, &offset); 
    if (ec != APR_SUCCESS)
    {
        goto Done;
    }

Done:
    return (DWORD32) ec;
}

/*************************************************************************
**  Method: SEQREAD::GetDataPointer
**
**  Description: Gets a pointer to bytes read.  When crossing read buffers
**      this pointer goes to a temp buffer.  Data is only copied when a 
**      read crosses the read buffers.
**
**  Paramaters:
**      ppv     returns pointer to buffer with bytes of data.
**      cb      number of bytes requested.
**
**  Returns:
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::GetDataPointer( void ** ppv, DWORD cb, DWORD* pcbRead )
{
    apr_status_t ec = APR_SUCCESS;
    
    if (ppv == NULL)
    {
        ec = APR_BADARG;
        goto Done;
    }

    if (m_pFile == NULL)
    {
        ec = APR_EGENERAL;
        goto Done;
    }

    apr_size_t cbRead;

    ec = apr_file_read_full(m_pFile, m_rgReadBuf, cb, &cbRead);
    if (ec != APR_SUCCESS && ec != APR_EOF)
    {
        goto Done;
    }

    ec = APR_SUCCESS;

    *ppv = m_rgReadBuf;

    if (pcbRead != NULL)
    {
        *pcbRead = (DWORD) cbRead;
    }

Done:
    return (DWORD32) ec;
}

/*************************************************************************
**  Method: SEQREAD::GetData
**
**  Description: Copies read bytes into passed buffer.
**
**  Paramaters:
**      pv      buffer to copy data into.
**      cb      number of bytes requested.
**
**  Returns:
**
**  Notes:
**
**************************************************************************/
DWORD32 
SEQREAD::GetData( __out_bcount(dwNumBytes) void * pv, DWORD cb )
{
    apr_status_t ec = APR_SUCCESS;
    
    if (pv == NULL)
    {
        ec = APR_BADARG;
        goto Done;
    }

    if (m_pFile == NULL)
    {
        ec = APR_EGENERAL;
        goto Done;
    }

    apr_size_t cbRead;
    ec = apr_file_read_full(m_pFile, pv, cb, &cbRead);
    if (ec != APR_SUCCESS && ec != APR_EOF)
    {
        goto Done;
    }

    ec = APR_SUCCESS;

Done:
    return (DWORD32) ec;
}

#endif

/*************************************************************************
**  Method: SEQWRITE::SEQWRITE
**
**  Description: Constructs a SEQWRITE object
**
**************************************************************************/
void SEQWRITE::Construct()
{
    m_bUsedIAlloc = FALSE;
    m_bOpened = FALSE;
    m_pBuffer = NULL;
    m_cbUsedSize = 0;
    m_cbMaxSize = 0;
}

SEQWRITE::SEQWRITE()
{
    Construct();
}

SEQWRITE::~SEQWRITE()
{
    DoDispose();
}

/*************************************************************************
**  Method: SEQWRITE::DoDispose
**
**  Description: Destructs and frees memory for a SEQWRITE object
**
**************************************************************************/
void SEQWRITE::DoDispose() 
{
    Close();

    if (!m_bUsedIAlloc)
    {
        // Not using IAllocator.
        if (m_pBuffer != NULL)
        {
            free(m_pBuffer);
            m_pBuffer = NULL;
        }
    }
}

/*************************************************************************
**  Method: SEQWRITE::DoInit
**
**  Description: Initializes the object.  Allocate memory, open file and 
**      prepare for writing.
**
**  Paramaters:
**      szFile          name of file to open
**      cbWrite         size of the write buffer
**      maxWrites       (Not used)
**
**  Returns:
**
**  Notes:
**      File is always created and will overwrite existing files.
**
**************************************************************************/
DWORD32 SEQWRITE::DoInit(
    IAllocator* pAlloc, 
    int cbWrite, 
    int maxWrites)
{
    UNREFERENCED_PARAMETER(maxWrites);
    Construct();

    DWORD32 ec = APR_SUCCESS;
    MallocAllocator mallocalloc;
    IAllocator *pMemAlloc = &mallocalloc;

    if (pAlloc != NULL)
    {
        m_bUsedIAlloc = TRUE;
        pMemAlloc = pAlloc;
    }
    else
    {
        m_bUsedIAlloc = FALSE;
    }

    m_pBuffer = pMemAlloc->Allocate(cbWrite);
    if (m_pBuffer == NULL)
    {
        ec = APR_ENOMEM;
        goto Done;
    }

    m_bOpened = FALSE;
    m_cbUsedSize = 0;
    m_cbMaxSize = cbWrite;
    memset(m_wszFile, 0, sizeof(m_wszFile));

Done:
    return ec;
}
    
DWORD32 SEQWRITE::Open(const TCHAR* wszFile)
{
    DWORD32 ec = APR_SUCCESS;
    PALFileHandle fd = PAL_INVALID_HANDLE;

    m_bOpened = FALSE;
    if (PALString::Copy(m_wszFile, ARRAYSIZE(m_wszFile), wszFile))
    {
        ec = APR_BADARG;
        goto Done;
    }
        
    fd = OpenFile(TRUE);
    if (fd < 0)
    {
        ec = APR_BADARG;
        goto Done;
    }

    m_cbUsedSize = 0;
    m_bOpened = TRUE;
Done:
    if (PAL_FILE_HANDLE_VALID(fd))
    {
        CloseFile(fd);
    }

    return ec;
}

DWORD32 SEQWRITE::Close()
{
    DWORD32 ec = Flush();
    m_bOpened = FALSE;
    return ec;
}

DWORD32 SEQWRITE::Flush()
{
    DWORD32 ec = APR_SUCCESS;
    PALFileHandle fd = PAL_INVALID_HANDLE;

    if (m_bOpened == FALSE)
    {
        ec = APR_BADARG;
        goto Done;
    }

    if (m_cbUsedSize > 0)
    {
        fd = OpenFile(FALSE);
        if (!PAL_FILE_HANDLE_VALID(fd))
        {
            ec = APR_BADARG;
            goto Done;
        }

        ec = WriteToFile(fd, m_pBuffer, m_cbUsedSize);
        if (ec != APR_SUCCESS)
        {
            goto Done;
        }
        m_cbUsedSize = 0;
    }

Done:
    if (PAL_FILE_HANDLE_VALID(fd))
    {
        CloseFile(fd);
    }
    return ec;
}

#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

/*************************************************************************
**  Method: SEQWRITE::Write
**
**  Description: Calls WriteInternal for the right size buffers.
**
**  Paramaters:
**      pbWrite         data to write
**      cbWrite         number of bytes of data to write
**
**  Returns:
**
**  Notes:
**
**************************************************************************/
DWORD32 SEQWRITE::Write(const void* pbWrite, Size_t cbWrite)
{
    DWORD32 ec = APR_SUCCESS;
    PALFileHandle fd = PAL_INVALID_HANDLE;

    if (pbWrite == NULL)
    {
        ec = APR_BADARG;
        goto Done;
    }

    if (m_bOpened == FALSE)
    {
        ec = APR_BADARG;
        goto Done;
    }

    while (cbWrite > 0)
    {
        // Copy to buffer.
        Size_t cbToBeWritten = min(cbWrite, m_cbMaxSize - m_cbUsedSize);
        memcpy(((BYTE *) m_pBuffer) + m_cbUsedSize, pbWrite, cbToBeWritten);
        pbWrite = ((BYTE *) pbWrite) + cbToBeWritten; 
        m_cbUsedSize += cbToBeWritten;
        cbWrite -= cbToBeWritten;

        // If buffer is full, write it out.
        if (m_cbUsedSize == m_cbMaxSize)
        {
            if (!PAL_FILE_HANDLE_VALID(fd))
            {
                fd = OpenFile(FALSE);
                if (!PAL_FILE_HANDLE_VALID(fd))
                {
                    ec = APR_BADARG;
                    goto Done;
                }
            }

            ec = WriteToFile(fd, m_pBuffer, m_cbUsedSize);
            if (ec != APR_SUCCESS)
            {
                goto Done;
            }

            m_cbUsedSize = 0;
        }
    }
    
Done:
    if (PAL_FILE_HANDLE_VALID(fd))
    {
        CloseFile(fd);
    }

    return ec;
}

void SEQWRITE::CloseFile(PALFileHandle fd)
{
#if APR_HAVE_WINDOWS_H
    CloseHandle(fd);
#else
    close(fd);
#endif
}


PALFileHandle SEQWRITE::OpenFile(BOOL bDeleteExisting)
{
#if APR_HAVE_WINDOWS_H
    // not using FILE_FLAG_NO_BUFFERING because it requires memory buffer alingment and I guess buffering is ok since we close the file before leaving the lock
    return CreateFile(m_wszFile, FILE_APPEND_DATA, 0, NULL, bDeleteExisting ? CREATE_ALWAYS : OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
#else
    char szFile[MAX_PATH];
    if (PALString::Convert(szFile, ARRAYSIZE(szFile), m_wszFile, TSTRLEN(m_wszFile)) != 0) // expensive for every 64k write ?
    {
        return PAL_INVALID_HANDLE;
    }

    return open(szFile,
        O_RDWR | O_CREAT | (bDeleteExisting ? O_TRUNC : O_APPEND),
        S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
#endif
}

DWORD32 SEQWRITE::WriteToFile(PALFileHandle fd, const void* pbWrite, Size_t cbWrite)
{
    DWORD32 ec = APR_SUCCESS;
    while (cbWrite > 0)
    {
#if APR_HAVE_WINDOWS_H
        DWORD dw = 0;
        Size_t cbwritten = WriteFile(fd, pbWrite, (DWORD)cbWrite, &dw, NULL) ? dw : -1;
#else
        ssize_t cbwritten = write(fd, pbWrite, cbWrite);
#endif
        if (cbwritten == -1 || cbwritten > cbWrite)
        {
            ec = APR_BADARG;
            goto Done;
        }

        cbWrite -= cbwritten;
        pbWrite = (BYTE *)pbWrite + cbwritten;
    }

Done:
    return ec;
}

