/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/

#ifndef __DISKIO
#define __DISKIO

#include "basic_types.h"
#include "CritSec.h"
#include "shared.h"

#if APR_HAVE_WINDOWS_H
// READBUFFER struct containing data for a single read
//
class READBUFFER
{
public:
    OVERLAPPED  m_overlap;
    void *      m_pbRead;   
    DWORD       m_cbRead;
    BOOL        m_fWait;
    BOOL        m_fNotEof;
};

/*************************************************************************
**  Class: SEQREAD
**
**  Description: Sequential file reader class.  Efficiently reads the file
**      sequentially using unbuffered and asynch I/O.  Can issue many reads 
**      in order to keep many disks busy and efficiently read the disk 
**      sequentially.
**
**************************************************************************/
class SEQREAD
{
public:
    static const int        c_maxReads = 64;
    static const int        c_maxReadsDefault = 4;
    static const DWORD32    c_readBufSize = 64*1024;
    static const int        c_cbReadCopy = c_readBufSize;
    static const UInt32     c_dwMaxSharingViolationRetries = 5;
    static const UInt32     c_dwSharingViolationDelay = 250;

private:
    int         m_numReads;
    int         m_iRead;
    READBUFFER  m_rgReads[c_maxReads];
    DDWORD      m_offsetNext;
    HANDLE      m_hfile;
    DDWORD      m_size;
    DWORD32     m_readBufSize;

    DWORD       m_cbLeft;
    BYTE*       m_pbRead;
    DWORD       m_cbSkip;
    BYTE        *m_rgb;

    DWORD32 IssueRead(READBUFFER *pread);
    DWORD32 ReadNext(BYTE** ppbRead, DWORD *pcbRead);

public:
    SEQREAD();
    ~SEQREAD();

    DWORD32 WaitOutstandingAsyncRead();
    DWORD32 DoInit(const TCHAR* wszFile, int maxReads = c_maxReadsDefault, DWORD32 readBufSize = c_readBufSize, BOOL readwrite = false);
    void    DoDispose();
    DWORD32 Reset(DWORD64 pos = 0);
    
    DWORD32 GetDataPointer( void ** ppvBuffer, DWORD dwNumBytes, DWORD* pcbRead=NULL );
    DWORD32 GetData(__out_bcount(dwNumBytes) void * pvBuffer, DWORD dwNumBytes );
    DWORD32 Skip(DWORD dwNumBytes);

    DWORD64 FileSize() const {return m_size.ddw;} 
    HANDLE  FileHandle() const {return m_hfile;}
};


// WRITEBUFFER struct containing data for a single read
//
class WRITEBUFFER
{
public:
    OVERLAPPED  m_overlap;
    void *      m_pb;   
    DWORD       m_cb;
    BOOL        m_fWait;
};

#else
// Linux platform.


/*************************************************************************
**  Class: SEQREAD
**
**  Description: Sequential file reader class.  Efficiently reads the file
**      sequentially using unbuffered and asynch I/O.  Can issue many reads 
**      in order to keep many disks busy and efficiently read the disk 
**      sequentially.
**
**************************************************************************/
class SEQREAD
{
public:
    static const int        c_maxReads = 64;
    static const int        c_maxReadsDefault = 4;
    static const DWORD32    c_readBufSize = 64*1024;
    static const int        c_cbReadCopy = c_readBufSize;
    static const UInt32     c_dwMaxSharingViolationRetries = 5;
    static const UInt32     c_dwSharingViolationDelay = 250;

public:
    SEQREAD();
    ~SEQREAD();

    DWORD32 DoInit(const TCHAR* wszFile, int maxReads = c_maxReadsDefault, DWORD32 readBufSize = c_readBufSize, BOOL readwrite = false);
    void    DoDispose();
    DWORD32 Reset(DWORD64 pos = 0);
    
    DWORD32 GetDataPointer( void ** ppvBuffer, DWORD dwNumBytes, DWORD* pcbRead=NULL );
    DWORD32 GetData(__out_bcount(dwNumBytes) void * pvBuffer, DWORD dwNumBytes );

    DWORD64 FileSize() const {return m_size;} 

private:
    apr_pool_t *m_pPool;
    apr_file_t *m_pFile;
    UInt64      m_size;

    void       *m_rgReadBuf;
    DWORD32     m_readBufSize;
};

#endif

#if APR_HAVE_WINDOWS_H
typedef HANDLE PALFileHandle;
#define PAL_INVALID_HANDLE INVALID_HANDLE_VALUE
#define PAL_FILE_HANDLE_VALID(h) (PAL_INVALID_HANDLE != h)
#else    
typedef int PALFileHandle;
#define PAL_INVALID_HANDLE -1
#define PAL_FILE_HANDLE_VALID(h) (h >= 0)
#endif


/*************************************************************************
**  Class: SEQWRITE
**
**  Description: Sequential file writer class.  Efficiently writes the file
**      sequentially using unbuffered and asynch I/O.  
**
**************************************************************************/
class SEQWRITE
{
public:
    static const int        c_maxWrites = 64;
    static const int        c_maxWritesDefault = 2;
    static const DWORD32    c_writeBufSizeDefault = 128*1024;
    static const int        c_cbWriteMax = 1024;

public:
    SEQWRITE();
    ~SEQWRITE();
    void Construct();

    DWORD32 DoInit(IAllocator* pAlloc = NULL,
                   int cbWrite = c_writeBufSizeDefault, 
                   int maxWrites = c_maxWritesDefault);
	void	DoDispose();

    DWORD32 Open(const TCHAR* wszFile);
    DWORD32 Close();
    DWORD32 Flush();
    DWORD32 Write(const void* pbWrite, Size_t cbWrite);

private:
    TCHAR   m_wszFile[MAX_PATH];
    BOOL    m_bUsedIAlloc;
    BOOL    m_bOpened;
    void   *m_pBuffer;
    Size_t  m_cbUsedSize;
    Size_t  m_cbMaxSize;
    PALFileHandle OpenFile(BOOL bDeleteExisting);
    void CloseFile(PALFileHandle fd);
    DWORD32 WriteToFile(PALFileHandle fd, const void* pbWrite, Size_t cbWrite);
};

#endif // __DISKIO
