/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "DeltaMerge.h"
#include "UncompressingReader.h"
#include "CompressingWriter.h"
#include "DeltaFile.h"
#include "SnapshotFile.h"
#include "SitemapOutput.h"
#include "RobotsUpdater.h"
#include "SimpleHashTable.h"
#include "ServiceLogging.h"
#include "counters.h"
#include "DiskSpaceTracker.h"
#include "SSPTypes.h"
#include "SSPHash.h"
#include "SSPCommon.h"
#include "SSPRobots.h"
#include "PALString.h"
#include "PALFile.h"

#if APR_HAVE_WINDOWS_H
#include <windows.h>
#include <WinHTTP.h>
#endif

// BUGBUG: these globals will cause problems if the service ever uses multiple threads for sitemap generation.
// However, using multiple threads for sitemap generation seems a bad idea as the main bottleneck would be disk IO.
static FILE* g_fpSiteLog;
Counters g_counters;

#define SSP_MAX(a,b) (((a)>(b)) ? (a) : (b))

#define MAX_PING_URL 2048

void SiteLog(const char* szFormat, ...)
{
    if (!g_fpSiteLog) return;
  
    struct tm tms;
    PluginGetLocalTime(&tms);
    fprintf(g_fpSiteLog, "%04u/%02u/%02u %02u:%02u:%02u ", 1900 + tms.tm_year, 1 + tms.tm_mon, tms.tm_mday, tms.tm_hour, tms.tm_min, tms.tm_sec);

    va_list args;
    va_start(args, szFormat);
    vfprintf(g_fpSiteLog, szFormat, args);
    va_end(args);
    
    fputs("\n", g_fpSiteLog);
}

class SigInfo
{
public:
    Signature sigFirst;
    Signature sigLast;
    unsigned tFirst; // time of first sigFirst instance
    unsigned tSecond; // time sigFirst changed to something else
    unsigned tLast; // time sigLast was introduced (not necessarily its last instance)
    VisitCount visitCount;

    SigInfo()
    {
        memset(this, 0, sizeof(*this));
    }
    BOOL Initialized() const
    {
        return sigLast ? TRUE : FALSE;
    }
    unsigned LastMod() const
    {
        return (tLast != tFirst) ? tLast : 0;
    }
    void Initialize(Signature sig, unsigned t32, VisitCount count)
    {
        sigFirst = sig;
        sigLast = sig;
        tFirst = t32;
        tSecond = t32;
        tLast = t32;
        visitCount = count;
    }
};

typedef struct
{ // BUGBUG: using about 40 bytes to track signature changes
    SigInfo normal;
    SigInfo bots;
    unsigned short cbUrl;
    unsigned uUrlOffset;
    unsigned char changePeriod;
} MergeCacheEntry;

class FileStats
{
public:
    Size_t cEntries;
    Size_t cbUrls;

    FileStats() { memset(this, 0, sizeof(*this)); }
};

class DeltaCache
{
    SimpleHashTable<MergeCacheEntry> m_ht;
    char* m_pcUrls;
    Size_t m_cbUrls;
    Size_t m_cbUsed;
public:
    DeltaCache() : m_pcUrls(NULL) {}
    ~DeltaCache() { if (m_pcUrls) delete[] m_pcUrls; }
    BOOL Initialize(Size_t cbMem);
    BOOL Add(unsigned flags, Signature, VisitCount, unsigned t32, const char* pcUrl, Size_t cbUrl);
    MergeCacheEntry* Find(const char* pcUrl, Size_t cbUrl);
    const char* Url(unsigned offset) { return m_pcUrls + offset; }
    BOOL Enumerate(void* ctx, BOOL(*pfn)(void* ctx, MergeCacheEntry* p)) { return m_ht.Enumerate(ctx, pfn); }
    BOOL CheckCapacity(FileStats*);
};

BOOL DeltaCache::Initialize(Size_t cbMem)
{
    BOOL fRet = FALSE;

    const Size_t cAvgUrlLength = 64;
    Size_t cEntries = cbMem / (cAvgUrlLength + m_ht.BytesPerEntry());
    Size_t cbHT = m_ht.MemoryNeeded(cEntries);
    if (cbHT > cbMem) 
    {
        goto Exit;
    }
    m_cbUrls = cbMem - cbHT;
#if !defined(_X86_) // 64-bit - is there a better way to test that ?
    if (m_cbUrls > 65536ULL * 65536) m_cbUrls = 65536ULL * 65536; // protect 32-bit uUrlOffset
#endif
    if (!m_ht.Initialize(cEntries)) SGS_ERROR;
    if (NULL == (m_pcUrls = new char[m_cbUrls])) SGS_ERROR;

    m_cbUsed = 0;
    fRet = TRUE;
Exit:
    return fRet;
}

#define CHANGEPERIOD_UNKNOWN 255

BOOL DeltaCache::Add(unsigned flags, Signature sig, VisitCount visitCount, unsigned t32, const char* pcUrl, Size_t cbUrl)
{
    unsigned short cb = (unsigned short)cbUrl;
    if (m_cbUsed + cb > m_cbUrls) return FALSE;

    SigInfo si;
    si.Initialize(sig, t32, visitCount);
    
    MergeCacheEntry e;
    e.cbUrl = cb;
    e.uUrlOffset = (unsigned)m_cbUsed;
    e.changePeriod = CHANGEPERIOD_UNKNOWN;
    if (flags & DELTA_FLAG_BOT)
    {
        e.bots = si;
    }
    else
    {
        e.normal = si;
    }

    memcpy(m_pcUrls + m_cbUsed, pcUrl, cb);  m_cbUsed += cb;
    if (!m_ht.Add(SSPHash(pcUrl, cb), e)) return FALSE;

    return TRUE;
}

MergeCacheEntry* DeltaCache::Find(const char* pcUrl, Size_t cbUrl)
{
    return m_ht.Get(SSPHash(pcUrl, cbUrl));
}

static unsigned short AddVisitCount(unsigned short c1, unsigned short c2)
{
    unsigned c = (unsigned)c1 + (unsigned)c2;
    if (c > 65535) c = 65535;
    return (unsigned short)c;
}

static void UpdateChangePeriod(unsigned char* pPeriod, unsigned flags, unsigned t32Old, unsigned t32New)
{
    if (flags & SNAPSHOT_FLAG_FIRSTDISCOVERED) // previous modification time unknown
    {
        return;
    }

    unsigned uDays = (t32New - t32Old) / T32_PER_DAY;
    if (uDays > CHANGEPERIOD_UNKNOWN - 1)
        uDays = CHANGEPERIOD_UNKNOWN - 1;
    if (*pPeriod > uDays)
        *pPeriod = (unsigned char)uDays;
}

static BOOL DeltaFileStats(const TCHAR* wszFile, FileStats* pStats)
{
    MyFileReader reader;
    BOOL fRet = FALSE;
    unsigned flags;
    Signature sig;
    VisitCount visitCount;
    char pcUrl[2048];
    Size_t cbUrl;
    Size_t cEntries = 0, cbUrls = 0;

    if (!reader.Initialize(wszFile, c_rDeltaMagic)) SGS_ERROR;

    while (cbUrl = sizeof(pcUrl), ReadDeltaFileEntry(&reader, &flags, &sig, &visitCount, pcUrl, &cbUrl))
    {
        cEntries++;
        cbUrls += cbUrl;
    }
    
    pStats->cEntries = cEntries;
    pStats->cbUrls = cbUrls;
    return TRUE;
Exit:
    return FALSE;
}

BOOL DeltaCache::CheckCapacity(FileStats* p)
{
    return (m_cbUsed + p->cbUrls <= m_cbUrls) && (p->cEntries < m_ht.SpaceLeft());
}

class DeltaInputEntry
{
public:
    unsigned flags;
    Signature sig;
    VisitCount visitCount;
    char pcUrl[2048];
    Size_t cbUrl;
    BOOL Read(MyFileReader* pReader);
    BOOL AddToCache(DeltaCache* pCache, unsigned timestamp);
};

BOOL DeltaInputEntry::Read(MyFileReader* pReader)
{
    cbUrl = sizeof(pcUrl);
    return ReadDeltaFileEntry(pReader, &flags, &sig, &visitCount, pcUrl, &cbUrl);

}
BOOL DeltaInputEntry::AddToCache(DeltaCache* pCache, unsigned timestamp)
{
    MergeCacheEntry* c = pCache->Find(pcUrl, cbUrl);
    if (c) // already exists in cache
    {
        unsigned tPrevChange = SSP_MAX(c->normal.LastMod(), c->bots.LastMod());
        SigInfo* e = (flags & DELTA_FLAG_BOT) ? &c->bots : &c->normal;
        if (e->Initialized())
        {
            if (e->sigLast != sig) // signature change
            {
                if (e->sigFirst == e->sigLast) // this is the first change we see within the delta batch
                { // BUGBUG: if signature changes A=>B=>A=>X, we won't UpldateChangePeriod() for X
                    e->tSecond = timestamp;
                    // UpdateChangePeriod logic happens after we load the snapshot values
                }
                if (tPrevChange) // already had a signature change within this delta batch
                {
                    UpdateChangePeriod(&c->changePeriod, 0, tPrevChange, timestamp);
                }
                e->sigLast = sig;
                e->tLast = timestamp;
            }
            e->visitCount = AddVisitCount(e->visitCount, visitCount); // BUGBUG: happens twice for files not completely merged when the cache is full
        }
        else // signature not yet initialized
        {
            e->Initialize(sig, timestamp, visitCount);
        }
    }
    else
    {
        if (!pCache->Add(flags, sig, visitCount, timestamp, pcUrl, cbUrl)) return FALSE; // this is expected when the cache is full
    }
    return TRUE;
}


static BOOL LoadDeltaFile(const TCHAR* wszFile, DeltaCache* pCache, time_t tFile)
{
    unsigned timestamp = (unsigned)(tFile / SEC_PER_T32);
    MyFileReader reader;
    BOOL fRet = FALSE;
    DeltaInputEntry e;

    if (!reader.Initialize(wszFile, c_rDeltaMagic)) SGS_ERROR;
    while (e.Read(&reader))
    {
        if (!e.AddToCache(pCache, timestamp))
        {
            break; // no SGS_ERROR because this is expected when the cache is full
        }
    }

    fRet = TRUE;
Exit:
    return fRet;
}


static const TCHAR *c_wszSnapshot = TEXT("snapshot.dat"), *c_wszTmp = TEXT("snapshot.tmp");
static const TCHAR* c_wszDeltaSnapshot = TEXT("DeltaSnapshot.dat"), *c_wszDeltaTmp = TEXT("DeltaSnapshot.tmp");

class SnapshotEntry : public BaseSnapshotEntry
{
public:
    SnapshotEntry(const MergeCacheEntry* p, const char* pcUrl);
    SnapshotEntry() { }
    BOOL MergeWithCache(DeltaCache* pCache, unsigned t32Now, SSPConfig* pConfig, SitemapOutput* pDeltaSitemap, VisitCount* pDeltaVisitCount);
    unsigned MergeWithSigInfo(Signature* pSig, const SigInfo* c, unsigned* ptLastSeen);
};
   
static void ApplyNonZeroMin(unsigned* pDst, unsigned src)
{
    if (*pDst > src && src)
        *pDst = src;
}

SnapshotEntry::SnapshotEntry(const MergeCacheEntry* p, const char* szUrl)
{ // generate snapshot entry from cache delta entries - happens when a given URL is added to the snapshot for the first time
    m_sig = p->normal.sigLast;
    m_sigBots = p->bots.sigLast;
    m_visitCount = p->normal.visitCount;
    m_lastSeen = SSP_MAX(p->normal.tLast, p->bots.tLast); // technically this is last modified, but lastSeen doesn't need great precision

    unsigned lastMod = p->normal.LastMod(),
       botsLastMod = p->bots.LastMod();
    m_lastMod = 0;
    m_flags = 0;
    if (lastMod < botsLastMod)
    {
        lastMod = botsLastMod;
        m_flags |= SNAPSHOT_FLAG_ROBOTS;
    }
    m_lastMod = lastMod;

    if (!m_lastMod) // no signature change during the initial batch of delta entries for this URL
    {
        m_lastMod = 0xffffffff;
        ApplyNonZeroMin(&m_lastMod, p->normal.tLast);
        ApplyNonZeroMin(&m_lastMod, p->bots.tLast);

        m_flags |= SNAPSHOT_FLAG_FIRSTDISCOVERED; // means it's "first discovered" rather than "last modified"
        if (!p->normal.tLast)
        {
            m_flags |= SNAPSHOT_FLAG_ROBOTS;
        }
    }
         
    m_changePeriod = p->changePeriod;
    m_cbUrl = p->cbUrl;
    if (m_cbUrl > sizeof(m_pbUrl))
        m_cbUrl = sizeof(m_pbUrl);
    memcpy(m_pbUrl, szUrl, m_cbUrl);
}

typedef BOOL (*SnapshotEntryCallback)(void* pv, SnapshotEntry* p);

static BOOL EnumerateSnapshotFile(const TCHAR* wszFile, void* pv, SnapshotEntryCallback pfn)
{
    MyFileReader reader;
    SnapshotEntry s;
    BOOL fRet = FALSE;

    if (!reader.Initialize(wszFile, c_rSnapshotMagic))
    {
        SGS_ERROR;
    }
    do
    {
        if (!s.Read(&reader)) SGS_ERROR;
        if (s.IsEmpty())
        {
            break;
        }

        if (!pfn(pv, &s))
        {
            goto Exit;
        }
    } while (1);

    fRet = TRUE;
Exit:
    return fRet;
}

const char* ChangeFreq(unsigned char changePeriod)
{
    if (CHANGEPERIOD_UNKNOWN == changePeriod) return NULL;

    if (changePeriod == 0) return "hourly";
    if (changePeriod <= 3) return "daily";
    if (changePeriod <= 15) return "weekly";
    if (changePeriod <= 100) return "monthly";
    return "yearly";
}

static float Priority(VisitCount visitCount, unsigned uUnity)
{

    if(uUnity == 0)
    {
        // default value = 0.5
        return 0.5f;
    }

    // the algorithm is to make sure URLs with visitCount around avg visit count will get 0.5
    float f = visitCount / (2 * (float)uUnity);
    if (f > 1.0) f = 1.0;
    if (f < 0.1) f = 0.1f; // Supposedly some parsers cannot handle 0.0
    return f;
}

typedef struct
{
    DeltaCache* pCache;
    CompressingFileWriter* pWriter;
    CompressingFileWriter* pDeltaWriter;
    SitemapOutput* pDeltaSitemap;
    SSPConfig* pConfig;
    Size_t cbSnapshot;
    Robots* pRobots;
}
CacheDumpContext;

BOOL DumpCacheEntry(void* pvContext, MergeCacheEntry* p)
{
    BOOL fRet = FALSE;
    const char* pcUrl;
    Size_t cb;

    CacheDumpContext* pctx = (CacheDumpContext*)pvContext;
    VisitCount overallVisitCount = AddVisitCount(p->normal.visitCount, p->bots.visitCount);
    VisitCount visitCount = p->normal.visitCount;
    if (!overallVisitCount) return TRUE; // handled during merge with snapshot

    pcUrl = pctx->pCache->Url(p->uUrlOffset);
    if (!pctx->pRobots->Allowed(pcUrl, p->cbUrl))
    {
        g_counters.cExcludedLater.Increment();
        return TRUE; // TRUE means continue enumerating other cache entries
    }

    SnapshotEntry s(p, pcUrl);

    if (0 == (cb = s.Write(pctx->pWriter))) SGS_ERROR;

    if (pctx->pDeltaWriter)
    {
        if (0 == (cb = s.Write(pctx->pDeltaWriter))) SGS_ERROR;
    }

    // for delta sitemap, given we don't have global picture, use fixed 0.5 as global priority
    if (!pctx->pDeltaSitemap->AddEntry(p->normal.sigLast, p->bots.sigLast, pcUrl, p->cbUrl, SSP_MAX(p->normal.tLast, p->bots.tLast), 
        p->bots.tLast > p->normal.tLast, ChangeFreq(p->changePeriod), 0.5, Priority(visitCount, pctx->pConfig->uNumVisitsPerDeltaPriority))) SGS_ERROR;
    g_counters.cUrlsOutputted.m_val++;

    pctx->cbSnapshot += cb;
    if (pctx->cbSnapshot > 1048576 * (Size_t)pctx->pConfig->uMaxSnapshotFileMB) SGS_ERROR;

    fRet = TRUE;
Exit:
    return fRet;
}

unsigned SnapshotEntry::MergeWithSigInfo(Signature* pSig, const SigInfo* c, unsigned* ptLastSeen)
{ // returns last change time if signature change detected
    if (!c->Initialized())
    {
        return 0;
    }

    unsigned tLastMod = 0;
    if (*pSig)
    {
        // Even though we loaded cache entries first, chronologically Snapshot entry is earlier because cache entries came from delta file.
        // Compute the interval between the last modified time from the snapshot file and the first time a different signature appeared in the current batch of delta files.
        // (we could not have considered this interval while loading delta files because we hadn't loaded the snapshot file at that point)
        tLastMod = c->tLast;
        if (*pSig != c->sigFirst)
        { // first delta entry signature already different from the snapshot file signature
            UpdateChangePeriod(&m_changePeriod, m_flags, m_lastMod, c->tFirst);
        }
        else if (c->sigFirst != c->sigLast)
        { // first delta entry had the same signature as snapshot, but the signature changed during a later delta entry
            UpdateChangePeriod(&m_changePeriod, m_flags, m_lastMod, c->tSecond); // BUGBUG wasting 4 bytes (tSecond) just to get this case right
        }
        else
        { // this batch of delta files introduces no signature change
            tLastMod = 0;
        }
    }
    
    *pSig = c->sigLast;
    if (*ptLastSeen < c->tLast && *pSig != DEADLINK_SIGNATURE) // do not update last seen if it is deadlink so that the deadlink can be removed from the sitemap after the grace period
    {
        *ptLastSeen = c->tLast; // technically tLast is not lastSeen, but it's close enough for the way lastSeen is used
    }

    return tLastMod;
}

BOOL SnapshotEntry::MergeWithCache(DeltaCache* pCache, unsigned t32Now, SSPConfig* pConfig, SitemapOutput* pDeltaSitemap, VisitCount* pDeltaVisitCount)
{
    MergeCacheEntry* c = pCache->Find(m_pbUrl, m_cbUrl);
    if (c) // Merge snapshot entry with cache entry
    {
        unsigned tLastMod = MergeWithSigInfo(&m_sig, &c->normal, &m_lastSeen),
           tLastModRobots = MergeWithSigInfo(&m_sigBots, &c->bots, &m_lastSeen);
        
        BOOL fLastModRobots = FALSE;
        if (tLastMod < tLastModRobots)
        {
            tLastMod = tLastModRobots;
            fLastModRobots = TRUE;
        }

        if (m_changePeriod > c->changePeriod) // combine with change period observed while loading delta files
            m_changePeriod = c->changePeriod;

        VisitCount cVisitCount = c->normal.visitCount;
        *pDeltaVisitCount = cVisitCount;
        if (tLastMod)
        {
            pDeltaSitemap->AddEntry(c->normal.sigLast, c->bots.sigLast, m_pbUrl, m_cbUrl, tLastMod, fLastModRobots, 
                ChangeFreq(m_changePeriod), 0.5, Priority(cVisitCount, pConfig->uNumVisitsPerDeltaPriority));
            g_counters.cUrlsOutputted.m_val++;
            
            m_lastMod = tLastMod;
            m_flags = fLastModRobots ? SNAPSHOT_FLAG_ROBOTS : 0;
        }

        // the reason of decay is to avoid fresh URLs get too low priority
        m_visitCount = AddVisitCount((unsigned short)(m_visitCount * pConfig->flNumVisitsDecayFactor), cVisitCount);
        c->normal.visitCount = c->bots.visitCount = 0; // mark cache entry as not needing to be dumped
    }
    else if (m_lastSeen + pConfig->uVisitTimeoutSec / SEC_PER_T32 < t32Now)
    { // no delta entry for this URL in a long time - drop it
        return FALSE;
    } 
    return TRUE;
}

static void DeleteFilesInFolder(const TCHAR* wszFolder, const TCHAR* wszPattern)
{
    PALFind find;
    if (!find.Initialize(wszFolder, wszPattern)) return;
    if (find.Find() != PALFind::Success) return;
    do {
        if (find.GetAttrib() == PALFind::Directory) continue;
 
        TCHAR wszFile[MAX_PATH];
        if (!GetFullPath(wszFolder, find.GetName(), wszFile, ARRAYSIZE(wszFile))) continue;
        PALFile::Delete(wszFile);
    } while (find.Find() == PALFind::Success);
}

#if UNICODE
static void DeleteFilesInFolder(const TCHAR* wszFolder, const char* szPattern)
{
    TCHAR wszPattern[MAX_PATH];
    if (0 != PALString::Convert(wszPattern, ARRAYSIZE(wszPattern), szPattern, strlen(szPattern)))
    {
        return;
    }
    DeleteFilesInFolder(wszFolder, wszPattern);
}
#endif

#define CHILD_SITEMAP_SUFFIX "_0*" // e.g. _00000001.xml.gz

static void DeleteDeltaChildSitemaps(const TCHAR* wszFolder)
{
    DeleteFilesInFolder(wszFolder, DELTA_SITEMAP_NAME CHILD_SITEMAP_SUFFIX);
    DeleteFilesInFolder(wszFolder, DELTA_DEADLINK_SITEMAP_NAME CHILD_SITEMAP_SUFFIX);
}

static void DeleteFullChildSitemaps(const TCHAR* wszFolder)
{
    DeleteFilesInFolder(wszFolder, SITEMAP_NAME CHILD_SITEMAP_SUFFIX);
    DeleteFilesInFolder(wszFolder, DEADLINK_SITEMAP_NAME CHILD_SITEMAP_SUFFIX);
}

static BOOL GetPluginCountersFile(const TCHAR* wszDataFolderPrefix, unsigned u, __out_ecount(cchFile) TCHAR* wszFile, Size_t cchFile)
{
    TCHAR wszDataFolder[MAX_PATH];
    if (!GetDataFolderWithIndex(wszDataFolderPrefix, u, wszDataFolder, ARRAYSIZE(wszDataFolder))) return FALSE;
    if (!GetPluginCountersFile(wszDataFolder, wszFile, cchFile)) return FALSE;
    return TRUE;
}


static BOOL GetSitemapWWWPath(__out_ecount(cchPath) char* szPath, Size_t cchPath, const char* szWWWRoot, const char* szSubPath, BOOL gzip, const char* szSitemapName)
{
    const char* gz = gzip ? ".gz" : "";
    if (0 >= SPRINTF_S(szPath, cchPath, "%s%s/%s.xml%s", szWWWRoot, szSubPath, szSitemapName, gz)) return FALSE;
    return TRUE;
}

static BOOL DifferentEpoch(time_t t1, time_t t2, time_t period)
{
    return (t1 / period) != (t2 / period);
}

TCHAR g_wszFullCountersFile[MAX_PATH], g_wszAllTimeCountersFile[MAX_PATH];
static unsigned g_uMaxPluginIndex = 0;
static time_t FileLastModifiedTime(const TCHAR* wszFilename);

BOOL DumpSnapshotEntry(void* pv, SnapshotEntry* p)
{
    return 0 < p->Write((CompressingFileWriter*)pv);
}

static BOOL MergeWithSnapshot(DeltaCache* pCache, const TCHAR* wszFolderPrefix, const TCHAR* wszFolder, const TCHAR* wszRobotsFolder, const TCHAR* wszSitemapFolder, const char* szSubPath, const TCHAR* wszTempFolder, SSPConfig* pConfig, time_t tNow, PALEvent *pStopEvent, Robots* pRobots, BOOL fCrossServer)
{
    BOOL fRet = FALSE;
    SitemapOutput deltaSitemap;
    CompressingFileWriter w, wDelta;                        // for snapshot file
    unsigned dwMagic = SNAPSHOT_FILE_MAGIC_HEADER1;
    TCHAR wszSnapshot[MAX_PATH], wszTmp[MAX_PATH];
    TCHAR wszDeltaSnapshot[MAX_PATH], wszDeltaTmp[MAX_PATH];
    char szDeltaSitemapPath[MAX_PATH];
    MyFileReader reader;
    SnapshotEntry s;
    Size_t cb;
    CacheDumpContext ctx = { pCache, &w, NULL, &deltaSitemap, pConfig, 0, pRobots };
    Counters pluginCounters;
    Size_t cbSnapshot = 0;
    BOOL fDeltaSnapshot = pConfig->fMultiServerWorker && !fCrossServer;
    
    if (!GetFullPath(wszFolder, c_wszSnapshot, wszSnapshot, ARRAYSIZE(wszSnapshot))) SGS_ERROR;
    if (!GetFullPath(wszFolder, c_wszTmp, wszTmp, ARRAYSIZE(wszTmp))) SGS_ERROR;

    if (fDeltaSnapshot)
    {
        if (!GetFullPath(wszFolder, c_wszDeltaSnapshot, wszDeltaSnapshot, ARRAYSIZE(wszDeltaSnapshot))) SGS_ERROR;
        if (!GetFullPath(wszFolder, c_wszDeltaTmp, wszDeltaTmp, ARRAYSIZE(wszDeltaTmp))) SGS_ERROR;

        // A race condition exists between the updating of DeltaSnapshot.dat here and
        // the process that collects DeltaSnapshot.dat in a centralized place and deletes these local ones.
        if (NO_ERROR != wDelta.InitAndOpen(wszDeltaTmp)) SGS_ERROR;
        if (NO_ERROR != wDelta.Write(&dwMagic, 4)) SGS_ERROR;
        if (PALDirectory::PathExists(wszDeltaSnapshot))
        { // copy existing contents of delta snapshot - we aim to append
            if (!EnumerateSnapshotFile(wszDeltaSnapshot, &wDelta, DumpSnapshotEntry)) SGS_ERROR;
        }

        ctx.pDeltaWriter = &wDelta;
    }

    // Getting www path for the top level sitemap file, which is never expected to be gz file.
    // Checking earlier to avoid unnecessary disk IO then fail later
    if (!GetSitemapWWWPath(szDeltaSitemapPath, sizeof(szDeltaSitemapPath), pConfig->szwwwhost, szSubPath, FALSE, DELTA_SITEMAP_NAME)) SGS_ERROR;

    if (!deltaSitemap.Initialize(DELTA_SITEMAP_NAME, DELTA_DEADLINK_SITEMAP_NAME, wszTempFolder, wszSitemapFolder, pConfig->szwwwhost, szSubPath, pConfig->fgzip, TRUE, fCrossServer)) SGS_ERROR;

    if (NO_ERROR != w.InitAndOpen(wszTmp)) SGS_ERROR;
    if (NO_ERROR != w.Write(&dwMagic, 4)) SGS_ERROR;

    g_counters.cUrlsOutputted.m_val = 0;
    if (PALDirectory::PathExists(wszSnapshot))
    {
        BOOL fResetChangeFreq = DifferentEpoch(FileLastModifiedTime(wszSnapshot), tNow, pConfig->uChangeFreqResetPeriodSec);
        if (fResetChangeFreq)
        {
            SiteLog("resetting change frequencies for " STRFMT_TS " (every %u days)", wszFolder, pConfig->uChangeFreqResetPeriodSec / 86400);
        }

        if (!reader.Initialize(wszSnapshot, c_rSnapshotMagic)) SGS_ERROR;
        for (unsigned u = 0; 1; u++)
        {
            if (!s.Read(&reader)) SGS_ERROR;
            if (s.IsEmpty())
            {
                break;
            }
            if (!pRobots->Allowed(s.m_pbUrl, s.m_cbUrl))
            {
                g_counters.cExcludedLater.Increment();
                continue;
            }
            if (fResetChangeFreq)
            {
                s.m_changePeriod = CHANGEPERIOD_UNKNOWN;
            }
            
            VisitCount deltaVisitCount = 0;
            if (!s.MergeWithCache(pCache, (unsigned)(tNow / SEC_PER_T32), pConfig, &deltaSitemap, &deltaVisitCount))
            {
                continue;
            }

            if (0 == ( cb = s.Write(&w))) SGS_ERROR;
            cbSnapshot += cb; // not checking maxSnapshotFileMB here because snapshot cannot grow just by this copy/merge loop

            if (fDeltaSnapshot && deltaVisitCount)
            {
                SnapshotEntry d = s;
                d.m_visitCount = deltaVisitCount;
                if (0 == (cb = d.Write(&wDelta))) SGS_ERROR;
            }

            if (0 == u % pConfig->uStopEventCheckPeriodEntries)
            {
                if (pStopEvent != NULL)
                {
                    if (PALEvent::Signaled == pStopEvent->Wait(0)) SGS_ERROR;
                }
            }
        }
        reader.Dispose();  
    }
    
    ctx.cbSnapshot = cbSnapshot;
    if (!pCache->Enumerate(&ctx, DumpCacheEntry)) SGS_ERROR;

    memset(&s, 0, sizeof(s)); // dummy 0/0/0 as EOF marker
    if (!s.Write(&w)) SGS_ERROR;
    if (NO_ERROR != w.DoDispose()) SGS_ERROR;
    PALFile::SetModifiedTime(wszTmp);
    if (!PALFile::Move(wszTmp, wszSnapshot)) SGS_ERROR;

    if (fDeltaSnapshot)
    {
        if (!s.Write(&wDelta)) SGS_ERROR;
        if (NO_ERROR != wDelta.DoDispose()) SGS_ERROR;
        if (!PALFile::Move(wszDeltaTmp, wszDeltaSnapshot)) SGS_ERROR;
    }

    for (unsigned u = 0; u <= g_uMaxPluginIndex; u++)
    {
        TCHAR wszDeltaCountersFile[MAX_PATH];
        if (!GetPluginCountersFile(wszFolderPrefix, u, wszDeltaCountersFile, ARRAYSIZE(wszDeltaCountersFile))) SGS_ERROR;
        ReadCounters(wszDeltaCountersFile, &pluginCounters);
        AddCounters(&g_counters, &pluginCounters);
    }
    
    DeleteDeltaChildSitemaps(wszSitemapFolder);
    if (!deltaSitemap.Close(&g_counters, pConfig)) SGS_ERROR; // even if this fails, the snapshot has already been merged
        
    if(pConfig->fWriteToRobots && pConfig->fPublishSitemap)
    {
        if (!AddSitemapToRobots(wszRobotsFolder, wszTempFolder, szDeltaSitemapPath, pConfig->wszRobotsTxtFilename)) 
        {
            SGS_ERROR;
        }
    }
    
    for (unsigned u = 0; u <= g_uMaxPluginIndex; u++)
    {
        TCHAR wszDeltaCountersFile[MAX_PATH];
        if (GetPluginCountersFile(wszFolderPrefix, u, wszDeltaCountersFile, ARRAYSIZE(wszDeltaCountersFile)))
        {
            PALFile::Delete(wszDeltaCountersFile);
        }
    }

    AddCountersToFile(g_wszFullCountersFile, &g_counters);

    fRet = TRUE;
Exit:
    return fRet;
}

#if APR_HAVE_WINDOWS_H
static time_t FileLastModifiedTime(const TCHAR* wszFilename)
{
    struct _stat s;
    if (_tstat(wszFilename, &s))
    {
        return 0;
    }
    return s.st_mtime;
}
#else
static time_t FileLastModifiedTime(const TCHAR* wszFilename)
{
    apr_finfo_t fileinfo;
    char szFilename[MAX_PATH];
    apr_pool_t *pPool = NULL;
    
    if (apr_pool_create(&pPool, NULL) != APR_SUCCESS)
    {
        fileinfo.mtime = 0;
        goto Done;
    }

    PALString::Convert(szFilename, ARRAYSIZE(szFilename), wszFilename, TSTRLEN(wszFilename));

    if (apr_stat(&fileinfo, szFilename, APR_FINFO_MTIME, pPool) != APR_SUCCESS)
    {
        fileinfo.mtime = 0;
        goto Done;
    }
    
Done:
    if (pPool != NULL)
    {
        apr_pool_destroy(pPool);
    }

    return apr_time_sec(fileinfo.mtime);
}
#endif

BOOL g_fTestNoFullSitemap = FALSE;

static BOOL UpdateFullSitemap(const TCHAR* wszFolder, const TCHAR* wszRobotsFolder, const TCHAR* wszSitemapFolder, const char* szSubPath, const TCHAR* wszTempFolder, SSPConfig* pConfig, PALEvent *pStopEvent, time_t tNow, Robots* pRobots, BOOL fCrossServer)
{
    BOOL fRet = FALSE;
    TCHAR wszSitemapPath[MAX_PATH];
    TCHAR wszSnapshot[MAX_PATH];
    MyFileReader reader;
    SitemapOutput sitemap;
    SitemapOutput deltaSitemap;
    char szSitemapPath[MAX_PATH];
    SnapshotEntry s;
    unsigned avgVisitCount = 0;
    if (!SitemapOutput::GetSitemapFilePath(wszSitemapPath, ARRAYSIZE(wszSitemapPath), wszSitemapFolder, SITEMAP_NAME)) SGS_ERROR;

    // consider processing time, give 30 minutes beffer here
    if (g_fTestNoFullSitemap || (tNow < FileLastModifiedTime(wszSitemapPath) + pConfig->uSitemapGenerationPeriodHours * (time_t)3600 - (time_t)1800))
    {
        return TRUE;
    }

    if (!GetFullPath(wszFolder, c_wszSnapshot, wszSnapshot, ARRAYSIZE(wszSnapshot))) SGS_ERROR;

    if (!reader.Initialize(wszSnapshot, c_rSnapshotMagic)) SGS_ERROR;

    if (!sitemap.Initialize(SITEMAP_NAME, DEADLINK_SITEMAP_NAME, wszTempFolder, wszSitemapFolder, pConfig->szwwwhost, szSubPath, pConfig->fgzip, FALSE, fCrossServer)) SGS_ERROR;

    ReadCounters(g_wszFullCountersFile, &g_counters);

    // take the counters value instead of iterating snapshot file
    // although iterating snapshot file is more accurate, it is too expensive to do so

    // this calculation might get smaller avg number because cUrlsOutputted.m_val contains bot traffic only URLs
    if((g_counters.cUrlsOutputted.m_val != 0)
        && (g_counters.cValidTraffic.m_val > g_counters.cBotTraffic.m_val))
    {
        avgVisitCount = (g_counters.cValidTraffic.m_val - g_counters.cBotTraffic.m_val) / g_counters.cUrlsOutputted.m_val;
    }

    g_counters.cUrlsOutputted.m_val = 0;
       
    for (unsigned u = 0; 1; u++)
    {
        if (!s.Read(&reader)) SGS_ERROR;
        if (s.IsEmpty())
        {
            break;
        }
        if (!pRobots->Allowed(s.m_pbUrl, s.m_cbUrl))
        {
            g_counters.cExcludedLater.Increment();
            continue;
        }
        
        if (!sitemap.AddEntry(s.m_sig, s.m_sigBots, s.m_pbUrl, s.m_cbUrl, s.m_lastMod, s.m_flags & SNAPSHOT_FLAG_ROBOTS, 
            ChangeFreq(s.m_changePeriod), Priority(s.m_visitCount, avgVisitCount), Priority(s.m_visitCount, pConfig->uNumVisitsPerFullPriority))) SGS_ERROR;
        g_counters.cUrlsOutputted.m_val++;
        
        if (0 == u % pConfig->uStopEventCheckPeriodEntries)
        {
            if (pStopEvent != NULL)
            {
                if (PALEvent::Signaled == pStopEvent->Wait(0)) SGS_ERROR;
            }
        }
    }

    DeleteFullChildSitemaps(wszSitemapFolder);
    if (!sitemap.Close(&g_counters, pConfig)) SGS_ERROR;

    AddCountersToFile(g_wszAllTimeCountersFile, &g_counters);
    PALFile::Delete(g_wszFullCountersFile);
    ResetCounters(&g_counters);
    
    // Getting www path for the top level sitemap file, which is never expected to be gz file.
    if (!GetSitemapWWWPath(szSitemapPath, sizeof(szSitemapPath), pConfig->szwwwhost, szSubPath, FALSE, SITEMAP_NAME)) SGS_ERROR;
    if(pConfig->fWriteToRobots && pConfig->fPublishSitemap)
    {
        if (!AddSitemapToRobots(wszRobotsFolder, wszTempFolder, szSitemapPath, pConfig->wszRobotsTxtFilename)) 
        {
            SGS_ERROR;
        }
    }

    // Empty the delta sitemap, treating errors as non-fatal
    DeleteDeltaChildSitemaps(wszSitemapFolder);
    if (deltaSitemap.Initialize(DELTA_SITEMAP_NAME, DEADLINK_SITEMAP_NAME, wszTempFolder, wszSitemapFolder, pConfig->szwwwhost, szSubPath, pConfig->fgzip, FALSE, fCrossServer))
    {
        deltaSitemap.Close(NULL, NULL, TRUE);
    }

    fRet = TRUE;
Exit:
    return fRet;
}

class ListOfStrings
{
    Size_t m_cchCapacity;
    Size_t m_cchUsed;
    Size_t m_cchEnum;
    TCHAR* m_pch;

public:
    ListOfStrings() : m_pch(NULL) {}
    ~ListOfStrings() { if (m_pch) delete[] m_pch; }
    BOOL Initialize(Size_t cchCapacity);
    BOOL Add(const TCHAR* wsz);
    const TCHAR* Next();
};

BOOL ListOfStrings::Initialize(Size_t cchCapacity)
{
    m_cchUsed = 0;
    m_cchEnum = 0;
    m_cchCapacity = cchCapacity;
    m_pch = new TCHAR[m_cchCapacity];
    return m_pch != NULL;
}

BOOL ListOfStrings::Add(const TCHAR* wsz)
{
    Size_t cch = TSTRLEN(wsz) + 1;
    if (cch > m_cchCapacity - m_cchUsed) return FALSE;
    memcpy(m_pch + m_cchUsed, wsz, cch * sizeof(TCHAR));
    m_cchUsed += cch;
    return TRUE;
}

const TCHAR* ListOfStrings::Next()
{
    if (m_cchEnum >= m_cchUsed)
    {
        return NULL;
    }
    TCHAR* wszRet = m_pch + m_cchEnum;
    Size_t cch = TSTRLEN(wszRet) + 1;
    m_cchEnum += cch;
    return wszRet;
}


BOOL ParseNDigits(const TCHAR* pch, Size_t cch, WORD* pw)
{
    WORD w = 0;
    for (Size_t i = 0; i < cch; i++)
    {
        TCHAR c = pch[i];
        if ('0' <= c && c <= '9')
        {
            w = 10 * w + (c - '0');
        }
        else
        {
            return FALSE;
        }
    }
    *pw = w;
    return TRUE;
}

BOOL DateFromFolder(const TCHAR* wszFolder, WORD* pwYear, WORD* pwMonth, WORD* pwDay)
{
    if (8 != TSTRLEN(wszFolder)) return FALSE;
    if (!ParseNDigits(wszFolder + 0, 4, pwYear)) return FALSE;
    if (!ParseNDigits(wszFolder + 4, 2, pwMonth)) return FALSE;
    if (!ParseNDigits(wszFolder + 6, 2, pwDay)) return FALSE;
    return TRUE;
}

BOOL TimeFromFilename(const TCHAR* wszFilename, WORD* pwHour, WORD* pwMin, WORD* pwSec)
{
    if (8 >= TSTRLEN(wszFilename) || '.' != wszFilename[6]) return FALSE;
    if (!ParseNDigits(wszFilename + 0, 2, pwHour)) return FALSE;
    if (!ParseNDigits(wszFilename + 2, 2, pwMin)) return FALSE;
    if (!ParseNDigits(wszFilename + 4, 2, pwSec)) return FALSE;
    return TRUE;
}

typedef struct
{
    unsigned uDate;
    unsigned uPluginInstance;
} DeltaFolderIndex;

static int __cdecl DeltaFolderComparer(const void *elem1, const void *elem2)
{
    return ((DeltaFolderIndex*)elem1)->uDate - ((DeltaFolderIndex*)elem2)->uDate;
}

static int __cdecl IntComparer(const void *elem1, const void *elem2)
{
    return *((int*)elem1) - *((int*)elem2);
}

static unsigned DateToUnsigned(WORD wYear, WORD wMonth, WORD wDay)
{
    return (wYear * 12 + (wMonth - 1)) * 31 + (wDay - 1);
}
static void DateFromUnsigned(unsigned u, WORD *pwYear, WORD *pwMonth, WORD *pwDay)
{
    *pwDay = 1 + (WORD)(u % 31);   u /= 31;
    *pwMonth = 1 + (WORD)(u % 12);  u /= 12;
    *pwYear = (WORD)u;
}
static unsigned TimeToUnsigned(WORD wHour, WORD wMin, WORD wSec)
{
    return (wHour * 60 + wMin) * 60 + wSec;
}
static void TimeFromUnsigned(unsigned u, WORD *pwHour, WORD* pwMin, WORD* pwSec)
{
    *pwSec = (WORD)(u % 60);  u /= 60;
    *pwMin = (WORD)(u % 60);  u /= 60;
    *pwHour = (WORD)u;
}

static BOOL SplitDataFolderPath(const TCHAR* wszPath, __out_ecount(cchFolder) TCHAR* wszFolder, Size_t cchFolder, __out_ecount(cchFile)TCHAR* wszFile, Size_t cchFile)
{ // Separates "...\folder" into "..." and "folder"  (_splitpath annoyingly returns a drive letter separately)
    Size_t cchPath = TSTRLEN(wszPath);
    Size_t i = cchPath;
    while (i)
    {
        i--;
        if ('\\' == wszPath[i] || '/' == wszPath[i]) // BUGBUG: '\\' a valid filename character on Unix ?
        {
            if (i >= cchFolder)
            {
                return FALSE;
            }
            memcpy(wszFolder, wszPath, i * sizeof(TCHAR));
            wszFolder[i] = '\0';

            i++; // skip the slash

            if (cchPath - i >= cchFile)
            {
                return FALSE;
            }
            memcpy(wszFile, wszPath + i, (cchPath - i) * sizeof(TCHAR));
            wszFile[cchPath - i] = '\0';

            return TRUE;
        }
    }

    return FALSE;
}

// 3-level iteration over plugin index (integer folder suffix), subfolder name (date) and filename (time) in increasing time order
static BOOL LoadDeltaFiles(const TCHAR* wszDataFolderPrefix, SSPConfig* pConfig, DeltaCache* pCache, ListOfStrings* pFilesToDelete, PALEvent *pStopEvent, time_t tNow)
{
    BOOL fRet = FALSE;
    PALFind findFile, outerFindFile;
    BOOL fFull = FALSE;
    PALFind::EStatus findStatus, outerFindStatus;
    DeltaFolderIndex folders[1000];
    unsigned cFolders = 0;
    TCHAR wszCommonDataFolder[MAX_PATH], wszSiteSpecificPrefix[MAX_PATH];
    Size_t cchSiteSpecificPrefix;
    
    if (!SplitDataFolderPath(wszDataFolderPrefix, wszCommonDataFolder, ARRAYSIZE(wszCommonDataFolder), wszSiteSpecificPrefix, ARRAYSIZE(wszSiteSpecificPrefix))) SGS_ERROR;
    cchSiteSpecificPrefix = TSTRLEN(wszSiteSpecificPrefix);
    if (cchSiteSpecificPrefix + 2 >= ARRAYSIZE(wszSiteSpecificPrefix)) SGS_ERROR;
    memcpy(wszSiteSpecificPrefix + cchSiteSpecificPrefix, TEXT("_*"), 3 * sizeof(TCHAR));

    if (!outerFindFile.Initialize(wszCommonDataFolder, wszSiteSpecificPrefix)) // enumerate <commonfolder>/<prefix>_*
    {
        SGS_ERROR;
    }
    outerFindStatus = outerFindFile.Find();
    if (outerFindStatus != PALFind::Success)
    {
        if (outerFindStatus != PALFind::Error_FileNotFound)
        {
            SGS_ERROR;
        }

        fRet = TRUE; // nothing to do
        goto Exit;
    }
    do
    {
        if (outerFindFile.GetAttrib() != PALFind::Directory)
        {
            continue;
        }
        unsigned uPluginInstance;
        if (1 != STSCANF_S(outerFindFile.GetName() + cchSiteSpecificPrefix + 1, TEXT("%u"), &uPluginInstance)) // + 1 for underscore
        {
            continue;
        }
        
        if (g_uMaxPluginIndex < uPluginInstance)
        {
            g_uMaxPluginIndex = uPluginInstance;
        }

        TCHAR wszDataFolder[MAX_PATH];
        if (0 >= STPRINTF_S(wszDataFolder, ARRAYSIZE(wszDataFolder), STRFMT_TLS STRFMT_TDIRSLASH STRFMT_TLS, wszCommonDataFolder, outerFindFile.GetName())) SGS_ERROR;
        if (findFile.Initialize(wszDataFolder, TEXT("20*")) == FALSE) // BUGBUG - stop working in year 2100
        {
            fRet = FALSE;
            SGS_ERROR;
        }
        findStatus = findFile.Find();
        if (findStatus != PALFind::Success)
        {
            if (findStatus != PALFind::Error_FileNotFound)
            {
                SGS_ERROR;
            }
            continue;
        }
    
        do { // collect available directories
            if (findFile.GetAttrib() != PALFind::Directory)
            {
                continue;
            }
    
            WORD wYear, wMonth, wDay;
            if (!DateFromFolder(findFile.GetName(), &wYear, &wMonth, &wDay))
            {
                continue;
            }

            folders[cFolders].uDate = DateToUnsigned(wYear, wMonth, wDay);
            folders[cFolders].uPluginInstance = uPluginInstance;
            cFolders++;
        } while ((cFolders < ARRAYSIZE(folders)) && findFile.Find() == PALFind::Success);
        findFile.Destroy();
    } while ((cFolders < ARRAYSIZE(folders)) && outerFindFile.Find() == PALFind::Success);
    outerFindFile.Destroy();

    qsort(folders, cFolders, sizeof(folders[0]), DeltaFolderComparer); // sort directories by increasing date

    for (unsigned iFolder = 0; (iFolder < cFolders) && !fFull; iFolder++) // go through directories
    {
        PALFind findSubFile;
        WORD wYear, wMonth, wDay,  wHour, wMinute, wSecond;
        DateFromUnsigned(folders[iFolder].uDate, &wYear, &wMonth, &wDay);

        TCHAR wszSubFolder[MAX_PATH];

        if (0 >= SNTPRINTF_S(wszSubFolder, ARRAYSIZE(wszSubFolder), _TRUNCATE, STRFMT_TLS TEXT("_%u") STRFMT_TDIRSLASH TEXT("%04hu%02hu%02hu"), wszDataFolderPrefix, folders[iFolder].uPluginInstance, wYear, wMonth, wDay)) SGS_ERROR;

        if (findSubFile.Initialize(wszSubFolder, TEXT("*.dat")) == FALSE)
        {
            fRet = FALSE;
            SGS_ERROR;
        }

        if (findSubFile.Find() != PALFind::Success)
        {
            PALDirectory::DeleteDirectory(wszSubFolder);
            continue;
        }

        unsigned uTimes[4096], cTimes = 0;
        do { // collect available files
            if ('.' == findSubFile.GetName()[0])
            {
                continue;
            }
            if (findSubFile.GetAttrib() == PALFind::Directory)
            { // ??
                continue;
            }
            if (!TimeFromFilename(findSubFile.GetName(), &wHour, &wMinute, &wSecond))
            { // ??
                continue;
            }
            uTimes[cTimes++] = TimeToUnsigned(wHour, wMinute, wSecond);
        } while (cTimes < ARRAYSIZE(uTimes) && findSubFile.Find() == PALFind::Success);
        findSubFile.Destroy();

        qsort(uTimes, cTimes, sizeof(uTimes[0]), IntComparer); // sort files by increasing time

        for (unsigned iTime = 0; iTime < cTimes; iTime++) // go through files
        {
            TimeFromUnsigned(uTimes[iTime], &wHour, &wMinute, &wSecond);
            TCHAR wszFile[MAX_PATH];

            if (0 >= SNTPRINTF_S(wszFile, ARRAYSIZE(wszFile), _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH TEXT("%02hu%02hu%02hu.dat"), wszSubFolder, wHour, wMinute, wSecond)) SGS_ERROR;

            struct tm tms;
            memset(&tms, 0, sizeof(tms));
            tms.tm_isdst = -1;
            tms.tm_year = wYear - 1900;
            tms.tm_mon = wMonth - 1;
            tms.tm_mday = wDay;
            tms.tm_hour = wHour;
            tms.tm_min = wMinute;
            tms.tm_sec = wSecond;
            time_t tUTC = mktime(&tms);
            if ((time_t)-1 == tUTC)
            {
                continue;
            }
            if (tNow > tUTC + (time_t)(pConfig->uMaxAgeToMergeSec))
            {
                PALFile::Delete(wszFile);
                g_counters.cOldFilesDeleted.Increment();
                continue;
            }
            FileStats s;
            if (!DeltaFileStats(wszFile, &s)) SGS_ERROR;
            if (!pCache->CheckCapacity(&s) || !pFilesToDelete->Add(wszFile))
            { // consumed list full - don't consume any more in this session
                fFull = TRUE;
                break;
            }

            if (pStopEvent != NULL)
            {
                if (PALEvent::Signaled == pStopEvent->Wait(0)) SGS_ERROR;
            }
        
            if (!LoadDeltaFile(wszFile, pCache, tUTC)) SGS_ERROR;
        } // files
    } // direcotries
    
    fRet = TRUE;

Exit:
    return fRet;
}

static BOOL SnapshotFileStatsCallback(void* pv, SnapshotEntry* p)
{
    FileStats* pCtx = (FileStats*)pv;
    pCtx->cEntries++;
    pCtx->cbUrls += p->m_cbUrl;
    return TRUE;
}

static BOOL SnapshotFileStats(const TCHAR* wszFile, FileStats* pStats)
{
    return EnumerateSnapshotFile(wszFile, pStats, SnapshotFileStatsCallback);
}

static BOOL SnapshotFileLoadCallback(void* pv, SnapshotEntry* p)
{
    DeltaCache* pCache = (DeltaCache*)pv;
    BOOL fRet = FALSE;
    VisitCount visitLeft = p->m_visitCount;
    
    DeltaInputEntry e;
    e.cbUrl = p->m_cbUrl;
    if (e.cbUrl > sizeof(e.pcUrl)) SGS_ERROR; // not likely since the arrays are the same size
    memcpy(e.pcUrl, p->m_pbUrl, e.cbUrl);

    if (p->m_sigBots)
    {
        e.sig = p->m_sigBots;
        if (p->m_sig)
        {
            e.visitCount =  visitLeft / 2; // arbitrarily eat half - we really don't know
        }
        else
        {
            e.visitCount = visitLeft; // eat all
        }
        visitLeft -= e.visitCount;
        e.flags = DELTA_FLAG_BOT;
        e.AddToCache(pCache, p->m_lastMod); // error eaten because it is expected when cache is full
    }

    if (p->m_sig)
    {
        e.sig = p->m_sig;
        e.visitCount = visitLeft;
        e.flags = DELTA_FLAG_NONE;
        e.AddToCache(pCache, p->m_lastMod); // error eaten because it is expected when cache is full
    }
    
    fRet = TRUE;
Exit:
    return fRet;
}

static BOOL LoadSnapshotFile(const TCHAR* wszFile, DeltaCache* pCache)
{
    return EnumerateSnapshotFile(wszFile, pCache, SnapshotFileLoadCallback);
}

static BOOL LoadSnapshotFiles(const TCHAR* wszFolder, DeltaCache* pCache, ListOfStrings* pFilesToDelete)
{
    PALFind find;
    BOOL fRet = FALSE;
    PALFind::EStatus findStatus = PALFind::Success;
    if (!find.Initialize(wszFolder, TEXT("*.dat")))
    {
        SGS_ERROR;
    }

    findStatus = find.Find();
    if (PALFind::Error_FileNotFound == findStatus)
    {
        return TRUE;
    }

    if (findStatus != PALFind::Success)
    {
        SGS_ERROR;
    }

    do
    {
        if (find.GetAttrib() == PALFind::Directory)
        {
            continue;
        }

        TCHAR wszFile[MAX_PATH];
        if (!GetFullPath(wszFolder, find.GetName(), wszFile, ARRAYSIZE(wszFile))) SGS_ERROR;

        FileStats s;
        if (!SnapshotFileStats(wszFile, &s)) SGS_ERROR;
        if (!pCache->CheckCapacity(&s) || !pFilesToDelete->Add(wszFile))
        { // consumed list full - don't consume any more in this session
            break;
        }

        if (!LoadSnapshotFile(wszFile, pCache)) SGS_ERROR;
    } while (find.Find() == PALFind::Success);
    
    fRet = TRUE;
Exit:
    return fRet;
}

static BOOL LoadCountersFiles(const TCHAR* wszFolder, ListOfStrings* pFilesToDelete)
{
    PALFind find;
    BOOL fRet = FALSE;
    PALFind::EStatus findStatus = PALFind::Success;
    Counters counters;

    if (!find.Initialize(wszFolder, TEXT("*.tsv")))
    {
        SGS_ERROR;
    }

    findStatus = find.Find();
    if (PALFind::Error_FileNotFound == findStatus)
    {
        return TRUE;
    }

    if (findStatus != PALFind::Success)
    {
        SGS_ERROR;
    }

    SiteLog("loading counters files");

    do
    {
        if (find.GetAttrib() == PALFind::Directory)
        {
            continue;
        }

        TCHAR wszFile[MAX_PATH];
        if (!GetFullPath(wszFolder, find.GetName(), wszFile, ARRAYSIZE(wszFile))) SGS_ERROR;

        if (!pFilesToDelete->Add(wszFile))
        { // consumed list full - don't consume any more in this session
            break;
        }

        if (!ReadCounters(wszFile, &counters)) SGS_ERROR;

        AddCounters(&g_counters, &counters);

    } while (find.Find() == PALFind::Success);

    fRet = TRUE;
Exit:
    return fRet;
}

static void DeleteFilesInList(ListOfStrings* pList)
{
    const TCHAR* wszConsumedFile;
    while (NULL != (wszConsumedFile = pList->Next()))
    {
        PALFile::Delete(wszConsumedFile); // the emtpy directory is deleted during next cycle
    }
}


FILE* g_fpTestLog = NULL;
BOOL OpenMergeServiceLog(const TCHAR* wszPath, FILE** pfp)
{
    if (0 != PALFile::Open(pfp, wszPath, TEXT("at")))
    {
        if (g_fpTestLog)
        {
            *pfp = g_fpTestLog;
        }
        else
        {
            return FALSE;
        }
    }
    return TRUE;
}
void CloseMergeServiceLog(FILE* fp)
{
    if (fp && (fp != g_fpTestLog))
    {
        fclose(fp);
    }
}

static BOOL DeterminePublishPath(const TCHAR** wszRobotsFolder, const char* szSubPath, // szSubPath described in DeltaMerge.h
                                 __out_ecount(cchSitemapFolder) TCHAR* wszSitemapFolder, Size_t cchSitemapFolder,
                                 SSPConfig* pConfig, const TCHAR* wszDataFolder)
{
    if (!*wszRobotsFolder)
    { // test mode
        *wszRobotsFolder = pConfig->wszwwwroot;
    }

    if (pConfig->fPublishSitemap)
    { // sitemap goes to WWWRoot
        // Note: using szSubPath as is (including any forward slashes, which seem to work fine on Windows)
        if (0 >= STPRINTF_S(wszSitemapFolder, cchSitemapFolder, STRFMT_TLS STRFMT_THS, *wszRobotsFolder, szSubPath))
        {
            return FALSE;
        }
    }
    else
    { // sitemap goes to private folder
        if (0 >= STPRINTF_S(wszSitemapFolder, cchSitemapFolder, STRFMT_TLS STRFMT_TDIRSLASH TEXT("SitemapPreview"), wszDataFolder))
        {
            return FALSE;
        }
        if (!PALDirectory::EnsureDirectoryExists(wszSitemapFolder))
        {
            return FALSE;
        }
    }

    return TRUE;
}

int DeltaMergeCppLineCount();

#if APR_HAVE_WINDOWS_H
static int MBTWCwrap(const char* sz, WCHAR* wsz, int cch)
{
    return MultiByteToWideChar(CP_UTF8, 0, sz, -1, wsz, cch);
}

static void PingSearchEngine(const TCHAR* engineUrl, const char* wwwHost)
{


    BOOL  bResults = FALSE;
    char pcUrlToPing[2048];
    char pcEncoded[2048];
    size_t cbEncoded = sizeof(pcEncoded);

    if (0 >= SPRINTF_S(pcUrlToPing, "%s/trafficbasedsspsitemap.xml", wwwHost)) return;

    if (!SitemapOutput::EncodeURL(pcUrlToPing, strlen(pcUrlToPing), pcEncoded, &cbEncoded) || (cbEncoded > MAX_PING_URL))
    {
        // ignore errors
        return;
    }

    if (0 >= SPRINTF_S(pcUrlToPing, "/ping?sitemap=%s", pcEncoded)) return;

    HINTERNET hSession = NULL,
              hConnect = NULL,
              hRequest = NULL;



    // Use WinHttpOpen to obtain a session handle.
    hSession = WinHttpOpen(  L"BingServerSidePlugin/1.0", 
                             WINHTTP_ACCESS_TYPE_DEFAULT_PROXY, /* WINHTTP_ACCESS_TYPE_NAMED_PROXY */ 
                             WINHTTP_NO_PROXY_NAME, /* L"127.0.0.1:8888", */
                             WINHTTP_NO_PROXY_BYPASS, 0);

    // Specify an HTTP server.
    if (hSession)
        hConnect = WinHttpConnect( hSession, engineUrl,
                                   INTERNET_DEFAULT_HTTP_PORT, 0);

    int cchNeed = MBTWCwrap(pcUrlToPing, NULL, 0);
    if (cchNeed <= 0)
    {
        return;
    }

    WCHAR *pwszOut = new WCHAR[cchNeed];
    if (0 >= MBTWCwrap(pcUrlToPing, pwszOut, cchNeed))
    {
        delete[] pwszOut;
        return;
    }

    // Create an HTTP Request handle.
    if (hConnect)
        hRequest = WinHttpOpenRequest( hConnect, L"GET", 
                                       pwszOut, 
                                       NULL, WINHTTP_NO_REFERER, 
                                       WINHTTP_DEFAULT_ACCEPT_TYPES,
                                       0);

    // Send a Request.
    if (hRequest) 
        bResults = WinHttpSendRequest( hRequest, 
                                       WINHTTP_NO_ADDITIONAL_HEADERS,
                                       0, WINHTTP_NO_REQUEST_DATA, 0, 
                                       0, 0);

    // Report any errors.
    if (!bResults)
        SiteLog( "Error %d has occurred on ping.\n", GetLastError());

    // Close any open handles.
    if (hRequest) WinHttpCloseHandle(hRequest);
    if (hConnect) WinHttpCloseHandle(hConnect);
    if (hSession) WinHttpCloseHandle(hSession);
}
#endif
static BOOL DoMerge(const TCHAR* wszDataFolderPrefix, const TCHAR* wszRobotsFolder, const char* szSubPath, PALEvent *pStopEvent, const TCHAR* wszDataFolder, const TCHAR* wszInputFolder)
{

    TCHAR wszLog[MAX_PATH];
    g_fpSiteLog = NULL;
    TCHAR wszConfigFile[MAX_PATH];
    SSPConfig config;
    BOOL fRet = FALSE;
    ListOfStrings consumedFiles;
    DeltaCache cache;
    time_t tNow;
    TCHAR wszSitemapFolder[MAX_PATH];
    DiskSpaceTracker diskSpaceTracker;
    if (!szSubPath) szSubPath = "";
    TCHAR wszRobotsPath[MAX_PATH], wszBlacklistPath[MAX_PATH];
    Robots robots;
    BOOL fCrossServer = (wszInputFolder != NULL);

    if(!fCrossServer)
    {
        ResetCounters(&g_counters);
    }

    if (0 >= SNTPRINTF_S(wszLog, ARRAYSIZE(wszLog), _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH TEXT("ServiceLog.txt"), wszDataFolder)) SGS_ERROR;

    if (!OpenMergeServiceLog(wszLog, &g_fpSiteLog)) SGS_ERROR;

    if (!GetFullPath(wszDataFolder, TEXT("FullCounters.tsv"), g_wszFullCountersFile, ARRAYSIZE(g_wszFullCountersFile))) SGS_ERROR;
    if (!GetAllTimeCountersFile(wszDataFolder, g_wszAllTimeCountersFile, ARRAYSIZE(g_wszAllTimeCountersFile))) SGS_ERROR;
    if (!GetConfigFile(wszDataFolder, wszConfigFile, ARRAYSIZE(wszConfigFile))) SGS_ERROR;
    if (!ReadSSPConfig(wszConfigFile, &config)) SGS_ERROR;
    if (!diskSpaceTracker.DoInit(wszDataFolder, config.uDiskUsageCheckPeriodSec, config.uDiskMinFreeMB, config.uDiskMaxFreeMB)) SGS_ERROR;
    if (!diskSpaceTracker.CheckSpace()) SGS_ERROR; // BUGBUG: not counting delta files in folders having index other than "_0"
    if (!consumedFiles.Initialize(100000)) SGS_ERROR;
    if (!cache.Initialize(config.uMergeCacheMemoryMB * 1048576)) SGS_ERROR;
    tNow = PluginTime();
    TCHAR wszTempFolder[MAX_PATH];
    if (0 >= SNTPRINTF_S(wszTempFolder, ARRAYSIZE(wszTempFolder), _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH TEXT("temp"), wszDataFolder)) SGS_ERROR;
    if (!PALDirectory::EnsureDirectoryExists(wszTempFolder)) SGS_ERROR;
    
    if (fCrossServer) // cross-server merge
    {
        config.fPublishSitemap = TRUE; // always publish for cross-server merge to avoid configuration confusion
        config.fWriteToRobots = TRUE;
    }
    if (!DeterminePublishPath(&wszRobotsFolder, szSubPath, wszSitemapFolder, ARRAYSIZE(wszSitemapFolder), &config, wszDataFolder)) SGS_ERROR;
    SiteLog("processing " STRFMT_TS " version %u " STRFMT_TS " (%s)", wszDataFolder, DeltaMergeCppLineCount(), wszRobotsFolder, szSubPath);
    
    if (!GetFullPath(wszRobotsFolder, config.wszRobotsTxtFilename, wszRobotsPath, ARRAYSIZE(wszRobotsPath))) SGS_ERROR;
    if (!GetFullPath(wszDataFolder, BLACKLIST_FILE, wszBlacklistPath, ARRAYSIZE(wszBlacklistPath))) SGS_ERROR;

    if (!robots.Load(NULL, wszRobotsPath, wszBlacklistPath)) SGS_ERROR;

    // Load delta files to cache
    if (fCrossServer)
    {
        if (!LoadSnapshotFiles(wszInputFolder, &cache, &consumedFiles)) SGS_ERROR;
    }
    else
    {
        if (!LoadDeltaFiles(wszDataFolderPrefix, &config, &cache, &consumedFiles, pStopEvent, tNow)) SGS_ERROR;
    }
    
    // Merge cache to snapshot and create delta sitemap files
    if (!MergeWithSnapshot(&cache, wszDataFolderPrefix, wszDataFolder, wszRobotsFolder, wszSitemapFolder, szSubPath, wszTempFolder, &config, tNow, pStopEvent, &robots, fCrossServer)) SGS_ERROR;

    // failures after here non-fatal
    DeleteFilesInList(&consumedFiles);

    if (!UpdateFullSitemap(wszDataFolder, wszRobotsFolder, wszSitemapFolder, szSubPath, wszTempFolder, &config, pStopEvent, tNow, &robots, fCrossServer)) SGS_ERROR;
#if APR_HAVE_WINDOWS_H
    if (config.fPingSearchEngines)
    {
        PingSearchEngine(L"www.bing.com", config.szwwwhost);
        PingSearchEngine(L"www.google.com", config.szwwwhost);
    }
#endif
    fRet = TRUE;
    SiteLog("finished successfully");
Exit:
    CloseMergeServiceLog(g_fpSiteLog);
    return fRet;
}


static BOOL DoCrossServerCountersMerge(const TCHAR* wszDataFolderPrefix, const TCHAR* wszRobotsFolder, const char* szSubPath, PALEvent *pStopEvent, const TCHAR* wszDataFolder, const TCHAR* wszInputFolder)
{
    TCHAR wszLog[MAX_PATH];
    BOOL fRet = FALSE;
    ListOfStrings consumedFiles;

    if (0 >= SNTPRINTF_S(wszLog, ARRAYSIZE(wszLog), _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH TEXT("ServiceLog.txt"), wszDataFolder)) SGS_ERROR;
    if (!consumedFiles.Initialize(100000)) SGS_ERROR;

    if (!OpenMergeServiceLog(wszLog, &g_fpSiteLog)) SGS_ERROR;
    if (!LoadCountersFiles(wszInputFolder, &consumedFiles)) SGS_ERROR;

    // failures after here non-fatal
    DeleteFilesInList(&consumedFiles);

    fRet = TRUE;
    SiteLog("finished counters merge successfully");
Exit:
    CloseMergeServiceLog(g_fpSiteLog);
    return fRet;
}


void DoOneSite(const TCHAR* wszDataFolderPrefix, const TCHAR* wszRobotsFolder, const char* szSubPath, PALEvent *pStopEvent)
{
    TCHAR wszDataFolder[MAX_PATH];
    if (!GetDataFolderWithIndex(wszDataFolderPrefix, 0, wszDataFolder, ARRAYSIZE(wszDataFolder))) SGS_ERROR;
    
    DoMerge(wszDataFolderPrefix, wszRobotsFolder, szSubPath, pStopEvent, wszDataFolder, NULL);
Exit:
    ;
}

static BOOL FolderContainsFiles(const TCHAR* wszFolder, const TCHAR* wszPattern)
{
    PALFind find;
    return find.Initialize(wszFolder, wszPattern) && (PALFind::Success == find.Find());
}

static void MergeSnapshots(const TCHAR* wszInputFolder, const TCHAR* wszOutputFolder, const char* szSubPath)
{
    do
    {
        if (!DoCrossServerCountersMerge(NULL, wszOutputFolder, szSubPath, NULL, TEXT("."), wszInputFolder))
        {
            break;
        }
    } while (FolderContainsFiles(wszInputFolder, TEXT("*.tsv")));

    do
    {
        if (!DoMerge(NULL, wszOutputFolder, szSubPath, NULL, TEXT("."), wszInputFolder))
        {
            break;
        }
    } while (FolderContainsFiles(wszInputFolder, TEXT("*.dat")));
}

int CommandLineMerge(int argc, TCHAR **wszArgv)
{
    if (2 > argc)
    {
        printf("Usage: sspservice %s <input folder> <output folder> [sub_path]\n", MERGE_OPTION);
        puts("");
        puts("   <input folder> contains DeltaSnapshot.dat and FullCounters.tsv files - these are deleted as they are consumed.");
        puts("   <output folder> is where the generated sitemaps (and robots.txt additions) go.");
        puts("   <sub_path> is for the IIS scenario where the site root is not the hostname root.");
        puts("   Current working directory must be different from <input folder> and must contain config.ini.");
        puts("See tools/CrossServerMerge/readme.txt for a description of the context in which this tool is meant to be used.");
        return -1;
    }

    const char* szSubPath = NULL;
    char szTemp[MAX_PATH];
    if (argc > 2)
    {
#ifdef _UNICODE
         PALString::Convert(szTemp, ARRAYSIZE(szTemp), wszArgv[2], TSTRLEN(wszArgv[2]));
         szSubPath = szTemp;
#else
         szSubPath = wszArgv[2];
#endif
    }

    PAL_Initialize(true);
    MergeSnapshots(wszArgv[0], wszArgv[1], szSubPath);
    PAL_Terminate();
    return 0;
}

Size_t FolderPrefixSize(const TCHAR* wszFolder)
{ // If wszFolder is of the form "<prefix>_<digits>", return the length of <prefix>.  Otherwise 0.
    size_t i = TSTRLEN(wszFolder);
    if (i < 3)
    {
        return 0;
    }
    
    i--;
    if (!isdigit(wszFolder[i])) // BUGBUG: should use _istdigit for Unicode ?
    {
        return 0;
    }

    while (i && isdigit(wszFolder[i])) i--;

    if ('_' != wszFolder[i])
    {
        return 0;
    }
    return i;
}

int DeltaMergeCppLineCount()
{
    return __LINE__;
}
