/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "SitemapOutput.h"
#include "ServiceLogging.h"
#include "counters.h"
#include "DiskSpaceTracker.h"
#include "SSPCommon.h"
#include "PALString.h"
#include "PALFile.h"
#include "SSPHash.h"

extern Counters g_counters;

#define VERSION_NUMBER(foo) int foo();
#include "VersionNumberList.h"
#undef VERSION_NUMBER

static BOOL AppendVersionString(char* szOut, Size_t cchOut, Size_t* pcchOutUsed, const char* szAppend)
{
    if (0 != STRCAT_S(szOut + *pcchOutUsed, cchOut - *pcchOutUsed, szAppend))
    {
        return FALSE;
    }
    *pcchOutUsed = strlen(szOut);
    return TRUE;
}

static BOOL AddVersionNumber(char* szOut, Size_t cchOut, Size_t* pcchOutUsed, int* pVerTotal, int ver)
{
    char szNum[40];
    if (0 >= SPRINTF_S(szNum, ARRAYSIZE(szNum), "%u", ver))
    {
        return FALSE;
    }
    if (!AppendVersionString(szOut, cchOut, pcchOutUsed, *pcchOutUsed ? "+" : " version "))
    {
        return FALSE;
    }
    if (!AppendVersionString(szOut, cchOut, pcchOutUsed, szNum))
    {
        return FALSE;
    }

    *pVerTotal += ver;
    return TRUE;
}

static BOOL GetVersionString(char* szOut, Size_t cchOut, Size_t* pcchOutUsed)
{
    int verTotal = 0;
    szOut[0] = '\0';
#define VERSION_NUMBER(foo) \
    if (!AddVersionNumber(szOut, cchOut, pcchOutUsed, &verTotal, foo())) \
    { \
        return FALSE; \
    }
#include "VersionNumberList.h"
#undef VERSION_NUMBER

    char szTotal[40];
    if (0 >= SPRINTF_S(szTotal, ARRAYSIZE(szTotal), "=%u", verTotal))
    {
        return FALSE;
    }
    if (!AppendVersionString(szOut, cchOut, pcchOutUsed, szTotal))
    {
        return FALSE;
    }
    return TRUE;
}

static void GetVersionStringWrap(char* szOut, Size_t cchOut)
{
    Size_t cchOutUsed = 0;
    if (!GetVersionString(szOut, cchOut, &cchOutUsed))
    {
        szOut[cchOutUsed] = '\0'; // take the portion that succeeded
    }
}

const char* c_szNamespace = "ssp";

#if APR_HAVE_WINDOWS_H
static BOOL DeleteFiles(const TCHAR* wszFolder, const char* szPrefix)
{
    BOOL fRet = TRUE;
    TCHAR wszSearch[MAX_PATH];

    if (strlen(szPrefix) < 3 || '*' == szPrefix[0] || '.' == szPrefix[0])
    { // dangerous
        return FALSE;
    }

    if (0 >= SNTPRINTF_S(wszSearch, ARRAYSIZE(wszSearch), _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH TEXT("%hs*"), wszFolder, szPrefix))
    {
        return FALSE;
    }

    WIN32_FIND_DATA win32_find_data;
    HANDLE hFind = FindFirstFile(wszSearch, &win32_find_data);
    if (INVALID_HANDLE_VALUE == hFind)
    {
        return (ERROR_FILE_NOT_FOUND == GetLastError()) ? TRUE : FALSE;
    }
    do
    {
        if (win32_find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
        {   // ??
        }
        else
        {
            TCHAR wszFile[MAX_PATH];

            if ((0 >= SNTPRINTF_S(wszFile, ARRAYSIZE(wszFile), STRFMT_TLS STRFMT_TDIRSLASH STRFMT_TLS, wszFolder, win32_find_data.cFileName)) ||
                !PALFile::Delete(wszFile))
            {
                fRet = FALSE;
            }
        }
    } while (FindNextFile(hFind, &win32_find_data));
    FindClose (hFind);
    return fRet;
}
#else
static BOOL DeleteFiles(const TCHAR* wszFolder, const char* szPrefix)
{
    BOOL fRet = TRUE;
    char szFolder[MAX_PATH];
    char szFullPath[MAX_PATH];
    char szSearch[MAX_PATH];
    apr_pool_t *pPool = NULL;
    apr_dir_t *pAprDir = NULL;
    apr_finfo_t fInfo;

    if (strlen(szPrefix) < 3 || '*' == szPrefix[0] || '.' == szPrefix[0])
    { // dangerous
        fRet = FALSE;
        goto Done;
    }

    if (apr_pool_create(&pPool, NULL) != APR_SUCCESS)
    {
        fRet = FALSE;
        goto Done;
    }

    PALString::Convert(szFolder, ARRAYSIZE(szFolder), wszFolder, TSTRLEN(wszFolder));
    if (0 >= SNPRINTF(szSearch, ARRAYSIZE(szSearch), "%s*", szPrefix))
    {
        fRet = FALSE;
        goto Done;
    }

    if (apr_dir_open(&pAprDir, szFolder, pPool) != APR_SUCCESS)
    {
        fRet = FALSE;
        goto Done;
    }

    while(apr_dir_read(&fInfo, APR_FINFO_TYPE | APR_FINFO_NAME, pAprDir) == APR_SUCCESS)
    {
        if (fInfo.filetype == APR_REG)
        {
            if (apr_fnmatch(szSearch, fInfo.name, APR_SUCCESS) == 0)
            {
                // Make the full pathname.
                if (SNPRINTF(szFullPath, ARRAYSIZE(szFullPath), "%s/%s", szFolder, fInfo.name) > 0)
                {   
                    if (remove(szFullPath) != 0)
                    {
                        fRet = FALSE;
                    }
                }
            }
        }
    }

Done:
    if (pAprDir != NULL)
    {
        apr_dir_close(pAprDir);
    }

    if (pPool != NULL)
    {
        apr_pool_destroy(pPool);
    }

    return fRet;
}
#endif

static const char* EntityEscape(char c)
{
    if ('&' == c) return "&amp;";
    if ('\''  == c) return "&apos;";
    if ('"' == c) return "&quot;";
    if ('>' == c) return "&gt;";
    if ('<' == c) return "&lt;";
    return NULL;
}

static BOOL NeedsPercentEncoding(char c)
{ // mimicking .NET's Uri.EscapeUriString

    return (c >= 127 || c <= ' ' || c == '"' || c == '%' || c == '<' || c == '>' || c == '[' || c == '\\' || c == ']' || c == '^' || c == '`' || c == '{' || c == '|' || c == '}');
}

static char rchHex[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

BOOL SitemapOutput::EncodeURL(const char* pcUrl, size_t cbUrl, __out_bcount_part(*pcbOut,*pcbOut) char* pcOut, size_t* pcbOut)
{
    size_t cbOut = 0;
    for (size_t i = 0; i < cbUrl; i++)
    {
        char c = pcUrl[i];
        const char* szEntity = EntityEscape(c);
        if (szEntity)
        {
            size_t cbEntity = strlen(szEntity);
            if (cbOut + cbEntity > *pcbOut)
            {
                return FALSE;
            }
            memcpy(pcOut, szEntity, cbEntity);
            cbOut += cbEntity;
        }
        else if (NeedsPercentEncoding(c))
        {
            if (cbOut + 3 > *pcbOut)
            {
                return FALSE;
            }
            pcOut[cbOut++] = '%';
            pcOut[cbOut++] = rchHex[(unsigned char)c >> 4];
            pcOut[cbOut++] = rchHex[(unsigned char)c & 0xf];
        }
        else
        {
            if (cbOut >= *pcbOut)
            {
                return FALSE;
            }
            pcOut[cbOut++] = c;
        }
    }
    *pcbOut = cbOut;
    return TRUE;
}

SitemapOutput::SitemapOutput() : 
    m_fpIndex(NULL), 
    m_pContentWriter(NULL), 
    m_pDeadlinkWriter(NULL), 
    m_iContentFile(0), 
    m_iDeadlinkFile(0),
    m_fError(TRUE)
{
}

void SitemapOutput::Free()
{
    if (m_fpIndex) 
    {
        fclose(m_fpIndex); 
        m_fpIndex = NULL;
    }
    if (m_pContentWriter) 
    {
        delete m_pContentWriter;  
        m_pContentWriter = NULL;
    }
    if (m_pDeadlinkWriter)
    {
        delete m_pDeadlinkWriter;  
        m_pDeadlinkWriter = NULL;
    }
}

SitemapOutput::~SitemapOutput()
{
    Free();
}

BOOL SitemapOutput::GetChildFileName(__out_ecount(cchName) char* szName, size_t cchName, const char* prefix, unsigned i)
{
    BOOL ret = SPRINTF_S(szName, cchName, "%s_%08u.xml%s", prefix, i, m_gzip ? ".gz" : "");
    return 0 < ret;
}

const unsigned char c_szBOM[] = { 0xEF, 0xBB, 0xBF };

#define MAX_SITEMAP_URL 2048

int SitemapOutput::PrintToSitemap(CompressingFileWriter* pFileWriter, const char* format, ...)
{
    va_list args;
    int len;
    char temp[MAX_SITEMAP_URL + 80];

    va_start( args, format );
    len = VSPRINTF_S( temp, ARRAYSIZE(temp), format, args );
    if (len <= 0) return len;

    if (NO_ERROR != (pFileWriter)->Write(temp, len)) return -1;
    return len;
}

static BOOL GetTimeNow(__out_bcount(cb) char* sz, size_t cb)
{
    struct tm tms;
    PluginGetUTCTime(&tms);
    return 0 < SPRINTF_S(sz, cb, "%04d-%02d-%02dT%02d:%02d:%02dZ", 
        1900 + tms.tm_year, 1 + tms.tm_mon, tms.tm_mday, tms.tm_hour, tms.tm_min, tms.tm_sec);
}

static void GetHost(__out_ecount(cchHost) char* szHost, Size_t cchHost)
{
#if APR_HAVE_WINDOWS_H
    DWORD dw = (DWORD)cchHost;
    if (!GetComputerNameA(szHost, &dw)) // Windows gethostname() requires WSAStartup() and behaves undesirably in cluster situations
#else
    if (0 != gethostname(szHost, cchHost))
#endif
    {
        szHost[0] = '\0';
    }
    szHost[cchHost-1] = '\0'; // POSIX doesn't guarantee NULL termination even on success
}

static UInt64 GetHostHash()
{
    char szHost[256] = "";
    GetHost(szHost, ARRAYSIZE(szHost));
    return SSPHash(szHost, strlen(szHost));
}

static const char* CrossServerString(BOOL fCrossServer)
{
    return fCrossServer ? "(cross-server)" : "";
}

BOOL SitemapOutput::WriteHeader(CompressingFileWriter* pFileWriter)
{
    int cb;
    char szVersion[1024];
    GetVersionStringWrap(szVersion, ARRAYSIZE(szVersion));

    if (0 >= (cb = PrintToSitemap(pFileWriter, "%.*s", (int) ARRAYSIZE(c_szBOM), c_szBOM))) return FALSE;
    if (0 >= (cb = PrintToSitemap(pFileWriter, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"))) return FALSE;                                                    m_cbContentChild += cb;
    if (0 >= (cb = PrintToSitemap(pFileWriter, "<!-- Powered by the Bing XML Sitemap Plugin -->\r\n"))) return FALSE;                                       m_cbContentChild += cb;
    if (0 >= (cb = PrintToSitemap(pFileWriter, "<urlset xmlns=\"http://www.sitemaps.org/schemas/sitemap/0.9\"\r\n"))) return FALSE;                                       m_cbContentChild += cb;
    if (0 >= (cb = PrintToSitemap(pFileWriter, "        xmlns:%s=\"http://www.bing.com/schemas/sitemap-%s/0.2\"\r\n", c_szNamespace, c_szNamespace))) return FALSE;                                       m_cbContentChild += cb;
    if (0 >= (cb = PrintToSitemap(pFileWriter, ">\r\n"))) return FALSE;                                m_cbContentChild += cb;
    if (0 >= (cb = PrintToSitemap(pFileWriter, "<!-- Generated on %s by " STRFMT_I64 "X %s%s-->\r\n", m_szTimeNow, m_uHostHash, CrossServerString(m_fCrossServer), szVersion))) return FALSE;    m_cbContentChild += cb;
    return TRUE;
}

BOOL SitemapOutput::WriteEntry(Signature sig, Signature sigBots, const char* pcUrl, size_t cbUrl, unsigned t32LastMod, BOOL fLastModBots, const char* szChangeFreq, float flPriority, float flSSPPriority)
{
    BOOL ret = FALSE;
    int cb = 0;
    size_t cbSum = 0;
    char szLastMod[80] = "";
    CompressingFileWriter** pWriter;
    BOOL fDeadlink = (DEADLINK_SIGNATURE == sig);

    if (t32LastMod)
    {
        time_t t = t32LastMod * (time_t)SEC_PER_T32;
        struct tm tms;
        if (gmtime_s(&tms, &t)) 
        {
            goto EXIT;
        }

        if (0 >= SPRINTF_S(szLastMod, ARRAYSIZE(szLastMod), "%04u-%02u-%02u", 1900 + tms.tm_year, 1 + tms.tm_mon, tms.tm_mday)) 
        {
            goto EXIT;
        }
    }
    
    pWriter = GetChildFileWriter(fDeadlink);

    if (0 >= (cb = PrintToSitemap(*pWriter, " <url>\r\n"))) 
    {
        goto EXIT;                           
    }
    cbSum += cb;

    if (0 >= (cb = PrintToSitemap(*pWriter, "  <loc>%.*s</loc>\r\n", cbUrl, pcUrl))) 
    {
        goto EXIT;                           
    }
    cbSum += cb;

    if (szLastMod[0]) 
    { 
        if (0 >= (cb = PrintToSitemap(*pWriter, "  <lastmod>%s</lastmod>\r\n", szLastMod)))     
        {
            goto EXIT;                           
        }
        cbSum += cb;
    }
    if (szChangeFreq) { 
        if (0 >= (cb = PrintToSitemap(*pWriter, "  <changefreq>%s</changefreq>\r\n", szChangeFreq))) 
        {
            goto EXIT;                           
        }
        cbSum += cb;
    }
    if (flPriority) { 
        if (0 >= (cb = PrintToSitemap(*pWriter, "  <priority>%.1f</priority>\r\n", flPriority)))    
        {
            goto EXIT;                           
        }
        cbSum += cb;
    }

    if(szLastMod[0] || !fDeadlink || flSSPPriority)
    {
        if (0 >= (cb = PrintToSitemap(*pWriter, "  <ssp:ssp>\r\n"))) 
        {
            goto EXIT;                           
        }
        cbSum += cb;

        if (fLastModBots)
        {
            if (0 >= (cb = PrintToSitemap(*pWriter, "    <%s:lastmodbots>1</%s:lastmodbots>\r\n", c_szNamespace, c_szNamespace)))
            {
                goto EXIT;                           
            }
            cbSum += cb;
        }

        if (!fDeadlink)
        {
            if (0 >= (cb = PrintToSitemap(*pWriter, "    <%s:sig>%08X:%08X</%s:sig>\r\n", c_szNamespace, sig, sigBots, c_szNamespace)))
            {
                goto EXIT;                           
            }
            cbSum += cb;
        }

        if(flSSPPriority) {
            if (0 >= (cb = PrintToSitemap(*pWriter, "    <%s:ssppriority>%.1f</%s:ssppriority>\r\n", c_szNamespace, flSSPPriority, c_szNamespace)))    
            {
                goto EXIT;                           
            }
            cbSum += cb;
        }

        if (0 >= (cb = PrintToSitemap(*pWriter, " </ssp:ssp>\r\n"))) 
        {
            goto EXIT;                           
        }
        cbSum += cb;
    }
    if (0 >= (cb = PrintToSitemap(*pWriter, " </url>\r\n"))) 
    {
        goto EXIT;                           
    }
    cbSum += cb;
    
    ret = TRUE;

EXIT:
    if(fDeadlink)
    {
        m_cbDeadlinkChild += cbSum;
        m_cEntriesInCurDeadlinkChild++;
    }
    else
    {
        m_cbContentChild += cbSum;
        m_cEntriesInCurContentChild++;
    }

    return ret;

}

BOOL SitemapOutput::WriteFooter(BOOL fDeadlink)
{
    CompressingFileWriter** pWriter = GetChildFileWriter(fDeadlink);

    int cb;
    if (0 >= (cb = PrintToSitemap(*pWriter, "</urlset>\r\n"))) 
    {
        return FALSE;
    }
    if(fDeadlink)
    {
        m_cbDeadlinkChild += cb;
    }
    else
    {
        m_cbContentChild += cb;
    }
    return TRUE;
}

CompressingFileWriter** SitemapOutput::GetChildFileWriter(BOOL fDeadlink)
{
    if(fDeadlink)
    {
        return &m_pDeadlinkWriter;
    }
    
    return &m_pContentWriter;
    
}

static BOOL WriteIndexHeader(FILE* fp)
{
    int cb;
    fprintf(fp, "%.*s", (int) ARRAYSIZE(c_szBOM), c_szBOM);
    if (0 >= (cb = fprintf(fp, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"))) return FALSE;
    if (0 >= (cb = fprintf(fp, "<!-- Powered by the Bing XML Sitemap Plugin -->\n"))) return FALSE;             
    if (0 >= (cb = fprintf(fp, "<sitemapindex xmlns=\"http://www.sitemaps.org/schemas/sitemap/0.9\"\n"))) return FALSE;
    if (0 >= (cb = fprintf(fp, "              xmlns:%s=\"http://www.bing.com/schemas/sitemap-%s/0.2\">\n", c_szNamespace, c_szNamespace))) return FALSE;

    return TRUE;
}

static BOOL WriteIndexEntry(BOOL fComment, FILE* fp, const char* szWWWHost, const char* szSubPath, const char* szChild, Size_t* pcb = NULL)
{
    char szTimeNow[80];
    if (!GetTimeNow(szTimeNow, ARRAYSIZE(szTimeNow))) return FALSE;
    
    int cb;
    if (0 >= (cb = fprintf(fp, " %s <sitemap>\n", fComment ? "<!--" : ""))) return FALSE;                                if (pcb) *pcb += cb;
    if (0 >= (cb = fprintf(fp, "  <loc>%s%s/%s</loc>\n", szWWWHost, szSubPath, szChild))) return FALSE;    if (pcb) *pcb += cb;
    if (0 >= (cb = fprintf(fp, "  <lastmod>%s</lastmod>\n", szTimeNow))) return FALSE;      if (pcb) *pcb += cb;
    if (0 >= (cb = fprintf(fp, " </sitemap> %s\n", fComment ? "-->" : ""))) return FALSE;                               if (pcb) *pcb += cb;
    return TRUE;
}

const char* c_szIndexFooter = "</sitemapindex>";

static BOOL WriteIndexFooter(FILE* fp)
{
    int cb;
    if (0 >= (cb = fprintf(fp, "%s", c_szIndexFooter))) return FALSE;
    return TRUE;
}

BOOL SitemapOutput::CloseChild(BOOL fDeadlink)
{
    CompressingFileWriter** pWriter = GetChildFileWriter(fDeadlink);
    if(*pWriter)
    {
        if (!WriteFooter(fDeadlink))
        {
            return FALSE;
        }

        delete *pWriter; 
        *pWriter = NULL;
    }
    return TRUE;
}

BOOL SitemapOutput::NewChildSitemap(CompressingFileWriter** pFileWriter, const char* szChild, __out_ecount(cCurFile) TCHAR* wszCurFile, size_t cCurFile)
{

    if (0 >= SNTPRINTF_S(wszCurFile, cCurFile, _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH STRFMT_THS, m_wszTempFolder, szChild)) SGS_ERROR;
    
    if (NULL == (*pFileWriter = CompressingFileWriter::Create(wszCurFile, m_gzip, m_gzip))) SGS_ERROR;
        
    if (!WriteHeader(*pFileWriter)) SGS_ERROR;

    return TRUE;


Exit:
    return FALSE;
}

BOOL SitemapOutput::NewChildSitemap(BOOL fDeadlink)
{
    CompressingFileWriter* pWriter = NULL;
    char* szName;
    TCHAR* wszCurFile;
    size_t cCurFile;
    unsigned* iFile;

    if(fDeadlink)
    {
        szName = m_szDeadlinkName;
        wszCurFile = m_wszCurDeadlinkFile;
        cCurFile = ARRAYSIZE(m_wszCurDeadlinkFile);
        iFile = &m_iDeadlinkFile;

    }
    else
    {
        szName = m_szContentName;
        wszCurFile = m_wszCurContentFile;
        cCurFile = ARRAYSIZE(m_wszCurContentFile);
        iFile = &m_iContentFile;
    }
    
    char szChild[MAX_PATH];
    if (!GetChildFileName(szChild, ARRAYSIZE(szChild), szName, ++(*iFile))) 
    {
        SGS_ERROR;
    }

    if(!NewChildSitemap(&pWriter, szChild, wszCurFile, cCurFile))
    {
        SGS_ERROR;
    }

    if (!WriteIndexEntry(fDeadlink, m_fpIndex, m_szWWWHost, m_szSubPath, szChild, &m_cbIndex)) 
    {
        SGS_ERROR;
    }

    if(fDeadlink)
    {
        m_pDeadlinkWriter = pWriter;
        m_cEntriesInCurDeadlinkChild = 0;
        m_cbDeadlinkChild = 0;
    }
    else
    {
        m_pContentWriter = pWriter;
        m_cEntriesInCurContentChild = 0;
        m_cbContentChild = 0;
    }

    return TRUE;

Exit:
    return FALSE;
}

static unsigned ParseUint(const char* pcStart, const char* pcEnd)
{
    unsigned ret = 0;
    while ((pcStart < pcEnd) && isdigit(static_cast<unsigned char>(*pcStart))) 
    {
        ret = 10 * ret + *pcStart++ - '0';
    }
    return ret;
}

static BOOL ParseFileIndex(const char* szLine, unsigned* puIndex)
{
    const char* pcDot = strstr(szLine, ".xml");
    if (!pcDot)
    {
        return FALSE;
    }

    const char* pcDigit;
    for (pcDigit = pcDot; (pcDigit > szLine) && isdigit(static_cast<unsigned char>(pcDigit[-1])); pcDigit--);
    if (!isdigit(static_cast<unsigned char>(*pcDigit)))
    {
        return FALSE;
    }

    *puIndex = ParseUint(pcDigit, pcDot);
    return TRUE;
}

static BOOL CopyIndexFile(const TCHAR* wszIndexFile, FILE* fpOut, unsigned* piMaxIndex)
{
    BOOL fRet = FALSE;
    BOOL fEndFound = FALSE;
    size_t cbIndexFooter = strlen(c_szIndexFooter);
    FILE* fp = NULL;

    if (0 != PALFile::Open(&fp, wszIndexFile, TEXT("rt")) || !fp) SGS_ERROR;
    
    while (!feof(fp))
    {
        char s[1024];
        if (!fgets(s, sizeof(s), fp)) continue;
        if (!s[0]) break;
        size_t cb = strlen(s);
        if (cb >= cbIndexFooter && !memcmp(c_szIndexFooter, s, cbIndexFooter))
        {
            fEndFound = TRUE;
            break;
        }
        if (0 > fputs(s, fpOut)) SGS_ERROR;

        unsigned u;
        if (ParseFileIndex(s, &u) && u > *piMaxIndex)
        {
            *piMaxIndex = u;
        }
    }
    if (!fEndFound)
    {
        SGS_ERROR;
    }
    
    fRet = TRUE;
Exit:
    if (fp) fclose(fp);
    return fRet;
    
}

BOOL SitemapOutput::GetSitemapFilePath(__out_ecount(rchPath) TCHAR* wszPath, size_t rchPath, const TCHAR* wszFolder, const char* szName)
{
    return 0 < SNTPRINTF_S(wszPath, rchPath, _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH STRFMT_THS TEXT(".xml"), wszFolder, szName);
}

BOOL SitemapOutput::Initialize(const char* szName, const char* szDeadlinkName, const TCHAR* wszTempFolder, const TCHAR* wszOutFolder, const char* szWWWHost, const char* szSubPath, BOOL gzip, BOOL fIncremental, BOOL fCrossServer)
{
    char szVersion[1024];

    m_fIncremental = fIncremental;
    m_gzip = gzip;
    m_fCrossServer = fCrossServer;

    if (0 != PALString::Copy(m_szContentName, ARRAYSIZE(m_szContentName), szName)) SGS_ERROR;
    if (0 != PALString::Copy(m_szDeadlinkName, ARRAYSIZE(m_szDeadlinkName), szDeadlinkName))
    {
        SGS_ERROR;
    }
    if (0 != PALString::Copy(m_wszTempFolder, ARRAYSIZE(m_wszTempFolder), wszTempFolder)) SGS_ERROR;
    if (0 != PALString::Copy(m_wszOutFolder, ARRAYSIZE(m_wszOutFolder), wszOutFolder)) SGS_ERROR;
    if (0 != PALString::Copy(m_szWWWHost, ARRAYSIZE(m_szWWWHost), szWWWHost)) SGS_ERROR;
    if (0 != PALString::Copy(m_szSubPath, ARRAYSIZE(m_szSubPath), szSubPath)) SGS_ERROR;
    m_cbWWWHost = strlen(m_szWWWHost);

    if (!PALDirectory::EnsureDirectoryExists(m_wszTempFolder)) SGS_ERROR;
    
    DeleteFiles(m_wszTempFolder, m_szContentName);

    if (!GetTimeNow(m_szTimeNow, ARRAYSIZE(m_szTimeNow))) SGS_ERROR;
    
    // Get hash of host name
    m_uHostHash = GetHostHash();
    
    m_iContentFile = 0;
    m_cbIndex = 0;
    if (!GetSitemapFilePath(m_wszIndexFile, ARRAYSIZE(m_wszIndexFile), m_wszOutFolder, m_szContentName)) SGS_ERROR;
    if (!GetSitemapFilePath(m_wszTempIndexFile, ARRAYSIZE(m_wszTempIndexFile), m_wszTempFolder, m_szContentName)) SGS_ERROR;

    if (0 != PALFile::Open(&m_fpIndex, m_wszTempIndexFile, TEXT("wt")) || !m_fpIndex) SGS_ERROR;

    if (m_fIncremental && PALDirectory::PathExists(m_wszIndexFile))
    {
        if (!CopyIndexFile(m_wszIndexFile, m_fpIndex, &m_iContentFile)) SGS_ERROR;
    }
    else
    {
        if (!WriteIndexHeader(m_fpIndex)) SGS_ERROR;
    }

    GetVersionStringWrap(szVersion, ARRAYSIZE(szVersion));
    fprintf(m_fpIndex, "<!-- Generated on %s by " STRFMT_I64 "X %s%s-->\n", m_szTimeNow, m_uHostHash, CrossServerString(m_fCrossServer), szVersion);

    m_iFirstContentFile = m_iContentFile + 1;
    m_iFirstDeadlinkFile = m_iDeadlinkFile + 1;

    m_pContentWriter = NULL;
    m_cEntriesInCurContentChild = 0;
    m_cbContentChild = 0;

    m_pDeadlinkWriter = NULL;
    m_cbDeadlinkChild = 0;
    m_cEntriesInCurDeadlinkChild = 0;

    m_fError = FALSE;
    return TRUE;

Exit:
    Free();
    return FALSE;
}

BOOL SitemapOutput::AddEntry(Signature sig, Signature sigBots, const char* pcUrl, size_t cbUrl, unsigned t32LastMod, BOOL fLastModBots, const char* szChangeFreq, float flPriority,float flSSPPriority)
{   
    char pcUrlWithHost[2048];
    char pcEncoded[2048];
    size_t cbEncoded = sizeof(pcEncoded);
    size_t cbUrlWithHost;
    BOOL fDeadlink = (DEADLINK_SIGNATURE == sig);

    if (m_fError) 
    {
        g_counters.cGenerationFailure.Increment();
        return FALSE;
    }

    // setting up the right variables to use depending on child sitemap type
    CompressingFileWriter** pWriter = GetChildFileWriter(fDeadlink);
    size_t cbChild = 0;
    size_t cEntriesInCurChild = 0;
    if(fDeadlink)
    {
        cbChild = m_cbDeadlinkChild;
        cEntriesInCurChild = m_cEntriesInCurDeadlinkChild;
    }
    else
    {
        cbChild = m_cbContentChild;
        cEntriesInCurChild = m_cEntriesInCurContentChild;
    }

    if (cEntriesInCurChild >= 50000 || cbChild >= 10000000)
    {
        if (!CloseChild(fDeadlink))
        {
            m_fError = TRUE;
            SGS_ERROR;
        }
    }

    if (!*pWriter)
    {
        if (!NewChildSitemap(fDeadlink))
        {
            m_fError = TRUE;
            SGS_ERROR;
        }
    }


    cbUrlWithHost = m_cbWWWHost + cbUrl;
    if (cbUrlWithHost > sizeof(pcUrlWithHost))
    {
        g_counters.cOutputUrlTooLong.Increment();
        SGS_ERROR;
    }
    memcpy_s(pcUrlWithHost, sizeof(pcUrlWithHost), m_szWWWHost, m_cbWWWHost);
    memcpy_s(pcUrlWithHost + m_cbWWWHost, sizeof(pcUrlWithHost) - m_cbWWWHost, pcUrl, cbUrl);

    if (!EncodeURL(pcUrlWithHost, cbUrlWithHost, pcEncoded, &cbEncoded) || (cbEncoded > MAX_SITEMAP_URL))
    {
        g_counters.cOutputUrlTooLong.Increment();
        SGS_ERROR; // m_fError only set if error is expected to affect all future URLs
    }

    if (!WriteEntry(sig, sigBots, pcEncoded, cbEncoded, t32LastMod, fLastModBots, szChangeFreq, flPriority, flSSPPriority))
    {
        m_fError = TRUE;
        SGS_ERROR;
    }

    return TRUE;
Exit:
    return FALSE;
}

static void WriteSitemapCounters(FILE* fp, const Counters* pCounters)
{
    if (!pCounters) return;

#define COUNTER(name) fprintf(fp, "<!-- counter%s=" STRFMT_I32 "u -->\n", #name, pCounters->c##name.m_val);
#include "counterlist.h"
#undef COUNTER
}

BOOL SitemapOutput::MoveChildSiteMapFiles(const char* szSiteMapName, unsigned iFirstFileIdx, unsigned iLastFileIdx)
{
    for (unsigned u = iFirstFileIdx; u <= iLastFileIdx; u++)
    {
        char szChild[MAX_PATH];
        if (!GetChildFileName(szChild, ARRAYSIZE(szChild), szSiteMapName, u)) SGS_ERROR;
        TCHAR wszFrom[MAX_PATH];
        if (0 >= SNTPRINTF_S(wszFrom, ARRAYSIZE(wszFrom), _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH STRFMT_THS, m_wszTempFolder, szChild)) SGS_ERROR;
        TCHAR wszTo[MAX_PATH];
        if (0 >= SNTPRINTF_S(wszTo, ARRAYSIZE(wszTo), _TRUNCATE, STRFMT_TLS STRFMT_TDIRSLASH STRFMT_THS, m_wszOutFolder, szChild)) SGS_ERROR;

        if(!MoveSiteMapFile(wszFrom, wszTo))
        {
            goto Exit;
        }
    }

    return TRUE;
Exit:
    return FALSE;
}

BOOL SitemapOutput::Close(const Counters* pCounters, const SSPConfig* pConfig, BOOL fKeepIndexFile)
{
    // content sitemap
    if (!CloseChild(FALSE))
    {
        SGS_ERROR;
    }

    // deadlink sitemap
    if (!CloseChild(TRUE))
    {
        SGS_ERROR;
    }

    PrintSSPConfig(m_fpIndex, pConfig);
    if(pConfig != NULL && pConfig->fHelpSSP)
    {
        WriteSitemapCounters(m_fpIndex, pCounters);
    }
    WriteIndexFooter(m_fpIndex);
    fclose(m_fpIndex); m_fpIndex = NULL;
    PALFile::SetModifiedTime(m_wszTempIndexFile);

    if (m_iContentFile != 1 || m_iDeadlinkFile != 1|| m_fIncremental || fKeepIndexFile || pCounters) // i.e. pretty much always
    {
        if(!MoveSiteMapFile(m_wszTempIndexFile, m_wszIndexFile))
        {
            goto Exit;
        }
        
        // content sitemaps
        if(!MoveChildSiteMapFiles(m_szContentName, m_iFirstContentFile, m_iContentFile))
        {
            goto Exit;
        }

        // deadlink sitemaps
        if(!MoveChildSiteMapFiles(m_szDeadlinkName, m_iFirstDeadlinkFile, m_iDeadlinkFile))
        {
            goto Exit;
        }
    }
    else
    { // no index file needed - use the first child sitemap and delete the index file
        if(!MoveSiteMapFile(m_wszCurContentFile, m_wszIndexFile))
        {
            goto Exit;
        }

        PALFile::Delete(m_wszTempIndexFile);
    }

    return TRUE;
Exit:
    return FALSE;
}

BOOL SitemapOutput::SetFileAccessForWebAccess(const TCHAR* wszFilePath)
{
#if APR_HAVE_WINDOWS_H
    return SetFileAccess(wszFilePath, "Everyone", FILE_GENERIC_READ);
#else
    return SetFileAccess(wszFilePath);
#endif
}

BOOL SitemapOutput::MoveSiteMapFile(const TCHAR* wszSrcFile, const TCHAR* wszDstFile)
{
    if(!PALFile::Move(wszSrcFile, wszDstFile))
    {
        SGS_ERROR;
    }
    if (!SetFileAccessForWebAccess(wszDstFile))
    {
        SGS_ERROR;
    }

    return TRUE;

Exit:
    return FALSE;
}

int SitemapOutputCppLineCount()
{
    return __LINE__;
}

