/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "FileWriter.h"
#include "DiskSpaceTracker.h"
#include "SSPCommon.h"
#include "DeltaFile.h"
#include "PALString.h"
#include "PALFile.h"
#include "PlatformSpecificPluginGlobals.h"

BOOL FileWriter::Initialize(
    IAllocator* pAlloc,
    ICritSecDispenser* pCSD,
    const TCHAR* wszFolder,
    unsigned maxSecondsPerFile,
    Counters* pCounters,
    DiskSpaceTracker *pDiskSpaceTracker)
{
    m_cb = 0;
    m_pCounters = pCounters;
    m_pwszFolder = wszFolder;
    m_maxSecondsPerFile = maxSecondsPerFile;
    m_pDiskSpaceTracker = pDiskSpaceTracker;

    if (!GetFullPath(m_pwszFolder, TEXT("output.tmp"), m_wszTempFile, ARRAYSIZE(m_wszTempFile)))
    {
        return FALSE;
    }
    if (NO_ERROR != m_writer.DoInit(TRUE, FALSE, pAlloc))
    {
        return FALSE;
    }
    m_ics = pCSD->GetCS();
    return TRUE;
}

void FileWriter::Write(const char* pb, Size_t cb)
{
    if (!cb)
    {
        return;
    }

    AutoCriticalSection a(CritSecPtr(m_ics));
    if (!m_cb)
    {
        if (m_pDiskSpaceTracker != NULL && !m_pDiskSpaceTracker->CheckSpace())
        {
            m_pCounters->cLostDiskFull.Increment();
            return;
        }

        unsigned dwMagic = DELTA_FILE_MAGIC_HEADER1;
        if ((NO_ERROR != m_writer.Open(m_wszTempFile)) ||
            (NO_ERROR != m_writer.Write(&dwMagic, 4)))
        {
            m_writer.Close();
            m_pCounters->cEntryLost.Increment();
            return;
        }
        
        m_cb = 0;
        m_tOpenTime = PluginTime();
    }

    m_writer.Write(pb, (DWORD)cb);
    m_cb += cb;
    if (m_cb >= 4194304 || TimeExpired())
    {
        FlushHelper();
    }
}

BOOL FileWriter::TimeExpired()
{
    time_t tNow = PluginTime();
    return tNow > m_tOpenTime + (time_t)m_maxSecondsPerFile;
}

void FileWriter::Tick()
{
    if (TimeExpired())
    {
        AutoCriticalSection a(CritSecPtr(m_ics));
        if (TimeExpired())
        {
            FlushHelper();
            m_tOpenTime = PluginTime(); // prevent TimeExpired() from triggering repeatedly
        }
    }
}

void FileWriter::Flush()
{
    AutoCriticalSection a(CritSecPtr(m_ics));
    FlushHelper();
}

// Note: In linux, the same lock in the same thread can not be acquired
// more than once. Ie, nested locks are not supported. Functions that 
// would normally call Flush(), instead would call FlushHelper().
void FileWriter::FlushHelper()
{
    if (m_cb)
    {
        m_writer.Close();
        m_cb = 0;

        TCHAR wszFolder[MAX_PATH];
        struct tm tms;
        PluginGetLocalTime(&tms);
        if (0 >= SNTPRINTF_S(wszFolder, ARRAYSIZE(wszFolder), ARRAYSIZE(wszFolder) - 1,
            STRFMT_TLS STRFMT_TDIRSLASH TEXT("%04u%02u%02u"), m_pwszFolder,
            1900 + tms.tm_year, 1 + tms.tm_mon, tms.tm_mday))
        {
            m_pCounters->cFileLost.Increment();
            return;
        }
        if (!PALDirectory::EnsureDirectoryExists(wszFolder))
        {
            m_pCounters->cFileLost.Increment();
            return;
        }
        TCHAR wszFile[MAX_PATH];
        if (0 >= SNTPRINTF_S(wszFile, ARRAYSIZE(wszFile), ARRAYSIZE(wszFolder) - 1,
            STRFMT_TLS STRFMT_TDIRSLASH TEXT("%02u%02u%02u.dat"), wszFolder,
            tms.tm_hour, tms.tm_min, tms.tm_sec))
        {
            m_pCounters->cFileLost.Increment();
            return;
        }
        if (!PALFile::Move(m_wszTempFile, wszFile))
        {
            m_pCounters->cFileLost.Increment();
            return;
        }
    }
}

