/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "PlatformSpecificPluginGlobals.h"
#include "PluginLibCommon.h"
#include "PALString.h"
#include "PALSync.h"
#include "PluginLib.h"
#include "SSPCommon.h"

const Size_t c1MB = 1048576;

#ifdef __TARGET_IIS__

VirtualAllocator::VirtualAllocator() :
    m_pSharedMemBaseAddr(NULL), m_hMapping(NULL), m_fSharedMemory(FALSE)
{
}

VirtualAllocator::~VirtualAllocator()
{
    Free();
}

void VirtualAllocator::FreeSharedMemory()
{
    if (m_pSharedMemBaseAddr)
    {
        UnmapViewOfFile(m_pSharedMemBaseAddr);
        m_pSharedMemBaseAddr = NULL;
    }
    if (m_hMapping)
    {
        CloseHandle(m_hMapping);
        m_hMapping = NULL;
    }
}

void VirtualAllocator::FreeChild()
{
}

void GlobalLog(const char* fmt, ...)
{
    TCHAR wszFolder[MAX_PATH], wszFile[MAX_PATH];
    if (!GetDataFolder(wszFolder, ARRAYSIZE(wszFolder)))
    {
        return;
    }
    if (0 >= STPRINTF_S(wszFile, ARRAYSIZE(wszFile), STRFMT_TLS STRFMT_TDIRSLASH TEXT("GlobalPluginLog.txt"), wszFolder))
    {
        return;
    }

    va_list args;
    va_start(args, fmt);
    char s[1024];
    if (0 >= VSPRINTF_S(s, ARRAYSIZE(s), fmt, args))
    {
        return;
    }

    LogMessage(wszFile, s);
}

BOOL VirtualAllocator::InitializeSharedMemory(const ConfigPluginGlobals & config, Size_t *pcb)
{
    BOOL fRet = FALSE;
    *pcb = config.cbSharedAllocatorSize * 1048576;
    LARGE_INTEGER li;
    li.QuadPart = *pcb;

    TCHAR tszName[MAX_PATH];
    if (!GetIPCName(tszName, ARRAYSIZE(tszName), TEXT("SharedMemory"), 0)) 
    {
        goto Exit;
    }
    //wprintf(TEXT("Shared memory name '%s'\n"), tszName);
    m_hMapping = CreateFileMapping(
                 INVALID_HANDLE_VALUE,    // use paging file
                 NULL,                    // default security
                 PAGE_READWRITE,          // read/write access
                 li.HighPart,             // maximum object size (high-order DWORD)
                 li.LowPart,              // maximum object size (low-order DWORD)
                 tszName); // name of mapping object
    if (!m_hMapping)
    {
        GlobalLog("CreateFileMapping(%u) last error %u " STRFMT_TS, (unsigned)config.cbSharedAllocatorSize, GetLastError(), tszName);
        goto Exit;
    }

    void* pDesired = (void*)(1048576 * (Size_t)config.uSharedMemAddrMB);
    m_pSharedMemBaseAddr = MapViewOfFileEx(m_hMapping, FILE_MAP_ALL_ACCESS, 0, 0, *pcb, pDesired);
    if (!m_pSharedMemBaseAddr)
    {
        GlobalLog("MapViewOfFileEx(%u,%u) last error %u " STRFMT_TS, config.uSharedMemAddrMB, (unsigned)config.cbSharedAllocatorSize, GetLastError(), tszName);
        goto Exit;
    }

    fRet = TRUE;
Exit:
    if (!fRet)
    {
        FreeSharedMemory();
    }
    return fRet;
}

BOOL VirtualAllocator::InitializePrivateMemory(const ConfigPluginGlobals & config, Size_t *pcb)
{
    *pcb = config.cbPrivateAllocatorSize * c1MB;
    m_pSharedMemBaseAddr = VirtualAlloc(NULL, *pcb, MEM_RESERVE, PAGE_READWRITE);
    if (!m_pSharedMemBaseAddr)
    {
        GlobalLog("failed to MEM_RESERVE %u MB", (unsigned)config.cbPrivateAllocatorSize);
        return FALSE;
    }
    SYSTEM_INFO si;
    GetSystemInfo(&si);
    m_cbPage = si.dwPageSize;
    if (m_cbPage < c1MB) // don't bother MEM_COMMITting less than this amount
    {
        m_cbPage = c1MB;
    }

    if (!VirtualAlloc(m_pSharedMemBaseAddr, m_cbPage, MEM_COMMIT, PAGE_READWRITE))
    {
        FreePrivateMemory();
        return FALSE;
    }
    m_cbCommitted = m_cbPage;

    memset(m_pSharedMemBaseAddr, 0, sizeof(SMHeader)); // so that InitializeSMHeader() sets m_fInitializeShared to TRUE
    return TRUE;
}

BOOL VirtualAllocator::Commit(Size_t cb)
{
    if (cb <= m_cbCommitted)
    {
        return TRUE;
    }
    Size_t cbCommitted = (cb + (m_cbPage - 1)) / m_cbPage * m_cbPage;
    if (cbCommitted > m_pHeader->cbReserved)
    {
        return FALSE;
    }
    if (!VirtualAlloc(m_pSharedMemBaseAddr, cbCommitted, MEM_COMMIT, PAGE_READWRITE))
    {
        return FALSE;
    }
    m_cbCommitted = cbCommitted;
    return TRUE;
}

void VirtualAllocator::FreePrivateMemory()
{
    if (m_pSharedMemBaseAddr)
    {
        VirtualFree(m_pSharedMemBaseAddr, 0, MEM_RELEASE);
        m_pSharedMemBaseAddr = NULL;
    }
}
#endif

void VirtualAllocator::Free()
{
    if (m_fSharedMemory)
    {
        FreeSharedMemory();
    }
    else
    {
        FreePrivateMemory();
    }
}

BOOL VirtualAllocator::Initialize(const ConfigPluginGlobals & config, void *pPlatformMemPool, BOOL fUseSharedMemory)
{
    Size_t cb;
    BOOL fRet = FALSE;

#ifdef __TARGET_APACHE__
    Construct();

    m_pPool = (apr_pool_t *) pPlatformMemPool;
    if (m_pPool == NULL)
    {
        return FALSE;
    }

    m_cbPrivateMemory = config.cbPrivateAllocatorSize * c1MB;
#endif

    m_fSharedMemory = FALSE;
    if (fUseSharedMemory && InitializeSharedMemory(config, &cb))
    {
        m_fSharedMemory = TRUE;
    }
    else if (!InitializePrivateMemory(config, &cb))
    {
#ifdef __TARGET_APACHE__
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, APR_EGENERAL, m_pPool, "Unable to initialize private memory.");
#endif
        goto Exit;
    }

    InitializeSMHeader(cb);

#ifdef __TARGET_IIS_
    GlobalLog("%u ok", m_fSharedMemory ? m_fInitializeShared : 2);
#endif

    fRet = TRUE;
Exit:
    return fRet;
}

#ifdef __TARGET_IIS__

BOOL VirtualAllocator::InitializeChild(void *pPlatformMemPool)
{
    UNREFERENCED_PARAMETER(pPlatformMemPool);
    return TRUE;
}

BOOL VirtualAllocator::ShouldInitializeShared()
{
    return m_fInitializeShared;
}

void* VirtualAllocator::SharedAddr()
{
    return (BYTE*)m_pSharedMemBaseAddr + m_pHeader->cbSharedOffset;
}

///////////////////////////////////////////////////////////////////////

#else
// Shared memory must be set to anonymous to keep the base location
// the same in different processes.
static const char *SHAREDMEMNAME = NULL;
#endif

static const TCHAR *PROCMUTEXNAME = TEXT("ProcLock");

#ifdef __TARGET_APACHE__

VirtualAllocator::VirtualAllocator()
{
    Construct();
}

VirtualAllocator::~VirtualAllocator()
{
    // Free is not called here and must be called explicitly
    // because forked processes can accidentally free
    // the mutex when other processes are still running.
}

void VirtualAllocator::Construct() 
{
    m_pSharedMem = NULL;
    m_pSharedMemBaseAddr = NULL;
    m_pHeader = NULL;
    m_fSharedMemory = TRUE;
}

void VirtualAllocator::FreeSharedMemory()
{
    if (m_pSharedMem != NULL)
    {
        apr_status_t status = apr_shm_destroy(m_pSharedMem);

        m_pSharedMem = NULL;
    }
}

void VirtualAllocator::FreePrivateMemory()
{ // gets freed when the pool dies, so nothing to do here
}

void VirtualAllocator::FreeChild()
{
    if (m_fSharedMemory)
    {
        if (SHAREDMEMNAME != NULL)
        {
            apr_shm_detach(m_pSharedMem);
        }
    }
}

BOOL VirtualAllocator::InitializeSharedMemory(const ConfigPluginGlobals& config, Size_t *pcb)
{
    apr_status_t status;
    Size_t cbReserved;

    if (SHAREDMEMNAME != NULL)
    {
        status = apr_shm_remove(SHAREDMEMNAME, m_pPool);
        if (status != APR_SUCCESS)
        {
            DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to delete existing shared memory");
            return FALSE;
        }
    }

    cbReserved = APR_ALIGN_DEFAULT(config.cbSharedAllocatorSize * c1MB);

    status = apr_shm_create(&m_pSharedMem, cbReserved, SHAREDMEMNAME, m_pPool);
    if (status != APR_SUCCESS)
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to create shared memory");
        return FALSE;
    }
    
    // Get usable size.
    cbReserved = apr_shm_size_get(m_pSharedMem);

    // We keep the base address and check if the address is still the same after
    // for the child process.
    m_pSharedMemBaseAddr = apr_shm_baseaddr_get(m_pSharedMem);
    
    // Assigned the first part of shared memory for book keeping.
    if (sizeof(SMHeader) > cbReserved)
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Shared memory does not meet min size");
        return FALSE;
    }

    *pcb = cbReserved;

    return TRUE;
}

BOOL VirtualAllocator::InitializePrivateMemory()
{
    m_pSharedMemBaseAddr = apr_palloc(m_pPool, m_cbPrivateMemory);
    return m_pSharedMemBaseAddr != NULL;
}

BOOL VirtualAllocator::InitializePrivateMemory(const ConfigPluginGlobals & config, Size_t *pcb)
{
    *pcb = m_cbPrivateMemory;
    return InitializePrivateMemory();
}

BOOL VirtualAllocator::InitializeChild(void *pPlatformMemPool)
{
    // Parent and Child process' pPlatformMemPools are different.
    m_pPool = (apr_pool_t *) pPlatformMemPool;

    if (!m_fSharedMemory)
    { // already inherited private memory block allocated by parent
        return TRUE;
    }

    if (AttachSharedMemory())
    {
        return TRUE;
    }

    m_fSharedMemory = FALSE;
    if (!InitializePrivateMemory())
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, APR_EGENERAL, m_pPool, "Unable to initialize private memory.");
        return FALSE;
    }

    InitializeSMHeader(m_cbPrivateMemory);

    return TRUE;
}

BOOL VirtualAllocator::AttachSharedMemory()
{
    apr_status_t status;
    
    if (m_pSharedMem == NULL)
    {
        return FALSE;
    }

    if (SHAREDMEMNAME != NULL)
    {
        // Only attached to named shared memory.
        // NULL shared memory is anonymous and can not be attached.
        status = apr_shm_attach(&m_pSharedMem, SHAREDMEMNAME, m_pPool);
        if (status != APR_SUCCESS)
        {
            DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, status, m_pPool, "Unable to attach to shared memory");
            return FALSE;
        }
    }

    // We do not currently support platforms where the address is different after fork().
    // Several pieces of code would need significant modifications to support that.
    if (m_pSharedMemBaseAddr != apr_shm_baseaddr_get(m_pSharedMem))
    {
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, APR_EGENERAL, m_pPool, "Shared memory has different base address");
        return FALSE;
    }

    // Increment reference count for every child initialized.
    PALInterlocked::Increment((UInt32 *) (&m_pHeader->m_cRef));
    
    return TRUE;
}

#endif

BOOL VirtualAllocator::ShouldFreeShared()
{
    Int32 iRetVal = (Int32) PALInterlocked::Decrement((UInt32 *) (&m_pHeader->m_cRef));
    return 0 >= iRetVal;
}

void VirtualAllocator::InitializeSMHeader(Size_t cb)
{
    m_pHeader = (SMHeader *) m_pSharedMemBaseAddr;
#ifdef __TARGET_IIS__
    // http://msdn.microsoft.com/en-us/library/windows/desktop/aa366537(v=vs.85).aspx
    // "The initial contents of the pages in a file mapping object backed by the operating system paging file are 0 (zero)."
    // http://msdn.microsoft.com/en-us/library/windows/desktop/ms684122(v=vs.85).aspx
    // "The threads of different processes can use these functions if the variable is in shared memory"
    if (1 != (Int32) PALInterlocked::Increment((UInt32 *)  (&m_pHeader->m_cRef)))
    {
        m_fInitializeShared = FALSE;
        return;
    }
    m_fInitializeShared = TRUE;
    m_pHeader->cbSharedOffset = (Size_t)-1; // -1 to make sure we crash if this isn't updated - 0 offset would be valid but point to the wrong thing
#else
    m_pHeader->m_cRef = 0;
#endif
    m_pHeader->cbReserved = cb;
    m_pHeader->cbUsed = sizeof(SMHeader);
    m_pHeader->cbSaveUsed = m_pHeader->cbUsed;
    m_pHeader->uCritSecRoundRobin = 0;
}

unsigned* VirtualAllocator::CritSecRoundRobinAddr()
{
    return &(m_pHeader->uCritSecRoundRobin);
}

void *VirtualAllocator::AllocateShared(Size_t cb)
{
#ifdef __TARGET_IIS__
    m_pHeader->cbSharedOffset = m_pHeader->cbUsed;
#endif
    return Allocate(cb);
}

void *VirtualAllocator::Allocate(Size_t cb)
{
    if (m_pHeader->cbUsed + cb > m_pHeader->cbReserved)
    {
#ifdef __TARGET_APACHE__
        DEBUG_ap_log_perror(APLOG_MARK, APLOG_WARNING, APR_EGENERAL, m_pPool, "Out of shared memory %lu", m_pHeader->cbReserved);
#endif
        return NULL;
    }

#ifdef __TARGET_IIS__
    if (!m_fSharedMemory && !Commit(m_pHeader->cbUsed + cb))
    {
        return NULL;
    }
#endif

    void *pRet = (BYTE *) m_pSharedMemBaseAddr + m_pHeader->cbUsed;
    m_pHeader->cbUsed += cb;
    return pRet;
}

void VirtualAllocator::Bookmark()
{
    m_pHeader->cbSaveUsed = m_pHeader->cbUsed;
}

void VirtualAllocator::Unroll()
{
    m_pHeader->cbUsed = m_pHeader->cbSaveUsed;
}


CritSecManager::CritSecManager()
{
    Construct();
}

CritSecManager::~CritSecManager()
{
    // Free is not called here and must be called explicitly
    // because forked processes can accidentally free the
    // global lock when other processes are still running.
}

void CritSecManager::Construct()
{
    m_rCS = NULL;
    m_cCS = 0;
    m_piNext = NULL;
}

void CritSecManager::Free()
{
    if (m_rCS)
    {
        for (unsigned i = 0; i < m_cCS; i++)
        {
            m_rCS[i].DoDispose();
        }

        delete []m_rCS;
        m_rCS = NULL;
        m_cCS = 0;
    }
}

void CritSecManager::FreeChild()
{
#ifdef __TARGET_APACHE__
    if (m_rCS)
    {
        // Does not call DoDispose() as it should only be disposed by the parent process.
        delete []m_rCS;
        m_rCS = NULL;
        m_cCS = 0;
    }
#endif
}

BOOL CritSecManager::Initialize(unsigned nMaxCritSecs, void *pPlatformMemPool, BOOL fWorkAcrossProcesses)
{
    TCHAR szCritName[MAX_PATH];
    
    Construct();
#ifdef __TARGET_APACHE__
    if (pPlatformMemPool == NULL)
    {
        return FALSE;
    }
#endif
    m_cCS = nMaxCritSecs;
    
    m_rCS = new CRITSEC[m_cCS];
    if (!m_rCS)
    {
        return FALSE;
    }

    for (unsigned i = 0; i < m_cCS; i++)
    {
        GetIPCName(szCritName, ARRAYSIZE(szCritName), PROCMUTEXNAME, i);
        m_rCS[i].DoInit(szCritName, pPlatformMemPool, fWorkAcrossProcesses);
        // not checking m_rCS[i].WorkAcrossProcesses() on the assumption that if there is a problem with CreateMutex(), we would have caught it during the master lock initialization in pluginlib.cpp
    }

    return TRUE;
}

BOOL CritSecManager::InitializeChild(void *pPlatformMemPool)
{
    TCHAR szCritName[MAX_PATH];
    for (unsigned i = 0; i < m_cCS; i++)
    {
        GetIPCName(szCritName, ARRAYSIZE(szCritName), PROCMUTEXNAME, i);
        m_rCS[i].DoInitChild(szCritName, pPlatformMemPool);
    }
    
    return TRUE;
}

void CritSecManager::InitializeCSIndex(unsigned* piNext)
{
    m_piNext = piNext;
}

int CritSecManager::GetCS()
{
    int iRet = *m_piNext;
    *m_piNext = (*m_piNext + 1) % m_cCS; // round robin
    return iRet;
}

ICriticalSection* CritSecManager::CritSecPtr(int i)
{
    return m_rCS + i;
}

