/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#pragma once
#include "basic_types.h"
#include "CritSec.h"
#include "PluginLib.h"

class IUnrollableAllocator : public IAllocator
{
public:
    virtual void Bookmark() = 0;
    virtual void Unroll() = 0;
};

class VirtualAllocator : public IUnrollableAllocator
{
public:
    VirtualAllocator();
    ~VirtualAllocator();

    // Initializer for the parent and child processes
    BOOL Initialize(const ConfigPluginGlobals& config, void *pPlatformMemPool, BOOL fUseSharedMemory);
    BOOL InitializeChild(void *pPlatformMemPool);
#ifdef __TARGET_APACHE__
    BOOL InitializePrivateMemory();
    BOOL AttachSharedMemory();
#endif
    void Free();
    void FreeChild();
    BOOL ShouldFreeShared();
    
    unsigned* CritSecRoundRobinAddr();
#ifdef __TARGET_IIS__
    BOOL ShouldInitializeShared();
    void* SharedAddr();
#endif
    void FreeSharedMemory();
    void FreePrivateMemory();
#ifdef __TARGET_IIS__
    BOOL Commit(Size_t);
#endif

    // These calls are not multi-process or thread-safe.
    void* Allocate(Size_t);
    void* AllocateShared(Size_t);
    void Bookmark();
    void Unroll();
    BOOL WorkAcrossProcesses() { return m_fSharedMemory; }

private:
    BOOL InitializeSharedMemory(const ConfigPluginGlobals & config, Size_t *pcb);
    BOOL InitializePrivateMemory(const ConfigPluginGlobals & config, Size_t *pcb);
    void InitializeSMHeader(Size_t cb);
    struct SMHeader
    {
        Size_t cbReserved;
        Size_t cbUsed;
        Size_t cbSaveUsed;
        unsigned uCritSecRoundRobin;
        volatile Int32 m_cRef;
#ifdef __TARGET_IIS__
        Size_t cbSharedOffset; // offset of PluginShared
#endif
    };

    BOOL   m_fSharedMemory;
#ifdef __TARGET_IIS__
    HANDLE m_hMapping;
    BOOL   m_fInitializeShared; // is this the lucky process to intialize shared memory ?
    Size_t m_cbCommitted;
    Size_t m_cbPage;
#else
    Size_t       m_cbPrivateMemory;
    apr_pool_t  *m_pPool;
    apr_shm_t   *m_pSharedMem; 
#endif
    void        *m_pSharedMemBaseAddr;
    SMHeader    *m_pHeader;

    void Construct();
};


class CritSecManager : public ICritSecDispenser
{
public:
    CritSecManager();
    ~CritSecManager();

    // Initializer for the parent and child processes
    BOOL Initialize(unsigned nMaxCritSecs, void *pPlatformMemPool, BOOL fWorkAcrossProcesses);
    BOOL InitializeChild(void *pPlatformMemPool);
    void InitializeCSIndex(unsigned* piNext);
    void Free();
    void FreeChild();

    int GetCS();
    ICriticalSection* CritSecPtr(int);

private:
    void Construct();
    BOOL Init1(void *pPlatformMemPool);
    void Init2(void *pPlatformMemPool);

    unsigned m_cCS;
    unsigned *m_piNext;
    CRITSEC *m_rCS;
};

ICriticalSection* CritSecPtr(int);

void GlobalLog(const char* fmt, ...);

