/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/

#include "PluginGlobals.h"
#include "PluginLibCommon.h"

const static TCHAR* PLUGINPROCMUTEXNAME = TEXT("GlobalLock");

void PluginGlobals::Free(void *pPlatformMemPool)
{
#ifdef __TARGET_APACHE__
    apr_pool_t *pPool = (apr_pool_t *) pPlatformMemPool;

    if (m_pAllocator->WorkAcrossProcesses())
    {
        // We need to check if one of the child processes have
        // freed the shared memory. Wait a few cycles before giving up.
        const int retry = 3;
        for (int i = 0; i < retry; i++)
        {
            m_pcs->Enter();
            if (m_pShared && m_pShared->IsInit())
            {
                DEBUG_ap_log_perror(APLOG_MARK, APLOG_NOTICE, 0, pPool, "[%d] Shared memory still not freed, waiting...", getpid());
                m_pcs->Leave();

                Sleep(500);

                if (i == retry - 1)
                {
                    DEBUG_ap_log_perror(APLOG_MARK, APLOG_CRIT, APR_EGENERAL, pPool, "Failed waiting for child process to free shared memory.");
                }    

                continue;
            }
            else
            {
                DEBUG_ap_log_perror(APLOG_MARK, APLOG_NOTICE, 0, pPool, "[%d] Shared memory freed by child process.", getpid());
                m_pcs->Leave();
                break;        
            }
        }
    }
#else
    UNREFERENCED_PARAMETER(pPlatformMemPool);
#endif


    if (m_pLockManager != NULL)
    {
        m_pLockManager->Free();
        delete m_pLockManager;
        m_pLockManager = NULL;
    }

    if (m_pcs != NULL)
    {
        m_pcs->DoDispose();
        delete m_pcs;
        m_pcs = NULL;
    }

    if (m_pAllocator != NULL)
    {
        m_pAllocator->Free();
        delete m_pAllocator;
        m_pAllocator = NULL;
    }
}

void PluginGlobals::FreeChild(void *pPlatformMemPool)
{  
    UNREFERENCED_PARAMETER(pPlatformMemPool);

    if (m_pThreadPool != NULL)
    {
        m_pThreadPool->Deinitialize();
        delete m_pThreadPool;
        m_pThreadPool = NULL;
    }

    // m_pShared is freed in the child processes because
    // they have different permissions than the parent
    // process (in Apache).
    // Only the final process should free the shared memory.
    if (m_pShared)
    {
        m_pcs->Enter();
        if (m_pAllocator->ShouldFreeShared())
        {
            m_pShared->Free();
        }
        else
        {
            m_pShared->Free(TRUE);
        }
        m_pShared = NULL;
        m_pcs->Leave();
    }

#if !APR_HAVE_WINDOWS_H
#ifndef _SINGLE_PROC_MODE
    // Linux specific.
    // TODO: Have to verify if this has the same behavior in Apache
    // for windows.
    if (m_pLockManager != NULL)
    {
        m_pLockManager->FreeChild();
        delete m_pLockManager;
        m_pLockManager = NULL;
    }

    if (m_pcs)
    {
        delete m_pcs;
        m_pcs = NULL;
    }

    if (m_pAllocator != NULL)
    {
        m_pAllocator->FreeChild();
        delete m_pAllocator;
        m_pAllocator = NULL;
    }
#endif
#endif
}

BOOL PluginGlobals::Initialize(const ConfigPluginGlobals &config, void *pPlatformMemPool)
{
    BOOL fRet = FALSE;
#ifdef __TARGET_IIS__
    BOOL fHaveLock = FALSE;
#endif

#ifdef __TARGET_IIS__
    DWORD cchUserName = ARRAYSIZE(g_tszUserName);
    if (GetUserName(g_tszUserName, &cchUserName))
    {
        SetBaseIPCName(IPCNameFromAppPoolName(g_tszUserName));
    }
#endif    

    memset(this, 0, sizeof(PluginGlobals));

    TCHAR tszName[MAX_PATH];
    if (!GetIPCName(tszName, ARRAYSIZE(tszName), PLUGINPROCMUTEXNAME, 0)) 
    {
        goto Exit;
    }
    m_pcs = new CRITSEC();
    if (!m_pcs)
    {
        DebugString("cannot allocate critical section");
        goto Exit;
    }
    m_pcs->DoInit(tszName, pPlatformMemPool, TRUE);

    m_fWorkAcrossProcesses = m_pcs->WorkAcrossProcesses();
#ifdef __TARGET_IIS__
    if (m_fWorkAcrossProcesses)
    {
        // Prevent another process from using shared memory before it is initialized
        m_pcs->Enter();
        fHaveLock = TRUE;
    }
#endif

    m_pAllocator = new VirtualAllocator();
    if (m_pAllocator == NULL)
    {
        DebugString("cannot allocate allocator");
        goto Exit;
    }
    if (!m_pAllocator->Initialize(config, pPlatformMemPool, m_fWorkAcrossProcesses))
    {
        DebugString("cannot initialize virtual allocator");
        goto Exit;
    }

    if (!m_pAllocator->WorkAcrossProcesses())
    {
        m_fWorkAcrossProcesses = FALSE;
    }

    m_pLockManager = new CritSecManager();
    if (m_pLockManager == NULL)
    {
        DebugString("cannot allocate critical section manager");
        goto Exit;
    }
    if (!m_pLockManager->Initialize(config.nMaxGlobalLocks, pPlatformMemPool, m_fWorkAcrossProcesses))
    {
        DebugString("cannot initialize critical section manager");
        goto Exit;
    }

#ifdef __TARGET_IIS__
    if (m_pAllocator->ShouldInitializeShared())
#endif
    {
        m_pShared = (PluginShared*)m_pAllocator->AllocateShared(sizeof(PluginShared));
        if (!m_pShared || !m_pShared->Initialize(m_pAllocator))
        {
            m_pShared = NULL;
            goto Exit;
        }
    }
#ifdef __TARGET_IIS__
    else
    {
        m_pShared = (PluginShared*)m_pAllocator->SharedAddr(); // already initialized by another process
    }
#endif

    m_fIsInitialized = TRUE;
    fRet = TRUE;
Exit:
#ifdef __TARGET_IIS__
    if (fHaveLock)
    {
        m_pcs->Leave();
    }
#endif
    if (!fRet)
    {
        Free(pPlatformMemPool);
    }
    return fRet;
}

static void ProcessWorkWrap(void* pv)
{
    ResponseBuffer* p = (ResponseBuffer*)pv;
    p->pSite->ProcessFinishedResponse(p);
}


BOOL PluginGlobals::InitializeChild(void *pPlatformMemPool)
{
    BOOL fRet = FALSE;

    if (m_fIsInitialized == FALSE)
    {
        goto Exit;
    }

    m_fIsInitialized = FALSE;

#if defined(__TARGET_APACHE__) && defined(_DEBUG)
    g_pServerPool = (apr_pool_t * ) pPlatformMemPool;
#endif

#ifndef _SINGLE_PROC_MODE
    if (!m_pAllocator->InitializeChild(pPlatformMemPool))
    {
        DebugString("cannot initialize virtual allocator in child process");
        goto Exit;
    }

    if (!m_pLockManager->InitializeChild(pPlatformMemPool))
    {
        DebugString("cannot initialize critical section manager in child process");
        goto Exit;
    }
#endif

    m_pLockManager->InitializeCSIndex(m_pAllocator->CritSecRoundRobinAddr());

    PALThread::StaticSetCallbackFunction(ProcessWorkWrap);

    // TODO: Have to verify if this has the same behavior in Apache
    // for windows.
    m_pThreadPool = new PALThread();
    if (m_pThreadPool == NULL)
    {
        DebugString("cannot allocate thread pool");
        goto Exit;
    }
    if (m_pThreadPool->Initialize(pPlatformMemPool) == FALSE)
    {
        DebugString("cannot initialize thread pool");
        goto Exit;
    }

#ifndef _SINGLE_PROC_MODE
    TCHAR tszName[MAX_PATH];
    if (!GetIPCName(tszName, ARRAYSIZE(tszName), PLUGINPROCMUTEXNAME, 0)) 
    {
        goto Exit;
    }
    if (m_pcs == NULL)
    {
        goto Exit;
    }
    m_pcs->DoInitChild(tszName, pPlatformMemPool);
#endif

    m_fIsInitialized = TRUE;
    fRet = TRUE;

Exit:
    if (!fRet)
    {
        FreeChild(pPlatformMemPool);
    }
    return fRet;
}

PluginSite* PluginGlobals::GetSite(WSFilter *pFilter)
{
    return m_pShared->GetSite(m_pAllocator, m_pLockManager, pFilter, m_pcs);
}

