/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/

#include "PALString.h"
#include "PlatformSpecificPluginGlobals.h"
#include "PluginLib.h"
#include "PluginSite.h"
#include "PluginGlobals.h"

NotificationCounters g_nc;
PluginGlobals g_globals;

#if defined(__TARGET_APACHE__) && defined(_DEBUG)
apr_pool_t *g_pServerPool = NULL;
#endif

void WaitPendingWork()
{
    while (s_cPendingWork) Sleep(10);
}

void ProcessResponse(WSFilter *pFilter)
{
    g_nc.Inc(5);
    if (!g_globals.IsInitialized())
    {
        goto Exit;
    }

    if (PTR_FFFF == pFilter->GetContext())
    { // we've already considered this response and decided not to process it
        DebugString("context already FFFF");
        goto Exit;
    }

    g_nc.Inc(6);
    PluginSite* pSite;
    if (pFilter->GetContext())  // if FILTER_CONTEXT already has a response instance, use it
    {
        g_nc.Inc(7);
        pSite = ((ResponseBuffer*)(pFilter->GetContext()))->pSite; // shouldn't be NULL because of the way ResponseBuffer is initialized in OnSendResponse()
    }
    else
    {
        g_nc.Inc(8);
        pSite = g_globals.GetSite(pFilter); // creates PluginSite if it doesn't already exist
        if (!pSite)
        { // failed to create/initialize PluginSite
            g_nc.Inc(9);
            pFilter->SetContext(PTR_FFFF);
            goto Exit;
        }

        pSite->OnSendResponse(pFilter); // sets pFilter->Context if successful
        if (!pFilter->GetContext())
        {
            g_nc.Inc(10);
            pFilter->SetContext(PTR_FFFF);
            goto Exit;
        }
    }

    g_nc.Inc(12);
    pSite->OnRawData(pFilter); // loops over data blocks (if any) internally
    g_nc.Inc(13);

    if (pFilter->NoMore())
    {
        g_nc.Inc(14);
        pSite->OnEndRequest(pFilter);
    }

Exit:
    if (pFilter->NoMore())
    {   // http://forums.iis.net/t/1025179.aspx and http://support.microsoft.com/kb/239738 say
        // pFilterContext is maintained over a session, which can have multiple requests.
        pFilter->SetContext(NULL); // clear any PTR_FFFF
    }
}

#ifdef __TARGET_IIS__
void OnSendResponse(PHTTP_FILTER_CONTEXT pfc, HTTP_FILTER_SEND_RESPONSE* pResponse)
{
    WSFilter filter(pfc);
    filter.m_pResponse = pResponse;
    ProcessResponse(&filter);
}

void OnRawData(PHTTP_FILTER_CONTEXT pfc, HTTP_FILTER_RAW_DATA* pData)
{
    WSFilter filter(pfc);
    filter.m_pDataBuffer = pData;
    ProcessResponse(&filter);
}

void OnEndRequest(PHTTP_FILTER_CONTEXT pfc)
{
    WSFilter filter(pfc);
    filter.m_fNoMore = TRUE;
    ProcessResponse(&filter);
}
#endif

BOOL InitializePlugin(const ConfigPluginGlobals &config, void *pPlatformMemPool)
{
    return InitPluginGlobals(config, pPlatformMemPool) && InitPluginGlobalsChild(pPlatformMemPool);
}

void FreePlugin(void *pPlatformMemPool)
{
    FreePluginGlobalsChild(pPlatformMemPool);
    FreePluginGlobals(pPlatformMemPool);
}

BOOL InitPluginGlobals(const ConfigPluginGlobals &config, void *pPlatformMemPool)
{
    return g_globals.Initialize(config, pPlatformMemPool);
}

BOOL InitPluginGlobalsChild(void *pPlatformMemPool)
{
    return g_globals.InitializeChild(pPlatformMemPool);
}

void FreePluginGlobals(void *pPlatformMemPool)
{
    DebugString("TerminateFilter");
    g_globals.Free(pPlatformMemPool);
}

void FreePluginGlobalsChild(void *pPlatformMemPool)
{
    g_globals.FreeChild(pPlatformMemPool);
}

void ConfigPluginGlobals::SetDefault()
{
    cbSharedAllocatorSize = 256;

#ifdef __TARGET_IIS__
    cbPrivateAllocatorSize = 1024; // Windows address space - not committed unless needed
    // Note that document buffers are allocated from separate heap, so the main consumer of this is m_cache / bloomfilter.
    // 1024 might be too taxing on x86 - reduce if so...

    DWORD dw = 0;
    if (PluginRegQueryDWORD(NULL, TEXT("SharedMemorySizeMB"), &dw))
    {
        cbSharedAllocatorSize = dw;
    }
    if (PluginRegQueryDWORD(NULL, TEXT("PrivateMemorySizeMB"), &dw))
    {
        cbPrivateAllocatorSize = dw;
    }
#else
    cbPrivateAllocatorSize = 64;
#endif

    nMaxGlobalLocks = 20;

#ifdef __TARGET_IIS__
#ifdef _X86_
    uSharedMemAddrMB = 1536;
#else
    uSharedMemAddrMB = 2048;
#endif
    if (PluginRegQueryDWORD(NULL, TEXT("SharedMemoryAddrMB"), &dw))
    {
        uSharedMemAddrMB = dw;
    }
#endif    
}