/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/

#include "PluginShared.h"
#include "PALString.h"
#include "SSPHash.h"

extern NotificationCounters g_nc;

static void DeleteSite(PluginSite* pSite)
{
    if (pSite)
    {
        pSite->Close();
    }
}

static void FlushSite(PluginSite* pSite)
{
    if (pSite)
    {
        pSite->Flush();
    }
}

static BOOL DeleteSiteTable(void* pvContext, PluginSite** ppSite)
{
    UNREFERENCED_PARAMETER(pvContext);
    DeleteSite(*ppSite);
    return TRUE;
}

static BOOL FlushSiteTable(void* pvContext, PluginSite** ppSite)
{
    UNREFERENCED_PARAMETER(pvContext);
    FlushSite(*ppSite);
    return TRUE;
}

static BOOL DeleteSiteMapping(void* pvContext, void* pUser)
{
    UNREFERENCED_PARAMETER(pvContext);
    if (pUser != PTR_FFFF)
    {
        DeleteSite((PluginSite*)pUser);
    }
    return TRUE;
}

static BOOL FlushSiteMapping(void* pvContext, void* pUser)
{
    UNREFERENCED_PARAMETER(pvContext);
    if (pUser != PTR_FFFF)
    {
        FlushSite((PluginSite*)pUser);
    }
    return TRUE;
}


void PluginShared::Free(BOOL bFlushOnly)
{
    BOOL (*pfnSiteMapping) (void *, void *) = bFlushOnly ? FlushSiteMapping : DeleteSiteMapping;
    BOOL (*pfnSiteTable) (void *, PluginSite **) = bFlushOnly ? FlushSiteTable : DeleteSiteTable;

    if (m_fUseHostMapping)
    {
        m_hostMapping.Enumerate(NULL, pfnSiteMapping);
    }
    else
    {
        m_sites.Enumerate(NULL, pfnSiteTable);

        if (!bFlushOnly)
        {
            m_sites.Free();
        }
    }

    if (!bFlushOnly)
    {
        m_bInit = FALSE;
    }
}


BOOL PluginShared::GetMappingFile(__out_ecount(cchMappingFile)TCHAR* wszMappingFile, Size_t cchMappingFile)
{
    TCHAR wszDataFolder[MAX_PATH];
    if(!GetDataFolder(wszDataFolder, ARRAYSIZE(wszDataFolder)))
    {
        return FALSE;
    }
    int returnedBufferLength = 0;
    if(SNTPRINTF_S(wszMappingFile, cchMappingFile, cchMappingFile - 1, STRFMT_TLS STRFMT_TDIRSLASH HOST_MAPPING_FILE, wszDataFolder) == -1)
    {
        return FALSE;
    }

    return TRUE;
}

BOOL PluginShared::Initialize(IAllocator* pAlloc)
{

    TCHAR wszMappingFile[MAX_PATH];
    m_bInit = FALSE;

    if(!GetMappingFile(wszMappingFile, ARRAYSIZE(wszMappingFile)))
    {
        return FALSE;
    }

    m_fUseHostMapping = m_hostMapping.Load(pAlloc, wszMappingFile); // BUGBUG: should we fail if the mapping exists but is invalid/corrupt ?
    if (!m_fUseHostMapping)
    {
        if (!m_sites.Initialize(10000, pAlloc))
        {
            DebugString("cannot create site table");
            return FALSE;
        }
    }

    m_bInit = TRUE;
    return TRUE;
}

static BOOL g_fOnlyHostQueried = FALSE;
static TCHAR g_wszOnlyHost[MAX_PATH] = TEXT("");

PluginSite* PluginShared::GetSite(IUnrollableAllocator* pAlloc, ICritSecDispenser* pCSD, WSFilter *pFilter, ICriticalSection* pcs)
{
    PluginSite* pSite = NULL;
    TCHAR wszHost[80], wszPort[PORT_STR_SIZE + 1];
    Size_t nHostLen, nPortLen;
    UInt64 hash;
    g_nc.Inc(20);

    nHostLen = ARRAYSIZE(wszHost);
    if (!pFilter->GetServerName(wszHost, nHostLen))
    {
        g_nc.Inc(21);
        DebugString("cannot get SERVER_NAME");
        return NULL;
    }

#ifdef __TARGET_IIS__
    if (!g_fOnlyHostQueried) // race - don't really care for now ?
    {   // assumption: if reg query fails, it leaves the output string buffer intact

        PluginRegQueryValue(NULL, TEXT("OnlyHostname"), g_wszOnlyHost, ARRAYSIZE(g_wszOnlyHost));
        g_fOnlyHostQueried = TRUE;
    }

    if (g_wszOnlyHost[0] && (0 != PALString::Compare(wszHost, g_wszOnlyHost)))
    {
        g_nc.Inc(22);
        DebugString("host different from specified one");
        return NULL;
    }
#endif

    nPortLen = ARRAYSIZE(wszPort);
    if (!pFilter->GetServerPort(wszPort, nPortLen))
    {
        g_nc.Inc(23);
        DebugString("cannot get SERVER_PORT");
        return NULL;
    }


    TCHAR wszApplPath[MAX_PATH];
    Size_t nLen = ARRAYSIZE(wszApplPath);
    if (!pFilter->GetApplPhysicalPath(wszApplPath, nLen))
    {
        g_nc.Inc(24);
        DebugString("cannot get APPL_PHYSICAL_PATH");
        return FALSE;
    }
    // MSDN says some versions of IIS include trailing slash and some don't
    for (int i = (int) nLen - 1; i >= 0 && (wszApplPath[i] == '\\' || wszApplPath[i] == '/'); i--) 
    {
        wszApplPath[i] = '\0';
    }

    g_nc.Inc(25);
    pcs->Enter();
    g_nc.Inc(26);
    HostMapping::Entry* pme = NULL;
    TCHAR wszHostAndPort[MAX_PATH + PORT_STR_SIZE + 2] = {0};
    if (m_fUseHostMapping)
    {
        g_nc.Inc(30);
        pme = m_hostMapping.Map(wszHost, wszPort, wszApplPath);
        if (pme)
        {
            g_nc.Inc(31);
            if (PTR_FFFF == pme->m_pUser)
            {
                g_nc.Inc(32);
                pSite = NULL;
            }
            else if (!pme->m_pUser)
            {
                g_nc.Inc(33);
                TCHAR wszMappedApplPath[MAX_PATH] = {0};
                if (pme->GetHostAndPort(wszHostAndPort, ARRAYSIZE(wszHostAndPort)) && pme->GetApplPath(wszMappedApplPath, ARRAYSIZE(wszMappedApplPath)))
                {
                    g_nc.Inc(34);
                    pAlloc->Bookmark();
                    pSite = PluginSite::Create(pAlloc, pCSD, wszHostAndPort, wszMappedApplPath);
                    if (pSite)
                    {
                        g_nc.Inc(35);
                        pme->m_pUser = pSite;
                    }
                    else
                    {
                        g_nc.Inc(36);
                        pAlloc->Unroll();
                        pme->m_pUser = PTR_FFFF; // don't try creating it again
                    }
                }
                else
                {
                    g_nc.Inc(37);
                    DebugString("cannot retrieve host/port/wwwroot from mapped entry");
                }
            }
            else
            {
                g_nc.Inc(38);
                pSite = (PluginSite*)(pme->m_pUser);
            }
        }
        else
        {
            g_nc.Inc(39);
            DebugString("host/port/wwwroot not found in the global mapping");
        }
    }
    else
    {
        g_nc.Inc(40);
        PluginSite** ppSite = NULL;
        hash = SSPHash(wszHost, nHostLen * sizeof(TCHAR), SSPHash(wszPort, nPortLen * sizeof(TCHAR)));
        ppSite = m_sites.Get(hash);
        if (ppSite)
        {
            g_nc.Inc(41);
            pSite = *ppSite;
        }
        else
        {
            if (CombineHostAndPort(wszHostAndPort, ARRAYSIZE(wszHostAndPort), wszHost, wszPort))
            {
                g_nc.Inc(42);
                pAlloc->Bookmark();
                pSite = PluginSite::Create(pAlloc, pCSD, wszHostAndPort, wszApplPath);
                // even if pSite is NULL, add it so we don't try creating it again
                if (!m_sites.Add(hash, pSite) && pSite)
                {
                    g_nc.Inc(43);
                    pSite->Free();
                    pSite = NULL;
                }
                if (!pSite)
                {
                    g_nc.Inc(44);
                    pAlloc->Unroll();
                }
            }
            else
            {
                g_nc.Inc(43);
                DebugString("cannot attach colon to port");
            }
        }
    }
    pcs->Leave();

    if (pSite) pSite->Tick();
    return pSite;
}

