/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/

#include "PluginSite.h"
#include "PALFile.h"
#include "PALString.h"
#include "FileAccess.h"
#include "DeltaFile.h"
#include "DocSignature.h"
#include "SSPHash.h"
#include "PluginGlobals.h"
#include "PlatformSpecificPluginGlobals.h"

extern NotificationCounters g_nc;
extern PluginGlobals g_globals;

ICriticalSection* CritSecPtr(int i)
{
    return g_globals.CritSecPtr(i);
}

void PluginLogDelegate::PluginLog(const char* szFormat, ...)
{
    va_list args;
    va_start(args, szFormat);
    m_pSite->PluginLogV(szFormat, args);
    va_end(args);
}

void PluginSite::Zero()
{
    m_wszLogFile[0] = 0;
    m_lHeapUsedKB = 0;
    m_tLastCacheSave = 0;
    m_tLastCountersSave = 0;
#if APR_HAVE_WINDOWS_H
    m_hLockFile = INVALID_HANDLE_VALUE;
#endif
}

void PluginSite::Free()
{
#if APR_HAVE_WINDOWS_H
    if (INVALID_HANDLE_VALUE != m_hLockFile) { CloseHandle(m_hLockFile); m_hLockFile = INVALID_HANDLE_VALUE; }
#endif
}

void PluginSite::Tick()
{
    m_outFile.Tick();
    ExecuteIfTime(PluginTime(), &m_tLastCountersSave, m_config.uCountersSavePeriodSec, &PluginSite::SaveCountersWrap);
}

void PluginSite::PluginLog(const char* szFormat, ...)
{
    va_list args;
    va_start(args, szFormat);
    PluginLogV(szFormat, args);
    va_end(args);
}

void PluginSite::PluginLogV(const char* szFormat, va_list args)
{
    char s[1024];
    VSPRINTF_S(s, ARRAYSIZE(s), szFormat, args);
    LogMessage(m_wszLogFile, s);
}

void PluginSite::LoadCache()
{
    TCHAR wszCacheFile[MAX_PATH];
    if (!GetCacheFile(m_wszSubfolder, wszCacheFile, ARRAYSIZE(wszCacheFile)))
    {
        return;
    }

    FILE* fp = NULL;
    if ((0 != PALFile::Open(&fp, wszCacheFile, TEXT("rb"))) || !fp)
    {
        return;
    }

    while (1)
    {
        char buf[CACHE_REC_BYTES * 1024];
        Size_t cb = fread(buf, 1, sizeof(buf), fp);

        const char* p = buf;
        for (unsigned u = 0; u < cb / CACHE_REC_BYTES; u++)
        {
            Hash       hash = *((Hash*)p); p += sizeof(hash);
            Signature   sig = *((Signature*)p);  p += sizeof(sig);
            FileWriteTime t = *((FileWriteTime*)p);  p += sizeof(t);
            VisitCount    c = *((VisitCount*)p);  p += sizeof(c);

            CacheEntry* e = m_cache.Access(hash);

            e->sig = sig;
            e->lastWrite = t;
            e->visitCount = c;
        }

        if (cb < sizeof(buf))
        {
            break;
        }
    }
    fclose(fp);
}

void SaveCacheCallback(void* pv, Hash hash, CacheEntry* p)
{
    CacheSaveContext* ctx = (CacheSaveContext*)pv;
    if (ctx->cb + CACHE_REC_BYTES > sizeof(ctx->buf))
    {
        fwrite(ctx->buf, 1, ctx->cb, ctx->fp);
        ctx->cb = 0;
    }

    *((Hash*)         (ctx->buf + ctx->cb)) = hash;          ctx->cb += sizeof(hash);
    *((Signature*)    (ctx->buf + ctx->cb)) = p->sig;        ctx->cb += sizeof(p->sig);
    *((FileWriteTime*)(ctx->buf + ctx->cb)) = p->lastWrite;  ctx->cb += sizeof(p->lastWrite);
    *((VisitCount*)   (ctx->buf + ctx->cb)) = p->visitCount; ctx->cb += sizeof(p->visitCount);
}

void PluginSite::SaveCache()
{
    AutoCriticalSection a(CritSecPtr(m_icsCache));

    TCHAR wszCacheFile[MAX_PATH];
    if (!GetCacheFile(m_wszSubfolder, wszCacheFile, ARRAYSIZE(wszCacheFile)))
    {
        return;
    }

    TCHAR wszCacheTemp[MAX_PATH];
    if (!GetFullPath(m_wszSubfolder, TEXT("cache.tmp"), wszCacheTemp, ARRAYSIZE(wszCacheTemp)))
    {
        return;
    }

    CacheSaveContext ctx;
    ctx.cb = 0;
    ctx.fp = NULL;
    if ((0 != PALFile::Open(&ctx.fp, wszCacheTemp, TEXT("wb"))) || !ctx.fp)
    {
        return;
    }

    m_cache.Enumerate(&ctx, SaveCacheCallback);
    fwrite(ctx.buf, 1, ctx.cb, ctx.fp);
    fclose(ctx.fp);

    PALFile::Move(wszCacheTemp, wszCacheFile);
}

PluginSite* PluginSite::Create(IAllocator* pAlloc, ICritSecDispenser* pCSD, const TCHAR* wszHostAndPort, const TCHAR* wszApplPath)
{
    PluginSite* pSite = (PluginSite*)(pAlloc->Allocate(sizeof(PluginSite)));
    if (!pSite)
        return NULL;

    pSite->Zero();
    if (!pSite->Initialize(pAlloc, pCSD, wszHostAndPort, wszApplPath))
    {
        pSite->Free();
        return NULL;
    }
    return pSite;
}

#if APR_HAVE_WINDOWS_H
static BOOL CreateFolderWithRestrictedPermissions(const TCHAR* wszFolderToCreate, IPluginLog* pLog)
{ // create a folder with ACL Allowed <user associated with this process> + "Administrators" entries
    BOOL fRet = FALSE;
    DWORD dw;
    HANDLE hToken = NULL;
    TOKEN_USER* pUser = NULL;
    ACL* pNewDacl = NULL;
    void* pSidSystem = NULL, *pSidAdministrators = NULL;
    DWORD cbNewACL;
    SECURITY_DESCRIPTOR sd;
    SECURITY_ATTRIBUTES attr;
    DWORD mask;

    //
    // Figure out current user's SID
    //
    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken) || !hToken) CALLEE_ERROR;
    if (GetTokenInformation(hToken, TokenUser, NULL, 0, &dw)) CALLEE_ERROR;
    if (ERROR_INSUFFICIENT_BUFFER != GetLastError()) CALLEE_ERROR;
    pUser = (TOKEN_USER*)malloc(dw); if (!pUser) CALLEE_ERROR;
    if (!GetTokenInformation(hToken, TokenUser, pUser, dw, &dw)) CALLEE_ERROR;
    CloseHandle(hToken);  hToken = NULL;

    if (!GetAccountSID("Administrators", &pSidAdministrators)) CALLEE_ERROR;
    if (!GetAccountSID("SYSTEM", &pSidSystem)) CALLEE_ERROR;

    //
    // Create new folder's access control list based on the template folder's
    //
    cbNewACL = 2048;
    pNewDacl = (ACL*)malloc(cbNewACL);  if (!pNewDacl) CALLEE_ERROR;
    if (!InitializeAcl(pNewDacl, cbNewACL, ACL_REVISION)) CALLEE_ERROR;
    mask = STANDARD_RIGHTS_ALL | SPECIFIC_RIGHTS_ALL;
    if (!AddAccessAllowedAceEx(pNewDacl, ACL_REVISION, CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE, mask, pUser->User.Sid)) CALLEE_ERROR;
    if (!AddAccessAllowedAceEx(pNewDacl, ACL_REVISION, CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE, mask, pSidAdministrators)) CALLEE_ERROR;
    if (!AddAccessAllowedAceEx(pNewDacl, ACL_REVISION, CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE, mask, pSidSystem)) CALLEE_ERROR;

    free(pUser);
    pUser = NULL;

    //
    // Create new folder using the access control list we just created
    //
    if (!InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION)) CALLEE_ERROR;
    if (!SetSecurityDescriptorDacl(&sd, TRUE, pNewDacl, FALSE)) CALLEE_ERROR;

    attr.nLength = sizeof(SECURITY_ATTRIBUTES);
    attr.lpSecurityDescriptor = &sd;
    attr.bInheritHandle = FALSE;
    if (!CreateDirectory(wszFolderToCreate, &attr)) CALLEE_ERROR;

    fRet = TRUE;

Exit:
    if (hToken) CloseHandle(hToken);
    if (pUser) free(pUser);
    if (pSidSystem) free(pSidSystem);
    if (pSidAdministrators) free(pSidAdministrators);
    if (pNewDacl) free(pNewDacl);
    return fRet;
}
#else
static BOOL CreateFolderWithRestrictedPermissions(const TCHAR* wszFolderToCreate, IPluginLog* pLog)
{
    return PALDirectory::CreateDirectory(wszFolderToCreate);
}
#endif

static BOOL IsMasterOrSleep(unsigned iInstance)
{
    if (0 == iInstance)
    {
        return TRUE;
    }
    else
    { // instance 0 must be running (has the "lock" file held in folder 0 - otherwise we'd be instance 0) - unless it crashed
        Sleep(5000); // give instance 0 a chance to write the file before we try reading after this function returns
        return FALSE;
    }
}

BOOL PluginSite::Initialize(IAllocator* pAlloc, ICritSecDispenser* pCSD, const TCHAR* wszHostAndPort, const TCHAR* wszApplPath)
{
    // this is just in case a log sneaks in before m_wszLogFile is properly initialized below based on m_wszSubfolder
    if (PALString::Copy(m_wszLogFile, ARRAYSIZE(m_wszLogFile), TEXT("log.txt")) != 0)
    {
        DebugString("cannot copy log string");
        return FALSE;
    }

    TCHAR wszHostPath[HOST_PATH_LENGTH];
    if (0 >= SNTPRINTF_S(wszHostPath, ARRAYSIZE(wszHostPath), _TRUNCATE, TEXT("http://") STRFMT_TLS, wszHostAndPort))
    {
        DebugString("cannot attach http:// to host");
        return FALSE;
    }

    if (PALString::Convert(m_szHostPath, sizeof(m_szHostPath), wszHostPath, TSTRLEN(wszHostPath)) != 0)
    {
        DebugString("cannot convert host path to UTF8");
        return FALSE;
    }

    TCHAR* wszHost = NULL;
    TCHAR wszTmpHost[MAX_PATH];
    TCHAR wszPort[PORT_STR_SIZE + 1];

    const TCHAR* portPos = TCSSTR(wszHostAndPort, TEXT(":"));
    if(portPos != NULL)
    {
        memset(wszTmpHost, 0, sizeof(wszTmpHost));
        if(PALString::Copy(wszTmpHost, ARRAYSIZE(wszTmpHost), wszHostAndPort, portPos - wszHostAndPort) != 0)
        {
            return FALSE;
        }

        wszHost = wszTmpHost;
        portPos++;
        if(PALString::Copy(wszPort, ARRAYSIZE(wszPort), portPos) != 0)
        {
            return FALSE;
        }
    }
    else
    {
        wszHost = const_cast<TCHAR*>(wszHostAndPort);
        if(PALString::Copy(wszPort, ARRAYSIZE(wszPort), TEXT("80")) != 0)
        {
            return FALSE;
        }
    }

    TCHAR wszFolderPrefix[MAX_PATH];
    if (!GetSiteFolderPrefixByHostScheme(wszHost, wszPort, wszFolderPrefix, ARRAYSIZE(wszFolderPrefix)))
    {
        DebugString("cannot generate working subfolder");
        return FALSE;
    }

#define MAX_INSTANCES 1000 // TODO ? read this from registry
    unsigned iInstance;
    for (iInstance = 0; iInstance < MAX_INSTANCES; iInstance++)
    {
        if (!GetDataFolderWithIndex(wszFolderPrefix, iInstance, m_wszSubfolder, ARRAYSIZE(m_wszSubfolder)))
        {
            DebugString("cannot attach plugin instance index");
            return FALSE;
        }

        //PluginLog(m_wszSubfolder);
        if (!PALDirectory::PathExists(m_wszSubfolder)) // existence test
        {
            PluginLogDelegate logger(this);
            if (!CreateFolderWithRestrictedPermissions(m_wszSubfolder, &logger))
            {
                DebugString("Cannot create subfolder");
                return FALSE;
            }
        }

#if APR_HAVE_WINDOWS_H
        TCHAR wszLockFile[MAX_PATH];
        if (!GetFullPath(m_wszSubfolder, TEXT("lock"), wszLockFile, ARRAYSIZE(wszLockFile)))
        {
            DebugString("cannot set subfolder log file name");
            return FALSE;
        }

        m_hLockFile = CreateFile(wszLockFile,
            GENERIC_WRITE,
            FILE_SHARE_READ,
            NULL,
            CREATE_ALWAYS,
            FILE_ATTRIBUTE_NORMAL,
            NULL);
        if (INVALID_HANDLE_VALUE == m_hLockFile)
        {
            if (ERROR_SHARING_VIOLATION == GetLastError())
            {
                continue; // try n+1
            }

            PluginLog("cannot create lock file " STRFMT_TS, wszLockFile);
            return FALSE;
        }

        // write host:port into the lock file for diagnosability - not fatal if this write fails
        char szLockString[1024];
        if (0 < SPRINTF_S(szLockString, ARRAYSIZE(szLockString), "%s %u " STRFMT_TS, m_szHostPath, GetCurrentProcessId(), g_tszUserName))
        {
            DWORD dwWritten = 0;
            WriteFile(m_hLockFile, szLockString, (DWORD)strlen(szLockString), &dwWritten, NULL);
        }
#endif
        break;
    }

    if (iInstance >= MAX_INSTANCES)
    {
        PluginLog("%s cannot work - another plugin instance is already using all available indexes for data folder " STRFMT_TS, m_szHostPath, m_wszSubfolder);
        return FALSE;
    }

    if (!GetFullPath(m_wszSubfolder, TEXT("log.txt"), m_wszLogFile, ARRAYSIZE(m_wszLogFile)))
    {
        DebugString("cannot set subfolder log file name");
        return FALSE;
    }

    if (!GetPluginCountersFile(m_wszSubfolder, m_wszCountersFile, ARRAYSIZE(m_wszCountersFile)))
    {
        DebugString("cannot set counters file name");
        return FALSE;
    }

    PluginLog("plugin initialization started");

    TCHAR wszConfigFolder[MAX_PATH];
    if (!GetDataFolderWithIndex(wszFolderPrefix, 0, wszConfigFolder, ARRAYSIZE(wszConfigFolder)))
    {
        DebugString("cannot attach plugin instance index 0 for config folder");
        return FALSE;
    }

    TCHAR wszConfigFile[MAX_PATH];
    if (!GetConfigFile(wszConfigFolder, wszConfigFile, ARRAYSIZE(wszConfigFile)))
    {
        DebugString("Cannot get config file");
        return FALSE;
    }
    //PluginLog(szConfigFile);

    char szApplPath[1024]; // MAX_PATH * ~4        
    if (PALString::Convert(szApplPath, ARRAYSIZE(szApplPath), wszApplPath, TSTRLEN(wszApplPath)) != 0)
    {
        DebugString("cannot convert APPL_PATH to UTF8");
        return FALSE;
    }

    m_config.Initialize();
    if (!PALDirectory::PathExists(wszConfigFile))
    {
        if (IsMasterOrSleep(iInstance))
        {
            PluginLog("writing config file");

            FILE* fp;
            if (0 != PALFile::Open(&fp, wszConfigFile, TEXT("at")))
            {
                DebugString("cannot write config file");
                return FALSE;
            }

            fprintf(fp, "%s=%s\n", c_szwwwroot, szApplPath);
            fprintf(fp, "%s=%s\n", c_szwwwhost, m_szHostPath);
            WriteDefaultConfig(fp);
            fclose(fp);
        }
    }
    else
    {
        for (unsigned u = 0; ;u++)
        {
            if (ReadSSPConfig(wszConfigFile, &m_config) && (0 == PALString::Compare(wszApplPath, m_config.wszwwwroot)))
            {
                break; // success, wwwroot setting is correct
            }
            if (u > 3)
            {
                PluginLog("unable to read config file (or wwwroot is incorrect)");
                return FALSE;
            }
            if (IsMasterOrSleep(iInstance))
            {
                PluginLog("wwwroot changed to %s - updating config file", szApplPath);
                if (!UpdateSSPConfig(wszConfigFile, c_szwwwroot, szApplPath))
                {
                    PluginLog("unable to update config file");
                    return FALSE;
                }
            }
        }
    }

    TCHAR wszUserAgentsFile[MAX_PATH];
    if (!GetFullPath(wszConfigFolder, TEXT("UserAgents.tsv"), wszUserAgentsFile, ARRAYSIZE(wszUserAgentsFile)))
    {
        PluginLog("Cannot get user agents file");
        return FALSE;
    }

    if (!PALDirectory::PathExists(wszUserAgentsFile))
    {
        if (IsMasterOrSleep(iInstance))
        {
            PluginLog("writing User Agents file");

            FILE* fp;
            if (0 != PALFile::Open(&fp, wszUserAgentsFile, TEXT("at")))
            {
                DebugString("cannot write config file");
                return FALSE;
            }

            fprintf(fp, "# This file configures which user agents are treated as human/browser (1) vs bot (0).\n");
            fprintf(fp, "# Document changes are tracked separately for humans vs bots in case the server\n");
            fprintf(fp, "# returns different content for bots.\n");
            fprintf(fp, "%s\t%u\n", "bot", 0);
            fprintf(fp, "%s\t%u\n", "MSIE", 1);
            fprintf(fp, "%s\t%u\n", "Firefox", 1);
            fprintf(fp, "%s\t%u\n", "Chrome", 1);
            fprintf(fp, "%s\t%u\n", "Safari", 1);
            fprintf(fp, "%s\t%u\n", "Opera", 1);
            fclose(fp);
        }
    }

    TCHAR wszRobotsPath[MAX_PATH], wszBlacklistPath[MAX_PATH], wszNormalizerPath[MAX_PATH];
    Size_t cbTotal;
    Size_t cbBloomFilters;
    void* pCacheMem;
    PluginLogDelegate logger(this);

    if (!m_userAgents.Load(pAlloc, wszUserAgentsFile, &logger))
    {
        PluginLog("cannot load User Agents file");
        goto Error;
    }

    m_outFile.Initialize(pAlloc, pCSD, m_wszSubfolder, m_config.uMaxSecondsPerFile, &m_pluginCounters, &m_diskSpaceTracker);

    // Using "robots.txt" here instead of m_config.wszRobotsTxtFileName means we end up applying Disallow rules from both robots files
    if (!GetFullPath(wszApplPath, TEXT("robots.txt"), wszRobotsPath, ARRAYSIZE(wszRobotsPath)) ||
        !GetFullPath(wszConfigFolder, BLACKLIST_FILE, wszBlacklistPath, ARRAYSIZE(wszBlacklistPath)) ||
        !GetFullPath(wszConfigFolder, TEXT("normalization.txt"), wszNormalizerPath, ARRAYSIZE(wszNormalizerPath)))
    {
        PluginLog("cannot compute file paths");
        goto Error;
    }

    if (!PALDirectory::PathExists(wszBlacklistPath))
    {
        if (IsMasterOrSleep(iInstance))
        {
            WriteBlacklistFile(wszBlacklistPath);
        }
    }
    if (0 == iInstance)
    {
        UpdateBlacklistFile(wszBlacklistPath);
    }
    if (!m_robots.Load(pAlloc, wszRobotsPath, wszBlacklistPath))
    {
        PluginLog("cannot initialize robots");
        goto Error;
    }

    if (!PALDirectory::PathExists(wszNormalizerPath))
    {
        if (IsMasterOrSleep(iInstance))
        {
            WriteNormalizerFile(wszNormalizerPath);
        }
    }
    if (!m_normalizer.Load(pAlloc, wszNormalizerPath, 10000))
    {
        PluginLog("cannot initialize normalizer");
        goto Error;
    }

    if (!m_diskSpaceTracker.DoInit(m_wszSubfolder, m_config.uDiskUsageCheckPeriodSec, m_config.uDiskMinFreeMB, m_config.uDiskMaxFreeMB))
    {
        PluginLog("Cannot initialize disk space tracker");
        goto Error;
    }

    if (m_config.uMaxMemoryCacheMB > 2047)
    {
        m_config.uMaxMemoryCacheMB = 2047;
    }
    cbTotal = m_config.uMaxMemoryCacheMB * (Size_t)1048576;
    cbBloomFilters = cbTotal / 8;
    m_cbCache = cbTotal - cbBloomFilters;

    if (!m_doubleBloomFilter.Initialize(m_config.uSignatureComputationPeriodSec, cbBloomFilters / 2, pAlloc))
    {
        PluginLog("cannot allocate bloomfilter");
        goto Error;
    }

    pCacheMem = pAlloc->Allocate(m_cbCache);
    if (!pCacheMem)
    {
        PluginLog("failed to allocate buffer for cache");
        goto Error;
    }
    if (!m_cache.Initialize(pCacheMem, (unsigned)m_cbCache))
    {
        PluginLog("failed to initialize cache");
        goto Error;
    }
    m_icsCache = pCSD->GetCS();

    LoadCache();

    PluginLog("plugin initialized");
    return TRUE;

Error:
    return FALSE;
}

void* PluginSite::HeapAllocWrap(Size_t cb)
{
    UInt32 cKB = (LONG)(cb / 1024);
    if (m_lHeapUsedKB + cKB > m_config.uMaxHeapMB * 1024)
    {
        m_pluginCounters.cHeapOverLimit.Increment();
        DebugString("Heap over limit");
        return NULL;
    }

    void* pv = malloc(cb);
    if (pv)
    {
        PALInterlocked::Add(&m_lHeapUsedKB, cKB);
    }
    return pv;
}

void PluginSite::HeapFreeWrap(void* pv, Size_t cb)
{
    free(pv);
    PALInterlocked::Add(&m_lHeapUsedKB, -(LONG)(cb/1024));
}

void PluginSite::WriteEntry(unsigned flags, Signature sig, UInt16 visitCount, const char* pbUrl, Size_t cbUrl, FileWriter* pFile)
{
    char buf[2048]; // if you increase this, increase it in LoadDeltaFile() too
    Size_t cbBuf = sizeof(buf);
    if (!GenerateDeltaFileEntry(flags, sig, visitCount, pbUrl, cbUrl, buf, &cbBuf))
    {
        m_pluginCounters.cDroppedTooLong.Increment();
        return;
    }

    pFile->Write(buf, cbBuf);
}

BOOL PluginSite::CacheLookup(Hash hash, Signature sig, UInt16 tNow, UInt16* pVisitCount, BOOL *pfWrite)
{
    // Lookup URL from cache and decide if the URL should be written to the delta entry
    // 1. If URL is not cached and it has 404 status, it is not considered for cache nor write.
    // 2. If URL is cached with 200 status and it is now 404, the cache entry is flagged to be considered for 404 list with max number to signature.

    *pfWrite = FALSE;

    BOOL fDeadLink = FALSE;
    if(sig == DEADLINK_SIGNATURE)
    {
        fDeadLink = TRUE;
    }

    // if fDeadLink is TRUE, a new entry will be created if key doesn't exist. 
    // Otherwise, Access() returns NULL
    CacheEntry* p = m_cache.Access(hash, !fDeadLink);

    if(!p && fDeadLink)
    {
        // If cache entry doesn't already exist and HTTP status is 404, cut it short and return with no write flag.
        *pfWrite = FALSE;
        return FALSE;
    }

    if((p->sig == DEADLINK_SIGNATURE) && fDeadLink)
    {
        // If the cache entry already indicates the deadlink and the new URL is a deadlink, do nothing.
        *pfWrite = FALSE;
        return FALSE;
    }

    p->visitCount++;
    BOOL fRet = FALSE;
    if (0 == p->sig)
    {
        m_pluginCounters.cWriteNotInCache.Increment();
        p->sig = sig;
        *pfWrite = TRUE;
    }
    else if (p->sig != sig)
    {
        m_pluginCounters.cWriteSigChanged.Increment();
        p->sig = sig;
        *pfWrite = TRUE;
    }
    else if ((unsigned)(tNow - p->lastWrite) * SEC_PER_T32 > m_config.uDeltaEntryPeriodSec)
    {
        m_pluginCounters.cWriteForcedByPeriod.Increment();
        *pfWrite = TRUE;
    }
    else if (p->visitCount > 65000)
    {
        m_pluginCounters.cWriteForcedByVisitCount.Increment();
        *pfWrite = TRUE;
    }

    if (*pfWrite)
    {
        *pVisitCount = p->visitCount;
        p->visitCount = 0;
        p->lastWrite = tNow;
    }

    return TRUE;
}

void PluginSite::SaveCountersWrap()
{
    AddCountersToFile(m_wszCountersFile, &m_pluginCounters);
    ResetCounters(&m_pluginCounters);
}

void PluginSite::ExecuteIfTime(time_t tNow, volatile UInt32 *pt, unsigned uIntervalSec, PFNVOIDVOID pfn)
{
    if (!uIntervalSec) // uninitialized
    {
        return;
    }

    UInt32 lNow = (LONG)tNow;
    if (0 == PALInterlocked::CompareSet(pt, lNow, 0))
    {
        return;
    }

    UInt32 l = *pt;
    if ((unsigned)(lNow - l) > uIntervalSec)
    {
        if (l == PALInterlocked::CompareSet(pt, lNow, l))
        {
            (this->*pfn)();
        }
    }
}

BOOL PluginSite::StripHostAndPort(const char** ppbUrl, Size_t* pcbUrl)
{
    if ('/' == **ppbUrl)
    { // URL already has no host
        return TRUE;
    }

    Size_t cbOurHost = strlen(m_szHostPath);
    if (*pcbUrl <= cbOurHost) // "<=" because we want at least '/' after hostname to have a proper non-0-length URL path
    {
        return FALSE;
    }

    // It's possible pbUrl has ":80" after hostname, but we carefully kept ":80" out of m_szHostPath when we generated it
    // So if the canonical URL doesn't start with m_szHostPath, we can safely drop it
    if (memcmp(*ppbUrl, m_szHostPath, cbOurHost))
    {
        return FALSE;
    }

    (*ppbUrl) += cbOurHost;
    (*pcbUrl) -= cbOurHost;

    if ('/' == **ppbUrl)
    {
        return TRUE;
    }
    if ((*pcbUrl >= 4) && !memcmp(*ppbUrl, ":80/", 4))
    { // skip the redundant ":80"
        (*ppbUrl) += 3;
        (*pcbUrl) -= 3;
        return TRUE;
    }

    return FALSE;
}

void * PluginSite::zlib_alloc(void *opaque, UINT size, UINT count)
{
    if (!opaque) return NULL;
    size_t cb = 4 + size * count; // 4 for our own bookkeeping to feed size to HeapFreeWrap()
    if ((cb & 0xffffffff) != cb) return NULL;
    PluginSite* pSite = (PluginSite*)opaque;
    void* pv = pSite->HeapAllocWrap(cb);
    if (!pv) return NULL;
    *((unsigned*)pv) = (unsigned)cb;
    return (BYTE*)pv + 4;
}

void PluginSite::zlib_free(void *opaque, void *data)
{
    if (!opaque) return;
    PluginSite* pSite = (PluginSite*)opaque;
    void* pBase = (BYTE*)data - 4;
    pSite->HeapFreeWrap(pBase, *((unsigned*)pBase));
}

static BOOL FindCompressedHeader(const BYTE* pbIn, Size_t cbIn, Size_t* pcbOffset)
{
    if (cbIn > 4096) // don't search too far
        cbIn = 4096;

    for (Size_t cbOffset = 0; cbOffset + 3 <= cbIn; cbOffset++)
    {
        if (0x1f == pbIn[cbOffset+0] && 0x8b == pbIn[cbOffset+1] && 0x08 == pbIn[cbOffset+2])
        { // gzip header
            *pcbOffset = cbOffset;
            return TRUE;
        }

        if ((0 == cbOffset || '\n' == pbIn[cbOffset-1]) && // assuming compressed data starts on newline
            (0x78 == pbIn[cbOffset+0]) && // deflate header
            (0x9c == pbIn[cbOffset+1] || 0xda == pbIn[cbOffset+1] || 0x01 == pbIn[cbOffset+1] || 0x5e == pbIn[cbOffset+1]))
        {
            *pcbOffset = cbOffset;
            return TRUE;
        }
    }
    return FALSE;
}

BOOL PluginSite::Decompress(const BYTE* pbIn, Size_t cbIn, BYTE** ppbOut, Size_t* pcbOut, Size_t* pcbOutAlloc)
{
    BOOL fRet = FALSE;
    BYTE* pbOut = NULL;
    Size_t cbInOffset = 0, cbOut = 0, cbOutAlloc = 0;
    z_stream zstream;
    BOOL fInflateInitialized = FALSE;

    if (!FindCompressedHeader(pbIn, cbIn, &cbInOffset))
    {
        m_pluginCounters.cDecompressionHeaderFailure.Increment();
        goto Exit;
    }

    zstream.opaque = this;
    zstream.zalloc = (alloc_func)zlib_alloc;
    zstream.zfree = (free_func)zlib_free;

    if (Z_OK != inflateInit2(&zstream, 32+MAX_WBITS)) // 32 means either gzip or deflate header
    {
        m_pluginCounters.cDecompressionZLibInitFailure.Increment();
        goto Exit;
    }
    fInflateInitialized = TRUE;

    zstream.next_in = (Bytef *)(pbIn + cbInOffset);
    zstream.avail_in = (unsigned)(cbIn - cbInOffset);

    cbOutAlloc = cbIn * 10;
    pbOut = (BYTE*)HeapAllocWrap(cbOutAlloc);
    if (!pbOut)
    {
        m_pluginCounters.cDecompressionHeapFailure.Increment();
        goto Exit;
    }

    cbOut = 0;
    do
    {
        zstream.next_out = (Bytef *)(pbOut + cbOut);
        zstream.avail_out = (unsigned)(cbOutAlloc - cbOut);

        int nRet = inflate(&zstream, Z_FINISH);
        cbOut = cbOutAlloc - zstream.avail_out;
        if (nRet == Z_BUF_ERROR && zstream.avail_out == 0)
        {
            Size_t cbNewAlloc = cbOutAlloc * 2;
            if (cbNewAlloc > m_config.uMaxDocumentSizeBytes)
                cbNewAlloc = m_config.uMaxDocumentSizeBytes;

            if (cbNewAlloc <= cbOutAlloc)
            {
                m_pluginCounters.cDocumentTooBig.Increment();
                goto Exit;
            }

            BYTE* pbNew = (BYTE*)HeapAllocWrap(cbNewAlloc);
            if (!pbNew)
            {
                m_pluginCounters.cDecompressionHeapFailure.Increment();
                goto Exit;
            }
            memcpy(pbNew, pbOut, cbOutAlloc);
            HeapFreeWrap(pbOut, cbOutAlloc);
            pbOut = pbNew;
            cbOutAlloc = cbNewAlloc;
            continue;
        }
        if (nRet != Z_STREAM_END)
        {
            if (Z_MEM_ERROR == nRet)
            {
                m_pluginCounters.cDecompressionHeapFailure.Increment();
            }
            else
            {
                m_pluginCounters.cDecompressionZLibFailure.Increment();
            }

            goto Exit;
        }
        break;
    } while (1);
    fRet = TRUE;

Exit:
    if (fInflateInitialized)
    {
        inflateEnd(&zstream);
    }

    if (fRet)
    {
        //PluginLog("successfully decompressed header 0x%02X at offset %u", pbIn[cbInOffset], cbInOffset);
        m_pluginCounters.cDecompressionSuccess.Increment();
        *ppbOut = pbOut;
        *pcbOut = cbOut;
        *pcbOutAlloc = cbOutAlloc;
    }
    else if (pbOut)
    {
        HeapFreeWrap(pbOut, cbOutAlloc);
    }
    return fRet;
}

void PluginSite::ProcessFinishedResponse(ResponseBuffer* p)
{
    g_nc.Inc(90);
    DebugString("ProcessFinishedResponse()");
    m_pluginCounters.cWorkDispatched.Increment();
    if (p->cbUsed < sizeof(*p) + p->cbUrl)
    {
        g_nc.Inc(91);
        m_pluginCounters.cWorkInvalid.Increment();
        return;
    }

    const char* pbUrl = (char*)(p+1);
    const char* pbDoc = pbUrl + p->cbUrl;
    Size_t cbDoc = p->cbUsed - (sizeof(*p) + p->cbUrl);
    BOOL fDecompressed = FALSE;
    Size_t cbDecompressedAlloc = 0;
    Size_t cbUrl = 0;
    BOOL fProcess = TRUE;
    //PluginLog("%.*s %u", p->cbUrl, pbUrl, cbDoc);
    if (p->flags & RESPONSE_FLAG_COMPRESSED)
    {
        BYTE* pbDecompressed = NULL;
        Size_t cbDecompressed = 0;
        if (Decompress((const BYTE*)pbDoc, cbDoc, &pbDecompressed, &cbDecompressed, &cbDecompressedAlloc))
        {
            pbDoc = (char*)pbDecompressed;
            cbDoc = cbDecompressed;
            fDecompressed = TRUE;
        }
        else
        {
            goto Exit; // decompression failure counters incremented inside Decompress()
        }
    }

    pbUrl = ExtractCanonicalLink(pbDoc, cbDoc, &cbUrl);
    if (pbUrl && cbUrl)
    {
        if (!StripHostAndPort(&pbUrl, &cbUrl))
        {
            g_nc.Inc(92);
            fProcess = FALSE; // the other option we're discussing is to use the non-canonical URL in this case
        }
        else if (m_doubleBloomFilter.Check(cbUrl, pbUrl))
        { // URL from canonical tag did not undergo the bloomfilter check in OnSendResponse(), so do it now
            g_nc.Inc(93);
            m_pluginCounters.cCanonicalBloomFilterDeduped.Increment();
            fProcess = FALSE;
        }
    }
    else
    { // no canonical tag - use URL from response header
        g_nc.Inc(94);
        pbUrl = (char*)(p+1);
        cbUrl = p->cbUrl;
    }

    if (fProcess)
    {
        g_nc.Inc(95);
        DebugString("ComputeSignature...");
        Hash hash = SSPHash(pbUrl, cbUrl);
        if (p->flags & RESPONSE_FLAG_BOT)
        {
            hash ^= 1;
        }

        Signature sig;

        if(p->uiHttpStatusCode == 404)
        {
            sig = DEADLINK_SIGNATURE;
        }
        else
        {
            sig = ComputeSignature(pbDoc, cbDoc);
        }

        m_pluginCounters.cSignatureComputed.Increment();

        time_t tNow = PluginTime();

        UInt16 visitCount = 0;
        BOOL fWrite = FALSE;
        BOOL fDeadlink = FALSE;

        CritSecPtr(m_icsCache)->Enter();
        BOOL fSuccess = CacheLookup(hash, sig, (UInt16)(tNow / SEC_PER_T32), &visitCount, &fWrite);
        CritSecPtr(m_icsCache)->Leave();

        g_nc.Inc(96);
        if (fSuccess)
        {
            g_nc.Inc(97);
            if (fWrite)
            {
                g_nc.Inc(98);
                m_pluginCounters.cEntriesWritten.Increment();
                WriteEntry((p->flags & RESPONSE_FLAG_BOT) ? DELTA_FLAG_BOT : 0, sig, visitCount, pbUrl, cbUrl, &m_outFile);
            }
            ExecuteIfTime(tNow, &m_tLastCacheSave, m_config.uCacheSavePeriodSec, &PluginSite::SaveCache);
        }
    }
Exit:
    HeapFreeWrap(p, p->cbAlloc);
    if (fDecompressed)
    {
        HeapFreeWrap((void*)pbDoc, cbDecompressedAlloc);
    }

    PALInterlocked::Decrement(&s_cPendingWork);
}

BOOL PluginSite::CheckHttpCode(int nStatus)
{
    if (nStatus < 200 || (nStatus >= 300 && nStatus != 404))
    {
        m_pluginCounters.cBadHttp.Increment();
        if (m_pluginCounters.cBadHttp.m_val < 10)
        {
            PluginLog("BadHttp %u", nStatus);
        }
#if DBG
        char temp[80];
        sprintf_s(temp, ARRAYSIZE(temp), "HTTP %u", nStatus);
        DebugString(temp);
#endif
        return FALSE;
    }

    return TRUE;
}

static BOOL HttpCodeValid(unsigned u)
{
    return u;
}

void PluginSite::UpdateVisitCounters(const char* szUserAgent, BOOL fBot)
{
    // It is an overhead to compare user-agent string twice
    // The reason of doing that m_userAgents are built from config file
    // and config file can be changed by user
    // Here we want to get more deterministic result, thus compare again

    m_pluginCounters.cValidTraffic.Increment();
    if (fBot)
    {
        m_pluginCounters.cBotTraffic.Increment();
        g_nc.Inc(57);
    }

    UserAgentType::Type userAgentType = UserAgentType::GetUserAgentType(szUserAgent);

    switch (userAgentType)
    {
    case UserAgentType::BingBot:
        {
            m_pluginCounters.cBingBotVisitCount.Increment();
            break;
        }
    case UserAgentType::GoogleBot:
        {
            m_pluginCounters.cGoogleBotVisitCount.Increment();
            break;
        }    


    case UserAgentType::Unknown:
        {
            // do nothing here since it is unknown type
            break;
        }
    default:
        {
            if(fBot)
            {
                m_pluginCounters.cOtherBotVisitCount.Increment();
            }

            break;
        }
    }
}

void PluginSite::OnSendResponse(WSFilter *pFilter)
{
    m_pluginCounters.cBeginResponse.Increment();
    int nStatus;
    DebugString("OnSendResponse()");

    g_nc.Inc(50);
    if (!pFilter->GetHTTPStatus(nStatus))
    {
        g_nc.Inc(51);
        DebugString("Can not retrieve HTTP status");
        return;
    }

    //nStatus = 0; // simulate .aspx problem
    if (HttpCodeValid(nStatus)) // IIS has cases where nStatus is not ready during SendResponse
    {
        if (!CheckHttpCode(nStatus))
        {
            g_nc.Inc(52);
            return;
        }
    }

    Size_t nLen;
#if 1
    TCHAR wszContentType[2048];
    nLen = ARRAYSIZE(wszContentType);
    if (!pFilter->GetContentType(wszContentType, nLen))
    {
        g_nc.Inc(53);
        m_pluginCounters.cNoContentType.Increment();
        DebugString("no Content-Type header");
        return;
    }

    const TCHAR* wszTextHtml = TEXT("text/");
    Size_t cchTextHtml = TSTRLEN(wszTextHtml);
    if (TSTRLEN(wszContentType) < cchTextHtml || memcmp(wszContentType, wszTextHtml, cchTextHtml * sizeof(TCHAR)))
    {
        g_nc.Inc(54);
        m_pluginCounters.cNotTextHtml.Increment();
        if (m_pluginCounters.cNotTextHtml.m_val < 5)
        {
            PluginLog("Bad ContentType " STRFMT_LS, wszContentType);
        }
        DebugString("Content-Type not text/");
        return;
    }

    char szUserAgent[1024];
    nLen = ARRAYSIZE(szUserAgent);
    if (!pFilter->GetUserAgent(szUserAgent, nLen))
    {
        g_nc.Inc(55);
        DebugString("no User-agent");
        return;
    }

    if (nLen >= m_config.uMaxUserAgent)
    {
        g_nc.Inc(56);
        m_pluginCounters.cBadUserAgent.Increment();
        return;
    }

    BOOL fBot = !m_userAgents.Allowed(szUserAgent);

    UpdateVisitCounters(szUserAgent, fBot);

#endif

    TCHAR wszUrl[2048];
    nLen = ARRAYSIZE(wszUrl);
    char szUrl[2048];

    if (!pFilter->GetURL(wszUrl, nLen) ||
        (PALString::Convert(szUrl, ARRAYSIZE(szUrl), wszUrl, nLen) != 0))
    {
        g_nc.Inc(58);
        DebugString("no URL");
        return;
    }

    if (!m_robots.Allowed(szUrl))
    {
        g_nc.Inc(59);
        m_pluginCounters.cExcluded.Increment();
        return;
    }

    char pcNormalized[2048];
    Size_t cbNormalized = sizeof(pcNormalized);
    if (!m_normalizer.NormalizeUrl(szUrl, pcNormalized, &cbNormalized))
    {
        g_nc.Inc(60);
        m_pluginCounters.cNormalizeFailed.Increment();
        return;
    }
    if (cbNormalized != strlen(szUrl))
    {
        m_pluginCounters.cNormalized.Increment();
    }

    //char szTemp[1024];
    //sprintf_s(szTemp, sizeof(szTemp), "%I64X SEND_RESPONSE %s", (UInt64)pfc, sz);
    //DebugString(szTemp);

#if !DBG
    if (m_doubleBloomFilter.Check(cbNormalized, pcNormalized))
    {
        g_nc.Inc(61);
        m_pluginCounters.cBloomFilterDeduped.Increment();
        return;
    }
#endif

    if (sizeof(ResponseBuffer) + cbNormalized > m_config.uMinDocumentSizeBytes)
    {
        g_nc.Inc(62);
        DebugString("URL too big");
        return;
    }

    pFilter->SetContext(HeapAllocWrap(m_config.uMinDocumentSizeBytes));
    if (!pFilter->GetContext())
    {
        g_nc.Inc(63);
        m_pluginCounters.cHeapAllocFailed.Increment();
        return;
    }

    ResponseBuffer* pBuf = (ResponseBuffer*)(pFilter->GetContext());
    pBuf->pSite = this;
    pBuf->cbAlloc = m_config.uMinDocumentSizeBytes;
    pBuf->cbUrl = (unsigned)cbNormalized;
    pBuf->cbUsed = (unsigned)(sizeof(ResponseBuffer) + cbNormalized);
    pBuf->uiHttpStatusCode = (unsigned)nStatus;
    pBuf->flags = fBot ? RESPONSE_FLAG_BOT : 0;
    memcpy(pBuf + 1, pcNormalized, cbNormalized);
    m_pluginCounters.cResponseStarted.Increment();
    g_nc.Inc(64);
}

#ifdef __TARGET_IIS__

static BOOL IsSpace(char c) // ctype.h isspace() considers \r and \n as space, which we don't want
{
    return ' ' == c || '\t' == c;
}

// Better place for this function ?
static unsigned ParseHttpCode(__out_bcount(cb) BYTE* pb, Size_t cb)
{ // HTTP/[\d\.]*\s+(\d+)
    unsigned uRet = 0;

    BYTE* pbEnd = pb + cb;
    const char* szHTTP = "HTTP/";
    Size_t cbHTTP = strlen(szHTTP);
    if (pb + cbHTTP > pbEnd)
    {
        return 0;
    }
    if (PALString::CompareInsensitive((char*)pb, szHTTP, cbHTTP))
    {
        return 0;
    }
    pb += cbHTTP;

    while ((pb < pbEnd) && !IsSpace(*pb))
    {
        if ('.' == *pb || isdigit(*pb))
        {
            pb++;
            continue;
        }
        return 0;
    }

    if (pb >= pbEnd)
    {
        return 0;
    }
    while ((pb < pbEnd) && IsSpace(*pb))
    {
        pb++;
    }

    if (pb >= pbEnd)
    {
        return 0;
    }
    while ((pb < pbEnd) && isdigit(*pb))
    { // not using scanf because pb is not NULL-terminated
        uRet = uRet * 10 + (*pb - '0');
        pb++;
    }
    return uRet;
}

#endif

void PluginSite::OnRawData(WSFilter *pFilter)
{
    DebugString("OnRawData()");
    ResponseBuffer* pBuf = (ResponseBuffer*)(pFilter->GetContext());
    void *pData;
    Size_t cbSize;
    BOOL bProcessContent = TRUE;

    while (pFilter->GetNextDataSet(pData, cbSize) == TRUE)
    {
        // In Apache, we need to loop through the data set to
        // see if it hits end of file.
        if (!bProcessContent)
        {
            continue;
        }
#if DBG
        char szTemp[4096];
        if (memcmp(pData, "HTTP", 4))
        {
            sprintf_s(szTemp, sizeof(szTemp), "SEND_RAW_DATA %u '%.*s...%.*s'", cbSize, 20, (char*)pData, 20, (BYTE*)pData + cbSize - 20);
        }
        else
        {
            sprintf_s(szTemp, sizeof(szTemp), "%.*s", cbSize, (char*)pData);
        }
        DebugString(szTemp);
        if (!pData)
        {
            DebugString("NULL data");
        }
#endif

        // Just for the first buffer, assuming the headers would fit into the first buffer
        if (sizeof(ResponseBuffer) + pBuf->cbUrl == pBuf->cbUsed)
        {   // For compressed content, we will skip the URL entirely
            if (pFilter->IsHTTPCompressedContent())
            {
                m_pluginCounters.cBodyCompressed.Increment();
                if (pBuf->pSite->Config()->fDecompress)
                {
                    pBuf->flags |= RESPONSE_FLAG_COMPRESSED;
                }
                else
                {
                    DebugString("ignoring compressed content");
                    HeapFreeWrap(pFilter->GetContext(), pBuf->cbAlloc);
                    pFilter->SetContext(PTR_FFFF);
                    m_pluginCounters.cDecompressionNotAttempted.Increment();
                    return;
                }
            }
#ifdef __TARGET_IIS__
            if (!HttpCodeValid(pBuf->uiHttpStatusCode)) // some .aspx pages
            {
                pBuf->uiHttpStatusCode = ParseHttpCode((BYTE*)pData, cbSize);
#if DBG
                PluginLog("Parsed HTTP %u", pBuf->uiHttpStatusCode);
#endif
                if (!CheckHttpCode(pBuf->uiHttpStatusCode))
                {
                    HeapFreeWrap(pFilter->GetContext(), pBuf->cbAlloc);
                    pFilter->SetContext(PTR_FFFF);
                    return;
                }
            }
#endif        
            if (404 == pBuf->uiHttpStatusCode)
            {
                m_pluginCounters.cHTTP404Content.Increment();

                // For HTTP status 404, we will skip the rest as the content doesn't matter.
                // NOTE: The buffer should still be kept because EndResponse() checks if the buffer is empty or not.
                DebugString("Skipping HTTP 404");
                bProcessContent = FALSE;
                continue;
            }
        }

        if (pBuf->cbAlloc < pBuf->cbUsed)
        {
            DebugString("confused");
            return;
        }

        Size_t cbNewAlloc = pBuf->cbAlloc;
        while ((cbNewAlloc - pBuf->cbUsed < cbSize) && (cbNewAlloc < m_config.uMaxDocumentSizeBytes))
        {
            cbNewAlloc *= 4;
        }
        if (cbNewAlloc > m_config.uMaxDocumentSizeBytes)
            cbNewAlloc = m_config.uMaxDocumentSizeBytes;

        if (cbNewAlloc - pBuf->cbUsed < cbSize)
        {
            m_pluginCounters.cDocumentTooBig.Increment();
            HeapFreeWrap(pFilter->GetContext(), pBuf->cbAlloc);
            pFilter->SetContext(PTR_FFFF);
            return;
        }

        if (cbNewAlloc > pBuf->cbAlloc)
        {
            void* pNew = HeapAllocWrap(cbNewAlloc);
            if (!pNew)
            {
                m_pluginCounters.cHeapReallocFailed.Increment();
                HeapFreeWrap(pFilter->GetContext(), pBuf->cbAlloc);
                pFilter->SetContext(PTR_FFFF);
                return;
            }

            memcpy(pNew, pBuf, pBuf->cbUsed);
            HeapFreeWrap(pFilter->GetContext(), pBuf->cbAlloc);
            pFilter->SetContext(pNew);
            pBuf = (ResponseBuffer*)pNew;
            pBuf->cbAlloc = (unsigned)cbNewAlloc;
        }

        memcpy(((char*)pBuf) + pBuf->cbUsed, pData, cbSize);
        pBuf->cbUsed += cbSize;
    }
}

void PluginSite::OnEndRequest(WSFilter *pFilter)
{
    g_nc.Inc(80);
    DebugString("OnEndRequest()");
    m_pluginCounters.cEndResponse.Increment();
    ResponseBuffer* pBuf = (ResponseBuffer*)(pFilter->GetContext());

    if (!pBuf || PTR_FFFF == pBuf)
    {
        g_nc.Inc(81);
        DebugString("END_REQUEST uninitialized");
        return;
    }

    if (pBuf->cbUsed < sizeof(ResponseBuffer) + pBuf->cbUrl)
    {
        g_nc.Inc(82);
        DebugString("Response buffer invalid");
        return;
    }

    if (!HttpCodeValid(pBuf->uiHttpStatusCode))
    {
        g_nc.Inc(83);
        DebugString("Http code not parsed");
        return;
    }

    m_pluginCounters.cResponseCompleted.Increment();

#if 0
    char szTemp[1024];
    char* pbBody = (char*)(pBuf + 1) + pBuf->cbUrl;
    Size_t cbBody = pBuf->cbUsed - sizeof(*pBuf) - pBuf->cbUrl;
    SPRINTF_S(szTemp, sizeof(szTemp), "END_REQUEST %.*s %u \"%.*s...%.*s\"", pBuf->cbUrl, pBuf + 1, cbBody, 4, pbBody, 4, pbBody + cbBody - 4);
    DebugString(szTemp);
    PluginLog(szTemp);
#endif

    if (g_globals.m_pThreadPool->DispatchWork(pFilter->GetContext()) == TRUE)
    {
        g_nc.Inc(84);
        m_pluginCounters.cWorkSubmitted.Increment();
        PALInterlocked::Increment(&s_cPendingWork);
    }
    else
    {
        g_nc.Inc(85);
        m_pluginCounters.cWorkCreationFailed.Increment();
        HeapFreeWrap(pFilter->GetContext(), ((ResponseBuffer*)(pFilter->GetContext()))->cbAlloc);
    }
    pFilter->SetContext(PTR_FFFF); // we shouldn't see this again
    g_nc.Inc(86);
}

void PluginSite::Close()
{
    Flush();
    PluginLog("plugin terminated gracefully");

    Free();
}

void PluginSite::Flush()
{
    m_outFile.Flush();
    SaveCache();
    SaveCountersWrap();
    PluginLog("plugin flushed");
}
