/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/

#pragma once
#include "DiskSpaceTracker.h"
#include "DoubleBloomFilter.h"
#include "FileWriter.h"
#include "PluginLibCommon.h"
#include "UrlNormalizer.h"
#include "SSPConfig.h"
#include "SSPLRUCache.h"
#include "SSPRobots.h"
#include "UserAgents.h"
#include "WSFilter.h"

class PluginSite;
typedef struct
{
    PluginSite* pSite;
    Size_t cbAlloc;
    Size_t cbUsed;
    Size_t cbUrl;
    unsigned uiHttpStatusCode;
    unsigned flags;
} ResponseBuffer;

class PluginSite
{
    typedef void (PluginSite::*PFNVOIDVOID)();

    Counters m_pluginCounters;
    TCHAR m_wszSubfolder[MAX_PATH];
    TCHAR m_wszLogFile[MAX_PATH];
    TCHAR m_wszCountersFile[MAX_PATH];
    SSPConfig m_config; 
    DiskSpaceTracker m_diskSpaceTracker;
    DoubleBloomFilter m_doubleBloomFilter;
    Robots m_robots;
    UrlNormalizer m_normalizer;
    volatile UInt32 m_lHeapUsedKB;
    Size_t m_cbCache;
    LRUCache<CacheEntry> m_cache;
    int m_icsCache;
    FileWriter m_outFile;
    char m_szHostPath[HOST_PATH_LENGTH];
    UserAgents m_userAgents;
    volatile UInt32 m_tLastCacheSave;
    volatile UInt32 m_tLastCountersSave;
#if APR_HAVE_WINDOWS_H
    HANDLE m_hLockFile;
#endif

    void PluginLog(const char* szFormat, ...);
    void LoadCache();
    void SaveCache();
    void WriteEntry(unsigned flags, Signature sig, UInt16 visitCount, const char* pbUrl, size_t cbUrl, FileWriter* pFile);
    // Lookup URL from cache and decide if the URL should be written to the delta entry
    // 1. If URL is not cached and it has 404 status, it is not considered for cache nor write.
    // 2. If URL is cached with 200 status and it is now 404, the cache entry is flagged to be considered for 404 list with max number to signature.
    BOOL CacheLookup(Hash hash, Signature sig, UInt16 tNow, UInt16* pVisitCount, BOOL *pfWrite);
    void SaveCountersWrap();
    void ExecuteIfTime(time_t tNow, volatile UInt32 *pt, unsigned uIntervalSec, PFNVOIDVOID pfn);
    BOOL StripHostAndPort(const char** ppbUrl, Size_t* pcbUrl);
    BOOL CheckHttpCode(int nStatus);
    BOOL Decompress(const BYTE* pbIn, Size_t cbIn, BYTE** ppbOut, Size_t* pcbOut, Size_t* pcbOutAlloc);
    void UpdateVisitCounters(const char* szUserAgent, BOOL fBot);

public:
    void Zero();
    void Free();
    static PluginSite* Create(IAllocator* pAlloc, ICritSecDispenser* pCSD, const TCHAR* wszHostAndPort, const TCHAR* wszApplPath);
    BOOL Initialize(IAllocator* pAlloc, ICritSecDispenser* pCSD, const TCHAR* wszHostAndPort, const TCHAR* wszApplPath);
    void Flush();
    void Close();
    void Tick();
    void PluginLogV(const char* szFormat, va_list);
    SSPConfig* Config() { return &m_config; }

    void OnSendResponse(WSFilter *pFilter);
    void OnRawData(WSFilter *pFilter);
    void OnEndRequest(WSFilter *pFilter);

    void ProcessFinishedResponse(ResponseBuffer* p);

    void* HeapAllocWrap(Size_t cb);
    void HeapFreeWrap(void* pv, Size_t cb);
    static void *zlib_alloc(void *opaque, UINT items, UINT size);
    static void zlib_free(void *opaque, void *address);
};


class PluginLogDelegate : public IPluginLog
{
    PluginSite* m_pSite;
public:
    PluginLogDelegate(PluginSite* pSite) : m_pSite(pSite) { }
    void PluginLog(const char* szFormat, ...);
};