/***********************************************************************
*                                                                      *
*   Copyright (c) Microsoft Corporation                                *
*                                                                      *
*   All rights reserved.                                               *
*                                                                      *
*   Licensed under the Apache License, Version 2.0 (the "License");    *
*   you may not use this file except in compliance with the License.   *
*   You may obtain a copy of the License at                            *
*   http://www.apache.org/licenses/LICENSE-2.0                         *
*                                                                      *
*   THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR         *
*   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING       *
*   WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE,  *
*   FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR                *
*   NON-INFRINGEMENT.                                                  *
*                                                                      *
*   See the Apache Version 2.0 License for specific language           *
*   governing permissions and limitations under the License.           *
*                                                                      *
***********************************************************************/
#include "SSPCommon.h"
#include "WSFilter.h"
#include "PALString.h"

#ifdef __TARGET_IIS__

#ifdef _UNICODE
#define SVRVAR_PREFIX "UNICODE_"
#else
#define SVRVAR_PREFIX
#endif

WSFilter::WSFilter(PHTTP_FILTER_CONTEXT p) :
    m_pfc(p),
    m_pResponse(NULL),
    m_pDataBuffer(NULL),
    m_fNoMore(FALSE),
    m_fCompressed(FALSE)
{
}

WSFilter::~WSFilter()
{
}

BOOL WSFilter::NoMore() const
{
    return m_fNoMore;
}

void *WSFilter::GetContext() const
{
    return m_pfc->pFilterContext;
}

void WSFilter::SetContext(void *pContext)
{
    m_pfc->pFilterContext = pContext;
}

void *WSFilter::GetPlatformMemPool()
{
    return NULL;
}

BOOL WSFilter::GetNextDataSet(void *&pData, Size_t &cbSize)
{
    if (m_pDataBuffer)
    {
        m_fCompressed = IsCompressed();
        pData = m_pDataBuffer->pvInData;
        cbSize = m_pDataBuffer->cbInData;
        m_pDataBuffer = NULL;
        return TRUE;
    }

    pData = NULL;
    cbSize = 0;
    
    return FALSE;
}

BOOL WSFilter::GetApplPhysicalPath(TCHAR *tszString, Size_t &nLen)
{
    return GetServerVariable(SVRVAR_PREFIX "APPL_PHYSICAL_PATH", tszString, nLen);
}

BOOL WSFilter::GetContentType(__out_ecount(nLen)TCHAR *tszString, Size_t &nLen)
{
    return GetHeaderVariable("Content-Type:", tszString, nLen);
}

BOOL WSFilter::GetServerName(TCHAR *tszString, Size_t &nLen)
{
    // use SERVER_NAME (Host HTTP header) instead of HTTP_HOST, which returns invalid or empty buffer during SF_NOTIFY_SEND_RESPONSE notification.
    // SF_NOTIFY_SEND_RESPONSE: Sent after the request is processed, but before any headers are sent back to the client. 
    return GetServerVariable(SVRVAR_PREFIX "SERVER_NAME", tszString, nLen);
}

BOOL WSFilter::GetServerPort(TCHAR *tszString, Size_t &nLen)
{
    return GetServerVariable(SVRVAR_PREFIX "SERVER_PORT", tszString, nLen);
}

BOOL WSFilter::GetHTTPStatus(int &nStatus)
{
    HTTP_FILTER_SEND_RESPONSE *pResponse = m_pResponse;

    if (pResponse != NULL)
    {
        nStatus = pResponse->HttpStatus;
        return TRUE;
    }
    
    return FALSE;
}

BOOL WSFilter::GetURL(TCHAR *tszString, Size_t &nLen)
{
    return GetServerVariable(SVRVAR_PREFIX "URL", tszString, nLen);
}

BOOL WSFilter::GetUserAgent(char *szString, Size_t &nLen)
{   // "It is not possible to retrieve unicode representations of request headers that are retrieved using the "HTTP_" or "HEADER_" prefix."
    // (http://msdn.microsoft.com/en-us/library/ms525335(v=vs.90).aspx)
    return GetServerVariableA("HTTP_USER_AGENT", szString, nLen);
}

BOOL WSFilter::IsHTTPCompressedContent() const
{    
    return m_fCompressed;
}


BOOL WSFilter::GetServerVariable(LPSTR lpszVariableName, TCHAR *tszString, Size_t &nLen)
{
    // BUGBUG: Need to check fo overflow.
    DWORD cbSize = (DWORD) nLen * sizeof(TCHAR);
    BOOL bRetVal =  m_pfc->GetServerVariable(m_pfc, lpszVariableName, (LPVOID) tszString, &cbSize);

    if (bRetVal == TRUE)
    {
        nLen = cbSize / sizeof(TCHAR) - 1;
    }

    return bRetVal;
}

BOOL WSFilter::GetServerVariableA(LPSTR lpszVariableName, char *szString, Size_t &nLen)
{
    // BUGBUG: Need to check fo overflow.
    DWORD cbSize = (DWORD) nLen;
    BOOL bRetVal =  m_pfc->GetServerVariable(m_pfc, lpszVariableName, (LPVOID) szString, &cbSize);

    if (bRetVal == TRUE)
    {
        nLen = cbSize;
    }

    return bRetVal;
}

BOOL WSFilter::GetHeaderVariable(LPSTR lpszVariableName, __out_ecount(nLen) TCHAR *tszString, Size_t &nLen)
{
    char szHeaderValue[2048];
    DWORD cbSize = ARRAYSIZE(szHeaderValue);
    HTTP_FILTER_SEND_RESPONSE *pResponse = m_pResponse;

    BOOL bRetVal =  pResponse->GetHeader(m_pfc, lpszVariableName, (LPVOID) szHeaderValue, &cbSize);

    if (bRetVal == TRUE)
    {
        if (nLen < cbSize * sizeof(TCHAR))
        {
            return FALSE;
        }

        if (PALString::Convert(tszString, nLen, szHeaderValue, cbSize - 1) == 0)
        {
            nLen = cbSize - 1;
        }
    }

    return bRetVal;
}

BOOL WSFilter::StringContains(const void* pv, size_t cb, const char* szSearch) const
{
    size_t cbSearch = strlen(szSearch);
    if (!cbSearch)
    {
        return TRUE;
    }

    for (const char* p = (const char*)pv; p + cbSearch <= (const char*)pv + cb; p++)
    {
        if (*p == *szSearch && 0 == memcmp(p, szSearch, cbSearch))
        {
            return TRUE;
        }
    }
    return FALSE;
}

BOOL WSFilter::IsCompressed() const
{
    HTTP_FILTER_RAW_DATA *pRawData = m_pDataBuffer;

    if (pRawData != NULL)
    {
        if((pRawData->cbInData > 4) &&
            !memcmp(pRawData->pvInData, "HTTP", 4) &&
            StringContains(pRawData->pvInData, pRawData->cbInData, "Content-Encoding:"))
        {
            return TRUE;
        }
        else
        {
            return FALSE;
        }
    }

    return FALSE;
}

#else
WSFilter::WSFilter() :
    m_pfc(NULL),
    m_pBrigade(NULL),
    m_pCurrBucket(NULL),
    m_pContext(NULL),
    m_fNoMore(FALSE)
{
}

WSFilter::~WSFilter()
{
}

void WSFilter::SetFilter(ap_filter_t *pFilter, apr_bucket_brigade *pBrigade)
{
    m_pfc = pFilter;
    m_pBrigade = pBrigade;
    m_pCurrBucket = NULL;
    m_fNoMore = FALSE;
}

void *WSFilter::GetContext() const
{
    return m_pContext;
}

void WSFilter::SetContext(void *pContext)
{
    m_pContext = pContext;
}

void *WSFilter::GetPlatformMemPool()
{
    return (void *) (m_pfc->r->pool);
}

BOOL WSFilter::GetNextDataSet(void *&pData, Size_t &cbSize)
{
    apr_status_t status;
    BOOL bRetVal = FALSE;
    pData = NULL;
    cbSize = 0;

    // No data in this data set. Wait for next one.
    if (m_pBrigade == NULL || APR_BRIGADE_EMPTY(m_pBrigade))
    {
        return FALSE;
    }

    if (m_pCurrBucket == NULL)
    {
        m_pCurrBucket = APR_BRIGADE_FIRST(m_pBrigade);
    }

    if (m_pCurrBucket == APR_BRIGADE_SENTINEL(m_pBrigade))
    {
        // There are no more buckets left in the brigade.
        // Need to get the next brigade.
        return FALSE;
    }

    if (APR_BUCKET_IS_EOS(m_pCurrBucket))
    {
        // No more data, we can now perform the site computation.
        m_fNoMore = TRUE;
        return FALSE;
    }

    if (APR_BUCKET_IS_FLUSH(m_pCurrBucket))
    {
        // Flush signal, move to next data set.
        bRetVal = TRUE;
        goto Exit;
    }

#if APR_HAVE_WINDOWS_H
    // Unfortunately the read() fn pointer is not correctly defined in apr_buckets.h because
    // it does not explicitly define the calling convention which fails at runtime for stdcall
    // (since the function is compiled in __cdecl). We solve this problem by casting using the
    // correct calling convention.
    status = ((apr_status_t (__cdecl *) (apr_bucket *b, const char **str, apr_size_t *len, apr_read_type_e block))
        ((void *) m_pCurrBucket->type->read))(m_pCurrBucket, (const char **) &pData, &cbSize, APR_BLOCK_READ);
#else
    // For Linux, we are building using the c calling convention.
    status = apr_bucket_read(m_pCurrBucket, (const char **) &pData, &cbSize, APR_BLOCK_READ);
#endif

    if (status != APR_SUCCESS)
    {
        // This is an unexpected error.
        DEBUG_ap_log_rerror(APLOG_MARK, APLOG_ERR, status, m_pfc->r, "Unable read from bucket");
        pData = NULL;
        cbSize = 0;
        goto Exit;
    }
           
    bRetVal = TRUE;
    
Exit:
    m_pCurrBucket = APR_BUCKET_NEXT(m_pCurrBucket);
    return bRetVal;
}

BOOL WSFilter::NoMore() const
{
    return m_fNoMore;
}

BOOL WSFilter::GetApplPhysicalPath(TCHAR *tszString, Size_t &nLen)
{
    if (m_pfc->r->subprocess_env == NULL)
    {
        return FALSE;
    }

#ifndef _TEST_MODE
    return CopyCheckString(tszString, nLen, ap_document_root(m_pfc->r));
#else
    return CopyCheckString(tszString, nLen, apr_table_get(m_pfc->r->subprocess_env, "DOCUMENT_ROOT"));
#endif
}

BOOL WSFilter::GetContentType(TCHAR *tszString, Size_t &nLen)
{
    return CopyCheckString(tszString, nLen, m_pfc->r->content_type);
}

BOOL WSFilter::GetServerName(TCHAR *tszString, Size_t &nLen)
{
    return CopyCheckString(tszString, nLen, m_pfc->r->hostname);
}

BOOL WSFilter::GetServerPort(TCHAR *tszString, Size_t &nLen)
{
    // Need to be larger than PORT_STR_SIZE (because of null terminated string) to hold 16-bit int.
    if (nLen <= PORT_STR_SIZE)
    {
        return FALSE;
    }

#ifndef _TEST_MODE
    if (SNTPRINTF_S(tszString, nLen, nLen - 1, TEXT("%u"), ap_get_server_port(m_pfc->r)) < 0)
#else
    if (SNTPRINTF_S(tszString, nLen, nLen - 1, TEXT("%u"), m_pfc->r->server->port) < 0)
#endif
    {
        return FALSE;
    }

    nLen = TSTRLEN(tszString);

    return TRUE;
}

BOOL WSFilter::GetHTTPStatus(int &nStatus)
{
    nStatus = m_pfc->r->status;
    return TRUE;
}

BOOL WSFilter::GetURL(TCHAR *tszString, Size_t &nLen)
{
    // Returns the URI.
    // Example: http://host/index.html will return "/index.html"
    return CopyCheckString(tszString, nLen, m_pfc->r->uri);
}

BOOL WSFilter::GetUserAgent(char *szString, Size_t &nLen)
{
    if (m_pfc->r->headers_in == NULL)
    {
        return FALSE;
    }

    const char* sz = apr_table_get(m_pfc->r->headers_in, "User-Agent");
    if (!sz)
    {
        return FALSE;
    }

    Size_t n = strlen(sz);
    if (n >= nLen)
    {
        return FALSE;
    }
    memcpy(szString, sz, n + 1);
    nLen = n;

    return TRUE;
}

BOOL WSFilter::IsHTTPCompressedContent() const
{
    if (m_pfc->r->protocol == NULL ||
        strncmp(m_pfc->r->protocol, "HTTP", 4) != 0 ||
        m_pfc->r->content_encoding == NULL) // To do we may need to check the content_encoding content.
    {
        return FALSE;
    }
    
    return TRUE;
}

BOOL WSFilter::CopyCheckString(TCHAR *tszString, Size_t &nLen, const char *szSrc)
{
    if (szSrc == NULL)
    {
        return FALSE;
    }

    Size_t nSrcLen = strlen(szSrc);

    if (nLen < nSrcLen + 1)
    {
        return FALSE;
    }

    if (PALString::Convert(tszString, nLen, szSrc, nSrcLen) != 0)
    {
        return FALSE;
    }

    nLen = nSrcLen;

    return TRUE;
}
#endif
